/*
 * ArrayIndexDialog.java
 *
 * Created on June 5, 2002, 8:23 PM
 */

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;

/**
 *
 * @author  _
 * @version 
 */
public class ArrayIndexDialog extends JDialog
    implements ActionListener, ChangeListener {

    // Because a reference is passed, all methods on this operand
    // will affect the original object...so no need to return the changed object.
    public ArrayIndexDialog( JFrame jfFrame, Operand operThisOperand,
            SortableDeclarationsTableModel sdtm ) {
        super( jfFrame, "Array Index Selector: " + operThisOperand.getOperandName(),
            B_MODAL );
                
        Container cPane;
        GridBagLayout gblLayoutGrid;
        
        sdtmDataObjects = sdtm;
        
        gblLayoutGrid = new GridBagLayout();
        cPane = this.getContentPane();
        jpDialogPanel = new JPanel();
        cPane.add( jpDialogPanel );
        jpDialogPanel.setLayout( gblLayoutGrid );
        
        fontLabel = new Font( "SansSerif", Font.BOLD, 14 );
        fontMenuItem = new Font( "monospaced", Font.PLAIN, 14 );

        bgRadioButtonGroup = new ButtonGroup();
        jrbNumber = createRadioButton( "Number: " );
        jrbDataObject = createRadioButton( "Data Object" );
        
        jbOK = createButton( "OK" );
        jbCancel = createButton( "Cancel" );
        jbHelp = createButton( "Help" );
        
        jlArrayIndexValue = createLabel( "" );
        jsArraySlider = createSlider( operThisOperand );
                
        jcbAvailableDataObjects = createComboBox();
        
        ptvhdHelp = new PTVHelpDialog( jfFrame, "Help >> General >> Array Index Selector" );
        ptvhdHelp.setHelpFile( "help_general_array.html" );
        
        this.setSize( 412, 194 );
        createPanel();
        
        this.addWindowListener( new WindowAdapter() {
            public void windowClosing( WindowEvent weEvent ) {
                closeThisWindow();
            }
        } );                
    }

    public Object displayDialog() {
        this.show();  // does not return until OK or Cancel
        return objReturnObject;
    }
       
    //----------------- ActionListener Implementation -----------\\
    public void actionPerformed( ActionEvent aeEvent ) {
        Object objThisObject;
        int iValue;
        
        objThisObject = aeEvent.getSource();

        if ( objThisObject == jrbNumber ) {
            if ( jrbNumber.isSelected() == true ) {
                jsArraySlider.setEnabled( true );
                jcbAvailableDataObjects.setEnabled( false );
            }
        }
        else if ( objThisObject == jrbDataObject ) {
            if ( jrbDataObject.isSelected() == true ) {
                jsArraySlider.setEnabled( false );
                jcbAvailableDataObjects.setEnabled( true );
            }
        }
        else if ( objThisObject == jbOK ) {
            if ( jrbNumber.isSelected() == true ) {
                System.out.println( "index = " + jsArraySlider.getValue() );
                objReturnObject = new Integer( jsArraySlider.getValue() );
                closeThisWindow();
            }
            else if ( jrbDataObject.isSelected() == true ) {
                objReturnObject = new String( (String)
                    jcbAvailableDataObjects.getSelectedItem() );
                closeThisWindow();
            }
            else {
                objReturnObject = new Status( I_INTERNAL_ERROR, "Internal error." );
            }
        }
        else if ( objThisObject == jbCancel ) {
            objReturnObject = new Status( I_CANCELLED_OPERATION, "Operation cancelled." );
            closeThisWindow();
        }
        else if ( objThisObject == jbHelp ) {
            ptvhdHelp.show();            
        }
    }
    
    //----------------- ChangeListener Implementation -----------\\
    public void stateChanged( ChangeEvent ceEvent ) {
        Object objThisObject;
                
        objThisObject = ceEvent.getSource();
        
        if ( objThisObject == jsArraySlider ) {
            jlArrayIndexValue.setText( Integer.toString( jsArraySlider.getValue() ) );
        }
    }
    
    /////////////////// PRIVATE METHODS \\\\\\\\\\\\\\\\\\\\\\\\\
    private void createPanel() {
        addComponentToPanel( 0, 0, 2, 1, jrbDataObject );
        addComponentToPanel( 2, 0, 3, 1, jcbAvailableDataObjects );        
                
        addComponentToPanel( 0, 1, 5, 1, new JLabel( "     " ) );

        addComponentToPanel( 0, 2, 2, 1, jrbNumber );      
        addComponentToPanel( 2, 2, 4, 2, jsArraySlider );
        
        addComponentToPanel( 0, 3, 1, 1, jlArrayIndexValue ); 
        addComponentToPanel( 1, 3, 1, 1, new JLabel( "     " ) );
        jlArrayIndexValue.setHorizontalAlignment( SwingConstants.CENTER );
        
        addComponentToPanel( 0, 4, 5, 1, new JLabel( "     " ) );        
        
        addComponentToPanel( 0, 5, 1, 1, new JLabel( "     " ) );
        addComponentToPanel( 1, 5, 1, 1, jbOK );
        addComponentToPanel( 2, 5, 1, 1, jbCancel );
        addComponentToPanel( 3, 5, 1, 1, jbHelp );
        addComponentToPanel( 4, 5, 1, 1, new JLabel( "     " ) );        
    }
    
    private void addComponentToPanel( int iGridXPos, int iGridYPos, int iGridWidth,
                                      int iGridHeight, Component coComponent ) {
        
        GridBagConstraints gbcLayoutConstraints;

        gbcLayoutConstraints = new GridBagConstraints();
        gbcLayoutConstraints.weightx = 100;
        gbcLayoutConstraints.weighty = 100;
        gbcLayoutConstraints.gridx = iGridXPos;
        gbcLayoutConstraints.gridy = iGridYPos;
        gbcLayoutConstraints.gridwidth = iGridWidth;
        gbcLayoutConstraints.gridheight = iGridHeight;
        gbcLayoutConstraints.anchor = GridBagConstraints.SOUTHWEST;
        gbcLayoutConstraints.insets = new Insets( 2, 5, 2, 3 );
        gbcLayoutConstraints.fill =  GridBagConstraints.HORIZONTAL;
        jpDialogPanel.add( coComponent, gbcLayoutConstraints );
    }
    
    private JButton createButton( String stText ) {
        JButton jbThisButton;
        
        jbThisButton = new JButton( stText );
        jbThisButton.addActionListener( this );
        return jbThisButton;
    }
    
    private JLabel createLabel( String stText ) {
        JLabel jlThisLabel;
        
        jlThisLabel = new JLabel( stText );
        jlThisLabel.setFont( fontLabel );
        jlThisLabel.setForeground( Color.black );
        return jlThisLabel;
    }
    
    private JRadioButton createRadioButton( String stText ) {
        JRadioButton jrbThisRadioButton;
     
        jrbThisRadioButton = new JRadioButton( stText, false );
        jrbThisRadioButton.setFont( fontLabel );
        jrbThisRadioButton.setForeground( Color.black );
        jrbThisRadioButton.addActionListener( this );
        bgRadioButtonGroup.add( jrbThisRadioButton );
        return jrbThisRadioButton;
    }

    private JSlider createSlider( Operand operThisOperand ) {    
        JSlider jsThisSlider;
        int iMinimum, iMaximum, iMinorTick, iMajorTick;
 
        iMinimum = 0;
        iMaximum = operThisOperand.getOperandArrayLength() - 1;
        iMajorTick = iMaximum / 10;
        iMinorTick = iMaximum / 30;
        
        if ( iMajorTick < 1 ) {
            iMinorTick = 0;
            iMajorTick = 1;
        }
        else if ( iMinorTick < 1 ) {
            iMinorTick = 1;
        }
        
        jsThisSlider = new JSlider( SwingConstants.HORIZONTAL, iMinimum, 
            iMaximum, iMinimum );
        jsThisSlider.setFont( fontLabel );
        jsThisSlider.setForeground( Color.black );
                
        jsThisSlider.setMajorTickSpacing( iMajorTick );
        jsThisSlider.setMinorTickSpacing( iMinorTick );
        jsThisSlider.setPaintTicks( true );
        jsThisSlider.setPaintLabels( true );
        jsThisSlider.addChangeListener( this );
        
        return jsThisSlider;
    }
    
    private JComboBox createComboBox() {
        JComboBox jcbComboBox;
        String stObjectName, stStructure, stType;
        int iIndex;
    
        jcbComboBox = new JComboBox();
        jcbComboBox.setFont( fontMenuItem );
        jcbComboBox.setForeground( Color.black );
        jcbComboBox.setEditable( false );
        
        //System.out.println( "Number of objects declared = " + sdtmDataObjects.getRowCount() );
        for ( iIndex = 0; iIndex < sdtmDataObjects.getRowCount();
            iIndex++ ) {
            
            stObjectName = (String)sdtmDataObjects.getValueAt( 
                iIndex, DeclarationsTableModel.I_COLUMN_NAME );                
            stStructure = (String)sdtmDataObjects.getValueAt( 
                iIndex, DeclarationsTableModel.I_COLUMN_STRUCTURE );
            stType = (String)sdtmDataObjects.getValueAt( 
                iIndex, DeclarationsTableModel.I_COLUMN_TYPE );

            //System.out.println( "Object name: " + stObjectName );
            
            if ( ( stStructure.startsWith( "Single" ) )  &&
                   ( stType.startsWith( "Integer" ) ) ) {
                jcbComboBox.addItem( stObjectName );
                //System.out.println( "Added name: " + stObjectName );
            }            
        }
        
        for ( iIndex = SystemDeclarations.I_SINGLE_REGISTER_START;
              iIndex <= SystemDeclarations.I_SINGLE_REGISTER_END; iIndex++ ) {
              jcbComboBox.addItem( 
                SystemDeclarations.st1dDataObjectNames[ iIndex ] );                
        }                    
        
        
        //System.out.println( "Elements in combo box = " + jcbComboBox.getItemCount() );
        if ( jcbComboBox.getItemCount() > 0 ) {
            jcbComboBox.setSelectedIndex( 0 );
        }
        else {
            jrbDataObject.setEnabled( false );
            jcbComboBox.setEnabled( false );
        }        
        return jcbComboBox;
    }
    
    private void closeThisWindow() {
        this.hide();
        this.dispose();
    }
    
    // instance variables
    private JFrame jfParentFrame;
    private JLabel jlArrayIndexValue;
    private JSlider jsArraySlider;
    private JButton jbOK, jbCancel, jbHelp;
    private JRadioButton jrbNumber, jrbDataObject;
    private Font fontLabel, fontMenuItem;
    private PTVHelpDialog ptvhdHelp;
    private Object objReturnObject;
    private SortableDeclarationsTableModel sdtmDataObjects;
    private ButtonGroup bgRadioButtonGroup;
    private JComboBox jcbAvailableDataObjects;
    private JPanel jpDialogPanel;
    
    // class constants
    private static final boolean B_MODAL = true;
    
    public static final int I_OK = 0;
    public static final int I_CANCELLED_OPERATION = -1;
    public static final int I_ILLEGAL_VALUE = -2;
    public static final int I_INTERNAL_ERROR = -3;
}