/*
 * ArrayRangeDialog.java
 *
 * Created on June 9, 2002, 8:23 PM
 */

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.*;

/**
 *
 * @author  _
 * @version 
 */
public class ArrayRangeDialog extends JDialog
    implements ActionListener, ChangeListener {

    // Because a reference is passed, all methods on this operand
    // will affect the original object...so no need to return the changed object.
    public ArrayRangeDialog( JFrame jfFrame, Operand operThisOperand,
        int iMaxLength ) {
            
        super( jfFrame, "Array Range Selector: " + operThisOperand.getOperandName(),
            B_MODAL );
                        
        GridBagLayout gblLayoutGrid;
        Container cPane;
        int iStart, iEnd;
        
        gblLayoutGrid = new GridBagLayout();
        cPane = this.getContentPane();
        jpPanel = new JPanel();       
        cPane.add( jpPanel );
        jpPanel.setLayout( gblLayoutGrid );
        
        iMaximumAllowedLength = iMaxLength;
        
        fontLabel = new Font( "SansSerif", Font.BOLD, 14 );
        
        jbOK = createButton( "OK" );
        jbCancel = createButton( "Cancel" );
        jbHelp = createButton( "Help" );
        
        iStart = operThisOperand.getStartingArrayIndex();
        jlStartingArrayIndexLabel = createLabel( "Starting Array Index:" );
        jlStartingArrayIndexValue = createLabel( Integer.toString( iStart ) );
        
        iEnd = operThisOperand.getEndingArrayIndex();
        jlEndingArrayIndexLabel = createLabel( "Ending Array Index:" );
        jlEndingArrayIndexValue = createLabel( Integer.toString( iEnd ) );
        
        jlArrayRangeLabel = createLabel( "Array Range Size:" );
        jlArrayRangeLabel.setHorizontalAlignment( SwingConstants.RIGHT );
        jlArrayRangeValue = createLabel(
            Integer.toString( operThisOperand.getArrayLength() ) );
        jlArrayRangeValue.setHorizontalAlignment( SwingConstants.LEFT );
        
        jsStartingArraySlider = createSlider( operThisOperand );
        jsEndingArraySlider = createSlider( operThisOperand );
        jsEndingArraySlider.setValue( iEnd );
        jsStartingArraySlider.setValue( iStart );
        
        ptvhdHelp = new PTVHelpDialog( jfFrame, "Help >> Panel >> Move Characters" );
        ptvhdHelp.setHelpFile( "help_panel_move_characters.html" );
                
        this.setSize( 419, 245 );
        createPanel();
        
        this.addWindowListener( new WindowAdapter() {
            public void windowClosing( WindowEvent weEvent ) {
                hide();
                dispose();
                objReturnObject = new Status( I_CANCELLED_OPERATION, 
                    "Operation cancelled." );
            }
        } );                
    }

    public Object displayDialog() {
        this.show();  // does not return until OK or Cancel
        return objReturnObject;
    }
       
    //----------------- ActionListener Implementation -----------\\
    public void actionPerformed( ActionEvent aeEvent ) {
        Object objThisObject;
        Vector vReturnIndices;
        
        objThisObject = aeEvent.getSource();
        
        if ( objThisObject == jbOK ) {
            try {
                vReturnIndices = new Vector( 2 );
                vReturnIndices.add( new Integer( jlStartingArrayIndexValue.getText() ) );
                vReturnIndices.add( new Integer( jlEndingArrayIndexValue.getText() ) );
                objReturnObject = vReturnIndices;
                this.hide();
                this.dispose();            
            }
            catch ( NumberFormatException nfeException ) {
                objReturnObject = new Status( I_ILLEGAL_VALUE, "Invalid array index." );                
            }
        }
        else if ( objThisObject == jbCancel ) {
            objReturnObject = new Status( I_CANCELLED_OPERATION, "Operation cancelled." );
            this.hide();
            this.dispose();            
        }
        else if ( objThisObject == jbHelp ) {
            ptvhdHelp.show();            
        }
    }
    
    //----------------- ChangeListener Implementation -----------\\
    public void stateChanged( ChangeEvent ceEvent ) {
        Object objThisObject;
        int iStartingIndex, iEndingIndex, iRange;
                
        objThisObject = ceEvent.getSource();
        
        if ( objThisObject instanceof JSlider ) {
            iStartingIndex = jsStartingArraySlider.getValue();
            iEndingIndex = jsEndingArraySlider.getValue();
            iArrayRangeValue = iEndingIndex - iStartingIndex + 1;
            
            if ( iStartingIndex > iEndingIndex ) {
                jsEndingArraySlider.setValue( iStartingIndex );
            }
            
            if ( iEndingIndex < iStartingIndex ) {
                jsStartingArraySlider.setValue( iEndingIndex );
            }

            jlArrayRangeValue.setText( Integer.toString( iArrayRangeValue ) );
        }
        
        if ( objThisObject == jsStartingArraySlider ) {
            jlStartingArrayIndexValue.setText( 
                Integer.toString( jsStartingArraySlider.getValue() ) );
        }
        else if ( objThisObject == jsEndingArraySlider ) {
            jlEndingArrayIndexValue.setText( 
                Integer.toString( jsEndingArraySlider.getValue() ) );        
        }        
    }
    
    /////////////////// PRIVATE METHODS \\\\\\\\\\\\\\\\\\\\\\\\\
    private void createPanel() {
        addComponentToPanel( 0, 0, 2, 1, jlStartingArrayIndexLabel );
        addComponentToPanel( 2, 0, 3, 2, jsStartingArraySlider );        
        addComponentToPanel( 0, 1, 1, 1, jlStartingArrayIndexValue );
        
        addComponentToPanel( 0, 2, 5, 1, new JLabel( "                   " ) );      
                
        addComponentToPanel( 0, 3, 2, 1, jlEndingArrayIndexLabel );
        addComponentToPanel( 2, 3, 3, 2, jsEndingArraySlider );
        addComponentToPanel( 0, 4, 1, 1, jlEndingArrayIndexValue );
        
        addComponentToPanel( 0, 5, 7, 1, new JLabel( "                   " ) );      
        
        addComponentToPanel( 0, 6, 1, 1, new JLabel( "     " ) );                
        addComponentToPanel( 1, 6, 2, 1, jlArrayRangeLabel );
        addComponentToPanel( 3, 6, 1, 1, jlArrayRangeValue );
        addComponentToPanel( 4, 6, 2, 1, new JLabel( "     " ) );
        
        addComponentToPanel( 0, 7, 5, 1, new JLabel( "                   " ) );        
        
        addComponentToPanel( 0, 8, 1, 1, new JLabel( "     " ) );
        addComponentToPanel( 1, 8, 1, 1, jbOK );
        addComponentToPanel( 2, 8, 1, 1, jbCancel );
        addComponentToPanel( 3, 8, 1, 1, jbHelp );
        addComponentToPanel( 4, 8, 1, 1, new JLabel( "     " ) );
    }
    
    private void addComponentToPanel( int iGridXPos, int iGridYPos, int iGridWidth,
                                      int iGridHeight, Component coComponent ) {
        
        GridBagConstraints gbcLayoutConstraints;

        gbcLayoutConstraints = new GridBagConstraints();
        gbcLayoutConstraints.weightx = 100;
        gbcLayoutConstraints.weighty = 100;
        gbcLayoutConstraints.gridx = iGridXPos;
        gbcLayoutConstraints.gridy = iGridYPos;
        gbcLayoutConstraints.gridwidth = iGridWidth;
        gbcLayoutConstraints.gridheight = iGridHeight;
        gbcLayoutConstraints.anchor = GridBagConstraints.SOUTHWEST;
        gbcLayoutConstraints.insets = new Insets( 2, 5, 2, 3 );
        gbcLayoutConstraints.fill =  GridBagConstraints.HORIZONTAL;
        jpPanel.add( coComponent, gbcLayoutConstraints );
    }

    private JButton createButton( String stText ) {
        JButton jbThisButton;
        
        jbThisButton = new JButton( stText );
        jbThisButton.addActionListener( this );
        return jbThisButton;
    }
    
    private JLabel createLabel( String stText ) {
        JLabel jlThisLabel;
        
        jlThisLabel = new JLabel( stText );
        jlThisLabel.setFont( fontLabel );
        jlThisLabel.setForeground( Color.black );
        return jlThisLabel;
    }
    
    private JSlider createSlider( Operand operThisOperand ) {
        JSlider jsThisSlider;
        int iMaximum, iMinorTick, iMajorTick;
        
        iMaximum = operThisOperand.getOperandArrayLength() - 1;
        jsThisSlider = new JSlider( SwingConstants.HORIZONTAL, 0, iMaximum, 0 );
        jsThisSlider.setFont( fontLabel );
        jsThisSlider.setForeground( Color.black );
        jsThisSlider.setPaintTicks( true );
        jsThisSlider.setPaintLabels( true );
        jsThisSlider.addChangeListener( this );

        iMajorTick = iMaximum / 10;
        iMinorTick = iMaximum / 30;
        
        if ( iMajorTick < 1 ) {
            iMinorTick = 0;
            iMajorTick = 1;
        }
        else if ( iMinorTick < 1 ) {
            iMinorTick = 1;
        }
                
        jsThisSlider.setMajorTickSpacing( iMajorTick );
        jsThisSlider.setMinorTickSpacing( iMinorTick );
        return jsThisSlider;
    }

    
    // instance variables
    private JLabel jlStartingArrayIndexLabel, jlStartingArrayIndexValue,
        jlEndingArrayIndexLabel, jlEndingArrayIndexValue, jlArrayRangeLabel, 
        jlArrayRangeValue;
    private JPanel jpPanel;
    private JSlider jsStartingArraySlider, jsEndingArraySlider;
    private JButton jbOK, jbCancel, jbHelp;
    private Font fontLabel;
    private PTVHelpDialog ptvhdHelp;
    private Object objReturnObject;
    private int iArrayRangeValue, iMaximumAllowedLength;
    
    // class constants
    private static final boolean B_MODAL = true;
    private static final int I_UNDEFINED = 0;
    
    public static final int I_NO_LIMIT = -1;
 
    public static final int I_INDEX_STARTING = 0;
    public static final int I_INDEX_ENDING = 1;
    
    public static final int I_OK = 0;
    public static final int I_CANCELLED_OPERATION = -1;
    public static final int I_ILLEGAL_VALUE = -2;
}