/*
 * AssemblyProgramView.java
 *
 * Created on April 28, 2002, 2:40 PM
 */

import java.io.*;
import java.lang.*;

/**
 *
 * @author  _
 * @version 
 */

public class AssemblyProgramView {

    public static Status translateCodeToAssemblyView( String stFilepath, 
        Program prgThisProgram ) {
            
        Instruction instThisInstruction;
        int iOpcode, iType, iRestriction, iSize, iCodeLineCount, iDeclarationCount,
            iThisLineNumber, iLeftOperandType, iRightOperandType, iStructure, 
            iLeftCharacterArrayLength, iRightCharacterArrayLength, iPreviousOpcode,
            iLeftOperandStartingIndex, iLeftOperandEndingIndex, iRightOperandStartingIndex,
            iRightOperandEndingIndex, iThisDeclarationNumber, iValue, iLeftStructure,
            iRightStructure, iIndex, iTempLength, iLeftArrayElementSetting,
            iRightArrayElementSetting;
        Code codeThisCode;
        char cValue;
        Declarations deDeclarations;
        DataObject doTemp;
        Operand operLeft, operRight, operLabel, operPreviousRight, operPreviousLeft;
        String stLabel, stOpcode, stLeftOperand, stRightOperand, stComments,
            stTempLeftOperand, stBranchMask, stLeftOperandStartingIndex, 
            stRightOperandStartingIndex, stTemp, stValue, stBranchOperand, stMessage, 
            stProgramName, stOperandLabel, stShiftOne, stExtraComments;
        NameAndDataObjectWrapper nadowLabel, nadowDeclaration;
        Opcode opOpcode;
        Object objThisObject;
        boolean bLeftOperandIsRegister, bRightOperandIsRegister, b64BitOperation,
            bValid, bNewComparison, bUsesBranchCompareOpcode, 
            bLeftOperandArrayAccessForBranchCompare, bRightOperandArrayAccessForBranchCompare;
        Status statInformation;
        Character charThisChar;
        Integer intThisInt;
        char chThisChar;

        statInformation = new Status( I_OK );
        stComments = "";
        stTemp = "";
        stLeftOperand = "";
        stRightOperand = "";
        
        try {
            fwOutputFile = new FileWriter( stFilepath );
        }
        catch ( IOException ioeException ) {
            stMessage = "The specified file " + stFilepath + " does not exist.";
            return new Status( I_INVALID_FILE, stMessage );
        }
        pwOutputStream = new PrintWriter( fwOutputFile, B_FLUSH );
        
        stProgramName = prgThisProgram.getProgramName();
        deDeclarations = prgThisProgram.getDeclarations();
        codeThisCode = prgThisProgram.getCode();
        iCodeLineCount = codeThisCode.getNumberOfLines() - 1;  // empty Instruction at end
        operPreviousLeft = new Operand();
        operPreviousRight = new Operand();
        iPreviousOpcode = Opcode.I_OPCODE__UNDEFINED;
        
        //*** Assume that proper validity checking was done for each object  ***\\
        //*** ====== ==== ====== ======== ======== === ==== === ==== ======  ***\\

        //------------ Generate Assembly Language Instructions ----------------\\
        // required opening lines
        createAssemblyLineInstruction( B_OPCODE, stProgramName, "CSECT", 
            ST_NO_LEFT_OPERAND, ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND, 
            ST_NO_COMMENTS );
        createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL, "USING", 
            stProgramName, ",", 
            SystemDeclarations.st1dDataObjectNames[ SystemDeclarations.I_DO_NAME_SREG15 ],
            ST_NO_COMMENTS );
        bValid = true;
        
        for ( iThisLineNumber = 0; 
              ( iThisLineNumber < iCodeLineCount )  &&  ( bValid == true ); 
              iThisLineNumber++ ) {
            instThisInstruction = 
                (Instruction)codeThisCode.getSingleLine( iThisLineNumber );
            opOpcode = instThisInstruction.getOpcode();
            iOpcode = opOpcode.getOpcode();
            stComments = instThisInstruction.getComments();
            b64BitOperation = false;
            nadowLabel = instThisInstruction.getLabel();
            stLabel = nadowLabel.getName();
            bValid = instThisInstruction.isValidInstruction();
            bLeftOperandArrayAccessForBranchCompare = false;
            bRightOperandArrayAccessForBranchCompare = false;
            
            // actual instruction marked like this /*X*/
            if ( ( iOpcode == Opcode.I_OPCODE__COMMENTS )  && 
                 ( bValid == true ) ) {                
/*X*/            createAssemblyLineInstruction( B_COMMENT_LINE, stLabel, ST_NO_OPCODE, 
                    ST_NO_LEFT_OPERAND, ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND,
                    stComments );
            }
            else if ( bValid == true ) {
                operLeft = instThisInstruction.getLeftOperand();
                operRight = instThisInstruction.getRightOperand();
                iLeftOperandType = operLeft.getOperandType();
                iRightOperandType = operRight.getOperandType();
                iLeftStructure = operLeft.getOperandStructure();
                iRightStructure = operRight.getOperandStructure();
                iRightArrayElementSetting = operRight.getArrayElementIndexSetting();
                iLeftArrayElementSetting = operLeft.getArrayElementIndexSetting();

                if ( ( iOpcode >= Opcode.I_OPCODE__BRANCH_FIRST )  &&
                     ( iOpcode <= Opcode.I_OPCODE__BRANCH_LAST ) ) {
                    bUsesBranchCompareOpcode = true;
                    System.out.println( "Uses compare/branch" );
                }
                else {
                    bUsesBranchCompareOpcode = false;
                    System.out.println( "--DOES NOT USE--- compare/branch" );
                }
                
                if ( isCharacterType( iRightOperandType ) == true ) {
                    if ( iRightOperandType == DataObject.I_SYSTEM_TYPE_KEYBOARD ) {
                        createAssemblyLineInstruction( B_OPCODE, stLabel,
                            "XREAD", "KEYBOARD", ",", "80", ST_NO_COMMENTS );
                        stLabel = ST_NO_LABEL;
                    }  // else if single character or character array
                }
                
                
                if ( operRight.getOperandRestriction() == 
                     DataObject.I_RESTRICTION_LITERAL ) {
                    System.out.println( "translateCodeToAssemblyView: Right operand is a literal." );    
                    objThisObject = getLiteralRepresentation( operRight );
                    if ( objThisObject instanceof String ) {
                        stTemp = (String)objThisObject;
                        stRightOperand = adjustConstantForSpecialCharacters( 
                            stTemp, B_CHANGE_STRING );
                    }
                    else {
                        statInformation = (Status)objThisObject;
                        bValid = false;  // end loop
                    }
                }
                else if ( iRightArrayElementSetting == Operand.I_UNDEFINED ) {
                    stRightOperand = getThisOperandName( operRight );
                }
                else if ( iOpcode == Opcode.I_OPCODE__MOVE_PARTS ) {
                    stRightOperand = getThisOperandName( operRight );
                }
                else {  // PROHIBIT ELEMENT SELECTION FROM DISPLAY AND KEYBOARD
                    if ( ( iRightArrayElementSetting == 
                           Operand.I_ARRAY_INDEX_DATA_OBJECT )  || 
                         ( iRightArrayElementSetting == 
                           Operand.I_ARRAY_INDEX_REGISTER ) ) {
                        if ( bUsesBranchCompareOpcode == false ) {
                            stRightOperand = createArrayElmentAccessInstructions( 
                                operRight, iRightOperandType, B_USE_REGISTER10,
                                stLabel );
                            stLabel = ST_NO_LABEL;
                        }
                        else {
                            bRightOperandArrayAccessForBranchCompare = true;
                            System.out.println( "RIGHT OPERAND uses array index." );
                        }
                    }
                    else if ( iRightArrayElementSetting == Operand.I_ARRAY_INDEX_NUMBER ) {                        
                        iSize = operRight.getOperandTypeSize();
                        stRightOperand = getThisOperandName( operRight ) + "+" + 
                            operRight.getArrayElementIndex() * iSize;
                    }
                }
                
                if ( operLeft.getOperandRestriction() == DataObject.I_RESTRICTION_LITERAL ) {
                    System.out.println( "translateCodeToAssemblyView: Left operand is a literal." );
                    objThisObject = getLiteralRepresentation( operLeft );
                    if ( objThisObject instanceof String ) {
                        stTemp = (String)objThisObject;
                        stLeftOperand = adjustConstantForSpecialCharacters( 
                            stTemp, B_CHANGE_STRING );
                    }
                    else {
                        statInformation = (Status)objThisObject;
                        bValid = false;  // end loop
                    }
                }
                else if ( iLeftArrayElementSetting == Operand.I_UNDEFINED ) {
                    stLeftOperand = getThisOperandName( operLeft );
                }
                else if ( iOpcode == Opcode.I_OPCODE__MOVE_PARTS ) {
                    stLeftOperand = getThisOperandName( operLeft );
                }
                else {  // PROHIBIT ELEMENT SELECTION FROM DISPLAY AND KEYBOARD
                    if ( ( iLeftArrayElementSetting == 
                           Operand.I_ARRAY_INDEX_DATA_OBJECT )  || 
                         ( iLeftArrayElementSetting == 
                           Operand.I_ARRAY_INDEX_REGISTER ) ) {                    
                        if ( bUsesBranchCompareOpcode == false ) {
                            stLeftOperand = createArrayElmentAccessInstructions( 
                                operLeft, iLeftOperandType, B_USE_REGISTER09,
                                stLabel );
                            stLabel = ST_NO_LABEL;
                        }
                        else {
                            bLeftOperandArrayAccessForBranchCompare = true;
                            System.out.println( "LEFT OPERAND uses array index." );
                        }
                    }
                    else if ( iLeftArrayElementSetting == Operand.I_ARRAY_INDEX_NUMBER ) {                        
                        iSize = operLeft.getOperandTypeSize();
                        stLeftOperand = getThisOperandName( operLeft ) + "+" + 
                            operLeft.getArrayElementIndex() * iSize;
                    }
                }

                System.out.println( "Line = " + iThisLineNumber + ":  LEFT = "
                    + stLeftOperand + "     RIGHT = " + stRightOperand );
                
                // If both operands are not registers, then load a number into
                // the register from memory, do the math operation, and store
                // the result into memory.  It will be the user's responsibility 
                // to make sure that no "register pairs" are used for math operations.

                // ONLY 32-BIT NUMBERS WILL BE USED
                if ( ( iLeftOperandType == DataObject.I_SYSTEM_TYPE_SINGLE_REGISTER )  ||
                     ( iLeftOperandType == DataObject.I_SYSTEM_TYPE_DOUBLE_REGISTER ) ) {
                    bLeftOperandIsRegister = true;
                }
                else {
                    bLeftOperandIsRegister = false;
                }

                if ( ( iRightOperandType == DataObject.I_SYSTEM_TYPE_SINGLE_REGISTER )  ||
                     ( iRightOperandType == DataObject.I_SYSTEM_TYPE_DOUBLE_REGISTER ) ) {
                    bRightOperandIsRegister = true;
                }
                else {
                    bRightOperandIsRegister = false;
                }
                                
        	//---------------------- Math Opcodes -----------------------\\
                if ( ( iOpcode >= Opcode.I_OPCODE__MATH_FIRST )  &&
                     ( iOpcode <= Opcode.I_OPCODE__MATH_LAST ) ) {
                    // left is destination, right is source
                    // left can only be a register
                    if ( bLeftOperandIsRegister == false ) {  // no register
                        stTempLeftOperand = SystemDeclarations.st1dDataObjectNames[ 
                            SystemDeclarations.I_DO_NAME_SREG11 ];
                        createAssemblyLineInstruction( B_OPCODE, 
                            stLabel, "L", stTempLeftOperand, ",", 
                            stLeftOperand, ST_NO_COMMENTS );
                        stLabel = ST_NO_LABEL;
                    }
                    else {  // is register
                        stTempLeftOperand = stLeftOperand;
                    }

                    switch ( iOpcode ) {
                        case Opcode.I_OPCODE__MATH_ADD:
                            if ( bRightOperandIsRegister == true ) {
                                stOpcode = "AR";
                            }
                            else {
                                stOpcode = "A";
                            }
                            break;

                        case Opcode.I_OPCODE__MATH_SUBTRACT:
                            if ( bRightOperandIsRegister == true ) {
                                stOpcode = "SR";
                            }
                            else {
                                stOpcode = "S";
                            }
                            break;

                        case Opcode.I_OPCODE__MATH_MULTIPLY:
                            if ( bRightOperandIsRegister == true ) {
                                stOpcode = "MR";
                            }
                            else {
                                stOpcode = "M";
                            }
                            b64BitOperation = true;
                            break;

                        case Opcode.I_OPCODE__MATH_DIVIDE:
                        case Opcode.I_OPCODE__MATH_MODULUS:
                            if ( bRightOperandIsRegister == true ) {
                                stOpcode = "DR";
                            }
                            else {
                                stOpcode = "D";
                            }
                            b64BitOperation = true;
                            break;
                            
                        default:
                            stOpcode = "";
                    }

                    // Since multiply and divide are 64-bit, move to another
                    // location and transfer results to original 32-bit register
                    if ( b64BitOperation == true ) {
                        createAssemblyLineInstruction( B_OPCODE, stLabel, 
                            "LR", SystemDeclarations.st1dDataObjectNames[ 
                            SystemDeclarations.I_DO_NAME_SREG13 ],
                            ",", stTempLeftOperand, "Move object to register " +
                            "pair for 64-bit operation" );
                        stLabel = ST_NO_LABEL;
                        
                        // setup register pair properly for division
                        if ( ( iOpcode == Opcode.I_OPCODE__MATH_DIVIDE )  ||
                             ( iOpcode == Opcode.I_OPCODE__MATH_MODULUS ) ) {
                            createAssemblyLineInstruction( B_OPCODE, stLabel,
                                "M", SystemDeclarations.st1dDataObjectNames[ 
                                SystemDeclarations.I_DO_NAME_SREG12 ], ",",
                                SystemDeclarations.st1dDataObjectNames[ 
                                SystemDeclarations.I_DO_NAME_ONE ], ST_NO_COMMENTS );
                        }
                        
/*X*/                   createAssemblyLineInstruction( B_OPCODE, stLabel,
/* M,MR,D,DR */             stOpcode, SystemDeclarations.st1dDataObjectNames[ 
                            SystemDeclarations.I_DO_NAME_SREG12 ],
                            ",", stRightOperand, stComments );
                        /*createAssemblyLineInstruction( B_COMMENT_LINE, ST_NO_LABEL, 
                            ST_NO_OPCODE, ST_NO_LEFT_OPERAND, ST_NO_OPERAND_SEPARATOR,
                            ST_NO_RIGHT_OPERAND, "END OF CONVERSION" );*/

                        if ( iOpcode == Opcode.I_OPCODE__MATH_MULTIPLY ) {
                            createAssemblyLineInstruction( B_OPCODE, stLabel,
                                "D", SystemDeclarations.st1dDataObjectNames[ 
                                SystemDeclarations.I_DO_NAME_SREG12 ],
                                ",", SystemDeclarations.st1dDataObjectNames[ 
                                SystemDeclarations.I_DO_NAME_ONE ], 
                                "Quotient in odd register" );
                            stLabel = ST_NO_LABEL;
                            createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL,
                                "LR", stTempLeftOperand, ",",
                                SystemDeclarations.st1dDataObjectNames[ 
                                SystemDeclarations.I_DO_NAME_SREG13 ],
                                ST_NO_COMMENTS );
                        }
                        else if ( iOpcode == Opcode.I_OPCODE__MATH_DIVIDE ) {
                            createAssemblyLineInstruction( B_OPCODE, stLabel,
                                "LR", stTempLeftOperand, ",", 
                                SystemDeclarations.st1dDataObjectNames[ 
                                SystemDeclarations.I_DO_NAME_SREG13 ],
                                "Quotient in odd register" );
                            stLabel = ST_NO_LABEL;
                        }
                        else if ( iOpcode == Opcode.I_OPCODE__MATH_MODULUS ) {
                            createAssemblyLineInstruction( B_OPCODE, stLabel,
                                "LR", stTempLeftOperand, ",",
                                SystemDeclarations.st1dDataObjectNames[ 
                                SystemDeclarations.I_DO_NAME_SREG12 ], 
                                "Modulus in even register" );
                            stLabel = ST_NO_LABEL;
                        }
                    }
                    else {
/*X*/                   createAssemblyLineInstruction( B_OPCODE, 
                            stLabel, stOpcode, stTempLeftOperand, 
                            ",", stRightOperand, stComments );
                    }

                    if ( bLeftOperandIsRegister == false ) {
                        createAssemblyLineInstruction( B_OPCODE, 
                            stLabel, "ST", stTempLeftOperand, ",",
                            stLeftOperand, ST_NO_COMMENTS );
                        stLabel = ST_NO_LABEL;
                    }
                } // end of math instruction
		//---------------- Compare/Branch Opcodes ---------------------\\
                else if ( ( iOpcode >= Opcode.I_OPCODE__BRANCH_FIRST )  &&
                          ( iOpcode <= Opcode.I_OPCODE__BRANCH_LAST ) ) {
                    if ( iOpcode != Opcode.I_OPCODE__BRANCH_ALWAYS ) {                        
                        if ( ( operPreviousLeft.isSameOperandFromLastComparison( 
                               operLeft ) == true )  &&
                             ( operPreviousRight.isSameOperandFromLastComparison(
                               operRight ) == true ) ) {
                             // do nothing
                        }                        
                        else {
                            if ( bLeftOperandArrayAccessForBranchCompare == true ) {
                                stLeftOperand = createArrayElmentAccessInstructions( 
                                    operLeft, iLeftOperandType, B_USE_REGISTER09, stLabel );
                                stLabel = ST_NO_LABEL;
                            }

                            if ( bRightOperandArrayAccessForBranchCompare == true ) {
                                stRightOperand = createArrayElmentAccessInstructions( 
                                    operRight, iRightOperandType, B_USE_REGISTER10, stLabel );
                                stLabel = ST_NO_LABEL;
                            }
                            
                    System.out.println( "LEFT = " + stLeftOperand + "   RIGHT = " + stRightOperand );                            
                                                                                    
			    // since operands are of the same type, 
		  	    // only check one operand's type
                            if ( isIntegerType( iLeftOperandType ) == true ) {
                                if ( iLeftOperandType == DataObject.I_TYPE_INTEGER ) {
                                    stTemp = SystemDeclarations.st1dDataObjectNames[ 
                                        SystemDeclarations.I_DO_NAME_SREG11 ];
                                    createAssemblyLineInstruction( B_OPCODE, stLabel,
                                        "L", stTemp, ",", stLeftOperand, ST_NO_COMMENTS );
                                    stLabel = ST_NO_LABEL;
                                }
                                else {
                                    stTemp = stLeftOperand;
                                }
                                
                                if ( iRightOperandType == DataObject.I_TYPE_INTEGER ) {
/*X*/				    createAssemblyLineInstruction( B_OPCODE, stLabel,
				        "C", stTemp, ",", stRightOperand, stComments );
                                    stComments = ST_NO_COMMENTS;
                                    bNewComparison = true;
                                    stLabel = ST_NO_LABEL;
                                }
				else {   // comparing two registers
/*X*/				    createAssemblyLineInstruction( B_OPCODE,
					stLabel, "CR", stTemp, ",",
					stRightOperand, stComments );
                                    stComments = ST_NO_COMMENTS;
                                    stLabel = ST_NO_LABEL;
				}
			    } // end of integer type check
                            else if ( isCharacterType( iLeftOperandType ) == true ) {
				// character in operand should have single quotes around it
				if ( ( iRightStructure == DataObject.I_STRUCTURE_SINGLE_VALUE )  &&
				     ( operRight.getOperandRestriction() == 
                                       DataObject.I_RESTRICTION_LITERAL ) ) {
                                    chThisChar = 
                                        ((String)operRight.getOperandValue()).charAt( 0 );
                                    if ( chThisChar == '\'' ) {
                                        stRightOperand = "C''''";
                                    }
                                    else if ( chThisChar == '&' ) {
                                        stRightOperand = "C'&&'";
                                    }
                                    else {
                                        stRightOperand = "C'" +  chThisChar + "'";
                                    }

/*X*/				    createAssemblyLineInstruction( B_OPCODE,
				        stLabel, "CLI", stLeftOperand, ",",
					stRightOperand, stComments );
                                    stComments = ST_NO_COMMENTS;
                                    stLabel = ST_NO_LABEL;
				}
				else {  // make sure the minimum character length is taken
                                    if ( ( iLeftArrayElementSetting ==
                                           Operand.I_UNDEFINED )  ||
                                         ( iRightArrayElementSetting ==
                                           Operand.I_UNDEFINED ) ) {
                                            
                                        iLeftCharacterArrayLength = 
                                            operLeft.getOperandArrayLength();
                                        iRightCharacterArrayLength = 
                                            operRight.getOperandArrayLength();                                            
                                            
                                        if ( iLeftCharacterArrayLength < 
                                             iRightCharacterArrayLength ) {
                                            iTempLength = iLeftCharacterArrayLength;
                                        }
                                        else {  // array element from character array
                                            iTempLength = iRightCharacterArrayLength;
                                        }
                                    }
                                    else {
                                        iTempLength = 1;
                                    }

                                    if ( ( iLeftArrayElementSetting == 
                                           Operand.I_ARRAY_INDEX_DATA_OBJECT )  ||
                                         ( iLeftArrayElementSetting == 
                                           Operand.I_ARRAY_INDEX_REGISTER ) ) {
                                        createAssemblyLineInstruction( B_OPCODE, 
/*X*/                                       stLabel, "CLC", stLeftOperand.substring( 0, 2 )
                                            + iTempLength + ","
                                            + stLeftOperand.substring( 2, 9 ), 
                                            ",", stRightOperand, stComments );
                                        stComments = ST_NO_COMMENTS;
                                        stLabel = ST_NO_LABEL;
                                    }
                                    else {
                                        createAssemblyLineInstruction( B_OPCODE, 
/*X*/                                       stLabel, "CLC", stLeftOperand + "(" 
                                            + iTempLength + ")", ",", 
                                            stRightOperand, stComments );
                                        stComments = ST_NO_COMMENTS;
                                        stLabel = ST_NO_LABEL;
                                    }                                        
				}  // end of character array check
                            }  // end of character type check
		            else {
			        stMessage = "Comparison requires operands of the "
                                    + "character or integer data type.  "
                                    + "No conversion is allowed.";
		                statInformation = new Status( 
                                    I_NO_TYPE_CONVERSION_ALLOWED, stMessage );
                                bValid = false;  // end loop
                            }  // no type available
                        } // end of different operands block                        
                        
                        operPreviousLeft = (Operand)operLeft.clone();
                        operPreviousRight = (Operand)operRight.clone();
                        iPreviousOpcode = iOpcode;
                    }  // end of all but BRANCH_ALWAYS                  

                    // branch code
                    operLabel = instThisInstruction.getBranchLabelOperand();
                    stOperandLabel = operLabel.getOperandName();

                    if ( stOperandLabel != null ) {                    
                        switch ( iOpcode ) {
                            case Opcode.I_OPCODE__BRANCH_LESS_THAN:
                                stExtraComments = "If less than, branch.  ";
                                stBranchMask = "0100";
                                break;

                            case Opcode.I_OPCODE__BRANCH_LESS_THAN_OR_EQUAL:
                                stExtraComments = "If less than or equal, branch.  ";
                                stBranchMask = "1100";
                                break;

                            case Opcode.I_OPCODE__BRANCH_EQUAL:
                                stExtraComments = "If equal, branch.  ";
                                stBranchMask = "1000";
                                break;

                            case Opcode.I_OPCODE__BRANCH_GREATER_THAN:
                                stExtraComments = "If greater than, branch.  ";
                                stBranchMask = "0010";
                                break;

                            case Opcode.I_OPCODE__BRANCH_GREATER_THAN_OR_EQUAL:
                                stExtraComments = "If greater than or equal, branch.  ";
                                stBranchMask = "1010";
                                break;

                            case Opcode.I_OPCODE__BRANCH_NOT_EQUAL:
                                stExtraComments = "If not equal, branch.  ";
                                stBranchMask = "0110";
                                break;

                            case Opcode.I_OPCODE__BRANCH_ALWAYS:
                                stExtraComments = "Always branch.  ";
                                stBranchMask = "1111";
                                break;
                                
                            default:
                                stExtraComments = "";
                                stBranchMask = "";
                        }

    /*X*/               createAssemblyLineInstruction( B_OPCODE, stLabel, "BC", 
                            "B'" + stBranchMask + "'", ",", stOperandLabel,
                            stExtraComments + stComments );
                    }
                    else {  // missing label instruction
                        stMessage = "Branch operation requires a label to "
                            + "jump to after a comparison.";
                        statInformation = new Status( I_MISSING_LABEL, stMessage );
                        bValid = false;  // end loop
                    }
                }// end of branch instruction
                //------------------- Move Parts Opcode -----------------------\\
                else if ( iOpcode == Opcode.I_OPCODE__MOVE_PARTS ) {
                    iLeftOperandStartingIndex = operLeft.getStartingArrayIndex();
                    iLeftOperandEndingIndex = operLeft.getEndingArrayIndex();
                    iRightOperandStartingIndex = operRight.getStartingArrayIndex();
                    iRightOperandEndingIndex = operRight.getEndingArrayIndex();
                    iLeftCharacterArrayLength = iLeftOperandEndingIndex - 
                        iLeftOperandStartingIndex + 1;


                    if ( ( operLeft.getArrayElementIndexSetting() == 
                             Operand.I_ARRAY_INDEX_RANGE )  ||
                         ( operRight.getArrayElementIndexSetting() == 
                             Operand.I_ARRAY_INDEX_RANGE ) ) {

                        if ( ( iLeftOperandEndingIndex - iLeftOperandStartingIndex ) ==
                             ( iRightOperandEndingIndex - iRightOperandStartingIndex ) ) {

//                            createAssemblyLineInstruction( B_OPCODE, stLabel,
//    /*X*/                      "MVC", stLeftOperand + "+" + iLeftOperandStartingIndex +
//                                "(" + Integer.toString( iLeftCharacterArrayLength ) + ")", 
//                                ",", stRightOperand + "+" + iRightOperandStartingIndex,
//                                stComments );
                            createAssemblyLineInstruction( B_OPCODE, stLabel,
    /*X*/                      "MVC", stLeftOperand + "(" + Integer.toString( 
                                iLeftCharacterArrayLength ) + ")", ",", 
                                stRightOperand, stComments );

                        }
                        else {
                            // not the same length
                            stMessage = "Move parts operation requires the number of "
                                + "characters moved between operands to be the same.";
                            statInformation = new Status( I_DIFFERING_ARRAY_LENGTHS,
                                stMessage );
                            bValid = false;  // end loop                        
                        }
                    }
                    else {  // array element from character array
                        createAssemblyLineInstruction( B_OPCODE, stLabel,
/*X*/                      "MVC", stLeftOperand + "(1)", ",", stRightOperand,
                            stComments );

//                        createAssemblyLineInstruction( B_OPCODE, stLabel,
///*X*/                      "MVC", stLeftOperand + "+" + iLeftOperandStartingIndex +
//                            "(1)", ",", stRightOperand + "+" + iRightOperandStartingIndex,
//                            stComments );
                    }                        
                }  // end of move parts instruction
                //------------------- Move All Opcode -----------------------\\
                else if ( iOpcode == Opcode.I_OPCODE__MOVE_ALL ) {
                        // source is RIGHT operand; destination is LEFT operand
                        // based on InstructionTableCellRenderer
                    
                    if ( isIntegerType( iRightOperandType ) == true ) {
                        if ( iRightOperandType == DataObject.I_TYPE_INTEGER ) {
                            if ( iLeftOperandType == 
                                DataObject.I_SYSTEM_TYPE_SINGLE_REGISTER ) {
/* X */                         createAssemblyLineInstruction( B_OPCODE, stLabel, "L",
                                    stLeftOperand, ",", stRightOperand, stComments );
                            }
                            else {
                                stTemp = SystemDeclarations.st1dDataObjectNames[ 
                                    SystemDeclarations.I_DO_NAME_SREG11 ];
                                createAssemblyLineInstruction( B_OPCODE, stLabel, "L", 
                                    stTemp, ",", stRightOperand, ST_NO_COMMENTS );
                                stLabel = ST_NO_LABEL;
                            }
                        }
                        else if ( iRightOperandType == 
                                  DataObject.I_SYSTEM_TYPE_SINGLE_REGISTER ) {
                            stTemp = stRightOperand;
                        }
                        
                        if ( ( ( iLeftStructure == DataObject.I_STRUCTURE_1D_ARRAY )  &&
                             ( iLeftOperandType == DataObject.I_TYPE_CHARACTER ) )  ||
                             ( iLeftOperandType == DataObject.I_SYSTEM_TYPE_DISPLAY ) ) {
                            if ( iLeftOperandType == DataObject.I_SYSTEM_TYPE_DISPLAY ) {
                                // blanking a text area before printing
                                createAssemblyLineInstruction( B_OPCODE, stLabel,
                                    "MVI", stLeftOperand, ",", "C' '", ST_NO_COMMENTS );
                                stLabel = ST_NO_LABEL;
                                createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL, 
                                    "MVC", stLeftOperand + "+1(" + String.valueOf( 
                                    operLeft.getOperandArrayLength() ) + ")", ",", 
                                    stLeftOperand, "Conversion from p.132 of Assembler "
                                        + "Lang. with Assist + Assist/I" );
                            }
                                
                            // +1 on left operand
                            if ( iLeftOperandType == DataObject.I_SYSTEM_TYPE_DISPLAY ) {
                                createAssemblyLineInstruction( B_OPCODE, stLabel, "XDECO",
                                    stTemp, ",", stLeftOperand + "+1", ST_NO_COMMENTS );
                            }
                            else {
    /*X*/                       createAssemblyLineInstruction( B_OPCODE, stLabel, "XDECO",
                                    stTemp, ",", stLeftOperand, stComments );
                            }
                        }
                    }
                    else if ( isCharacterType( iRightOperandType ) == true ) {
/*                        if ( iRightOperandType == DataObject.I_SYSTEM_TYPE_KEYBOARD ) {
                            createAssemblyLineInstruction( B_OPCODE, stLabel,
                                "XREAD", stRightOperand, ",", String.valueOf( 
                                operRight.getOperandArrayLength() ), ST_NO_COMMENTS );
                            stLabel = ST_NO_LABEL;
                        }  // else if single character or character array*/
                        
                        if ( iRightOperandType != DataObject.I_SYSTEM_TYPE_DISPLAY ) {
                            if ( isCharacterType( iLeftOperandType ) == true ) {
                                if ( ( iLeftArrayElementSetting == 
                                       Operand.I_UNDEFINED )  ||
                                     ( iRightArrayElementSetting == 
                                       Operand.I_UNDEFINED ) ) {
                                    iLeftCharacterArrayLength = 
                                        operLeft.getOperandArrayLength();
                                    iRightCharacterArrayLength = 
                                        operRight.getOperandArrayLength();
                                }
                                else {
                                    iLeftCharacterArrayLength = 1;
                                    iRightCharacterArrayLength = 1;
                                }
                                                                
                                // takes care of single character and character array;
                                // by default, single character array length is 1
                                
                                // +1 used
                                if ( iLeftOperandType == DataObject.I_SYSTEM_TYPE_DISPLAY ) {
                                    // blanking a text area before printing
                                    createAssemblyLineInstruction( B_OPCODE, stLabel,
                                        "MVI", stLeftOperand, ",", "C' '", ST_NO_COMMENTS );
                                    stLabel = ST_NO_LABEL;
                                    createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL, 
                                        "MVC", stLeftOperand + "+1(" + String.valueOf( 
                                        operLeft.getOperandArrayLength() ) + ")", ",", 
                                        stLeftOperand, "Conversion from p.132 of Assembler "
                                            + "Lang. with Assist + Assist/I" );
                                    
                                    if ( ( iRightArrayElementSetting != 
                                           Operand.I_UNDEFINED )  &&  
                                         ( operRight.getOperandType() ==
                                           DataObject.I_TYPE_CHARACTER ) &&
                                         ( operRight.getOperandStructure() ==
                                           DataObject.I_STRUCTURE_1D_ARRAY ) ) {
                                        iTempLength = 1;
                                    }
                                    else {
                                        iTempLength = iRightCharacterArrayLength;
                                    }
                                    
                                    createAssemblyLineInstruction( B_OPCODE, 
                                        ST_NO_LABEL, "MVC", stLeftOperand + "+1" 
                                        + "(" + iTempLength + ")", ",", 
                                        stRightOperand, ST_NO_COMMENTS );
                                }
                                else {
                                    if ( iRightCharacterArrayLength < 
                                         iLeftCharacterArrayLength ) {
                                        iTempLength = iRightCharacterArrayLength;
                                    }
                                    else {
                                        iTempLength = iLeftCharacterArrayLength;
                                    }

                                    if ( ( iLeftArrayElementSetting == 
                                           Operand.I_ARRAY_INDEX_DATA_OBJECT )  ||
                                         ( iLeftArrayElementSetting == 
                                           Operand.I_ARRAY_INDEX_REGISTER ) ) {
                                        createAssemblyLineInstruction( B_OPCODE, 
                                            stLabel, "MVC", stLeftOperand.substring( 0, 2 )
                                            + iTempLength + ","
                                            + stLeftOperand.substring( 2, 9 ), 
                                            ",", stRightOperand, stComments );
                                    }
                                    else {
                                        createAssemblyLineInstruction( B_OPCODE, 
                                            stLabel, "MVC", stLeftOperand + "(" 
                                            + iTempLength + ")", ",", 
                                            stRightOperand, stComments );
                                    }                                        
                                }
                            }
                        } // end of checking right operand                        
                    }  // end of checking right operand
                    
                    if ( isIntegerType( iLeftOperandType ) == true ) {
                        if ( isCharacterArray( iRightOperandType, iRightStructure )
                            == true ) {                            
                            if ( iLeftOperandType == 
                                 DataObject.I_SYSTEM_TYPE_SINGLE_REGISTER ) {
                                stTemp = stLeftOperand;
                            }
                            else if ( iLeftOperandType == DataObject.I_TYPE_INTEGER ) {
                                stTemp = SystemDeclarations.st1dDataObjectNames[ 
                                    SystemDeclarations.I_DO_NAME_SREG11 ];
                            }
                                
                            createAssemblyLineInstruction( B_OPCODE, stLabel,
    /*X*/                       "XDECI", stTemp, ",", stRightOperand, stComments );                            

                            if ( iLeftOperandType == DataObject.I_TYPE_INTEGER ) {
/* NEW */                       if ( isIntegerType( iRightOperandType ) == false ) {
                                    stComments = ST_NO_COMMENTS;
                                }
                                createAssemblyLineInstruction( B_OPCODE, stLabel,
                                    "ST", stTemp, ",", stLeftOperand, stComments );
                                stLabel = ST_NO_LABEL;
                            }                        
                        }
                        else {  // integer
                            if ( iLeftOperandType == DataObject.I_TYPE_INTEGER ) {
/* NEW */                       if ( isIntegerType( iRightOperandType ) == false ) {
                                    stComments = ST_NO_COMMENTS;
                                }
                                createAssemblyLineInstruction( B_OPCODE, stLabel,
                                    "ST", stTemp, ",", stLeftOperand, stComments );
                                stLabel = ST_NO_LABEL;
                            }
                            else if ( ( iRightOperandType != DataObject.I_TYPE_INTEGER )  && 
                                      ( iLeftOperandType == 
                                        DataObject.I_SYSTEM_TYPE_SINGLE_REGISTER ) ) {
    /*X*/                       createAssemblyLineInstruction( B_OPCODE, stLabel,
                                    "LR", stLeftOperand, ",", stTemp, stComments );
                            }
                        }
                    }
                    else if ( iLeftOperandType == DataObject.I_SYSTEM_TYPE_DISPLAY ) {
                        createAssemblyLineInstruction( B_OPCODE, stLabel, "XPRNT", 
                            stLeftOperand, ",", String.valueOf( 
                            operLeft.getOperandArrayLength() + 1 ), stComments );
                        stLabel = ST_NO_LABEL;
                    }
                }  // end of move all instruction
                else {
                    stMessage = "Opcode cannot be translated to assembler.";
                    statInformation = new Status( I_OPCODE_CANNOT_BE_TRANSLATED, stMessage );
                    bValid = false;  // end loop
                }  // unable to translate instruction
            } // non-comment-only line
            else {
                bValid = false;
                stMessage = "Illegal instruction.  Code generation stopped.";
                statInformation = new Status( I_ILLEGAL_INSTRUCTION, stMessage );
            }
        } // end of loop
        
        createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL, "XPRNT", 
            "=C'0Press the enter key to continue.'", ",", "32",
            ST_NO_COMMENTS );
        createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL, "XREAD", 
            "KEYBOARD", ",", "1", ST_NO_COMMENTS );
        
        // required instruction for exiting program properly
        createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL, "BCR", 
            "B'1111'", ",", SystemDeclarations.st1dDataObjectNames[ 
            SystemDeclarations.I_DO_NAME_SREG14 ],
            "Exit to operating system" );

        //-------------- System Declarations Generation -----------------\\
        if ( bValid == true ) {
            createAssemblyLineInstruction( B_COMMENT_LINE, ST_NO_LABEL, ST_NO_OPCODE, 
                ST_NO_LEFT_OPERAND, ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND,
                "" );
            createAssemblyLineInstruction( B_COMMENT_LINE, ST_NO_LABEL, ST_NO_OPCODE, 
                ST_NO_LEFT_OPERAND, ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND,
                "SYSTEM DECLARATIONS" );
            createAssemblyLineInstruction( B_COMMENT_LINE, ST_NO_LABEL, ST_NO_OPCODE, 
                ST_NO_LEFT_OPERAND, ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND,
                "" );
            
            createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL, "LTORG", 
                ST_NO_LEFT_OPERAND, ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND,
                ST_NO_COMMENTS );

            // predefined DataObjects, so exact contents are known, and error handling
            // will be minimal
            for ( iIndex = 0; ( iIndex < SystemDeclarations.I_SYSTEM_OBJECT_COUNT )  &&
                  ( bValid == true ); iIndex++ ) {
                doTemp = (DataObject)SystemDeclarations.getDeclaration( 
                    SystemDeclarations.st1dDataObjectNames[ iIndex ] );
                iType = doTemp.getType();
                stComments = doTemp.getDescription();

/*                if ( iType == DataObject.I_TYPE_CHARACTER ) {
                    createAssemblyLineInstruction( B_OPCODE,
                        SystemDeclarations.st1dDataObjectNames[ iIndex ], "DS", 
                        "CL" + String.valueOf( doTemp.getArrayDimensions().getXDimension() ),
                        ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND, stComments );
                }*/
                if ( iType == DataObject.I_TYPE_INTEGER ) {
                    objThisObject = doTemp.getValidatedValue( 0 );
                    
                    if ( objThisObject instanceof String ) {
                        createAssemblyLineInstruction( B_OPCODE,
                            SystemDeclarations.st1dDataObjectNames[ iIndex ], "DC", 
                            "F'" + (String)objThisObject + "'", ST_NO_OPERAND_SEPARATOR,
                            ST_NO_RIGHT_OPERAND, stComments );
                    }
                }
                else if ( ( iType == DataObject.I_SYSTEM_TYPE_SINGLE_REGISTER )  ||
                          ( iType == DataObject.I_SYSTEM_TYPE_DOUBLE_REGISTER ) ) {
                    stTemp = SystemDeclarations.st1dDataObjectNames[ iIndex ];
                    iSize = stTemp.length();

                    if ( stTemp.charAt( iSize - 2 ) == '0' ) {
                        stTemp = stTemp.substring( iSize - 1, iSize );
                    }
                    else {  // should be 1
                        stTemp = stTemp.substring( iSize - 2, iSize );
                    }                
                    createAssemblyLineInstruction( B_OPCODE, 
                        SystemDeclarations.st1dDataObjectNames[ iIndex ], "EQU",
                        stTemp, ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND, stComments );
                }                          
                else if ( iType == DataObject.I_SYSTEM_TYPE_DISPLAY ) {
                    createAssemblyLineInstruction( B_OPCODE,
                        SystemDeclarations.st1dDataObjectNames[ iIndex ], "DC", 
                        "C' '", ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND, 
                        "Single space all printed lines." );                
                    createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL, "DS", "CL"                         
                        + String.valueOf( doTemp.getArrayDimensions().getXDimension() ),
                        ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND, "Display buffer" );
                }
                else if ( iType == DataObject.I_SYSTEM_TYPE_KEYBOARD ) {
                    createAssemblyLineInstruction( B_OPCODE,
                        SystemDeclarations.st1dDataObjectNames[ iIndex ], "DS", 
                        "CL" + String.valueOf( doTemp.getArrayDimensions().getXDimension() ),
                        ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND, "Input buffer" );                
                }
                /*else if  ( iType == DataObject.I_SYSTEM_TYPE_DOUBLE_REGISTER ) {
                    stTemp = SystemDeclarations.st1dDataObjectNames[ iIndex ];
                    iSize = stTemp.length();

                    // always use even register as the equivalent for double registers
                    if ( stTemp.charAt( iSize - 4 ) == '0' ) {
                        stTemp = stTemp.substring( iSize - 3, iSize - 2 );
                    }
                    else {  // should be 1
                        stTemp = stTemp.substring( iSize - 4, iSize - 2 );
                    }                
                    createAssemblyLineInstruction( B_OPCODE, 
                        SystemDeclarations.st1dDataObjectNames[ iIndex ], "EQU",
                        stTemp, ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND, stComments );
                }*/
                else {
                    stMessage = "Expecting valid system object declarations.";
                    statInformation = new Status( I_INTERNAL_ERROR, stMessage );
                    bValid = false;
                }
            }  // end of system object declarations loop
        }

        //------------------ User Declarations Generation -----------------\\                
        // NOTE: ALL LITERALS ARE RETURNED AS STRINGS
        
        if ( bValid == true ) {
            createAssemblyLineInstruction( B_COMMENT_LINE, ST_NO_LABEL, ST_NO_OPCODE, 
                ST_NO_LEFT_OPERAND, ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND,
                "" );
            createAssemblyLineInstruction( B_COMMENT_LINE, ST_NO_LABEL, ST_NO_OPCODE, 
                ST_NO_LEFT_OPERAND, ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND,
                "USER DECLARATIONS" );
            createAssemblyLineInstruction( B_COMMENT_LINE, ST_NO_LABEL, ST_NO_OPCODE, 
                ST_NO_LEFT_OPERAND, ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND,
                "" );

            stLeftOperand = "";
            
            deDeclarations = prgThisProgram.getDeclarations();            
            iDeclarationCount = ((Integer)deDeclarations.getNumberOfObjects( 
                Declarations.I_ALL_OBJECTS, DataObject.ST_UNDEFINED )).intValue();                               
            
            for ( iIndex = 0; ( iIndex < iDeclarationCount )  &&  
                  ( bValid == true ); iIndex++ ) {
                objThisObject = deDeclarations.getFieldAtIndex( 
                    Declarations.I_ALL_OBJECTS, iIndex, DataObject.ST_UNDEFINED );
                
                if ( objThisObject instanceof NameAndDataObjectWrapper ) {
                    nadowDeclaration = (NameAndDataObjectWrapper)objThisObject;
                    stLabel = nadowDeclaration.getName();
                    doTemp = nadowDeclaration.getObject();
                    iType = doTemp.getType();
                    iRestriction = doTemp.getRestriction();
                    iStructure = doTemp.getStructure();
                    stComments = doTemp.getDescription();
                    
                    System.out.println( "object = " + stLabel );
                                                            
                    if ( iType == DataObject.I_TYPE_CHARACTER ) {
                        if ( iStructure == DataObject.I_STRUCTURE_SINGLE_VALUE ) {                            
                            if ( iRestriction == DataObject.I_RESTRICTION_CONSTANT ) {                                 
                                // p.47 for duplicates ('' and &&)
                                System.out.println( "Character, Single, Constant" );
                                stOpcode = "DC";
                                objThisObject = doTemp.getValidatedValue( 0 );
                                if ( objThisObject instanceof String ) {
                                    stTemp = (String)objThisObject;
                                    
                                    // not really verifying, but algorithm works
                                    // better this way
                                    stLeftOperand = 
                                        adjustConstantForSpecialCharacters( 
                                        stTemp, B_VERIFYING_LENGTH );
                                    stLeftOperand = "C'" +  stLeftOperand + "'";
                                    /*chThisChar = stTemp.charAt( 0 );
                                    
                                    if ( chThisChar == '\'' ) {
                                        stLeftOperand = "C''''";
                                    }
                                    else if ( chThisChar == '&' ) {
                                        stLeftOperand = "C'&&'";
                                    }
                                    else {
                                        stLeftOperand = "C'" +  chThisChar + "'";
                                    }*/
                                    createAssemblyLineInstruction( B_OPCODE, stLabel,
                                        stOpcode, stLeftOperand, ST_NO_OPERAND_SEPARATOR,
                                        ST_NO_RIGHT_OPERAND, stComments );                                    
                                }
                                else {  
                                    bValid = false;
                                    statInformation = (Status)objThisObject;
                                }
                            }
                            else if ( iRestriction == DataObject.I_RESTRICTION_VARIABLE ) {
                                System.out.println( "Character, Single, Variable" );
                                iSize = doTemp.getSize();
                                stLeftOperand = "CL" + iSize;
                                createDeclarationStorageLine( stLabel, stLeftOperand,
                                    stComments );
                            }                                
                            else {
                                stMessage = "Declarations must be a variable, constant, "
                                    + " or literal.";
                                statInformation = new Status( I_INVALID_DECLARATION, 
                                    stMessage );
                                bValid = false;
                            }                            
                        }  // end of character single object
                        else if ( iStructure == DataObject.I_STRUCTURE_1D_ARRAY ) {
                            if ( iRestriction == DataObject.I_RESTRICTION_CONSTANT ) {
                                System.out.println( "Character, Array, Constant" );
                                iSize = doTemp.getArrayDimensions().getXDimension();
                                
/*                                for ( iIndex = 0; ( iIndex < iSize )  && 
                                      ( bValid == true ); iIndex++ ) {*/
                                    objThisObject = doTemp.getValidatedValue( 0 );
                                    if ( objThisObject instanceof String ) {
                                        stTemp = (String)objThisObject;
                                        stLeftOperand = 
                                            adjustConstantForSpecialCharacters( 
                                            stTemp, B_CHANGE_STRING );

                                        /*chThisChar = stTemp.charAt( 0 );
                                        if ( chThisChar == '\'' ) {
                                            stLeftOperand = stLeftOperand + "''";
                                        }
                                        else if ( chThisChar == '&' ) {
                                            stLeftOperand = stLeftOperand + "&&";
                                        }
                                        else {
                                            stLeftOperand = stLeftOperand + chThisChar;
                                        }*/
                                    }
                                    else {  
                                        bValid = false;
                                        statInformation = (Status)objThisObject;
                                    }
                                /*}  // end of loop*/
                                
                                if ( bValid == true ) {
                                    stOpcode = "DC";
                                    stLeftOperand = "C'" + stLeftOperand + "'";
                                    /*stLeftOperand = "CL" + Integer.toString( iSize ) + "'"
                                        + stLeftOperand + "'";*/
                                    createAssemblyLineInstruction( B_OPCODE, stLabel,
                                        stOpcode, stLeftOperand, ST_NO_OPERAND_SEPARATOR,
                                        ST_NO_RIGHT_OPERAND, stComments );
                                }
                            }
                            else if ( iRestriction == DataObject.I_RESTRICTION_VARIABLE ) {
                                iSize = doTemp.getArrayDimensions().getXDimension();
                                stLeftOperand = "CL" + iSize;
                                createDeclarationStorageLine( stLabel, stLeftOperand,
                                    stComments );
                            }
                            else {
                                stMessage = "Declarations must be a variable or constant.";
                                statInformation = new Status( I_INVALID_DECLARATION, 
                                    stMessage );
                                bValid = false;
                            }
                        }  // end of 1-d character array
                        else {
                            stMessage = "An character declaration must have a single value "
                                + "or 1-D array structure.";
                            statInformation = new Status( I_INVALID_DECLARATION, stMessage );
                            bValid = false;
                        }
                    }  // end of character
                    else if ( iType == DataObject.I_TYPE_INTEGER ) {
                        if ( iStructure == DataObject.I_STRUCTURE_SINGLE_VALUE ) {
                            if ( iRestriction == DataObject.I_RESTRICTION_CONSTANT ) {
                                System.out.println( "Integer, Single, Constant" );
                                stOpcode = "DC";
                                objThisObject = doTemp.getValidatedValue( 0 );
                                System.out.println( "Object is " + objThisObject.getClass().toString() );
                                if ( objThisObject instanceof String ) {
                                    stTemp = (String)objThisObject;
                                    stLeftOperand = "F'" + stTemp + "'";
                                    createAssemblyLineInstruction( B_OPCODE, stLabel,
                                        stOpcode, stLeftOperand, ST_NO_OPERAND_SEPARATOR,
                                        ST_NO_RIGHT_OPERAND, stComments );
                                }
                                else {  
                                    bValid = false;
                                    statInformation = (Status)objThisObject;
                                }
                            }
                            else if ( iRestriction == DataObject.I_RESTRICTION_VARIABLE ) {
                                System.out.println( "Integer, Single, Variable" );                            
                                stLeftOperand = "F";
                                createDeclarationStorageLine( stLabel, stLeftOperand,
                                    stComments );
                            }                                
                            else {
                                stMessage = "Declarations must be a variable or constant.";
                                statInformation = new Status( I_INVALID_DECLARATION, 
                                    stMessage );
                                bValid = false;
                            }                            
                        }
                        else if ( iStructure == DataObject.I_STRUCTURE_1D_ARRAY ) {
                            if ( iRestriction == DataObject.I_RESTRICTION_CONSTANT ) {
                                System.out.println( "Integer, Array, Constant" );                                
                                stOpcode = "DC";                                
                                iSize = doTemp.getArrayDimensions().getXDimension();
                                stLeftOperand = iSize + "F'0'";
                                createAssemblyLineInstruction( B_OPCODE, stLabel,
                                    stOpcode, stLeftOperand, ST_NO_OPERAND_SEPARATOR,
                                    ST_NO_RIGHT_OPERAND, stComments );

                                stLeftOperand = stLabel;                                
                                createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL,
                                    "ORG", stLeftOperand, ST_NO_OPERAND_SEPARATOR, 
                                    ST_NO_RIGHT_OPERAND, ST_NO_COMMENTS );
                                

                                for ( int iIndex2 = 0; ( iIndex2 < iSize )  && 
                                      ( bValid == true ); iIndex2++ ) {
                                    objThisObject = doTemp.getValidatedValue( iIndex2 );
                                    if ( objThisObject instanceof String ) { 
                                        stTemp = (String)objThisObject;
                                        stLeftOperand = "F'" + stTemp + "'";
                                        createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL,
                                            stOpcode, stLeftOperand, ST_NO_OPERAND_SEPARATOR,
                                            ST_NO_RIGHT_OPERAND, stComments );                                        
                                    }
                                    else {  
                                        bValid = false;
                                        statInformation = (Status)objThisObject;
                                    }
                                }  // end of loop
                                
                                if ( bValid == true ) {
                                    createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL,
                                        "ORG", ST_NO_LEFT_OPERAND, ST_NO_OPERAND_SEPARATOR,
                                        ST_NO_RIGHT_OPERAND, ST_NO_COMMENTS );            
                                }
                            }
                            else if ( iRestriction == DataObject.I_RESTRICTION_VARIABLE ) {
                                System.out.println( "Integer, Array, Variable" );
                                iSize = doTemp.getArrayDimensions().getXDimension();
                                stLeftOperand = iSize + "F";
                                createDeclarationStorageLine( stLabel, stLeftOperand,
                                    stComments );
                            }
                            else {
                                stMessage = "Declarations must be a variable or constant.";
                                statInformation = new Status( I_INVALID_DECLARATION, 
                                    stMessage );
                                bValid = false;
                            }
                        }
                        else {
                            stMessage = "An integer declaration must have a single value "
                                + "or 1-D array structure.";
                            statInformation = new Status( I_INVALID_DECLARATION, stMessage );
                            bValid = false;
                        }
                    }  // end of integer
                    else if ( iType == DataObject.I_TYPE_LABEL ) {
                        // do nothing...just ignore...labels are not declared
                    }
                    else {
                        stMessage = "Declarations must be of type integer or character.";                        
                        statInformation = new Status( I_INVALID_DECLARATION, stMessage );
                        bValid = false;
                    }     
                }  // invalid declaration
                else {
                    statInformation = (Status)objThisObject;
                    bValid = false;
                }
            }
        }
        
        if ( bValid == true ) {
            // required closing line for code
            createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL, "END", stProgramName, 
                ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND, ST_NO_COMMENTS );
            pwOutputStream.println( "$ENTRY CONSOLE NOPROMPT" );
            statInformation = new Status( I_OK );
        }
        
//        try {
            pwOutputStream.close();
/*        }
        catch ( IOException ioeException ) {
            stMessage = "The specified file " + stFilepath + " cannot be closed.";
            return new Status( I_INVALID_FILE, stMessage );
        }        */
        return statInformation;
    }    
    
    public static String adjustConstantForSpecialCharacters( String stConstant,
        boolean bVerifyingLength ) {
            
        StringBuffer sbRevised;
        char chThisChar;
        int iIndex, iLength, iRevisedLength, iStart;

        iLength = stConstant.length();
        
        if ( bVerifyingLength == true ) {
            System.out.println( "* * *Verifying length..." );
            sbRevised = new StringBuffer();
            iRevisedLength = iLength;
            iStart = 0;
        }
        else {
            System.out.println( "= = = Converting string..." );
            sbRevised = new StringBuffer( stConstant.substring( 0, 3 ) );
            iRevisedLength = iLength - 1;
            iStart = 3;
        }
        
        // starts with =C' and ends with '        
        for ( iIndex = iStart; iIndex < iRevisedLength; iIndex++ ) {
            chThisChar = stConstant.charAt( iIndex );
            System.out.println( "Index = " + iIndex + ", char = " + chThisChar );
            if ( chThisChar == '\'' ) {
                sbRevised.append( "''" );
                System.out.println( "Doubling '" );
            }
            else if ( chThisChar == '&' ) {
                sbRevised.append( "&&" );
                System.out.println( "Doubling &" );
            }
            else {
                sbRevised.append( chThisChar );
                System.out.println( "Appending character" );
            }
        }
        
        if ( bVerifyingLength == false ) {
            sbRevised.append( stConstant.charAt( iLength - 1 ) );
        }
        
        System.out.println( "original string = " + stConstant );
        System.out.println( "revised string = " + sbRevised.toString() );
        System.out.println( "LEAVING adjustConstantForSpecialCharacters - - -" );

        return sbRevised.toString();        
    }        
    
    /////////////////// PRIVATE FUNCTIONS \\\\\\\\\\\\\\\\\\\\\\\\
    private static void createAssemblyLineInstruction( boolean bIsCommentLine,
        String stLabel, String stOpcode, String stLeftOperand, 
        String stOperandSeparator, String stRightOperand, String stComments ) {
        
        String stCurrentLine, stSeparator, stCommentPortion;
        int iExtraCommentLines, iFirstStringPosition, iLastStringPosition, iLine,
            iFirstLinePosition, iLastLinePosition, iOperandsSize, iCommentsLength,
            iRemainingLength, iShiftSpace, iCarraigeReturnLocation;
        StringBuffer sbEmptySpace;
                
        // create instruction
        sbEmptySpace = new StringBuffer( I_CHARACTERS_PER_LINE );
        iCommentsLength = stComments.length();
        System.out.println( "COMMENTS: " + stComments );
        System.out.println( "Length = " + iCommentsLength );
        
        if ( bIsCommentLine == true ) {
            // 69 + 1 for the "exclusive" replace end range
            iRemainingLength = 68;
            iFirstLinePosition = 2;
            stSeparator = "";
            stRightOperand = "* ";  // since this is a comment line, no right operand exists
            iOperandsSize = 0;
        }
        else {            
            // +1 for the space between comments and operands
            iOperandsSize = stLeftOperand.length() + stRightOperand.length()
                    + stOperandSeparator.length() + 1;
            
            // use "+1" for exclusive replace end range (?)
            iFirstLinePosition = iOperandsSize + I_LABEL_SIZE + I_OPCODE_SIZE + 1; 
            iRemainingLength = I_CHARACTERS_PER_LINE - iFirstLinePosition;
                        
            sbLabel.replace( 0, stLabel.length(), stLabel );
            sbOpcode.replace( 0, stOpcode.length(), stOpcode );
            stSeparator = " ";
        }

        System.out.println( "Remaining length = " + iRemainingLength );
        
        if ( iRemainingLength < 1 ) {
            // write comments first followed by instruction...using recursion
/*X*/       createAssemblyLineInstruction( B_COMMENT_LINE, ST_NO_LABEL, ST_NO_OPCODE, 
                ST_NO_LEFT_OPERAND, ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND,
                stComments );
            
            // comments do not need to be printed again
            stComments = ST_NO_COMMENTS;
            iCommentsLength = 0;
        }

        // make sure the correct area of stComments is printed
        
        if ( iCommentsLength > iRemainingLength ) {
            iLastStringPosition = iRemainingLength;
        }
        else {
            iLastStringPosition = iCommentsLength;
        }

        iCarraigeReturnLocation = stComments.indexOf( '\n' );        
        System.out.println( "Found carriage return at " + iCarraigeReturnLocation );        

        if ( ( iCarraigeReturnLocation != I_NOT_FOUND )  && 
             ( iCarraigeReturnLocation <= iLastStringPosition ) ) {
            System.out.println( "Old iLastStringPosition " + iLastStringPosition );
            iLastStringPosition = iCarraigeReturnLocation;
            System.out.println( "New iLastStringPosition " + iLastStringPosition );            
        }
        
        
        // output instruction
        if ( bIsCommentLine == true ) {
            stCurrentLine = stRightOperand + stComments.substring( 0, iLastStringPosition );
        }
        else {
            stCurrentLine = sbLabel.toString() + sbOpcode.toString() 
                + stLeftOperand + stOperandSeparator + stRightOperand
                + stSeparator + stSeparator + stComments.substring( 0, iLastStringPosition );
        }
        pwOutputStream.println( stCurrentLine );

        // allow for multi-line comments
        //iExtraCommentLines = ( iCommentsLength / iRemainingLength );
        //ExtraCommentLines = ( iCommentsLength % iRemainingLength );
        //iLastLinePosition = I_CHARACTERS_PER_LINE + 1;        
        
        if ( bIsCommentLine == true ) {
            iShiftSpace = 2;
        }
        else {
            iShiftSpace = iOperandsSize + I_LABEL_SIZE + I_OPCODE_SIZE + 1;
        }
                        
        while ( iCommentsLength > ( iLastStringPosition + 1 ) ) {
            sbEmptySpace.replace( 0, I_CHARACTERS_PER_LINE, ST_EMPTY_COMMENTS );
            iFirstStringPosition = iLastStringPosition;
            iLastStringPosition = iFirstStringPosition + iRemainingLength;
            
            if ( iCarraigeReturnLocation != I_NOT_FOUND ) {
                // skip over printing carriage return
                iFirstStringPosition++;
            }
            
            iCarraigeReturnLocation = stComments.indexOf( '\n', iFirstStringPosition );        
            System.out.println( "Found carriage return at " + iCarraigeReturnLocation );        
            if ( ( iCarraigeReturnLocation != I_NOT_FOUND )  && 
                 ( iCarraigeReturnLocation <= iLastStringPosition ) ) {
                System.out.println( "Old iLastStringPosition " + iLastStringPosition );
                iLastStringPosition = iCarraigeReturnLocation;
                System.out.println( "New iLastStringPosition " + iLastStringPosition );            
            }                        
            
            if ( iLastStringPosition > iCommentsLength ) {
                iLastStringPosition = iCommentsLength;
                iRemainingLength = iLastStringPosition - iFirstStringPosition + 1;
            }
                     
            System.out.println( "Remaining length..." + iRemainingLength );
            
            stCommentPortion = stComments.substring( iFirstStringPosition, 
                iLastStringPosition );
            sbEmptySpace.replace( iShiftSpace, iShiftSpace + iRemainingLength, 
                stCommentPortion );
            pwOutputStream.println( sbEmptySpace.toString() );
        }
        
        // clear buffers
        sbLabel.replace( 0, I_LABEL_SIZE, ST_EMPTY_LABEL );
        sbOpcode.replace( 0, I_OPCODE_SIZE, ST_EMPTY_OPCODE );
    }
        
    private static Object getLiteralRepresentation( Operand operLiteral ) {
        Object objThisObject, objReceivedObject;
        String stMessage, stValue;
        Status statInformation;
        int iType, iStructure;
        
        iType = operLiteral.getOperandType();
        iStructure = operLiteral.getOperandStructure();
        objReceivedObject = operLiteral.getOperandValue();

        if ( objReceivedObject instanceof Status ) {
            objThisObject = (Status)objReceivedObject;
        }
        else {
            stValue = (String)objReceivedObject;
            
            if ( iType == DataObject.I_TYPE_CHARACTER ) {
                objThisObject = new String( "=C'" + stValue + "'" );
            }
            else if ( iType == DataObject.I_TYPE_INTEGER ) {
                objThisObject = new String( "=F'" + stValue + "'" );            
            }
            else {
                stMessage = "No literal is available for the given type named "
                    + DataObjectPropertyTextView.getTypeName( iType );
                objThisObject = new Status( I_NO_LITERAL_AVAILABLE, stMessage );
            }
        }  // not likely to be anything else but these possibilities
        return objThisObject;        
    }
    
    public static boolean isCharacterType( int iOperandType ) {
        if ( ( iOperandType == DataObject.I_TYPE_CHARACTER )  ||
             ( iOperandType == DataObject.I_SYSTEM_TYPE_DISPLAY )  || 
             ( iOperandType == DataObject.I_SYSTEM_TYPE_KEYBOARD ) ) {
            return true;
        }
        else {
            return false;
        }
    }
    
    public static boolean isIntegerType( int iOperandType ) {
        if ( ( iOperandType == DataObject.I_TYPE_INTEGER )  ||
             ( iOperandType == DataObject.I_SYSTEM_TYPE_SINGLE_REGISTER )  || 
             ( iOperandType == DataObject.I_SYSTEM_TYPE_DOUBLE_REGISTER ) ) {
            return true;
        }
        else {
            return false;
        }
    }
    
    public static boolean isCharacterArray( int iType, int iStructure ) {
        if ( iStructure == DataObject.I_STRUCTURE_1D_ARRAY ) {
            if ( ( iType == DataObject.I_TYPE_CHARACTER )  ||
                 ( iType == DataObject.I_SYSTEM_TYPE_DISPLAY )  ||
                 ( iType == DataObject.I_SYSTEM_TYPE_KEYBOARD ) ) {
                return true;
            }
            else {
                return false;
            }
        }
        else {
            return false;
        }
    }
    
    private static void createDeclarationStorageLine( String stName, 
        String stOperand, String stComments ) {
        createAssemblyLineInstruction( B_OPCODE, stName, "DS", stOperand,
            ST_NO_OPERAND_SEPARATOR, ST_NO_RIGHT_OPERAND, stComments );                                
    }
    
    private static String getThisOperandName( Operand operThisOperand ) {
        int iStructure, iType, iStartingArrayIndex;
        String stShiftFactor, stName;
        DataObject doThisObject;
            
        stName = operThisOperand.getOperandName();
        iStructure = operThisOperand.getOperandStructure();
        iType = operThisOperand.getOperandType();
        iStartingArrayIndex = operThisOperand.getStartingArrayIndex();

        if ( ( iStructure == DataObject.I_STRUCTURE_1D_ARRAY )  &&
             ( iStartingArrayIndex != Operand.I_UNDEFINED ) ) {
            if ( iStartingArrayIndex == 0 ) {
               stShiftFactor = "";
            }
            else if ( iType == DataObject.I_TYPE_INTEGER ) {
                stShiftFactor = "+" + String.valueOf( 4 * iStartingArrayIndex );
            }
            else if ( ( iType == DataObject.I_TYPE_CHARACTER )  ||
                      ( iType == DataObject.I_SYSTEM_TYPE_KEYBOARD )  ||
                      ( iType == DataObject.I_SYSTEM_TYPE_DISPLAY ) ) {
                stShiftFactor = "+" + String.valueOf( iStartingArrayIndex );
            }
            else {
                stShiftFactor = "";
            }
        }
        else {
            stShiftFactor = "";
        }                
        return ( stName + stShiftFactor );
    }
        
    private static String createArrayElmentAccessInstructions( Operand operThisOperand,
        int iType, boolean bUseRegister10, String stThisLabel ) {
            
        String stObjectName, stArrayIndexObjectName, stSize, stRegister13,
            stRegister12, stBaseRegister;

        stObjectName = getThisOperandName( operThisOperand );
        stArrayIndexObjectName = operThisOperand.getArrayElementDataObject();
        stRegister12 = SystemDeclarations.st1dDataObjectNames[ 
            SystemDeclarations.I_DO_NAME_SREG12 ];
        stRegister13 = SystemDeclarations.st1dDataObjectNames[ 
            SystemDeclarations.I_DO_NAME_SREG13 ];
        
        if ( bUseRegister10 == true ) {
            stBaseRegister = SystemDeclarations.st1dDataObjectNames[ 
                SystemDeclarations.I_DO_NAME_SREG10 ];
        }
        else {
            stBaseRegister = SystemDeclarations.st1dDataObjectNames[ 
                SystemDeclarations.I_DO_NAME_SREG09 ];
        }

        createAssemblyLineInstruction( B_OPCODE, stThisLabel, "LA", 
            stBaseRegister, ",", stObjectName, "Load the address of data object" );

        if ( iType == DataObject.I_TYPE_CHARACTER ) {
            stSize = SystemDeclarations.st1dDataObjectNames[ 
                SystemDeclarations.I_DO_NAME_CHARSIZE ];
        }
        else if ( iType == DataObject.I_TYPE_INTEGER ) {
            stSize = SystemDeclarations.st1dDataObjectNames[ 
                SystemDeclarations.I_DO_NAME_INTSIZE ];
        }
        else {
            stSize = "";
        }

        createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL, "L", 
            stRegister13, ",", stSize, "Load the size of the data type" );

        if ( SystemDeclarations.isObjectDeclaredAsARegister( stArrayIndexObjectName ) 
             == true ) {
            createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL, "MR", 
                stRegister12, ",", stArrayIndexObjectName,
                "Calculate position based on type size and array index in register" );
        }
        else {
            createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL, "M", 
                stRegister12, ",", stArrayIndexObjectName,
                "Calculate position based on type size and array index number" );
        }

        createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL, "D", 
            stRegister12, ",", SystemDeclarations.st1dDataObjectNames[ 
            SystemDeclarations.I_DO_NAME_ONE ], "Quotient in odd register" );    
        createAssemblyLineInstruction( B_OPCODE, ST_NO_LABEL, "AR", 
            stBaseRegister, ",", SystemDeclarations.st1dDataObjectNames[ 
            SystemDeclarations.I_DO_NAME_SREG13 ],
            "Add the position to the base address" );

        return new String( "0(" + stBaseRegister + ")" );
    }
    
    
    // instance variables
    private static StringBuffer sbLabel, sbOpcode;
    private static FileWriter fwOutputFile;
    private static PrintWriter pwOutputStream;
    private static int iLiteralCharacterLength;  // as would appear on the screen
    
    // class constants
    private static final int I_LABEL_SIZE = 9;  // includes space
    private static final int I_OPCODE_SIZE = 6;  // includes space
    private static final int I_CHARACTERS_PER_LINE = 71;  // includes "* "
    
    private static final boolean B_FLUSH = true;
    
    private static final boolean B_COMMENT_LINE = true;
    private static final boolean B_OPCODE = false;
    
    private static final boolean B_USE_REGISTER10 = true;
    private static final boolean B_USE_REGISTER09 = false;
    
    public static final boolean B_VERIFYING_LENGTH = true;
    public static final boolean B_CHANGE_STRING = false;
    
    private static final String ST_EMPTY_LABEL = "         ";
    private static final String ST_EMPTY_OPCODE = "      ";
    private static final String ST_EMPTY_COMMENTS = 
        "*                                                                      ";
    private static final String ST_NO_LABEL = "";
    private static final String ST_NO_OPCODE = "";
    private static final String ST_NO_LEFT_OPERAND = "";
    private static final String ST_NO_OPERAND_SEPARATOR = "";
    private static final String ST_NO_RIGHT_OPERAND = "";
    private static final String ST_NO_COMMENTS = "";
    
    private static final int I_NOT_FOUND = -1;
        
        // error codes
        public static final int I_OK = 0;
        public static final int I_INVALID_FILE = -1;
        public static final int I_NO_TYPE_CONVERSION_AVAILABLE = -2;
        public static final int I_NO_TYPE_CONVERSION_ALLOWED = -3;
        public static final int I_NO_LITERAL_AVAILABLE = -4;
        public static final int I_OPCODE_CANNOT_BE_TRANSLATED = -5;
        public static final int I_DIFFERING_ARRAY_LENGTHS = -6;
        public static final int I_INTERNAL_ERROR = -7;
        public static final int I_INVALID_DECLARATION = -8;
        public static final int I_MISSING_LABEL = -9;
        public static final int I_ILLEGAL_INSTRUCTION = -10;
        
        private static final String ST_NO_RECORD = "";
        
    static {
        // p.32 of "Assembly Language..." book for column placement
        sbLabel = new StringBuffer( I_LABEL_SIZE );
        sbLabel.insert( 0, ST_EMPTY_LABEL );
        sbOpcode = new StringBuffer( I_OPCODE_SIZE );
        sbOpcode.insert( 0, ST_EMPTY_OPCODE );
        iLiteralCharacterLength = 0;
    }
}