/*
 * BranchAndComparePanel.java
 *
 * Created on June 13, 2002, 3:26 PM
 */

import java.awt.event.*;
import javax.swing.*;
import java.util.*;
import java.awt.*;

/**
 *
 * @author  _
 * @version 
 */
public class BranchAndComparePanel extends PTVPanel
    implements ActionListener, DataObjectAreaListener {

    public BranchAndComparePanel( JFrame jfParentFrame, String stDialogTitle, 
        Dimension diPanelSize, Program progThisProgram, 
        SortableDeclarationsTableModel sdtmModel ) {

        super( jfParentFrame, stDialogTitle, diPanelSize, progThisProgram, sdtmModel );

        iFirstClickedOperandID = I_UNDEFINED;
        super.setActionListener( this );
        super.setDataObjectAreaListener( this );
                
        createObjects();
        drawPanel();
        finishPanelConfiguration();        
    }

    public void setInstruction( Instruction instInstruction ) {
        int iOpcodeID;

        instThisInstruction = (Instruction)instInstruction.clone();        
        opBranchAndCompareOpcode = instThisInstruction.getOpcode();
        iOpcodeID = opBranchAndCompareOpcode.getOpcode();
        operLeft = instThisInstruction.getLeftOperand();
        operRight = instThisInstruction.getRightOperand();
        operResult = instThisInstruction.getBranchLabelOperand();

        // set up the panel appropriately
        dataObjectAreaChanged( iLeftOperandID, operLeft );
        super.updateDataObjectArea( iLeftOperandID, operLeft );
        dataObjectAreaChanged( iRightOperandID, operRight );
        super.updateDataObjectArea( iRightOperandID, operRight );
        dataObjectAreaChanged( iResultOperandID, operResult );
        super.updateDataObjectArea( iResultOperandID, operResult );
        jlLabelText.setText( instThisInstruction.getLabel().getName() );
        jtaComments.setText( instThisInstruction.getComments() );
        setCorrectButton( iOpcodeID );
    }    
    
    //------------------ ActionListener Implementation -----------------\\
    public void actionPerformed( ActionEvent aeEvent ) {
        Object objActionSource, objTemp;
        JToggleButton jtbThisButton;
        boolean bFound, bValid, bLabelUsed;
        int iIndex, iCurrentLine, iLabelUsedStatus;
        String stTemp;
        DataObject doThisDataObject;
        Status statInformation;
        
        super.actionPerformed( aeEvent );

        iLabelUsedStatus = super.I_OK;
        statInformation = new Status( iLabelUsedStatus );
        objActionSource = aeEvent.getSource();
        
        if ( objActionSource == jbApply ) {
            // since the Instruction parts were likely defined legally, it is
            // not necessary to do error checking...unless in "debugging mode"            
            
            bValid = false;
            
            if ( opBranchAndCompareOpcode.getOpcode() != Opcode.I_OPCODE__UNDEFINED ) {
                stTemp = this.getLabel().getText();

                if ( stTemp.equals( "" ) == false ) {
                    statInformation = super.isLabelUsed( stTemp );
                    iLabelUsedStatus = statInformation.getStatusNumber();
                    
                    if ( iLabelUsedStatus == super.I_OK ) {
                        objTemp = deDeclarations.getDeclaration( stTemp );

                        if ( objTemp instanceof DataObject ) {
                            doThisDataObject = (DataObject)objTemp;
                            instThisInstruction.setLabel( 
                                new NameAndDataObjectWrapper( stTemp, doThisDataObject ) );
                        }
                    }
                    else {
                        JOptionPane.showMessageDialog( this, 
                            statInformation.getStatusMessage(), 
                            ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );                    
                    }                                                
                }
                else {
                    doThisDataObject = new DataObject();
                    doThisDataObject.setType( DataObject.I_TYPE_LABEL );
                    instThisInstruction.setLabel( 
                        new NameAndDataObjectWrapper( "", doThisDataObject ) );
                }

                
                stTemp = this.getComments().getText();

                if ( stTemp.equals( "" ) == false ) {
                    instThisInstruction.setComments( stTemp );
                }
                //  else, ignore
                
                if ( iLabelUsedStatus == super.I_OK ) {
                    // operands are updated as the panel is clicked
                    if ( opBranchAndCompareOpcode.getOpcode()  != Opcode.I_OPCODE__BRANCH_ALWAYS ) {
                        if ( operLeft.getOperandType() != DataObject.I_UNDEFINED ) {
                            instThisInstruction.setLeftOperand( operLeft );

                            if ( operRight.getOperandType() != DataObject.I_UNDEFINED ) {
                                instThisInstruction.setRightOperand( operRight );
                                statInformation = instThisInstruction.validateInstruction();                            
                                bValid = true;
                            }
                            else {
                                JOptionPane.showMessageDialog( this, 
                                    "A right operand is required for this instruction.", 
                                    ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );                    
                            }
                        }
                        else {
                            JOptionPane.showMessageDialog( this, 
                                "A left operand is required for this instruction.", 
                                ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );                    
                        }
                    }  // end of checking operands if not I_BRANCH_ALWAYS
                    else {
                        // no operands are needed when solely branching
                        statInformation = instThisInstruction.validateInstruction();
                        bValid = true;
                    }
                }
                //System.out.println( "-1-" );
                if ( bValid == true ) {                
                    //System.out.println( "-2-" );
                    if ( operResult.getOperandType() == DataObject.I_TYPE_LABEL ) {
                        //System.out.println( "-3-" );
                        if ( statInformation.getStatusNumber() == Instruction.I_OK ) {
                            //System.out.println( "-4-" );
                            iCurrentLine = vctCodeTable.getCurrentRow();
                            //System.out.println( "-5-" );
                            if ( codeThisCode.getEditingMode() == Code.B_MODE_INSERT ) {
                                //System.out.println( "-6-" );
                                statInformation = codeThisCode.insertSingleLine( 
                                    iCurrentLine, instThisInstruction );
                                //System.out.println( "-7-" );
                            }
                            else {  // overwrite
                                //System.out.println( "-8-" );
                                statInformation = codeThisCode.changeSingleLine( 
                                    iCurrentLine, instThisInstruction );
                                //System.out.println( "-9-" );
                            }

                            if ( statInformation.getStatusNumber() != Instruction.I_OK ) {
                                //System.out.println( "-10-" );
                                JOptionPane.showMessageDialog( this, 
                                    statInformation.getStatusMessage(), 
                                    ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );                    
                                //System.out.println( "-11-" );
                            }
                            else {
                                //System.out.println( "-12,13-" );
                                clearAll();
                                //System.out.println( "-14-" );
                            }                            
                        }
                        else {
                            //System.out.println( "-15-" );
                            JOptionPane.showMessageDialog( this, 
                                statInformation.getStatusMessage(),
                                ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );                    
                            //System.out.println( "-16-" );
                        }
                    }
                    else {
                        //System.out.println( "-17-" );
                        JOptionPane.showMessageDialog( this, 
                            "A branching label is required for this instruction.", 
                            ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );
                        System.out.println( "-18-" );
                    }
                }  // else, error already issued
            }  // opcode found
            else {
                System.out.println( "-19-" );
                JOptionPane.showMessageDialog( this, "A function button must be "
                    + "pressed for the instruction to be valid.", 
                    ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );
                System.out.println( "-20-" );
            }
        }  // end of Apply
        else {
            // check button panel buttons
            bFound = false;
            iIndex = I_JBUTTON_LESS_THAN;
                        
            do {
                jtbThisButton = (JToggleButton)alGroupButtons.get( iIndex );
                
                if ( objActionSource == jtbThisButton ) {
                    bFound = true;
                }
                else {
                    iIndex++;
                }
            }
            while ( ( bFound == false )  &&  ( iIndex < I_JBUTTON_LAST_BUTTON ) );
            
            if ( bFound == true ) {
                super.enableLeftAndRightOperands( true );
                
                switch ( iIndex ) {
                    case I_JBUTTON_LESS_THAN:
                        opBranchAndCompareOpcode.setOpcode( Opcode.I_OPCODE__BRANCH_LESS_THAN );
                        break;
                        
                    case I_JBUTTON_LESS_THAN_OR_EQUAL_TO:
                        opBranchAndCompareOpcode.setOpcode( 
                            Opcode.I_OPCODE__BRANCH_LESS_THAN_OR_EQUAL );
                        break;
                        
                    case I_JBUTTON_GREATER_THAN:
                        opBranchAndCompareOpcode.setOpcode( 
                            Opcode.I_OPCODE__BRANCH_GREATER_THAN );
                        break;
                        
                    case I_JBUTTON_GREATER_THAN_OR_EQUAL_TO:
                        opBranchAndCompareOpcode.setOpcode( 
                            Opcode.I_OPCODE__BRANCH_GREATER_THAN_OR_EQUAL );
                        break;
                        
                    case I_JBUTTON_EQUAL_TO:
                        opBranchAndCompareOpcode.setOpcode( 
                            Opcode.I_OPCODE__BRANCH_EQUAL );
                        break;
                        
                    case I_JBUTTON_NOT_EQUAL_TO:
                        opBranchAndCompareOpcode.setOpcode( 
                            Opcode.I_OPCODE__BRANCH_NOT_EQUAL );
                        break;
                        
                    case I_JBUTTON_BRANCH:
                        opBranchAndCompareOpcode.setOpcode( 
                            Opcode.I_OPCODE__BRANCH_ALWAYS );
                        dataObjectAreaChanged( iLeftOperandID, new Operand() );
                        super.updateDataObjectArea( iLeftOperandID, operLeft );
                        dataObjectAreaChanged( iRightOperandID, new Operand() );
                        super.updateDataObjectArea( iRightOperandID, operRight );
                        super.enableLeftAndRightOperands( false );
                        break;
                }
                instThisInstruction.setOpcode( (Opcode)opBranchAndCompareOpcode.clone() );
                System.out.println( "-31-" );
            }
            else {
                super.actionPerformed( aeEvent );
        
                if ( objActionSource == super.jbClear ) {
                    clearAll();
                }        
            }
        }
        //  else ignore
    }
    
    //------------- DataObjectAreaListener Implementation ------------\\
    public void dataObjectAreaChanged( int iID, Operand operData ) {
        if ( iID == iLeftOperandID ) {
            operLeft = operData;
            updateRestrictionFlags( iID );
            instThisInstruction.setLeftOperand( operLeft );
        }
        else if ( iID == iRightOperandID ) {
            operRight = operData;
            updateRestrictionFlags( iID );
            instThisInstruction.setRightOperand( operRight );
        }
        else if ( iID == iResultOperandID ) {  
            // always a label regardless of what type other operands are
            operResult = operData;
            instThisInstruction.setBranchLabelOperand( operResult );
        }
        // else ignore
    }
    
    //////////////////////// PRIVATE METHODS \\\\\\\\\\\\\\\\\\\\\\\\\\
    private void createObjects() {
        ptvhdHelp = new PTVHelpDialog( jfParentFrame, "Help >> Panel >> Compare & Branch" );
        ptvhdHelp.setHelpFile( "help_panel_compare.html" );
        super.setHelpDialog( ptvhdHelp );
        
        jbApply = super.getApplyInstructionButton();
        jbApply.addActionListener( this );
        
        // just use display and keyboard and constant for one operand
        // ...no point in comparing two constants, two keyboards, two displays
        bsLeftOpRestrictionFlags = super.getNewDataObjectAreaRestrictionList();
        bsLeftOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_CONSTANT );
        bsLeftOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_VARIABLE );
        bsLeftOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_INT_SINGLE );
        bsLeftOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_CHARACTER_SINGLE );
        bsLeftOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_CHARACTER_ARRAY );
        bsLeftOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_DISPLAY_OUTPUT );
        bsLeftOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_KEYBOARD_INPUT );
        bsLeftOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_INT_ARRAY_ELEMENT );
        bsLeftOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_CHAR_ARRAY_ELEMENT );
        
        bsRightOpRestrictionFlags = super.getNewDataObjectAreaRestrictionList();
        bsRightOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_CONSTANT );
        bsRightOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_VARIABLE );
        bsRightOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_LITERAL );
        bsRightOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_INT_SINGLE );
        bsRightOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_CHARACTER_SINGLE );
        bsRightOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_CHARACTER_ARRAY );
        bsRightOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_DISPLAY_OUTPUT );
        bsRightOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_KEYBOARD_INPUT );
        bsRightOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_INT_ARRAY_ELEMENT );
        bsRightOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_CHAR_ARRAY_ELEMENT );
        
        bsResultOpRestrictionFlags = super.getNewDataObjectAreaRestrictionList();
        bsResultOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_LABEL );
        bsResultOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_ALL_LABELS );
        
        bsLeftOpBackupRestrictionFlags = (BitSet)bsLeftOpRestrictionFlags.clone();
        bsRightOpBackupRestrictionFlags = (BitSet)bsRightOpRestrictionFlags.clone();
                
        alGroupButtons = new ArrayList( 7 );
        
        alGroupButtons.add( I_JBUTTON_LESS_THAN, createButton( "less-than-icon.jpg", 
            "inv-less-than-icon.jpg", "Less Than" ) );
        alGroupButtons.add( I_JBUTTON_LESS_THAN_OR_EQUAL_TO, createButton( 
            "less-than-equal-icon.jpg", "inv-less-than-equal-icon.jpg", 
            "Less Than or Equal To" ) );
        alGroupButtons.add( I_JBUTTON_GREATER_THAN, createButton( 
            "greater-than-icon.jpg", "inv-greater-than-icon.jpg", 
            "Greater Than" ) );
        alGroupButtons.add( I_JBUTTON_GREATER_THAN_OR_EQUAL_TO, createButton( 
            "greater-than-equal-icon.jpg", "inv-greater-than-equal-icon.jpg",
            "Greater Than or Equal To" ) );
        alGroupButtons.add( I_JBUTTON_EQUAL_TO, createButton( "equal-icon.jpg", 
            "inv-equal-icon.jpg", "Equal To" ) );
        alGroupButtons.add( I_JBUTTON_NOT_EQUAL_TO, createButton( "not-equal-icon.jpg",
            "inv-not-equal-icon.jpg", "Not Equal To" ) );
        alGroupButtons.add( I_JBUTTON_BRANCH, createButton( "always-branch-icon.jpg",
            "inv-always-branch-icon.jpg", "Branch Always" ) );        
        
        operLeft = new Operand();
        operRight = new Operand();
        operResult = new Operand();
        opBranchAndCompareOpcode = new Opcode();
        instThisInstruction = new Instruction();
    }
    
    private void finishPanelConfiguration() {
        // assume that everything worked ok...since IDs should be valid
        super.setDataObjectAreaRestrictionList( iLeftOperandID, bsLeftOpRestrictionFlags );
        super.setDataObjectAreaRestrictionList( iRightOperandID, bsRightOpRestrictionFlags );
        super.setDataObjectAreaRestrictionList( iResultOperandID, bsResultOpRestrictionFlags );
    }
    
    private void drawPanel() {
        iLeftOperandID = super.drawDataObjectArea( new Point( 5, 7 ), operLeft );
        iRightOperandID = super.drawDataObjectArea( new Point( 297, 7 ), operRight );
        iResultOperandID = super.drawDataObjectArea( new Point( 297, 105 ), operResult );
        super.setExceptionOperand( iResultOperandID );

        super.drawButtonPanel( new Point( 189, 7 ), alGroupButtons );
        super.drawLabelAndCommentsAndActionButtons();
        super.drawLine( new Rectangle( 149, 40, 25, 3 ) );  // left line
        super.drawLine( new Rectangle( 270, 40, 22, 3 ) );  // right line
        super.drawRectangle( new Rectangle( 174, 34, 12, 15 ) );  // left rectangle
        super.drawRectangle( new Rectangle( 258, 33, 12, 15 ) );  // right rectangle 
        super.drawLine( new Rectangle( 217, 116, 3, 42 ) );  // vertical line from button panel
        super.drawLine( new Rectangle( 220, 155, 56, 3 ) );  // horiz. line from button panel
        super.drawArrowhead( new Rectangle( 276, 148, 16, 16 ), PTVPanel.I_RIGHT_TRIANGLE );
    }

    private JToggleButton createButton( String stStandardIconFile, String stPressedIconFile,
        String stButtonDescription ) {
            
        JToggleButton jtbThisButton;
        
        jtbThisButton = new JToggleButton( "", new ImageIcon( 
            ST_HOME_DIRECTORY + stStandardIconFile ) );
        jtbThisButton.setPressedIcon( new ImageIcon( 
            ST_HOME_DIRECTORY + stPressedIconFile ) );
        jtbThisButton.setSelectedIcon( new ImageIcon( 
            ST_HOME_DIRECTORY + stPressedIconFile ) );        
        jtbThisButton.setToolTipText( stButtonDescription );
        jtbThisButton.addActionListener( this );
        return jtbThisButton;
    }
    
    private void updateRestrictionFlags( int iClickedID ) {
        Operand operClickedOperand;
        int iClickedOperandType, iClickedOperandStructure, iClickedOperandRestriction;
        
        // IGNORE LABEL OPERAND CLICKS
        if ( iClickedID != iResultOperandID ) {
            System.out.println( "Changing flags..." );
            // BEFORE CLICK: one operand blank; AFTER CLICK: both operands blank
            if ( ( operLeft.getOperandType() == DataObject.I_UNDEFINED )  && 
                 ( operRight.getOperandType() == DataObject.I_UNDEFINED ) ) {
                iFirstClickedOperandID = I_UNDEFINED;
                bsLeftOpRestrictionFlags = (BitSet)bsLeftOpBackupRestrictionFlags.clone();
                bsRightOpRestrictionFlags = (BitSet)bsRightOpBackupRestrictionFlags.clone();
                System.out.println( "Both operands cleared." );
            }
            else {
                // BEFORE CLICK: both operands blank; AFTER CLICK: one operand blank
                if ( iFirstClickedOperandID == I_UNDEFINED ) {
                    iFirstClickedOperandID = iClickedID;
                    System.out.println( "Clicked operand ID: " + iFirstClickedOperandID );
                }                

                // clicked operand affects the unclicked operand
                if ( iFirstClickedOperandID == iClickedID ) {
                    if ( iClickedID == iLeftOperandID ) {
                        bsRightOpRestrictionFlags = makeRestrictionFlagChanges( operLeft );
                        bsRightOpRestrictionFlags.set( 
                            AllPurposeMenu.I_FLAG_LITERAL );
                        super.setDataObjectAreaRestrictionList( 
                            iRightOperandID, bsRightOpRestrictionFlags ); 
                        System.out.println( "Clicked left operand...update right operand flags." );
                    }
                    else if ( iClickedID == iRightOperandID ) {
                        bsLeftOpRestrictionFlags = makeRestrictionFlagChanges( operRight );
                        super.setDataObjectAreaRestrictionList( 
                            iLeftOperandID, bsLeftOpRestrictionFlags ); 
                        System.out.println( "Clicked right operand...update left operand flags." );
                    }
                } // else, no restriction change...ignore
            }  // end of else
        }
        // else, ignore if the label operand was clicked    
    }
    
    private BitSet makeRestrictionFlagChanges( Operand operFirstClickedOperand ) {
        BitSet bsRestrictionFlags;
        int iOperandType, iOperandStructure, iOperandArrayElement;
        
        bsRestrictionFlags = super.getNewDataObjectAreaRestrictionList();
        
        iOperandType = operFirstClickedOperand.getOperandType();
        iOperandStructure = operFirstClickedOperand.getOperandStructure();
        iOperandArrayElement = operFirstClickedOperand.getArrayElementIndexSetting();
        
        System.out.println( "Received...type = " + iOperandType + "   struct = " + iOperandStructure );
        
        bsRestrictionFlags.set( AllPurposeMenu.I_FLAG_CONSTANT );
        bsRestrictionFlags.set( AllPurposeMenu.I_FLAG_VARIABLE );
        //bsRestrictionFlags.set( AllPurposeMenu.I_FLAG_LITERAL );

        if ( ( iOperandType == DataObject.I_TYPE_INTEGER )  || 
             ( iOperandType == DataObject.I_SYSTEM_TYPE_SINGLE_REGISTER ) ) {
            bsRestrictionFlags.set( AllPurposeMenu.I_FLAG_INT_SINGLE );
            bsRestrictionFlags.set( AllPurposeMenu.I_FLAG_INT_ARRAY_ELEMENT );
            System.out.println( "SET INT" );
        }
        else if ( ( ( iOperandType == DataObject.I_TYPE_CHARACTER )  &&
                  ( ( iOperandStructure == DataObject.I_STRUCTURE_SINGLE_VALUE )  ||
                    ( iOperandArrayElement != Operand.I_UNDEFINED ) ) ) ) {
            bsRestrictionFlags.set( AllPurposeMenu.I_FLAG_CHARACTER_SINGLE );
            bsRestrictionFlags.set( AllPurposeMenu.I_FLAG_CHAR_ARRAY_ELEMENT );
            System.out.println( "SET SINGLE CHAR" );
        }
        else if ( ( ( iOperandType == DataObject.I_TYPE_CHARACTER )  ||
                    ( iOperandType == DataObject.I_SYSTEM_TYPE_DISPLAY )  || 
                    ( iOperandType == DataObject.I_SYSTEM_TYPE_KEYBOARD ) )  &&
                  ( iOperandStructure == DataObject.I_STRUCTURE_1D_ARRAY ) ) { 
            if ( iOperandArrayElement == Operand.I_UNDEFINED ) {
                bsRestrictionFlags.set( AllPurposeMenu.I_FLAG_CHARACTER_ARRAY );
                bsRestrictionFlags.set( AllPurposeMenu.I_FLAG_DISPLAY_OUTPUT );
                bsRestrictionFlags.set( AllPurposeMenu.I_FLAG_KEYBOARD_INPUT );
                System.out.println( "SET CHAR ARRAY" );
            }
        }
        return bsRestrictionFlags;        
    }

    ////////////////////////// PRIVATE METHOD \\\\\\\\\\\\\\\\\\\\\\
    private void setCorrectButton( int iOpcodeID ) {
        JToggleButton jtbButtonToPress;
        boolean bContinue;

        bContinue = true;
        super.enableLeftAndRightOperands( true );

        switch ( iOpcodeID ) {
            case Opcode.I_OPCODE__BRANCH_LESS_THAN:
                jtbButtonToPress = (JToggleButton)alGroupButtons.get( I_JBUTTON_LESS_THAN );
                break;
                
            case Opcode.I_OPCODE__BRANCH_LESS_THAN_OR_EQUAL:
                jtbButtonToPress = (JToggleButton)alGroupButtons.get( 
                    I_JBUTTON_LESS_THAN_OR_EQUAL_TO );
                break;
                
            case Opcode.I_OPCODE__BRANCH_EQUAL:
                jtbButtonToPress = (JToggleButton)alGroupButtons.get( I_JBUTTON_EQUAL_TO );
                break;
                
            case Opcode.I_OPCODE__BRANCH_GREATER_THAN:
                jtbButtonToPress = (JToggleButton)alGroupButtons.get( I_JBUTTON_GREATER_THAN );
                break;
                
            case Opcode.I_OPCODE__BRANCH_GREATER_THAN_OR_EQUAL:
                jtbButtonToPress = (JToggleButton)alGroupButtons.get( 
                    I_JBUTTON_GREATER_THAN_OR_EQUAL_TO );
                break;

            case Opcode.I_OPCODE__BRANCH_NOT_EQUAL:
                jtbButtonToPress = (JToggleButton)alGroupButtons.get( I_JBUTTON_NOT_EQUAL_TO );
                break;

            case Opcode.I_OPCODE__BRANCH_ALWAYS:
                jtbButtonToPress = (JToggleButton)alGroupButtons.get( I_JBUTTON_BRANCH );
                super.enableLeftAndRightOperands( false );
                break;
                
            default:
                bContinue = false;
                jtbButtonToPress = new JToggleButton();
        }
        
        if ( bContinue == true ) {
            jtbButtonToPress.setSelected( true );
        }        
    }
    
    private void clearAll() {       
        instThisInstruction = new Instruction();
        iFirstClickedOperandID = I_UNDEFINED;

        bsLeftOpRestrictionFlags = (BitSet)
            bsLeftOpBackupRestrictionFlags.clone();
        bsRightOpRestrictionFlags = (BitSet)
            bsRightOpBackupRestrictionFlags.clone();
        super.setDataObjectAreaRestrictionList( 
            iLeftOperandID, bsLeftOpRestrictionFlags ); 
        super.setDataObjectAreaRestrictionList( 
            iRightOperandID, bsRightOpRestrictionFlags ); 
        
        super.clearForm();        
    }
    
    
    // instance variables
    private Instruction instThisInstruction;
    private Opcode opBranchAndCompareOpcode;
    private Operand operLeft, operRight, operResult;
    private JButton jbApply;
    private ArrayList alGroupButtons;
    private int iLeftOperandID, iRightOperandID, iResultOperandID, iFirstClickedOperandID;
    private BitSet bsLeftOpRestrictionFlags, bsLeftOpBackupRestrictionFlags,
        bsRightOpRestrictionFlags, bsRightOpBackupRestrictionFlags,
        bsResultOpRestrictionFlags;
    private PTVHelpDialog ptvhdHelp;
    
    // constants
    private final int I_JBUTTON_LESS_THAN = 0;
    private final int I_JBUTTON_LESS_THAN_OR_EQUAL_TO= 1;
    private final int I_JBUTTON_GREATER_THAN = 2;
    private final int I_JBUTTON_GREATER_THAN_OR_EQUAL_TO = 3;
    private final int I_JBUTTON_EQUAL_TO = 4;
    private final int I_JBUTTON_NOT_EQUAL_TO = 5;
    private final int I_JBUTTON_BRANCH = 6;
    private final int I_JBUTTON_LAST_BUTTON = 7;
}