/*
 * Code.java
 *
 * Created on December 17, 2001, 5:08 PM
 */

//package panelsAndTools;
import java.io.*;
import java.util.*;


/**
 *
 * @author  _
 * @version 
 */
public class Code implements Cloneable, Serializable {

    public Code() {
        vInstructions = new Vector();
        alUsedLabelList = new ArrayList();
        alLineNumbersOfChangedInstructions = new ArrayList();
        alInstructionsToBeChanged = new ArrayList();
        stOldLabelName = "";
        stNewLabelName = "";    
        bEditingMode = B_MODE_INSERT;
        iNumberOfLines = vInstructions.size();
    }
    
    public Code( Vector vCodeBlock ) {
        int iIndex;
        Instruction instThisInstruction;
        NameAndDataObjectWrapper nadowThisObject;
        String stLabelName;
        
        vInstructions = (Vector)vCodeBlock.clone();
        alUsedLabelList = new ArrayList();
        iNumberOfLines = vInstructions.size();
        
        for ( iIndex = 0; iIndex < iNumberOfLines; iIndex++ ) {
            instThisInstruction = (Instruction)vInstructions.get( iIndex ); 
            nadowThisObject = instThisInstruction.getLabel();
            stLabelName = nadowThisObject.getName();
            alUsedLabelList.add( stLabelName );
        }
        stOldLabelName = "";
        stNewLabelName = "";
//        insertSingleLine( vInstructions.size() - 1, instBlank );
    }

    public void clearData() {
        deleteLines( 0, iNumberOfLines - 1 );
        alLineNumbersOfChangedInstructions.clear();
        alInstructionsToBeChanged.clear();
        stOldLabelName = "";
        stNewLabelName = "";
        insertSingleLine( 0, instBlank );
        vctCodeTable.moveCursor( 0 );
    }
       
    protected Object clone() {
        // strucutre based on Sun's online Java tutorial for
        // the Object.clone() method and pp.235-6 C of Core Java 2: ISBN 0-13-081933-6
        try {
            Code codeThisCode = (Code)super.clone();
            codeThisCode.vInstructions = (Vector)this.vInstructions.clone();
            codeThisCode.alUsedLabelList = (ArrayList)this.alUsedLabelList.clone();
            codeThisCode.alLineNumbersOfChangedInstructions = 
                (ArrayList)this.alLineNumbersOfChangedInstructions.clone();
            codeThisCode.alInstructionsToBeChanged = 
                (ArrayList)this.alInstructionsToBeChanged.clone();
            codeThisCode.stOldLabelName = new String( this.stOldLabelName );
            codeThisCode.stNewLabelName = new String( this.stNewLabelName );
            return codeThisCode;
        }
        catch ( CloneNotSupportedException cnseException ) {
            System.out.println( "Clone of Code is not supported." );
            return null;
        }
    }    
   
    // ******* Assume that Instruction and Code objects are VALID ******* \\
    
    public Status insertLines( int iStartingLineNumber, Code codeBlock ) {
        String stMessage, stLabelName;
        Status statInformation;
        boolean bValid;
        int iIndex, iTotalNumberOfLabels, iLastLine;
        ArrayList alCodeBlockLabelList;
               
        if ( iNumberOfLines == 0 ) {
            iStartingLineNumber = -1;  // used to make the conditional true
        }
        
        // assuming line numbers are from 0 to size - 1
        if ( ( iStartingLineNumber < iNumberOfLines )  || 
             ( iNumberOfLines == 0 ) ) {
            alCodeBlockLabelList = codeBlock.getUsedLabelDataObjectNames();
            iTotalNumberOfLabels = alCodeBlockLabelList.size();            
            iIndex = 0;
            bValid = true;
            stLabelName = "";
            
            while ( ( iIndex < iTotalNumberOfLabels )  &&  ( bValid == true ) ) {
                stLabelName = (String)alCodeBlockLabelList.get( iIndex );
                bValid = addLabelToList( stLabelName );
                
                if ( bValid == true ) {
                    iIndex++;
                }
            }
            
            if ( bValid == true ) {                
                iLastLine = codeBlock.getNumberOfLines() - 1;
                
                if ( iNumberOfLines > 0 ) {
                    vInstructions.addAll( iStartingLineNumber, 
                        (List)codeBlock.getLines( 0, iLastLine ) );
                    clCodeListener.insertLinesOfCode( iStartingLineNumber, iLastLine );
                }
                else {
                    vInstructions.addAll( (List)codeBlock.getLines( 0, iLastLine ) );
                    clCodeListener.insertLinesOfCode( 0, iLastLine );
                }
                vctCodeTable.moveCursor( iLastLine + 1 );
                plistProgramListener.programChanged();
                statInformation = new Status( I_OK );
            }
            else {
                stMessage = "The label object named " + stLabelName + " was already "
                    + "used once in the program.  No insertion occurred.";
                statInformation = new Status( I_DUPLICATE_LABEL, stMessage );
            }
        }        
        else {
            stMessage = "The specified insertion point is greater than the number "
                + "of lines in this program.  No insertion occurred.";
            statInformation = new Status( I_ILLEGAL_INSERTION_POINT, stMessage );
        }        
        iNumberOfLines = vInstructions.size();
        
        return statInformation;
    }

    // parameters specify an inclusive range
    public Object getLines( int iStartOfRange, int iEndOfRange ) {
        Vector vFromCodeBlock;
        String stMessage;
        Code codeToCodeBlock;
        Object objReturnObject;
        
        if ( ( iStartOfRange >= 0 )  &&  ( iStartOfRange < iNumberOfLines ) ) {
            if ( ( iEndOfRange >= iStartOfRange )  &&  ( iEndOfRange < iNumberOfLines ) ) {
                objReturnObject = vInstructions.subList( iStartOfRange, iEndOfRange + 1 );
            }
            else {
                stMessage = "Invalid ending point of range to get program lines.";
                objReturnObject = new Status( I_ILLEGAL_ENDING_RANGE_POINT, stMessage );
            }
        }
        else {
            stMessage = "Invalid starting point of range to get program lines.";
            objReturnObject = new Status( I_ILLEGAL_STARTING_RANGE_POINT, stMessage );
        }
        return objReturnObject;
    }
    
    // parameters specify an inclusive range
    public Status deleteLines( int iStartOfRange, int iEndOfRange ) {
        int iIndex, iLineNumber;
        Status statInformation;
        String stMessage, stLabelName;
        Instruction instThisInstruction;
        NameAndDataObjectWrapper nadowLabel;
        
        if ( ( iStartOfRange >= 0 )  &&  ( iStartOfRange < iNumberOfLines ) ) {
            if ( ( iEndOfRange >= iStartOfRange )  &&  ( iEndOfRange < iNumberOfLines ) ) {
                for ( iIndex = iStartOfRange; iIndex <= iEndOfRange; iIndex++ ) {
                    instThisInstruction = (Instruction)vInstructions.get( iIndex );
                    nadowLabel = instThisInstruction.getLabel();
                    stLabelName = nadowLabel.getName();
                    removeLabelFromList( stLabelName );                    
                }
                
                // because elements are shifted after removal, use same index
                for ( iLineNumber = iStartOfRange; iLineNumber <= iEndOfRange;
                    iLineNumber++ ) {
                    vInstructions.removeElementAt( iStartOfRange );
                }
                
                clCodeListener.deleteLinesOfCode( iStartOfRange, iEndOfRange );
                /* DO NOT MOVE CURSOR...deletion does shifting */
                plistProgramListener.programChanged();
                statInformation = new Status( I_OK );
            }
            else {
                stMessage = "Invalid ending point of range to delete program lines.";
                statInformation = new Status( I_ILLEGAL_ENDING_RANGE_POINT, stMessage );
            }
        }
        else {
            stMessage = "Invalid starting point of range to delete program lines.";
            statInformation = new Status( I_ILLEGAL_STARTING_RANGE_POINT, stMessage );
        }
        iNumberOfLines = vInstructions.size();
        return statInformation;        
    }
    
    public Status insertSingleLine( int iLineNumber, Instruction instThisInstruction ) {
        Status statInformation;
        String stMessage, stLabelName;
        NameAndDataObjectWrapper nadowLabel;
        boolean bValid;
                
        if ( iNumberOfLines == 0 ) {
            iLineNumber = 0;
        }
        
        System.out.println( "LINE NUMBER = " + iLineNumber );
        System.out.println( "# OF LINES = " + iNumberOfLines );
        
        if ( ( ( iLineNumber >= 0 )  &&  ( iLineNumber < iNumberOfLines ) )  ||
             ( iNumberOfLines == 0 ) ) {
            nadowLabel = instThisInstruction.getLabel();
            stLabelName = nadowLabel.getName();
            System.out.println( "INSERT SINGLE LINE...label = " + stLabelName );
            bValid = addLabelToList( stLabelName );
            
            System.out.println( "VALID = " + bValid );
            
            if ( bValid == true ) {
                if ( iNumberOfLines > 0 ) {
                    vInstructions.insertElementAt( instThisInstruction, iLineNumber );
                    clCodeListener.insertSingleLineOfCode( iLineNumber );
                }
                else {
                    vInstructions.add( instThisInstruction );
                    clCodeListener.insertSingleLineOfCode( 0 );
                }
                vctCodeTable.moveCursor( 1 );
                
                if ( instThisInstruction.getOpcode().getOpcode() 
                       != Opcode.I_OPCODE__IGNORE ) {
                    plistProgramListener.programChanged();
                }
                
                statInformation = new Status( I_OK );
            }
            else {
                stMessage = "The label named "
                    + instThisInstruction.getLabel().getName() + " is already used to "
                    + "mark another program line.";
                statInformation = new Status( I_INVALID_LINE, stMessage );
            }
        }
        else {
            stMessage = "The line cannot be inserted at this point in the program.";
            statInformation = new Status( I_INVALID_LINE, stMessage );
        }
        iNumberOfLines = vInstructions.size();
        return statInformation;
    }

    // use with Declarations
    public Status changeSingleLine( int iLineNumber, Instruction instThisInstruction ) {
        Status statInformation;
        String stMessage, stReplacedLabel, stNewLabel;
        Instruction instReplacedInstruction;
        NameAndDataObjectWrapper nadowLabel;
        boolean bValid;
        
        if ( ( iLineNumber >= 0 )  &&  ( iLineNumber < iNumberOfLines - 1 ) ) {
            instReplacedInstruction = (Instruction)vInstructions.get( iLineNumber );
            nadowLabel = instReplacedInstruction.getLabel();
            stReplacedLabel = nadowLabel.getName();
            removeLabelFromList( stReplacedLabel );
            
            nadowLabel = instThisInstruction.getLabel();
            stNewLabel = nadowLabel.getName();
            bValid = addLabelToList( stNewLabel );
            
            if ( bValid == true ) {
                vInstructions.setElementAt( instThisInstruction, iLineNumber );
                clCodeListener.changeSingleLineOfCode( iLineNumber );
                plistProgramListener.programChanged();
                statInformation = new Status( I_OK );
            }
            else {
                addLabelToList( stReplacedLabel );  // rollback
                stMessage = "Invalid position to change a line of the program.";
                statInformation = new Status( I_INVALID_LINE, stMessage );
            }
        }
        else if ( iLineNumber == iNumberOfLines - 1 ) {
            stMessage = "You cannot overwrite the last empty line.";
            statInformation = new Status( I_INVALID_LINE, stMessage );
        }
        else {
            stMessage = "The line cannot be changed at this point in the program.";
            statInformation = new Status( I_INVALID_LINE, stMessage );
        }
        return statInformation;
    }
        
    public Object getSingleLine( int iLineNumber ) {
        Object objReturnedObject;
        String stMessage;
        
        if ( ( iLineNumber >= 0 )  &&  ( iLineNumber < iNumberOfLines ) ) {
            objReturnedObject = (Instruction)vInstructions.get( iLineNumber );
        }
        else {
            stMessage = "There is no instruction in this part of the program.";
            objReturnedObject = new Status( I_INVALID_LINE, stMessage );
        }
        return objReturnedObject;        
    }
    
    public Status deleteSingleLine( int iLineNumber ) {
        Status statInformation;
        Instruction instThisInstruction;
        NameAndDataObjectWrapper nadowLabel;       
        String stMessage, stLabel;
        
        if ( ( iLineNumber >= 0 )  &&  ( iLineNumber < iNumberOfLines ) ) {
            instThisInstruction = (Instruction)vInstructions.elementAt( iLineNumber );
            nadowLabel = instThisInstruction.getLabel();
            vInstructions.removeElementAt( iLineNumber );
            stLabel = nadowLabel.getName();
            removeLabelFromList( stLabel );       
            clCodeListener.deleteSingleLineOfCode( iLineNumber );
            vctCodeTable.moveCursor( -1 );
            plistProgramListener.programChanged();
            statInformation = new Status( I_OK );
        }
        else {
            stMessage = "There is no instruction in this part of the program.";
            statInformation = new Status( I_INVALID_LINE, stMessage );
        }
        iNumberOfLines = vInstructions.size();
        return statInformation;        
    }
        
    public int getNumberOfLines() {
        return iNumberOfLines;
    }
    
    public ArrayList getUsedLabelDataObjectNames() {
        return (ArrayList)alUsedLabelList.clone();
    }
    
    public void setCodeListener( CodeListener clListener ) {
        clCodeListener = clListener;
    }
    
    public void setProgramCodeListener( ProgramListener plistListener ) {
        plistProgramListener = plistListener;
    }
    
    public boolean isLabelUsed( String stLabelName ) {
        if ( alUsedLabelList.indexOf( stLabelName ) == I_NOT_FOUND ) {
            //System.out.println( "Label is not used." );
            return false;            
        }
        else {
            //System.out.println( "Label is used." );
            return true;
        }
    }
    
    // if deleting...objData should be null
    public int getErrorCountWithDeclarationChange( int iDeclarationAction, 
        String stOldObjectName, Object objNewData,
        final Declarations deReadOnlyDeclarations ) {
            
        int iInstructionLines, iErrorCount, iOpcode;
        Instruction instThisInstruction, instInstructionToCheck;
        Operand operNewData, operThisData;
        NameAndDataObjectWrapper nadowThisData, nadowNewData;
        String stNewObjectName, stOperandLocation1, stOperandLocation2;
        boolean bFound;
        Status statInformationLeft, statInformationRight, statInformationInstruction;
        
        iAction = iDeclarationAction;
        alInstructionsToBeChanged.clear();
        alLineNumbersOfChangedInstructions.clear();        
        iInstructionLines = vInstructions.size();
        operThisData = new Operand();
        
        if ( objNewData == null ) {
            operNewData = new Operand();
            nadowNewData = new NameAndDataObjectWrapper();
            stNewObjectName = "";
        }
        else if ( objNewData instanceof Operand ) {
            operNewData = (Operand)objNewData;
            nadowNewData = operNewData.getOperandData();
            stNewObjectName = operNewData.getOperandName();
        }
        else if ( objNewData instanceof NameAndDataObjectWrapper ) {
            nadowNewData = (NameAndDataObjectWrapper)objNewData;
            stNewObjectName = nadowNewData.getName();
        }
        else {
            nadowNewData = new NameAndDataObjectWrapper();
            return -1;  // should not happen
        }        
        iErrorCount = 0;
        
        for ( int iLineNumber = 0; iLineNumber < iInstructionLines; iLineNumber++ ) {
            // (1) Check if there is a name match in the Instruction
            bFound = false;            
            System.out.println( "Line = " + iLineNumber );
            instThisInstruction = (Instruction)vInstructions.get( iLineNumber );
            iOpcode = instThisInstruction.getOpcode().getOpcode();
            instInstructionToCheck = (Instruction)instThisInstruction.clone();
            nadowThisData = instThisInstruction.getLabel();
            statInformationLeft = new Status( I_OK );

            //  NOTE: An object name can only represent an object as a 
            //  label type or non-label type
            if ( stOldObjectName.equals( nadowThisData.getName() ) == true ) {
                System.out.println( "*1*" );
                bFound = true;
                instInstructionToCheck.setLabel( nadowNewData );
                stOldLabelName = stOldObjectName;
                stNewLabelName = stNewObjectName;
                System.out.println( "Changed label" );
            }

            if ( ( iOpcode >= Opcode.I_OPCODE__BRANCH_FIRST )  && 
                 ( iOpcode <= Opcode.I_OPCODE__BRANCH_LAST ) ) {
                stOperandLocation1 = "Left";
                stOperandLocation2 = "Right";
            }
            else {
                stOperandLocation1 = "Destination";
                stOperandLocation2 = "Source";
            }
                        
            operThisData = (Operand)instThisInstruction.getLeftOperand().clone();            
            if ( stOldObjectName.equals( operThisData.getOperandName() ) == true ) {
                System.out.println( "*2*" );
                bFound = true;
                statInformationLeft = updateOperand( B_NO_CHANGE_TO_ARRAY_INDEX_NAME,
                    operThisData, nadowNewData, deReadOnlyDeclarations );

                if ( statInformationLeft.getStatusNumber() == I_OK ) {
                    System.out.println( "*3*" );
                    operThisData.clearFailedOperandInformation();
                    instInstructionToCheck.setLeftOperand( operThisData );                            
                    System.out.println( "Changed " + stOperandLocation1 + " operand" );
                }
                else {
                    System.out.println( "*4*" );
                    operThisData.setFailedOperandInformation( stOperandLocation1 + " operand "
                        + "unchanged.  " + statInformationLeft.getStatusMessage() );
                    instInstructionToCheck.setLeftOperand( operThisData );                            
                    System.out.println( "Changed " + stOperandLocation1 + " operand" );
                }
            }            
            else if ( hasArrayIndexDataObjectChanged( stOldObjectName, operThisData, 
                nadowNewData ) == true ) {
                System.out.println( "*5*" );
                bFound = true;
                statInformationLeft = updateOperand( B_CHANGED_ARRAY_INDEX_NAME,
                    operThisData, nadowNewData, deReadOnlyDeclarations );

                if ( statInformationLeft.getStatusNumber() == I_OK ) {
                    System.out.println( "*6*" );
                    operThisData.clearFailedOperandInformation();
                    instInstructionToCheck.setLeftOperand( operThisData );
                    System.out.println( "Changed " + stOperandLocation1 + " operand "
                        + "data object array index" );                            
                }
                else {
                    System.out.println( "*7*" );
                    operThisData.setFailedOperandInformation( stOperandLocation1 + " operand "
                        + "unchanged.  " + statInformationLeft.getStatusMessage() );
                    instInstructionToCheck.setLeftOperand( operThisData );
                    System.out.println( "Changed " + stOperandLocation1 + " operand "
                        + "data object array index" );                            
                }
            }

            statInformationRight = new Status( I_OK );
//            if ( statInformationLeft.getStatusNumber() == I_OK ) {

            System.out.println( "*8*" );
            operThisData = (Operand)instThisInstruction.getRightOperand().clone();
            if ( stOldObjectName.equals( operThisData.getOperandName() ) == true ) {
                System.out.println( "*9*" );
                bFound = true;
                statInformationRight = updateOperand( B_NO_CHANGE_TO_ARRAY_INDEX_NAME,
                    operThisData, nadowNewData, deReadOnlyDeclarations );

                if ( statInformationRight.getStatusNumber() == I_OK ) {
                    System.out.println( "*10*" );
                    operThisData.clearFailedOperandInformation();
                    instInstructionToCheck.setRightOperand( operThisData );
                    System.out.println( "Changed " + stOperandLocation2 + " operand" );                            
                }
                else {
                    System.out.println( "*11*" );
                    operThisData.setFailedOperandInformation( stOperandLocation2 + " operand "
                        + "unchanged.  " + statInformationRight.getStatusMessage() );
                    instInstructionToCheck.setRightOperand( operThisData );
                    System.out.println( stOperandLocation2 + " right operand" );                            
                }
            }
            else if ( hasArrayIndexDataObjectChanged( stOldObjectName, operThisData, 
                nadowNewData ) == true ) {
                System.out.println( "*12*" );
                bFound = true;                    
                statInformationRight = updateOperand( B_CHANGED_ARRAY_INDEX_NAME,
                    operThisData, nadowNewData, deReadOnlyDeclarations );

                if ( statInformationRight.getStatusNumber() == I_OK ) {
                    System.out.println( "*13*" );
                    operThisData.clearFailedOperandInformation();
                    instInstructionToCheck.setRightOperand( operThisData );
                    System.out.println( "Changed " + stOperandLocation2 + " operand array index data object" );
                }
                else {
                    System.out.println( "*14*" );
                    operThisData.setFailedOperandInformation( stOperandLocation2 + " operand "
                        + "unchanged.  " + statInformationRight.getStatusMessage() );
                    instInstructionToCheck.setRightOperand( operThisData );
                    System.out.println( "Changed " + stOperandLocation2 + " operand array index data object" );
                }                       
            }

            operThisData = (Operand)instThisInstruction.getBranchLabelOperand().clone();
            if ( stOldObjectName.equals( operThisData.getOperandName() ) ) {
                System.out.println( "*15*" );
                bFound = true;
                System.out.println( "Changed branch/label operand" );                        
                operThisData.clearFailedOperandInformation();
                operThisData.setOperandData( nadowNewData );
                System.out.println( "branch/label type = " + operThisData.getOperandType() );
                instInstructionToCheck.setBranchLabelOperand( operThisData );
            }
//        }
            
            if ( bFound == true ) {
                System.out.println( "*16*" );
                /*System.out.println( "CHANGED OPERAND..." );
                System.out.println( "Name = " + operThisData.getOperandName() );
                System.out.println( "Type = " + DataObjectPropertyTextView.getTypeName( operThisData.getOperandType() ) );
                System.out.println( "Restriction = " + DataObjectPropertyTextView.getRestrictionName( operThisData.getOperandRestriction() ) );
                System.out.println( "Structure = " + DataObjectPropertyTextView.getStructureName( operThisData.getOperandStructure() ) );*/
                alInstructionsToBeChanged.add( instInstructionToCheck );
                alLineNumbersOfChangedInstructions.add( new Integer( iLineNumber ) );
                
                System.out.println( "Lines to change = " + alInstructionsToBeChanged.size() );
                
                if ( ( statInformationLeft.getStatusNumber() != I_OK )  ||
                     ( statInformationRight.getStatusNumber() != I_OK ) ) {
                    System.out.println( "*17*" );
                    iErrorCount++;
                }
                else {
                    System.out.println( "*18*" );
                    statInformationInstruction = 
                        instInstructionToCheck.validateInstruction();
                        
                    if ( statInformationInstruction.getStatusNumber() != I_OK ) {
                        System.out.println( "*19*" );
                        iErrorCount++;
                    }
                    // else do nothing...instruction is ok
                }

            }
        }  // end of loop
        System.out.println( "leaving with " + iErrorCount + " error(s)." );
        return iErrorCount;
    }
    
    public void commitCodeChangesFromDeclarationChange() {
        int iLineNumber;
        Instruction instThisInstruction;
        
        System.out.println( "Committing changes..." );

        for ( int iChangeCount = 0; iChangeCount < alInstructionsToBeChanged.size();
            iChangeCount++ ) {
            instThisInstruction = 
                (Instruction)alInstructionsToBeChanged.get( iChangeCount );
            iLineNumber = 
                ((Integer)alLineNumbersOfChangedInstructions.get( iChangeCount )).intValue();
            instThisInstruction.validateInstruction();
            changeSingleLine( iLineNumber, instThisInstruction );
            //vInstructions.set( iLineNumber, instThisInstruction );
            //clCodeListener.changeSingleLineOfCode( iLineNumber );
        }
        
        /*if ( iChange == I_CHANGED_LABEL ) {
            if ( iAction == I_ACTION_CHANGE ) {
                removeLabelFromList( stOldLabelName );
                addLabelToList( stNewLabelName );                
            }
            else if ( iAction == I_ACTION_DELETE ) {
                removeLabelFromList( stOldLabelName );
            }
        }
        plistProgramListener.programChanged();        */
    }
    
    public void setEditingMode( boolean bMode ) {
        bEditingMode = bMode;
        vctCodeTable.updateTableCursor();
    }
    
    public boolean getEditingMode() {
        return bEditingMode;
    }
    
    public void setVisualCodeTable( VisualCodeTable vctThisCodeTable ) {
        vctCodeTable = vctThisCodeTable;
        insertSingleLine( 0, instBlank );
        vctCodeTable.moveCursor( 0 );
    }
    
    /////////////////////// FILE I/O METHODS \\\\\\\\\\\\\\\\\\\\\\\\\
    public Vector exportState() {
        Vector vData;
        Instruction instThisInstruction;
        
        vData = new Vector();
        vData.setSize( iNumberOfLines - 1 );
        
        // account for blank line which does not need to be stored
        for ( int iIndex = 0; iIndex < iNumberOfLines - 1; iIndex++ ) {
            instThisInstruction = (Instruction)vInstructions.get( iIndex );
            vData.set( iIndex, instThisInstruction.exportState() );
            //System.out.println( "CODE:   instThisInstruction.exportState()=" +  instThisInstruction.exportState() );
        }
        //System.out.println( "CODE:   vData=" + vData );
        return vData;
    }
    
    public void importState( Vector vData ) {
        Instruction instThisInstruction;
        Opcode opThisOpcode;
        Vector vInstructionData;
        
        for ( int iIndex = vData.size() - 1; iIndex >= 0; iIndex-- ) {
            //System.out.println( "Code: importState: vData size = " + vData.size() );
            vInstructionData = (Vector)vData.get( iIndex );
            //System.out.println( "-1-" );
            instThisInstruction = new Instruction();
            //System.out.println( "-2-" );
            instThisInstruction.importState( vInstructionData );
            //System.out.println( "-3-" );
            insertSingleLine( 0, instThisInstruction );
            //System.out.println( "-4-" );
        }        
        vctCodeTable.moveCursor( 0 );
    }
                    
    ////////////// CHECK FOR USED LABEL POINTS \\\\\\\\\\\\\\\\
    private void removeLabelFromList( String stLabelName ) {
        if ( isLabelUsed( stLabelName ) == true ) {
            alUsedLabelList.remove( stLabelName );
        }
    }
    
    private boolean addLabelToList( String stLabelName ) {
        //System.out.println( "Label received..." + stLabelName );
        if ( ( isLabelUsed( stLabelName ) == false )  &&
             ( stLabelName.equals( "" ) == false ) ) {
            alUsedLabelList.add( stLabelName );
            return true;
        }
        else if ( stLabelName.equals( "" ) == true ) {
            return true;
        }
        else {
            return false;
        }
    }

    private boolean hasArrayIndexDataObjectChanged( String stPreviousObjectName, 
        Operand operCurrent, NameAndDataObjectWrapper nadowNewOperandData ) {
            
        String stCurrentDataObjectIndex, stNewDataObjectIndex;
        int iIndexArraySetting, iCurrentProperty, iNewProperty;
        boolean bChange;
        DataObject doTemp;
        
        iIndexArraySetting = operCurrent.getArrayElementIndexSetting();
        doTemp = (DataObject)nadowNewOperandData.getObject();
        
        if ( ( iIndexArraySetting == Operand.I_ARRAY_INDEX_REGISTER )  ||
             ( iIndexArraySetting == Operand.I_ARRAY_INDEX_DATA_OBJECT ) ) {            

            stCurrentDataObjectIndex = operCurrent.getArrayElementDataObject();
            System.out.println( "Old object name = " + stPreviousObjectName );
            System.out.println( "New object name = " + stCurrentDataObjectIndex );
                        
            if ( stPreviousObjectName.equals( stCurrentDataObjectIndex ) == true ) {
                stNewDataObjectIndex = nadowNewOperandData.getName();            
                bChange = !( stNewDataObjectIndex.equals( stCurrentDataObjectIndex ) );

                iCurrentProperty = operCurrent.getOperandType();
                iNewProperty = doTemp.getType();
                bChange |= !( iCurrentProperty == iNewProperty );

                iCurrentProperty = operCurrent.getOperandStructure();
                iNewProperty = doTemp.getStructure();
                bChange |= !( iCurrentProperty == iNewProperty );
            }
            else {
                bChange = false;
            }
        }
        else {
            bChange = false;
        }        
        System.out.println( "Array Index Data Object Names changed = " + bChange );
        return bChange;
    }
    
    // nadowNewData is the changed data object; if the changed data object appears
    // in an array index for this instruction, then all of the information is
    // available and Declarations reference is unneeded    
    private Status updateOperand( boolean bArrayIndexNameModified,
        Operand operReadyForChange, NameAndDataObjectWrapper nadowNewOperandData, 
        final Declarations deDeclarationsReference ) {
    
        int iArrayIndexSetting, iIndex1, iIndex2;
        Status statInformation;
        DataObject doArrayIndexDataObject;
        Object objThisObject;
        String stArrayIndexName, stMessage;
        
        System.out.println( "UPDATE OPERAND" );
        iArrayIndexSetting = operReadyForChange.getArrayElementIndexSetting();
        
        if ( ( operReadyForChange.getOperandStructure() == 
               DataObject.I_STRUCTURE_1D_ARRAY )  &&
             ( nadowNewOperandData.getObject().getStructure()  !=
               DataObject.I_STRUCTURE_1D_ARRAY ) ) {
            // error 
            stMessage = "Current operand has an array index which cannot be used "
                + "with a single value structure.";
            statInformation = new Status( I_CANNOT_CONVERT_TO_SINGLE_VALUE, stMessage );
        }
        else {
            switch ( iArrayIndexSetting ) {
                case Operand.I_ARRAY_INDEX_DATA_OBJECT:
                case Operand.I_ARRAY_INDEX_REGISTER:
                    if ( bArrayIndexNameModified == true ) {
                        System.out.println( "=1=" );
                        statInformation = operReadyForChange.setArrayElementDataObject(
                            nadowNewOperandData.getName(), nadowNewOperandData.getObject().getType(),
                            nadowNewOperandData.getObject().getStructure() );
                    }
                    else {
                        System.out.println( "=2=" );
                        operReadyForChange.setOperandData( nadowNewOperandData );
                        stArrayIndexName = operReadyForChange.getArrayElementDataObject();

                        if ( iArrayIndexSetting == Operand.I_ARRAY_INDEX_DATA_OBJECT ) {
                            System.out.println( "=3=" );
                            objThisObject = deDeclarationsReference.getDeclaration( 
                                stArrayIndexName );
                        }
                        else {
                            System.out.println( "=4=" );
                            objThisObject = SystemDeclarations.getDeclaration( 
                                stArrayIndexName );
                        }

                        if ( objThisObject instanceof DataObject ) {
                            System.out.println( "=5=" );
                            doArrayIndexDataObject = (DataObject)objThisObject;
                            statInformation = operReadyForChange.setArrayElementDataObject(
                                stArrayIndexName, doArrayIndexDataObject.getType(),
                                doArrayIndexDataObject.getStructure() );
                        }
                        else if ( objThisObject instanceof Status ) {
                            System.out.println( "=6=" );
                            statInformation = (Status)objThisObject;
                        }
                        else {
                            System.out.println( "=7=" );
                            statInformation = new Status( Operand.I_ILLEGAL_ARRAY_INDEX,
                                "Illegal array index data object." );
                        }
                    }
                    break;

                case Operand.I_ARRAY_INDEX_NUMBER:
                    System.out.println( "=8=" );
                    iIndex1 = operReadyForChange.getArrayElementIndex();
                    operReadyForChange.setOperandData( nadowNewOperandData );
                    statInformation = operReadyForChange.setArrayIndex( iIndex1 );
                    break;

                case Operand.I_ARRAY_INDEX_RANGE:
                    System.out.println( "=9=" );
                    iIndex1 = operReadyForChange.getStartingArrayIndex();
                    iIndex2 = operReadyForChange.getEndingArrayIndex();
                    operReadyForChange.setOperandData( nadowNewOperandData );                
                    statInformation = operReadyForChange.setArrayIndices(
                        iIndex1, iIndex2 );
                    break;

                case Operand.I_UNDEFINED:
                    System.out.println( "=10=" );
                    operReadyForChange.setOperandData( nadowNewOperandData );
                    statInformation = new Status( I_OK );
                    break;

                default:
                    System.out.println( "=11=" );
                    statInformation = new Status( Operand.I_INTERNAL_ERROR, "???" );
                    break;
            }
        }
        return statInformation;
    }
    
    
    // instance variables
    private VisualCodeTable vctCodeTable;
    private Vector vInstructions;
    private ArrayList alUsedLabelList, alLineNumbersOfChangedInstructions,
        alInstructionsToBeChanged;
    private CodeListener clCodeListener;
    private ProgramListener plistProgramListener;
    private int iChange, iAction, iNumberOfLines;
    private String stOldLabelName, stNewLabelName;
    private boolean bEditingMode;
     
    // class constants
    public static final int I_NOT_FOUND = -1;
    
    public static final int I_ACTION_CHANGE = 0;
    public static final int I_ACTION_DELETE = 1;
 
    private static final Instruction instBlank;
    private static final Opcode opIgnore;
    
    private static final int I_CHANGED_LABEL = 0;
    private static final int I_CHANGED_LEFT_OPERAND = 1;
    private static final int I_CHANGED_RIGHT_OPERAND = 2;
    private static final int I_CHANGED_LEFT_OPERAND_ARRAY_INDEX_DATA_OBJECT = 3;
    private static final int I_CHANGED_RIGHT_OPERAND_ARRAY_INDEX_DATA_OBJECT = 4;
    private static final int I_CHANGED_BRANCH_AND_LABEL_OPERAND = 7;
 
    private static final boolean B_CHANGED_ARRAY_INDEX_NAME = true;
    private static final boolean B_NO_CHANGE_TO_ARRAY_INDEX_NAME = false;
    
    public static final boolean B_MODE_INSERT = true;
    public static final boolean B_MODE_OVERWRITE = false;
    
    
        // error codes
        public static final int I_OK = 0;
        public static final int I_ILLEGAL_INSERTION_POINT = -1;
        public static final int I_ILLEGAL_STARTING_RANGE_POINT = -2;
        public static final int I_ILLEGAL_ENDING_RANGE_POINT = -3;
        public static final int I_INVALID_LINE = -4;
        public static final int I_DUPLICATE_LABEL = -5;
        public static final int I_CANNOT_CONVERT_TO_SINGLE_VALUE = -6;
        
    static {
        opIgnore = new Opcode();
        opIgnore.setOpcode( Opcode.I_OPCODE__IGNORE );
        instBlank = new Instruction();
        instBlank.setOpcode( opIgnore );
        instBlank.validateInstruction();
    }    
}