/*
 * DataObject.java
 *
 * Created on January 28, 2002, 5:42 PM
 */

//package panelsAndTools;

import javax.swing.*;
import java.lang.*;
import java.util.*;
import java.io.*;

/**
 *
 * @author  _
 * @version 
 */

public class DataObject implements Cloneable, Serializable {

    public DataObject() {
        iObjectSize = 0;
        iTypeSize = 0;
        iType = I_UNDEFINED;
        iRestriction = I_UNDEFINED;
        iStructure = I_UNDEFINED;
        stDescription = "";
        stValueString = "";
        stIconFileName = "";
        stMenuIconFileName = "";
        iiIcon = new ImageIcon( "", "none" );
        iiMenuIcon = new ImageIcon( "", "none" );
        roRecord = new RecordObject();
        adArraySize = new ArrayDimensions();
        alValidatedValues = new ArrayList();
        bValidatedValueFlag = false;
    }
    
    protected Object clone() {
        // strucutre based on Sun's online Java tutorial for
        // the Object.clone() method and pp.235-6 C of Core Java 2: ISBN 0-13-081933-6
        try {
            Object objTempObject;
            
            DataObject doObject = (DataObject)super.clone();
            doObject.alValidatedValues = (ArrayList)this.alValidatedValues.clone();
            doObject.roRecord = (RecordObject)this.roRecord.clone();
            doObject.adArraySize = (ArrayDimensions)this.adArraySize.clone();
            doObject.iiIcon = getIcon();
            doObject.iiMenuIcon = getMenuIcon();
            doObject.stDescription = new String( this.stDescription );
            doObject.stValueString = new String( this.stValueString );
            doObject.stIconFileName = new String( this.stIconFileName );
            doObject.stMenuIconFileName = new String( this.stMenuIconFileName );
                        
            return doObject;
        }
        catch ( CloneNotSupportedException cnseException ) {
            System.out.println( "Clone of DataObject is not supported." );
            return null;
        }
    }    
            
    public int getType() {
        return iType;
    }
    
    public ImageIcon getIcon() {
        return new ImageIcon( new String( stHomeDirectory + stIconFileName ), 
            new String( iiIcon.getDescription() ) );
    }
    
    public ImageIcon getMenuIcon() {
        if ( stMenuIconFileName != "" ) {
            return new ImageIcon( new String( stHomeDirectory + stMenuIconFileName ), 
                new String( iiMenuIcon.getDescription() ) );
        }
        else {
            return new ImageIcon( new String( stHomeDirectory + stIconFileName ), 
                new String( iiIcon.getDescription() ) );
        }
    }
              
    public Status setType( int iThisType ) {
        Status statInformation;
        String stMessage;
        boolean bValidType;

        bValidType = true;
        iType = iThisType;
        
        switch ( iType ) {
            case I_SYSTEM_TYPE_KEYBOARD:
                iTypeSize = 1;
                setArrayDimensions( 80, 1 );
                break;
                    
            case I_SYSTEM_TYPE_SINGLE_REGISTER:
                iTypeSize = 1;
                break;
                    
            case I_SYSTEM_TYPE_DOUBLE_REGISTER:
                iTypeSize = 2;
                break;
                    
            case I_SYSTEM_TYPE_CONDITION_CODE:
                iTypeSize = 0;
                break;
                    
            case I_SYSTEM_TYPE_DISPLAY:
                iTypeSize = 1;
                setArrayDimensions( 80, 1 );
                break;
                   
            case I_TYPE_LABEL:
                iTypeSize = 0;
                break;
                    
            case I_TYPE_CHARACTER:
                iTypeSize = 1;
                break;
                    
            case I_TYPE_INTEGER:
                iTypeSize = 4;
                break;
                  
            case I_TYPE_DECIMAL:
                iTypeSize = 15;  // verify text places in decimal!!
                break;
                    
            case I_TYPE_RECORD:
                iTypeSize = 0; //depends on types/structures of record fields
                break;
                
            default:
                iTypeSize = I_UNDEFINED;
                iType = iThisType;
                bValidType = false;
                break;
        }                   
            
        if ( bValidType ) {
            recalculateObjectSize();
            setDataObjectIcon();
            statInformation = new Status( I_OK );
        }
        else {
            stMessage = "Type code " + iThisType + " is not a type property.";
            statInformation = new Status( I_UNDECLARED_TYPE, stMessage );
        }                

        return statInformation;
    }

    public int getRestriction() {
        return iRestriction;
    }    
        
    public Status setRestriction( int iThisRestriction ) {
        String stMessage;
        Status statInformation;
        
        if ( ( iThisRestriction >= I_FIRST_RESTRICTION_INDEX )  &&
             ( iThisRestriction <= I_LAST_RESTRICTION_INDEX ) ) {
            iRestriction = iThisRestriction;
            statInformation = new Status( I_OK );
        }
        else {
            stMessage = "Type code " + iThisRestriction + " is not a restriction property.";
            statInformation = new Status ( I_UNDECLARED_TYPE, stMessage );
        }
        
        return statInformation;
    }
    
    public int getStructure() {
        return iStructure;
    }

    public StructureAndDimensionsWrapper getStructureAndDimensions() {
        return new StructureAndDimensionsWrapper( 
            DataObjectPropertyTextView.getStructureName( iStructure ), getArrayDimensions() );
    }

    public Status setStructure( int iThisStructure ) {
        String stMessage;
        Status statInformation;
        
        if ( ( iThisStructure >= I_FIRST_STRUCTURE_INDEX )  &&
             ( iThisStructure <= I_LAST_STRUCTURE_INDEX ) ) {
            iStructure = iThisStructure;
            adArraySize.setArrayDimensions( 1, 1 );
            recalculateObjectSize();
            setDataObjectIcon();
            statInformation = new Status( I_OK );            
        }
        else {
            stMessage = "Type code " + iThisStructure + " is not a structure property.";
            statInformation = new Status ( I_UNDECLARED_TYPE, stMessage );
        }
        
        return statInformation;
    }

    public ArrayDimensions getArrayDimensions() {
        return adArraySize.getArrayDimensions();
    }
    
    public Status setArrayDimensions( int iXSize, int iYSize ) {
        Status statInformation;
        
        statInformation = new Status( I_OK );
        statInformation = adArraySize.setArrayDimensions( iXSize, iYSize );
        
        if ( statInformation.getStatusNumber() == I_OK ) {
            recalculateObjectSize();
        }
            
        return statInformation;
    }
    
    public Object getValue() {
        Object objThisObject;
        String stMessage;
               
        if ( bValidatedValueFlag == true ) {
            objThisObject = new String( stValueString );
        }
        else {
            stMessage = "The value has not been validated.";
            objThisObject = new Status( I_UNDEFINED_PROPERTY, stMessage );        
        }
            
        return objThisObject;
    }
        
    // only single arrays have been correctly implemented
    public Object getValidatedValue( int iXLocation ) {
        Object objThisObject;
        int iArrayLocation;
        String stMessage;

        System.out.println( "DO.getValidatedValue...with index: " + iXLocation );
        if ( bValidatedValueFlag == true ) {
            System.out.println( "Value has been validated!" );
            validationTest();
            if ( ( iXLocation >= 0 )  &&  ( iXLocation <= adArraySize.getXDimension() ) ) {
                objThisObject = 
                    SerializedClone.cloneObject( alValidatedValues.get( iXLocation ) );
                objThisObject = objThisObject.toString();
            }
            else {
                stMessage = "A validated value does not exist at position " 
                    + iXLocation + ".";
                objThisObject = new Status( I_ILLEGAL_VALUE, stMessage );
            }
        }
        else {
            stMessage = "The value has not been validated.";
            objThisObject = new Status( I_UNDEFINED_PROPERTY, stMessage );        
        }
        return objThisObject;   
    }
    
    public Object getValidatedCharacterArray() {
        int iIndex;
        String stMessage, stCharacterArray;
        Object objThisObject;
        
        if ( bValidatedValueFlag == true ) {
            objThisObject = new String( "" );        
            /*for ( iIndex = 0; iIndex < adArraySize.getXDimension(); iIndex++ ) {
                objThisObject = objThisObject +
                    ((Character)alValidatedValues.get( iIndex )).toString();
            }*/
            
            //used string instead since char array operated upon as a unit
            objThisObject = (String)alValidatedValues.get( 0 );
        }
        else {
            stMessage = "The value has not been validated.";
            objThisObject = new Status( I_UNDEFINED_PROPERTY, stMessage );        
        }
        return objThisObject;   
    }
    
    // consider changing to receive a String and then cast the type
    // validate stuff later
    public Status setValue( String stValue ) {
        stValueString = stValue;
        bValidatedValueFlag = false;
        
        return new Status( I_OK );
    }
    
    public String getDescription() {
        return new String( stDescription );
    }
        
    public Status setDescription( String stThisDescription ) {
        stDescription = stThisDescription;
        
        return new Status( I_OK );
    }    
    
    public int getSize() {
        return iObjectSize;
    }
    
    public DataObject getProperties() {  
        return (DataObject)this.clone();
    }    

    public boolean isRecord() {
        if ( iType == I_TYPE_RECORD )
            return true;
        else
            return false;
    }
    
    public Object getRecordObject() {
        Object objThisObject;
        String stMessage;
        
        objThisObject = new Object();
        
        if ( roRecord.getRecordSize() <= 0 ) {
            stMessage = "This data object currently has no record data.";
            objThisObject = new Status( I_EMPTY_RECORD, stMessage );
        }
        else {
            objThisObject = (RecordObject)roRecord.clone();
        }
        
        return objThisObject;
    }

    public Status setRecordObject( RecordObject roThisRecord ) {
        roRecord = roThisRecord;
        iTypeSize = roRecord.getRecordSize();
        recalculateObjectSize();       
    
        return new Status( I_OK );
    }
    
    public Object getRecordObjectFieldNames() {
        // already cloned
        return roRecord.getFieldNames();
    }

    public Object getRecordObjectFieldObjects() {
        // already cloned
        return roRecord.getFieldObjects();
    }
    
    public boolean isFieldInRecordObject( String stRecordName ) {
        if ( roRecord.isFieldInRecord( stRecordName ) == true ) 
            return true;
        else
            return false;
    }
    
    public Status validateValue() {
        Status statInformation;
        String stMessage;
        Object objValidatedValue;
        
        statInformation = new Status( I_OK );
        stMessage = "";
        
        alValidatedValues = new ArrayList( adArraySize.getXDimension() );
        
        /*if ( iType == I_TYPE_CHARACTER ) {
            alValidatedValues = new ArrayList( adArraySize.getXDimension() );            
        }
        else {
            alValidatedValues = 
                new ArrayList( adArraySize.getXDimension() * adArraySize.getYDimension() );
        }*/       
        objValidatedValue = new Object();

        /* ALWAYS VALIDATE VALUE REGARDLESS OF FLAG!!! */
        //if ( bValidatedValueFlag == false ) {
            if ( ( iRestriction == I_RESTRICTION_CONSTANT )  ||
                 ( iRestriction == I_RESTRICTION_LITERAL ) ) {
//          if ( iRestriction == I_RESTRICTION_CONSTANT ) {
                if ( iStructure == I_STRUCTURE_SINGLE_VALUE ) {
                    if ( iType == I_TYPE_CHARACTER ) {
                        objValidatedValue = validateAndGetCharacter( stValueString );
                        if ( objValidatedValue instanceof Character ) {
                            alValidatedValues.add( (Character)objValidatedValue );
                            bValidatedValueFlag = true;
                        }
                        // else: error already in statInformation
                    }
                    else {
                        // check the number
                        objValidatedValue = validateAndGetNumberString( stValueString );
                        
                        if ( objValidatedValue instanceof Integer ) {
                            alValidatedValues.add( (Integer)objValidatedValue );
                            bValidatedValueFlag = true;
                        }
                        else if ( objValidatedValue instanceof Float ) {
                            alValidatedValues.add( (Float)objValidatedValue );
                            bValidatedValueFlag = true;
                        }
                        // else: error already in statInformation                        
                    }
                }
                else if ( ( iStructure == I_STRUCTURE_1D_ARRAY )  &&
                          ( iType == I_TYPE_CHARACTER ) ){
                    objValidatedValue = validateCharacterString( stValueString );

                    if ( objValidatedValue instanceof String ) {
                        alValidatedValues.add( (String)objValidatedValue );
                        bValidatedValueFlag = true;
                    }
                }
                else if ( ( iStructure == I_STRUCTURE_1D_ARRAY ) &&
                          ( iType == I_TYPE_INTEGER ) ) {                    
                    statInformation = validateNumberArray();
                    System.out.println( "MESSAGE from validate array = " +statInformation.getStatusMessage() );
                    if ( statInformation.getStatusNumber() == I_OK ) {
                        System.out.println( "Validated array" );
                        bValidatedValueFlag = true;
                    }
                    else {
                        objValidatedValue = 
                            (Status)SerializedClone.cloneObject( statInformation );
                    }
                }
            }
            // else alValidatedValues will be of size of ArrayDimensions
            /*            else {
                bValidatedValueFlag = true;
                objValidatedValue = new String( "" );
            }*/
        //}
    
        if ( bValidatedValueFlag == true ) {
            statInformation = new Status( I_OK );            
        }
        else {
            if ( objValidatedValue instanceof Status ) {
                statInformation = (Status)SerializedClone.cloneObject( objValidatedValue );
            }
            else {
                statInformation = new Status( I_ILLEGAL_VALUE, stMessage );                
            }
        }
        //System.out.println( "MESSAGE = " +statInformation.getStatusMessage() );
        return statInformation;
    }
    
    public Object writeDataObject() {
        return new Object();
    }
    
    public int getTypeSize() {
        return iTypeSize;
    }
    
    /////////////////////// FILE I/O METHODS \\\\\\\\\\\\\\\\\\\\\\\
    // Assumes no records used...dependent on Declarations to validate constants/literals  
    public Vector exportState() {
        Vector vData;
        
        vData = new Vector();
        vData.setSize( I_TOTAL_FIELD_COUNT );
        vData.set( I_TYPE, new Integer( iType ) );
        //System.out.println( "DATA OBJECT: iType=" + iType );
        vData.set( I_RESTRICTION, new Integer( iRestriction ) );
        //System.out.println( "DATA OBJECT: iRestriction=" + iRestriction );
        vData.set( I_STRUCTURE, new Integer( iStructure ) );
        //System.out.println( "DATA OBJECT: iStructure=" + iStructure );
        //vData.set( I_VALIDATED_VALUES, alValidatedValues );
        //System.out.println( "DATA OBJECT: alValidatedValues=" + alValidatedValues );
        vData.set( I_ARRAY_SIZE, adArraySize );
        //System.out.println( "DATA OBJECT: adArraySize=" + adArraySize );
        vData.set( I_DESCRIPTION, stDescription );
        //System.out.println( "DATA OBJECT: stDescription=" + stDescription );
        vData.set( I_VALUE_STRING, stValueString );        
        //System.out.println( "DATA OBJECT: stValueString=" + stValueString );
        //vData.set( I_VALIDATED_VALUE_FLAG, new Boolean( bValidatedValueFlag ) );
        //System.out.println( "DATA OBJECT: bValidatedValueFlag=" + bValidatedValueFlag );
        //System.out.println( "DATA OBJECT: vData=" + vData );
        return vData;
    }
    
    public void importState( Vector vData ) {
        ArrayDimensions adTemp;

        setType( ((Integer)vData.get( I_TYPE )).intValue() );
        iRestriction = ((Integer)vData.get( I_RESTRICTION )).intValue();
        setStructure( ((Integer)vData.get( I_STRUCTURE )).intValue() );
        //alValidatedValues = (ArrayList)vData.get( I_VALIDATED_VALUES );
        
        adTemp = (ArrayDimensions)vData.get( I_ARRAY_SIZE );
        setArrayDimensions( adTemp.getXDimension(), adTemp.getYDimension() );

        stDescription = (String)vData.get( I_DESCRIPTION );
        setValue( (String)vData.get( I_VALUE_STRING ) );
        //validateValue();        
        //bValidatedValueFlag = ((Boolean)vData.get( I_VALIDATED_VALUE_FLAG )).booleanValue();
        //validationTest();
    }
    
    /////////////////////// VALIDATION HELPER FUNCTIONS \\\\\\\\\\\\\\\\\\\
    public void validationTest() {
        Object objTemp;
        int i;
        char c;
        String s, sType;
                
        System.out.println( "VALIDATION TEST..." );
        System.out.println( "Number of elements = " + alValidatedValues.size() );
        for ( int iIndex = 0; iIndex < alValidatedValues.size(); iIndex++ ) {
            objTemp = alValidatedValues.get( iIndex );
            
            if ( objTemp instanceof Character ) {
                sType = "Character";
                s = ((Character)objTemp).toString();
            }
            else if ( objTemp instanceof Integer ) {
                sType = "Integer";
                s = ((Integer)objTemp).toString();
            }
            else if ( objTemp instanceof String ) {
                sType = "String";
                s = (String)objTemp;
            }
            else {
                sType = "??";
                s = "?";
            }
            System.out.println( "("+sType+")  Element #" + iIndex + " = " + s );            
        }
    }        
    
    private Status validateNumberArray() {        
        int iArrayLengthLocation, iArrayWidthLocation, iStartingPosition, 
            iEndingPosition, iLength, iCurrentPosition;
        boolean bContinue;
        Object objThisObject;
        Status statInformation;
        String stMessage, stOuterString;
        Integer intCurrentPosition;
        char chDelimiter;
                
        stMessage = "";
        objThisObject = new Object();       
        intCurrentPosition = new Integer( 0 );
        iLength = adArraySize.getXDimension();
        iCurrentPosition = 0;        
        stOuterString = "";

        iArrayLengthLocation = 0;
        iArrayWidthLocation = 0;
        statInformation = new Status( I_OK );
        
        objThisObject = parseArrayBrace( iCurrentPosition, '{' );
        
        if ( objThisObject instanceof Integer ) {
            intCurrentPosition = (Integer)objThisObject;
            iCurrentPosition = intCurrentPosition.intValue();
            bContinue = true;
            System.out.println( "Found opening brace" );
            
            while ( ( iArrayLengthLocation < iLength )  && 
                    ( bContinue == true ) ) {
                        
                if ( bContinue == true ) {
                    if ( iArrayLengthLocation + 1 == iLength ) {
                        chDelimiter = '}';
                    }
                    else {
                        chDelimiter = ',';
                    }

                    System.out.println( "CURRENT POSITION = " + iCurrentPosition );
                    objThisObject = parseArrayElement( iCurrentPosition, stOuterString, 
                        chDelimiter, iArrayLengthLocation, iArrayWidthLocation, 
                        B_FIND_ELEMENT );

                    if ( objThisObject instanceof Integer ) {
                        intCurrentPosition = (Integer)objThisObject;
                        iCurrentPosition = intCurrentPosition.intValue();                        
                        System.out.println( "At [" + iArrayLengthLocation+" , "+iArrayWidthLocation+ "]");
                        iArrayLengthLocation++;
                    }
                    else {
                        // invalid array element
                        statInformation = (Status)objThisObject;
                        bContinue = false;      
                        System.out.println( "Invalid outer array element." );
                    }
                }
                // else bContinue is false
            }
            // end of outer array loop
                        
            if ( bContinue == true ) {
                System.out.println( "CURRENT POSITION = " + iCurrentPosition );
                objThisObject = parseArrayBrace( iCurrentPosition, '}' );

                if ( objThisObject instanceof Integer ) {
                    statInformation = new Status( I_OK );
                    System.out.println("Found closing brace...valid array." );
                }
                else {      
                    //statInformation = (Status)(((Status)objThisObject).clone());
                    if ( objThisObject instanceof Status )
                        System.out.println( "***> Before recasting...MESSAGE = " +((Status)objThisObject).getStatusMessage() );
                    
                    statInformation = (Status)objThisObject;
                    //statInformation = new Status( ((Status)objThisObject).getStatusNumber(),
                        //((Status)objThisObject).getStatusMessage() );
                    //statInformation = (Status)SerializedClone.cloneObject( (Status)objThisObject );
                    System.out.println( "Did not find closing brace." );
                }
            }
        }
        else {
            // did not find outer opening brace
            if ( objThisObject instanceof Status )
                System.out.println( "***> Before recasting...MESSAGE = " +((Status)objThisObject).getStatusMessage() );
            //statInformation = new Status( ((Status)objThisObject).getStatusNumber(),
            //    ((Status)objThisObject).getStatusMessage() );

            statInformation = (Status)objThisObject;
            //statInformation = (Status)(((Status)objThisObject).clone());
            //statInformation = (Status)SerializedClone.cloneObject( (Status)objThisObject );
            System.out.println( "Did not find opening brace." );
        }
        
        //System.out.println( "Leaving validate array...MESSAGE = " +statInformation.getStatusMessage() );
        return statInformation;
    }
    
    private Object parseArrayElement( int iCurrentPosition, String stArraySpace,
        char chMarker, int iArrayLengthPosition, int iArrayWidthPosition,
        boolean bOnlyFindMarker ) {
            
        int iStartingPosition, iEndingPosition;
        String stToken, stMessage;
        Object objThisObject, objValidatedObject;
        boolean bValidObject, bContinue, bContinueSearching;
        
        bValidObject = false;
        objValidatedObject = new Object();
        objThisObject = new Object();
        iStartingPosition = iCurrentPosition;
        stMessage = "";        
        bContinueSearching = false;
        
        do {
            iEndingPosition = stValueString.indexOf( chMarker, iStartingPosition );
            
            if ( ( iEndingPosition - 1 ) >= 0 ) {
                if ( stValueString.charAt( iEndingPosition - 1 ) == '/' ) {
                    bContinueSearching = true;
                    iStartingPosition = iEndingPosition;
                }
            }
        } while ( ( bContinueSearching == true )  &&  ( iEndingPosition != I_NOT_FOUND ) );
                
        if ( iEndingPosition == I_NOT_FOUND ) {
            stMessage = "A comma is missing between array elements" + 
                " or there are not enough array elements.";
            objThisObject = new Status( I_MISSING_COMMA, stMessage );
            bContinue = false;
        }
        else {
            bContinue = true;
        }
        
        if ( bContinue == true ) {
            if ( bOnlyFindMarker == false ) {
                stToken 
                    = stValueString.substring( iStartingPosition, iEndingPosition );
                //System.out.println( "Start="+iStartingPosition+"   End="+(iEndingPosition-1)+"   Token=" + stToken );                

                if ( ( iType == I_TYPE_CHARACTER )  && 
                     ( iStructure != I_STRUCTURE_SINGLE_VALUE ) ) {
                    System.out.println(" VALIDATING CHARACTER STRING IN ARRAY..." );
                    objThisObject = validateCharacterString( stToken );

                    if ( objThisObject instanceof String ) {
                        bValidObject = true;
                        objValidatedObject = (String)objThisObject;
                    }
                    else {
                        stMessage = "The element " + stToken.trim() + " in the " 
                            + stArraySpace + " array is not of the character type "
                            + "or there are too many array elements.";
                        objThisObject = new Status( I_INCORRECT_TYPE, stMessage );
                    }                
                }
                else {
                    //System.out.println( "Validating token " + stToken );
                    objThisObject = validateAndGetNumberString( stToken );

                    if ( objThisObject instanceof Integer ) {
                        bValidObject = true;
                        objValidatedObject = (Integer)objThisObject;
                    }
                    else if ( objThisObject instanceof Float ) {
                        bValidObject = true;
                        objValidatedObject = (Float)objThisObject;                        
                    }
                    else {                                    
                        stMessage = "The element " + stToken.trim() + " in the " 
                            + stArraySpace + " array is not of the " 
                            + DataObjectPropertyTextView.getTypeName( iType )  
                            + " type or there are too many array elements.";
                        objThisObject = new Status( I_INCORRECT_TYPE, stMessage );                        objThisObject = new Status( I_INCORRECT_TYPE, stMessage );
                    }
                }
            } // only find marker
            else {
                objThisObject = new Integer( iEndingPosition + 1 );
            }
        }

        if ( bValidObject == true ) {            
            if ( ( iType == I_TYPE_CHARACTER )  && 
                 ( iStructure != I_STRUCTURE_SINGLE_VALUE ) ) { 
                alValidatedValues.add( iArrayLengthPosition, objValidatedObject );
            }
            else {
                alValidatedValues.add( iArrayLengthPosition + 
                    ( iArrayWidthPosition * adArraySize.getXDimension() ),
                    objValidatedObject );
            }
            
            if ( chMarker == '}' ) {
                objThisObject = new Integer( iEndingPosition );
            }
            else {
                objThisObject = new Integer( iEndingPosition + 1 );
            }
        }
        if ( objThisObject instanceof Status )
            System.out.println( "===> Leaving parse array element...MESSAGE = " +((Status)objThisObject).getStatusMessage() );        
        
        return SerializedClone.cloneObject( objThisObject );
    }
        
    private Object parseArrayBrace( int iCurrentPosition, char chBrace ) {            
        int iEndingPosition;
        boolean bContinue;
        char chThisCharacter;
        Object objThisObject;
        String stMessage;
        
        objThisObject = new Object();        
        bContinue = true;        
        iEndingPosition = stValueString.length();
        
        while ( bContinue == true ) {            
            chThisCharacter = stValueString.charAt( iCurrentPosition );
            
            if ( ( chThisCharacter == ' ' ) || 
                 ( chThisCharacter == '\t' ) ||
                 ( chThisCharacter == '\b' ) ) {
                iCurrentPosition++;
            }
            else if ( chThisCharacter == chBrace ) {
                objThisObject = new Integer( iCurrentPosition + 1 );
                bContinue = false;
            }
            else if ( chThisCharacter == ',' ) {
                stMessage = "The number of array elements is greater than "
                    + "the specified length.";
                objThisObject = new Status( I_ILLEGAL_ARRAY_CHARACTER, stMessage );
                bContinue = false;                        
            }
            else {                
                stMessage = "The array cannot have any characters except those "
                    + "that are relevant to the array and the array elements.";
                objThisObject = new Status( I_ILLEGAL_ARRAY_CHARACTER, stMessage );
                bContinue = false;
            }
        }        
        return objThisObject;
    }

    private Object validateAndGetNumberString( String stNumberString ) {
        Object objThisObject;
        Status statInformation;
        String stMessage;
        char ch;
        int iLength;
        boolean bValid;
        
        bValid = false;
        objThisObject = new Object();
        stNumberString = stNumberString.trim();
        //System.out.println( "Received token = " +stNumberString );
        
        try {
            stMessage = "***";
            if ( iType == I_TYPE_INTEGER ) {
                iLength = stNumberString.length();
                
                if ( iLength > 10 ) {
                    if ( iLength == 11 ) {
                        ch = stNumberString.charAt( 0 );

                        if ( ( ch == '+' )  ||  ( ch == '-' ) ) {
                            bValid = true;
                        }                        
                    }
                }
                else {
                    bValid = true;
                }
                    
                if ( bValid == true ) {
                    objThisObject = new Integer( stNumberString );
                }
                else {
                    stMessage = "The integer cannot be greater than 10 digits "
                        + "and 1 optional sign character.";
                    objThisObject = new Status( I_ILLEGAL_VALUE, stMessage );
                }
            }
            else if ( iType == I_TYPE_DECIMAL ) {
                objThisObject = new Float( stNumberString );
            }
            else {
                stMessage = stNumberString + " is not a valid number.";
                objThisObject = new Status( I_ILLEGAL_VALUE, stMessage );
            }
        }
        catch ( NumberFormatException nfeException ) {            
            if ( iType == I_TYPE_INTEGER ) {
                stMessage = stNumberString + " is not a valid integer number.";
            }
            else {
                stMessage = stNumberString + " is not a valid decimal number.";
            }
            objThisObject = new Status( I_ILLEGAL_VALUE, stMessage );
        }

        //System.out.println( "validateAndGetNumberString: " +stMessage );
        return objThisObject;
    }

    
    private Object validateAndGetCharacter( String stCharacters ) {
        Object objThisObject;
        Status statInformation;
        String stMessage;
        boolean bError, bOkToCheckAhead2Chars, bOkToCheckAhead3Chars;
        int iStartingPosition, iLastCharacterPosition;
        
        bError = false;
        bOkToCheckAhead2Chars = false;
        bOkToCheckAhead3Chars = false;
        objThisObject = new Object();
        System.out.println( "TOKEN: " + stCharacters );
        stCharacters = stCharacters.trim();
        System.out.println( "TRIMMED TOKEN: " + stCharacters );
        stMessage = "***";
        
        iLastCharacterPosition = stCharacters.length() - 1;
        iStartingPosition = stCharacters.indexOf( '\'' );        
        
        if ( ( iStartingPosition + 2 ) <= iLastCharacterPosition ) {
            bOkToCheckAhead2Chars = true;
        }
        
        if ( ( iStartingPosition + 3 ) <= iLastCharacterPosition ) {
            bOkToCheckAhead3Chars = true;
        }
            
        if ( iStartingPosition != I_NOT_FOUND ) {
            // case: '\?'
            //       0123
            if ( ( bOkToCheckAhead3Chars == true )  && 
                 ( stCharacters.charAt( iStartingPosition + 3 ) == '\'' ) ) {
                if ( stCharacters.charAt( iStartingPosition + 1 ) == '/' ) {                
                    // valid character
                    objThisObject = 
                        new Character( stCharacters.charAt( iStartingPosition + 2 ) );
                }
                else {
                    // missing end quoted
                    stMessage = "Closing ' (single quote) missing from "
                        + "character constant.";                    
                    bError = true;
                }
            }
            // case: '?'
            //       012
            else if ( ( bOkToCheckAhead2Chars == true )  && 
                      ( stCharacters.charAt( iStartingPosition + 2 ) == '\'' ) ) {
                // valid character
                objThisObject = 
                    new Character( stCharacters.charAt( iStartingPosition + 1 ) );
            }
            // case: ''
            //       01
            else if ( stCharacters.charAt( iStartingPosition + 1 ) == '\'' ) {
                // valid character
                objThisObject = new Character( CH_EMPTY );
            }
            else {
                stMessage = "Closing ' (single quote) missing from character constant.";
                bError = true;      
            }
        }
        else {
            stMessage = "Opening ' (single quote) missing from character constant.";
            bError = true;
        }
        
        if ( bError == true ) {
            stMessage = stCharacters + " is not a valid character type value.";
            objThisObject = new Status( I_ILLEGAL_VALUE, stMessage );
        }
        //System.out.println( "validateAndGetCharacter: " +stMessage );
        return objThisObject;
    }
    
    private Object validateCharacterString( String stCharacterArray ) {
        Object objThisObject;
        String stMessage, stModifiedString;
        StringBuffer sbStringArrayBuffer;
        Status statInformation;
        int iCurrentPosition, iStringLength, iArraySize, iActualStringLength;
        char chThisCharacter;
        boolean bError, bEndQuote, bHavePreviousSlash;
        
        sbStringArrayBuffer = new StringBuffer();
        stMessage = "";
        
        if ( iStructure == I_STRUCTURE_1D_ARRAY ) {
            iArraySize = adArraySize.getXDimension();
        }
        else {
            iArraySize = adArraySize.getYDimension();
        }
        
        stCharacterArray = stCharacterArray.trim();
        iActualStringLength = 0;
        iCurrentPosition = 1;
        
        if ( stCharacterArray.charAt( 0 ) == '"' ) {
            iStringLength = stCharacterArray.length();
            
            bEndQuote = false;
            bHavePreviousSlash = false;
            
            while ( ( bEndQuote == false )  &&  
                    ( iCurrentPosition < iStringLength ) ) {
                chThisCharacter = stCharacterArray.charAt( iCurrentPosition );
                
                if ( bHavePreviousSlash == true ) {
                    sbStringArrayBuffer.append( chThisCharacter );
                    bHavePreviousSlash = false;
                }
                else if ( chThisCharacter == '"' ) {
                    bEndQuote = true;
                }
                else if ( chThisCharacter == '/' ) {
                    bHavePreviousSlash = true;
                }
                else {
                    sbStringArrayBuffer.append( chThisCharacter );
                    bHavePreviousSlash = false;
                }
                iCurrentPosition++;
            }
            
            if ( bEndQuote == true ) {                
                iActualStringLength = sbStringArrayBuffer.length();

                //System.out.println( "In DO, restriction=" + DataObjectPropertyTextView.getRestrictionName( iRestriction ) );
                if ( iRestriction != I_RESTRICTION_LITERAL ) {
                    if ( iActualStringLength == iArraySize ) {
                        // ok
                        bError = false;
                    }
                    else if ( iActualStringLength < iArraySize ) {
                        // string too short ( string length != array width )
                        stMessage = "The length of the character array is less than " 
                            + "the specified length.";
                        bError = true;
                    }                                
                    else {
                        // string too long ( string length != array width )
                        stMessage = "The length of the character array is greater than " 
                            + "the specified length.";
                        bError = true;
                    }
                }
                else {
                    iArraySize = iActualStringLength;
                    System.out.println( "Original length = " + iArraySize );
                    stModifiedString = 
                        AssemblyProgramView.adjustConstantForSpecialCharacters(
                            sbStringArrayBuffer.toString(),
                            AssemblyProgramView.B_VERIFYING_LENGTH );
                    
                    System.out.println( "New size = " + stModifiedString.length() );
                    
                    if ( stModifiedString.length() > 35 ) {
                        stMessage = "The character array cannot exceed 35 "
                            + "characters including duplicated characters ' and &.";
                        bError = true;
                    }
                    else {
                        setArrayDimensions( iArraySize, 1 );
                        bError = false;
                    }
                }
            }
            else {
                // missing closing quotes
                stMessage = "Closing double quotes (\") is missing from the character array.";
                bError = true;
            }
        }
        else {
            // missing opening quotes
            stMessage = "Opening double quotes (\") is missing from the character array.";
            bError = true;
        }
        
        if ( bError == true ) {
            objThisObject = new Status( I_ILLEGAL_VALUE, stMessage );
        }
        else {
            objThisObject = new String( sbStringArrayBuffer );
        }
        
        if ( objThisObject instanceof String ) {
            System.out.println( "Leaving from String with: " + (String)objThisObject );
        }
        else {
            System.out.println( "Leaving from String with ERROR: " + ((Status)objThisObject).getStatusMessage() );
        }
        return objThisObject;
    }
    
    ////////////////// PREPARATION FUNCTIONS \\\\\\\\\\\\\\\\\\\\\
    private void recalculateObjectSize() {
        int iX, iY;
        
        iX = adArraySize.getXDimension();
        //System.out.println( "Array size = " + iX );
        iY = adArraySize.getYDimension();
        iObjectSize = iTypeSize * iX * iY;        
    }
    
    private void setDataObjectIcon() {
        switch ( iType ) {
            case I_SYSTEM_TYPE_KEYBOARD:
                iiIcon = iiKeyboard;
                stIconFileName= "keyboard-icon.jpg";
                iiMenuIcon = iiMenuKeyboard;
                stMenuIconFileName = "keyboard-icon-16.jpg";
                break;
                    
            case I_SYSTEM_TYPE_SINGLE_REGISTER:
                iiIcon = iiSingleRegister;
                stIconFileName= "single-register-icon.jpg";
                iiMenuIcon = iiMenuSingleRegister;
                stMenuIconFileName = "single-register-icon-16.jpg";
                break;
                    
            case I_SYSTEM_TYPE_DOUBLE_REGISTER:
                iiIcon = iiDoubleRegister;
                stIconFileName= "double-register-icon.jpg";
                break;
                    
            case I_SYSTEM_TYPE_CONDITION_CODE:
                iiIcon = iiConditionCode;
                stIconFileName= "condition-code-icon.jpg";
                break;
                    
            case I_SYSTEM_TYPE_DISPLAY:
                iiIcon = iiDisplay;
                stIconFileName= "display-icon.jpg";
                iiMenuIcon = iiMenuDisplay;
                stMenuIconFileName = "display-icon-16.jpg";
                break;
                   
            case I_TYPE_LABEL:
                iiIcon = iiLabel;
                stIconFileName= "label-icon.jpg";
                iiMenuIcon = iiMenuLabel;
                stMenuIconFileName = "label-icon-16.jpg";                
                break;
                    
            case I_TYPE_CHARACTER:
                switch( iStructure ) {            
                    case I_STRUCTURE_SINGLE_VALUE:
                        iiIcon = iiSingleCharacter;
                        stIconFileName= "char-single-icon.jpg";
                        iiMenuIcon = iiMenuSingleCharacter;
                        stMenuIconFileName = "char-single-icon-16.jpg";                                        
                        break;

                    case I_STRUCTURE_1D_ARRAY:
                        iiIcon = ii1dArrayCharacter;
                        stIconFileName= "char-array1d-icon.jpg";
                        iiMenuIcon = iiMenu1dArrayCharacter;
                        stMenuIconFileName = "char-array1d-icon-16.jpg";                                        
                        break;

                    case I_STRUCTURE_2D_ARRAY:
                        iiIcon = ii2dArrayCharacter;
                        stIconFileName= "char-array2d-icon.jpg";
                        break;

                    default:
                        iiIcon = iiUnknown;
                        stIconFileName= "unknown-icon.jpg";
                        break;
                }      
                break;
                    
            case I_TYPE_INTEGER:
                switch( iStructure ) {            
                    case I_STRUCTURE_SINGLE_VALUE:
                        iiIcon = iiSingleInteger;
                        stIconFileName= "int-single-icon.jpg";
                        iiMenuIcon = iiMenuSingleInteger;
                        stMenuIconFileName= "int-single-icon-16.jpg";                        
                        break;

                    case I_STRUCTURE_1D_ARRAY:
                        iiIcon = ii1dArrayInteger;
                        stIconFileName= "int-array1d-icon.jpg";
                        iiMenuIcon = iiMenu1dArrayInteger;
                        stMenuIconFileName= "int-array1d-icon-16.jpg";                        
                        break;

                    case I_STRUCTURE_2D_ARRAY:
                        iiIcon = ii2dArrayInteger;
                        stIconFileName= "int-array2d-icon.jpg";
                        break;

                    default:
                        iiIcon = iiUnknown;
                        stIconFileName= "unknown-icon.jpg";
                        break;
                }      
                break;
                  
            case I_TYPE_DECIMAL:
                switch( iStructure ) {            
                    case I_STRUCTURE_SINGLE_VALUE:
                        iiIcon = iiSingleDecimal;
                        stIconFileName= "decimal-single-icon.jpg";
                        iiMenuIcon = iiMenuSingleDecimal;
                        stMenuIconFileName= "decimal-single-icon-16.jpg";                        
                        break;

                    case I_STRUCTURE_1D_ARRAY:
                        iiIcon = ii1dArrayDecimal;
                        stIconFileName= "decimal-array1d-icon.jpg";
                        iiMenuIcon = iiMenu1dArrayDecimal;
                        stMenuIconFileName= "decimal-array1d-icon-16.jpg";                        
                        break;

                    case I_STRUCTURE_2D_ARRAY:
                        iiIcon = ii2dArrayDecimal;
                        stIconFileName= "decimal-array2d-icon.jpg";
                        break;

                    default:
                        iiIcon = iiUnknown;
                        stIconFileName= "unknown-icon.jpg";
                        break;
                }      
                break;
                    
            case I_TYPE_RECORD:
                switch( iStructure ) {            
                    case I_STRUCTURE_SINGLE_VALUE:
                        iiIcon = iiSingleRecord;
                        stIconFileName= "record-single-icon.jpg";
                        break;

                    case I_STRUCTURE_1D_ARRAY:
                        iiIcon = ii1dArrayRecord;
                        stIconFileName= "record-array1d-icon.jpg";
                        break;

                    case I_STRUCTURE_2D_ARRAY:
                        iiIcon = ii2dArrayRecord;
                        stIconFileName= "record-array2d-icon.jpg";
                        break;

                    default:
                        iiIcon = iiUnknown;
                        stIconFileName= "unknown-icon.jpg";
                        break;
                }      
                break;
                
            default:
                iiIcon = iiUnknown;
                stIconFileName= "";
                break;
        }
    }
    
    /************* DEBUG FUNCTION **********/
    public void printDataObject() {
            //System.out.println( "Icon Address " + getIcon() );
            System.out.println( "Type: " + DataObjectPropertyTextView.getTypeName( iType ) );
            System.out.println( "Restriction: " + DataObjectPropertyTextView.getRestrictionName( iRestriction ) );
            System.out.println( "Structure: " + DataObjectPropertyTextView.getStructureName( iStructure ) );
            System.out.println( "Size: " + getSize() );
    }
    /*********** END OF DEBUGGING FUNCTIONS *********/
    
    
    // instance objects
    private int iObjectSize, iTypeSize, iType, iRestriction, iStructure;
    private ArrayList alValidatedValues;
    private ImageIcon iiIcon, iiMenuIcon;
    private RecordObject roRecord;
    private ArrayDimensions adArraySize;
    private String stDescription, stValueString, stIconFileName, stMenuIconFileName;
    private boolean bValidatedValueFlag;
    
    
    //class constants
        private final static int I_TYPE = 0;
        private final static int I_RESTRICTION = 1;
        private final static int I_STRUCTURE = 2;
        //private final static int I_VALIDATED_VALUES = 3;
        private final static int I_ARRAY_SIZE = 3;
        private final static int I_DESCRIPTION = 4;
        private final static int I_VALUE_STRING = 5;
        //private final static int I_VALIDATED_VALUE_FLAG = 6;
        private final static int I_TOTAL_FIELD_COUNT = 6;
    
        public final static int I_FIRST_NON_SYSTEM_TYPE_INDEX = 1;
        public final static int I_TYPE_CHARACTER = 1;        
        public final static int I_TYPE_INTEGER = 2;
        public final static int I_TYPE_LABEL = 3;
        public final static int I_LAST_NON_SYSTEM_TYPE_INDEX = 3;
        public final static int I_TYPE_DECIMAL = 4;
        public final static int I_TYPE_RECORD = 5;
        
        public final static int I_FIRST_SYSTEM_TYPE_INDEX = 6;
        public final static int I_SYSTEM_TYPE_CONDITION_CODE = 6;
        public final static int I_SYSTEM_TYPE_DISPLAY = 7;
        public final static int I_SYSTEM_TYPE_SINGLE_REGISTER = 8;
        public final static int I_SYSTEM_TYPE_DOUBLE_REGISTER = 9;
        public final static int I_SYSTEM_TYPE_KEYBOARD = 10;        
        public final static int I_LAST_SYSTEM_TYPE_INDEX = 10;
        
        public final static int I_PROPERTY_RESTRICTION = 20;
        public final static int I_FIRST_RESTRICTION_INDEX = 21;
        public final static int I_RESTRICTION_CONSTANT = 21;
        public final static int I_RESTRICTION_VARIABLE = 22;
        public final static int I_RESTRICTION_LITERAL = 23;        
        public final static int I_LAST_RESTRICTION_INDEX = 23;

        public final static int I_PROPERTY_STRUCTURE = 30;
        public final static int I_FIRST_STRUCTURE_INDEX = 31;
        public final static int I_STRUCTURE_SINGLE_VALUE = 31;
        public final static int I_STRUCTURE_1D_ARRAY = 32;        
        public final static int I_LAST_STRUCTURE_INDEX = 32;
        public final static int I_STRUCTURE_2D_ARRAY = 33;
        
        public final static int I_UNDEFINED = -100;
        public final static String ST_UNDEFINED = "";
        private final static String stHomeDirectory;
        
        private final static boolean B_LEGAL_PROPERTY_TYPE = true;
        private final static boolean B_ILLEGAL_PROPERTY_TYPE = false;
        private final static boolean B_FIND_ELEMENT = false;
        private final static boolean B_FIND_MARKER = true;
  
        private final static char CH_EMPTY = '\u0000';
        // type icons
        private final static ImageIcon iiSingleCharacter, ii1dArrayCharacter, 
            ii2dArrayCharacter, iiSingleInteger, ii1dArrayInteger, ii2dArrayInteger,
            iiSingleDecimal, ii1dArrayDecimal, ii2dArrayDecimal, iiSingleRecord,
            ii1dArrayRecord, ii2dArrayRecord, iiKeyboard, iiSingleRegister,
            iiDoubleRegister, iiConditionCode, iiDisplay, iiLabel, iiUnknown,
            iiMenuSingleCharacter, iiMenu1dArrayCharacter, iiMenuSingleDecimal,
            iiMenu1dArrayDecimal, iiMenuSingleInteger, iiMenu1dArrayInteger,
            iiMenuLabel, iiMenuDisplay, iiMenuKeyboard, iiMenuSingleRegister;
                        
        // error codes
        public final static int I_OK = 0;
        public final static int I_UNDECLARED_TYPE = -1;
        public final static int I_ILLEGAL_SIZE = -2;
        public final static int I_ILLEGAL_VALUE = -3;
        public final static int I_UNDEFINED_PROPERTY = -4;
        public final static int I_EMPTY_RECORD = -5;
        public final static int I_INCORRECT_TYPE = -6;
        public final static int I_MISSING_COMMA = -7;
        public final static int I_ILLEGAL_ARRAY_CHARACTER = -8;
                
        private final static int I_NOT_FOUND = -1;
        
        // class initialization
        static {            
            stHomeDirectory = "";/*D:\\forte4j\\Development\\panelsAndTools\\";*/
            
            iiSingleCharacter = 
                new ImageIcon( stHomeDirectory + "char-single-icon.jpg", 
                    "Single Character" );
            ii1dArrayCharacter =
                new ImageIcon( stHomeDirectory + "char-array1d-icon.jpg", 
                    "1-D Character" );
            ii2dArrayCharacter = 
                new ImageIcon( stHomeDirectory + "char-array2d-icon.jpg", 
                    "2-D Character" );            
            iiSingleInteger =
                new ImageIcon( stHomeDirectory + "int-single-icon.jpg", 
                    "Single Integer" );
            ii1dArrayInteger =
                new ImageIcon( stHomeDirectory + "int-array1d-icon.jpg", 
                    "1-D Integer" );
            ii2dArrayInteger =
                new ImageIcon( stHomeDirectory + "int-array2d-icon.jpg", 
                    "2-D Integer" );
            iiSingleDecimal = 
                new ImageIcon( stHomeDirectory + "decimal-single-icon.jpg", 
                    "Single Decimal" );
            ii1dArrayDecimal =
                new ImageIcon( stHomeDirectory + "decimal-array1d-icon.jpg", 
                    "1-D Decimal" );
            ii2dArrayDecimal =
                new ImageIcon( stHomeDirectory + "decimal-array2d-icon.jpg",
                    "2-D Decimal" );
            iiSingleRecord =
                new ImageIcon( stHomeDirectory + "record-single-icon.jpg",
                    "Single Record" ); 
            ii1dArrayRecord =
                new ImageIcon( stHomeDirectory + "record-array1d-icon.jpg",
                    "1-D Record" );
            ii2dArrayRecord =
                new ImageIcon( stHomeDirectory + "record-array2d-icon.jpg",
                    "2-D Record" );
            iiKeyboard = 
                new ImageIcon( stHomeDirectory + "keyboard-icon.jpg",
                    "Keyboard Input" );
            iiSingleRegister =
                new ImageIcon( stHomeDirectory + "single-register-icon.jpg",
                    "Single Register" );
            iiDoubleRegister =
                new ImageIcon( stHomeDirectory + "double-register-icon.jpg",
                    "Double Register" );
            iiConditionCode =
                new ImageIcon( stHomeDirectory + "condition-code-icon.jpg",
                    "Condition Code" );
            iiDisplay = 
                new ImageIcon( stHomeDirectory + "display-icon.jpg",
                    "Display Output" );
            iiLabel = 
                new ImageIcon( stHomeDirectory + "label-icon.jpg", "Label" );
            iiUnknown = 
                new ImageIcon( stHomeDirectory + "unknown-icon.jpg", "?" );
            
            iiMenuSingleCharacter = 
                new ImageIcon( stHomeDirectory + "char-single-icon-16.jpg", 
                    "Single Character" );
            iiMenu1dArrayCharacter =
                new ImageIcon( stHomeDirectory + "char-array1d-icon-16.jpg", 
                    "1-D Character" );                
            iiMenuSingleDecimal = 
                new ImageIcon( stHomeDirectory + "decimal-single-icon-16.jpg", 
                    "Single Decimal" );
            iiMenu1dArrayDecimal =
                new ImageIcon( stHomeDirectory + "decimal-array1d-icon-16.jpg", 
                    "1-D Decimal" );                
            iiMenuSingleInteger = 
                new ImageIcon( stHomeDirectory + "int-single-icon-16.jpg", 
                    "Single Integer" );
            iiMenu1dArrayInteger =
                new ImageIcon( stHomeDirectory + "int-array1d-icon-16.jpg", 
                    "1-D Integer" );                
            iiMenuLabel = 
                new ImageIcon( stHomeDirectory + "label-icon-16.jpg", 
                    "Label" );
            iiMenuDisplay =
                new ImageIcon( stHomeDirectory + "display-icon-16.jpg", 
                    "Display Output" );                
            iiMenuKeyboard = 
                new ImageIcon( stHomeDirectory + "keyboard-icon-16.jpg", 
                    "Keyboard Input" );
            iiMenuSingleRegister =
                new ImageIcon( stHomeDirectory + "single-register-icon-16.jpg", 
                    "Single Register" );                
        }
}