/*
 * Declarations.java
 *
 * Created on January 31, 2002, 2:04 PM
 */

//package panelsAndTools;

import java.util.*;
import java.lang.*;
import java.io.*;

/**
 *
 * @author  _
 * @version 
 */

public class Declarations implements Cloneable, Serializable {        
    public Declarations() {
        roGlobalObjects = new RecordObject();
        alRecordObjectNames = new ArrayList();
        alNonRecordObjectNames = new ArrayList();
        stProgramName = "";
    }
    
    public void clearData() {
        String stName;
        Status statInformation;
        int iIndex, iNumberOfNonRecords, iNumberOfRecords;
        ArrayList alDuplicatedRecordNames, alDuplicatedNonRecordNames;
        
        iNumberOfNonRecords = alNonRecordObjectNames.size();
        iNumberOfRecords = alRecordObjectNames.size();
        
        alDuplicatedRecordNames = new ArrayList( alRecordObjectNames );
        alDuplicatedNonRecordNames = new ArrayList( alNonRecordObjectNames );
        
        System.out.println( "INSIDE...Declarations.clearData " );
        for ( iIndex = 0; iIndex < iNumberOfRecords; iIndex++ ) {
            stName = (String)alDuplicatedRecordNames.get( iIndex );
            System.out.println( "Deleting record object " + stName + "  ( "
                + iIndex + " of " + iNumberOfRecords );
            statInformation = removeDeclaration( stName );
            System.out.println( "Result...status = " + statInformation.getStatusNumber() 
                + "   message = " + statInformation.getStatusMessage() );
        }

        for ( iIndex = 0; iIndex < iNumberOfNonRecords; iIndex++ ) {
            stName = (String)alDuplicatedNonRecordNames.get( iIndex );
            System.out.println( "Deleting non-record object " + stName + "  ( "
                + iIndex + " of " + iNumberOfNonRecords );
            statInformation = removeDeclaration( stName );
            System.out.println( "Result...status = " + statInformation.getStatusNumber() 
                + "   message = " + statInformation.getStatusMessage() );
        }
        stProgramName = "";        
    }        
    
    protected Object clone() {
        // strucutre based on Sun's online Java tutorial for
        // the Object.clone() method and pp.235-6 C of Core Java 2: ISBN 0-13-081933-6        
        try {
            Declarations decObject = (Declarations)super.clone();
            decObject.roGlobalObjects = (RecordObject)this.roGlobalObjects.clone();
            decObject.alRecordObjectNames = (ArrayList)this.alRecordObjectNames.clone();
            decObject.alNonRecordObjectNames = 
                (ArrayList)this.alNonRecordObjectNames.clone();
            decObject.stProgramName = new String( stProgramName );
            return decObject;
        }
        catch ( CloneNotSupportedException cnseException ) {
            System.out.println( "Clone of Declarations is not supported." );
            return null;
        }
    }
    
    public void setProgramName( String stName ) {
        stProgramName = stName;
    }
    
    public String getProgramName() {
        return stProgramName;
    }
            
    public Status addDeclaration( String stObjectName, 
                                  DataObject doThisDataObject ) {
        
        ArrayList alNamesToCheck, alRecordNamesOfDataObject;
        boolean bObjectRecordStatus;
        Object objThisObject;
        Status statInformation;
        String stMessage;

        if ( doThisDataObject.getType() == DataObject.I_TYPE_LABEL ) {
            statInformation = new Status( I_OK );
        }
        else {
            if ( doThisDataObject.getRestriction() == DataObject.I_RESTRICTION_VARIABLE ) {
                statInformation = new Status( I_OK );
            }
            else {
                statInformation = doThisDataObject.validateValue();
            }
        }

        if ( statInformation.getStatusNumber() == I_OK ) {
            alNamesToCheck = new ArrayList();
            alNamesToCheck.add( stObjectName );         
            objThisObject = doThisDataObject.getRecordObjectFieldNames();

            if ( objThisObject instanceof ArrayList ) {
                alRecordNamesOfDataObject = (ArrayList)objThisObject;
                alNamesToCheck.addAll( alRecordNamesOfDataObject );
            }

            statInformation = 
                checkNamesInDataObjects( alNamesToCheck, B_LIST_CONTAINS_DATA_OBJECT_NAME );

            if ( statInformation.getStatusNumber() == I_OK ) {
                statInformation = roGlobalObjects.addField( stObjectName, doThisDataObject );

                if ( statInformation.getStatusNumber() == I_OK ) {
                    bObjectRecordStatus = doThisDataObject.isRecord();

                    updateNameLists( stObjectName, ST_NOTHING, I_ADD_NAME_TO_LIST, 
                        bObjectRecordStatus );

                    System.out.println( "Declarations: Adding a declaration..." );
                    delDeclarationsListener.addedOrUpdatedDeclaration( stObjectName,
                        doThisDataObject, DeclarationsListener.B_ADD_OBJECT );
                    plistProgramListener.programChanged();
                }
                else if ( statInformation.getStatusNumber() == RecordObject.I_FIELD_DECLARED ) {
                    stMessage = "The object named " + stObjectName + " has already "
                        + "been declared by another object.";
                    statInformation = new Status( I_OBJECT_DECLARED, stMessage );
                }
                // else probably an invalid object name
                // ---OR---
                // else...object was successfully added into global scope
                // either way --> statInformation contains the information
            }
            // else...error comes from checkNamesInDataObjects
        }
        // else...error comes from validateValue
        System.out.println( "Leaving addDeclaration with status code " +
            statInformation.getStatusNumber() + " and message " + statInformation.getStatusMessage() );
        
        return statInformation;
    }
    
    public Status removeDeclaration( String stObjectName ) {
        Status statInformation;
        String stMessage, stThisRecordFieldName;
        Object objThisObject;
        DataObject doThisDataObject, doRecordFieldDataObject;
        ArrayList alRecordFieldNames;
        int iNumberOfRecordFieldNames, iRecordFieldIndex;
        boolean bObjectRecordStatus;
        RecordObject roThisRecordObject;

        objThisObject = roGlobalObjects.getField( stObjectName );
        statInformation = new Status( I_OK );
        
        if ( objThisObject instanceof DataObject ) {
            doThisDataObject = (DataObject)objThisObject;
            
            bObjectRecordStatus = doThisDataObject.isRecord();

            if ( alRecordObjectNames.contains( stObjectName ) ) {
                // move record field objects to global scope...
                // already checked for no duplication when fields
                // were added to the record object, so should be ok
                
                objThisObject = doThisDataObject.getRecordObjectFieldNames();
                
                if ( objThisObject instanceof ArrayList ) {

                    alRecordFieldNames = (ArrayList)objThisObject;
                    iNumberOfRecordFieldNames = alRecordFieldNames.size();
                    roThisRecordObject = (RecordObject)
                        doThisDataObject.getRecordObject();
                    
                    for ( iRecordFieldIndex = 0; 
                          iRecordFieldIndex < iNumberOfRecordFieldNames;
                          iRecordFieldIndex++ ) {
                        
                        stThisRecordFieldName = (String)
                            alRecordFieldNames.get( iRecordFieldIndex );
                        doRecordFieldDataObject = (DataObject)
                            roThisRecordObject.getField( stThisRecordFieldName );
                        roThisRecordObject.removeField( stThisRecordFieldName );
                        roGlobalObjects.addField( stThisRecordFieldName,
                            doRecordFieldDataObject );

                        bObjectRecordStatus = doRecordFieldDataObject.isRecord();
                        updateNameLists( stThisRecordFieldName, ST_NOTHING,
                            I_ADD_NAME_TO_LIST, bObjectRecordStatus );
                        
                        delDeclarationsListener.addedOrUpdatedDeclaration( 
                            stThisRecordFieldName, doRecordFieldDataObject,
                            DeclarationsListener.B_ADD_OBJECT );
                        plistProgramListener.programChanged();
                    }
                                        
                    roGlobalObjects.removeField( stObjectName );
                    updateNameLists( stObjectName, ST_NOTHING, 
                        I_REMOVE_NAME_FROM_LIST, bObjectRecordStatus );

                    delDeclarationsListener.removedDeclaration( 
                        stObjectName, doThisDataObject );
                    plistProgramListener.programChanged();
                }
                else if ( objThisObject instanceof Status ) {
                    statInformation = (Status)objThisObject;                    
                }
                else {
                    stMessage = "removeDeclaration: Failed to receive Status or "
                        + "ArrayList object.";
                    statInformation = new Status( I_UNKNOWN, stMessage );
                }
            }
            else {
                roGlobalObjects.removeField( stObjectName );
                updateNameLists( stObjectName, ST_NOTHING, 
                    I_REMOVE_NAME_FROM_LIST, bObjectRecordStatus );
                delDeclarationsListener.removedDeclaration( 
                        stObjectName, doThisDataObject );
                plistProgramListener.programChanged();
            }                        
        }
        else if ( objThisObject instanceof Status ) {
            statInformation = (Status)objThisObject;
            
            if ( statInformation.getStatusNumber() == RecordObject.I_FIELD_UNDECLARED ) {
                stMessage = "The object named " + stObjectName + " has not "
                    + "been declared.";
                statInformation = new Status( I_OBJECT_UNDECLARED, stMessage );
            }
            else {
                stMessage = "removeDeclaration: Failed to receive "
                    + "I_FIELD_UNDECLARED.";
                statInformation = new Status( I_UNKNOWN, stMessage );
            }
        }
        else {
            stMessage = "removeDeclaration: Failed to receive Status or "
                + "DataObject object.";
            statInformation = new Status( I_UNKNOWN, stMessage );
        }
        
        System.out.println( "Leaving removeDeclaration with status code " +
            statInformation.getStatusNumber() + " and message " + statInformation.getStatusMessage() );
        return statInformation;
    }
    
    // object name is modified
    public Status setDeclarationName( String stOldObjectName, 
                                      String stNewObjectName ) {
        Status statInformation;
        String stMessage;
        Object objThisObject;
        DataObject doThisDataObject;
        ArrayList alNamesToCheck;
        boolean bObjectRecordStatus;
        int iStatusNumber, iObjectIndex;
                
        objThisObject = roGlobalObjects.getField( stOldObjectName );

        if ( objThisObject instanceof DataObject ) {
            doThisDataObject = (DataObject)objThisObject;
            alNamesToCheck = new ArrayList();

            alNamesToCheck.add( stNewObjectName );
            
            statInformation = 
                checkNamesInDataObjects( alNamesToCheck, 
                    B_LIST_CONTAINS_DATA_OBJECT_NAME );
                        
            if ( statInformation.getStatusNumber() == I_OK ) {                                
                statInformation = 
                    roGlobalObjects.setFieldName( stOldObjectName, stNewObjectName );
                iStatusNumber = statInformation.getStatusNumber();
                
                if ( iStatusNumber == I_OK ) {
                    bObjectRecordStatus = doThisDataObject.isRecord();
                    updateNameLists( stOldObjectName, stNewObjectName,
                        I_UPDATE_NAME_IN_LIST, bObjectRecordStatus );

                    delDeclarationsListener.updatedDeclarationName( 
                        stOldObjectName, stNewObjectName, doThisDataObject );
                    plistProgramListener.programChanged();
                }
                else if ( iStatusNumber == RecordObject.I_FIELD_DECLARED ) { 
                    stMessage = "The object named " + stNewObjectName + " has already "
                        + "been declared.";
                    statInformation = new Status( I_OBJECT_DECLARED, stMessage );
                }
                else if ( iStatusNumber == RecordObject.I_FIELD_UNDECLARED ) {
                    stMessage = "The object named " + stOldObjectName + " has already "
                        + "been declared.";
                    statInformation = new Status( I_OBJECT_DECLARED, stMessage );
                }
                // else statInformation contains needed information                    
            }
            // else statInformation already contains error info
        }
        else if ( objThisObject instanceof Status ) {
            statInformation = (Status)objThisObject;
            
            if ( statInformation.getStatusNumber() == RecordObject.I_FIELD_UNDECLARED ) {
                stMessage = "The object named " + stOldObjectName + " has not "
                    + "been declared.";
                statInformation = new Status( I_OBJECT_UNDECLARED, stMessage );
            }
            else {
                statInformation = new Status( I_UNKNOWN, "Could not change name." );
            }
        }
        else {
            stMessage = "setDeclarationName: Failed to receive Status or "
                + "DataObject object.";
            statInformation = new Status( I_UNKNOWN, stMessage );
        }
        System.out.println( "Leaving setDeclarationName with status code " +
            statInformation.getStatusNumber() + " and message " + statInformation.getStatusMessage() );
        
        return statInformation;
    }
    
    // object properties are modified
    public Status setDeclarationProperties( String stObjectName,
                                            DataObject doNewDataObject ) {

        ArrayList alNamesToCheck, alFieldNamesOfOldDataObject, 
                  alFieldNamesOfNewDataObject;
        boolean bNamesMatch; 
        Object objThisObject;
        Status statInformation;
        String stMessage, stOldRecordName, stNewRecordName;
        DataObject doOldDataObject;
        int iOldRecordIndex, iNewRecordIndex, iNumberOfOldRecordFields,
            iNumberOfNewRecordFields;

        if ( doNewDataObject.getType() == DataObject.I_TYPE_LABEL ) {
            statInformation = new Status( I_OK );
        }
        else {
            if ( doNewDataObject.getRestriction() == DataObject.I_RESTRICTION_VARIABLE ) {
                statInformation = new Status( I_OK );
            }
            else {
                statInformation = doNewDataObject.validateValue();
            }
        }
                        
        if ( statInformation.getStatusNumber() == I_OK ) {        
            objThisObject = doNewDataObject.getRecordObjectFieldNames();

            if ( objThisObject instanceof ArrayList ) {
                alNamesToCheck = new ArrayList();

                alFieldNamesOfNewDataObject = (ArrayList)objThisObject;            
                objThisObject = roGlobalObjects.getField( stObjectName );

                if ( objThisObject instanceof DataObject ) {
                    doOldDataObject = (DataObject)objThisObject;
                    objThisObject = doOldDataObject.getRecordObjectFieldNames();

                    if ( objThisObject instanceof ArrayList ) {
                        alFieldNamesOfOldDataObject = (ArrayList)objThisObject;

                        // eliminate duplicates between oldFields and NewFields;
                        // then send that list for testing of all other names

                        // when names DO NOT MATCH between lists, add NEW NAME
                        // to the list

                        iNumberOfOldRecordFields = alFieldNamesOfOldDataObject.size();
                        iNumberOfNewRecordFields = alFieldNamesOfNewDataObject.size();

                        for ( iNewRecordIndex = 0, bNamesMatch = true;
                              ( ( iNewRecordIndex < iNumberOfNewRecordFields )  && 
                                ( bNamesMatch == true ) );
                              iNewRecordIndex++ ) {

                            bNamesMatch = false;
                            stNewRecordName = (String)
                                    alFieldNamesOfNewDataObject.get( iNewRecordIndex );

                            for ( iOldRecordIndex = 0;
                                  ( ( iOldRecordIndex < iNumberOfOldRecordFields )
                                  && ( bNamesMatch == false ) ); iOldRecordIndex++ ) {

                                stOldRecordName = (String)
                                    alFieldNamesOfOldDataObject.get( iOldRecordIndex );

                                if ( stOldRecordName.equals( stNewRecordName ) )
                                    bNamesMatch = true;                            
                            }

                            if ( bNamesMatch == false ) 
                                alNamesToCheck.add( stNewRecordName );
                        }
                    }
                    else {
                        // don't care about other options...these actions are 
                        // optional and dependent if there are fields in a record

                        alNamesToCheck.addAll( alFieldNamesOfNewDataObject );                    
                    }
                    // continue checking for duplicates
                    statInformation = 
                        checkNamesInDataObjects( alNamesToCheck, B_LIST_HAS_NO_DATA_OBJECT_NAME );

                    if ( statInformation.getStatusNumber() == I_OK ) {
                        statInformation = roGlobalObjects.setFieldProperties( stObjectName,
                            doNewDataObject );

                        if ( statInformation.getStatusNumber() == RecordObject.I_FIELD_UNDECLARED ) {
                            stMessage = "The object named " + stObjectName + " has not "
                                + "been declared.";
                            statInformation = new Status( I_OBJECT_UNDECLARED, stMessage );
                        }
                        // else statInformation has I_OK
                    }
                    // else...error comes from checkNamesInDataObjects

                }
                else if ( objThisObject instanceof Status ) {
                    statInformation = (Status)objThisObject;

                    if ( statInformation.getStatusNumber() == 
                        RecordObject.I_FIELD_UNDECLARED ) {
                        stMessage = "The object named " + stObjectName + " has not "
                           + "been declared.";
                        statInformation = new Status( I_OBJECT_UNDECLARED, stMessage );                                    
                    }
                    else {
                        statInformation = new Status( I_UNKNOWN, "setDeclarationProperties:"
                            + " I_FIELD_UNDECLARED failed to be returned." );
                    }
                }
                else {
                    stMessage = "setDeclarationProperties: Did not receive ArrayList"
                       + " or Status object.";
                    statInformation = new Status( I_UNKNOWN, stMessage );
                }                
            }        
            // optional in order to avoid name duplication
            // here, the new data object has no record fields

            statInformation = roGlobalObjects.setFieldProperties( stObjectName,
                doNewDataObject );

            if ( statInformation.getStatusNumber() == RecordObject.I_FIELD_UNDECLARED ) {
                stMessage = "The object named " + stObjectName + " has not "
                    + "been declared.";
                statInformation = new Status( I_OBJECT_UNDECLARED, stMessage );
            }        
            // else statInformation has I_OK
            else {
                delDeclarationsListener.addedOrUpdatedDeclaration( stObjectName, 
                    doNewDataObject, DeclarationsListener.B_UPDATE_OBJECT );
                plistProgramListener.programChanged();
            }
        }
        // else...validateValue contains the error
        System.out.println( "Leaving setDeclarationProperties with status code " +
            statInformation.getStatusNumber() + " and message " + statInformation.getStatusMessage() );
        
        return statInformation;
    }
    
/************** DEBUGGING METHODS *****************************/
                                                            /**/
    public void printDeclaration( String stObjectName ) {   /**/
        Object objThisObject;                               /**/
        DataObject doDataObject;                            /**/
        Status statInformation;                             /**/
        String stMessage;                                   /**/
        ArrayDimensions adDimensions;                        /**/
        int iDimension;
        
        objThisObject = getDeclaration( stObjectName );
        
        if ( objThisObject instanceof Status ) {
            statInformation = (Status)objThisObject;
            System.out.println( "ERROR: " + statInformation.getStatusMessage() );
        }
        else if ( objThisObject instanceof DataObject ) {
            doDataObject = (DataObject)objThisObject;
            System.out.println( "Name (Given): " + stObjectName );
            System.out.println( "Icon Address " + doDataObject.getIcon().getDescription() + " " + doDataObject.getIcon() );
            System.out.println( "Type: " + DataObjectPropertyTextView.getTypeName( doDataObject.getType() ) );
            System.out.println( "Restriction: " + DataObjectPropertyTextView.getRestrictionName( doDataObject.getRestriction() ) );
            System.out.println( "Structure: " + DataObjectPropertyTextView.getStructureName( doDataObject.getStructure() ) );
            System.out.println( "Size: " + doDataObject.getSize() );
        }
        else {
            System.out.println( "?" );
        }
    }
    
    /*********** END OF DEBUGGING METHODS *************/
    
    public Object getDeclaration( String stObjectName ) {
        Object objThisValue;
        String stMessage;
        Status statInformation;
        
        objThisValue = roGlobalObjects.getField( stObjectName );
        
        if ( objThisValue instanceof Status ) {
            statInformation = (Status)objThisValue;
            if ( statInformation.getStatusNumber() == 
                 RecordObject.I_FIELD_UNDECLARED ) {
                stMessage = "The object named " + stObjectName + " has not "
                    + "been declared.";
                objThisValue = new Status( I_OBJECT_UNDECLARED, stMessage );                
            }
            else {
                stMessage = "getDeclaration: Did not receive I_FIELD_UNDECLARED.";
                objThisValue = new Status( I_UNKNOWN, stMessage );
            }        
        }
        else if ( objThisValue instanceof DataObject ) {
            // do nothing...objThisValue already has DataObject
        }
        else {
            stMessage = "getDeclaration: Did not receive DataObject or Status.";
            objThisValue = new Status( I_UNKNOWN, stMessage );
        }
        
        return objThisValue;
    }
        
    public Status writeDeclarations() {
        return new Status( I_OK );
    }
    
    public Status readIntoDeclarations( StringBuffer sbData ) {
        return new Status( I_OK );
    }

    public Object getNumberOfObjects( int iScope, String stRecordName ) {
        Object objThisObject;
        DataObject doThisDataObject;
        RecordObject roThisRecord;
        Status statInformation;
        String stMessage;
        boolean bMajorError, bRecordError;
        
        statInformation = new Status( I_OK );
        
        switch ( iScope ) {
            case I_NON_RECORD_OBJECTS:
                objThisObject = new Integer( alNonRecordObjectNames.size() );
                break;
                
            case I_RECORD_OBJECTS:
                objThisObject = new Integer( alRecordObjectNames.size() );
                break;
                
            case I_ALL_OBJECTS:
                objThisObject = new Integer( roGlobalObjects.getNumberOfFields() );
                break;
                
            case I_RECORD_OBJECT_FIELDS:
                bMajorError = false;
                objThisObject = roGlobalObjects.getField( stRecordName );
                
                if ( objThisObject instanceof DataObject ) {
                    doThisDataObject = (DataObject)objThisObject;
                    objThisObject = doThisDataObject.getRecordObject();

                    if ( objThisObject instanceof RecordObject ) {                        
                        roThisRecord = (RecordObject)objThisObject;
                        objThisObject = 
                            new Integer( roThisRecord.getNumberOfFields() );
                    }
                    // else...error is in objThisObject
                }
                else if ( objThisObject instanceof Status ) {
                    statInformation = (Status)objThisObject;
                    stMessage = "There is no record named " + stRecordName + ".";
                    objThisObject = new Status( I_OBJECT_UNDECLARED, stMessage );
                }
                else {
                    bMajorError = true;
                }
                    
                if ( bMajorError == true ) {
                    stMessage = "getNumberOfObjects: Expecting Status or "
                       + "DataObject.";
                    objThisObject = new Status( 
                        statInformation.getStatusNumber(), stMessage );
                }
                break;
               
            default:
                objThisObject = new Status( I_ILLEGAL_SCOPE,
                    "getNumberOfObjects: Incorrect scope issued." );                
                break;
        }

        // all have new instances
        return objThisObject;
    }

    public Object getFieldAtIndex( int iScope, int iIndex, String stRecordName ) {
        Object objThisObject;
        DataObject doThisDataObject;
        String stObjectName, stMessage;
        boolean bIndexRangeError, bMajorError, bRecordError;
        int iStatusNumber;
        Status statInformation;
        RecordObject roThisRecord;
        
        bIndexRangeError = true;
        bMajorError = false;
        bRecordError = false;
        objThisObject = new Object();
        statInformation = new Status( I_OK );
        
        switch ( iScope ) {
            case I_NON_RECORD_OBJECTS:
                if ( ( iIndex >= 0 )  &&  ( iIndex <= alNonRecordObjectNames.size() ) ) {
                    stObjectName = (String)alNonRecordObjectNames.get( iIndex );
                    objThisObject = roGlobalObjects.getField( stObjectName );
                    bIndexRangeError = false;
                    
                    if ( objThisObject instanceof DataObject ) {
                        doThisDataObject = (DataObject)objThisObject;
                        objThisObject = 
                            new NameAndDataObjectWrapper( stObjectName, doThisDataObject );
                    }
                    else if ( objThisObject instanceof Status ) {
                        stMessage = "The index number " + iIndex + " that "
                            + "corresponds to the non-record object named "
                            + stObjectName + " does not exist.";
                        statInformation = (Status)objThisObject;
                        objThisObject = new Status( statInformation.getStatusNumber(),
                            stMessage );
                    }
                    // else...do nothing
                }                
                break;
                
            case I_RECORD_OBJECTS:
                if ( ( iIndex >= 0 )  &&  ( iIndex <= alRecordObjectNames.size() ) ) {
                    stObjectName = (String)alRecordObjectNames.get( iIndex );
                    objThisObject = roGlobalObjects.getField( stObjectName );
                    bIndexRangeError = false;
                    
                    if ( objThisObject instanceof DataObject ) {
                        doThisDataObject = (DataObject)objThisObject;
                        objThisObject = 
                            new NameAndDataObjectWrapper( stObjectName, doThisDataObject );
                    }
                    else if ( objThisObject instanceof Status ) {
                        stMessage = "The index number " + iIndex + " that "
                            + "corresponds to the record object named "
                            + stObjectName + " does not exist.";
                        statInformation = (Status)objThisObject;
                        objThisObject = new Status( statInformation.getStatusNumber(),
                            stMessage );
                    }
                    else {
                        bMajorError = true;                        
                    }
                }
                break;
                
            case I_ALL_OBJECTS:
                //System.out.println( "ALL OBJECTS" );
                objThisObject = roGlobalObjects.getFieldAtIndex( iIndex );
                //System.out.println( "objThisObject in Declarations = " + objThisObject );
                if ( objThisObject instanceof NameAndDataObjectWrapper ) {
                    //System.out.println( "Object is a WRAPPER" );
                    bIndexRangeError = false;
                    /**DEBUGGING**/
//                    NameAndDataObjectWrapper nadow = (NameAndDataObjectWrapper)
//                        objThisObject;
//                    DataObject doObj = nadow.getObject();
//                    System.out.println("IN DECLARATIONS AT " +iIndex+": " );
//                    doObj.printDataObject();                    
                    /**END EBUG**/
                    // will be stored in objThisObject
                }
                else if ( objThisObject instanceof Status ) {
                    statInformation = (Status)objThisObject;
                    iStatusNumber = statInformation.getStatusNumber();
                    
                    if ( iStatusNumber == RecordObject.I_INVALID_RECORD_INDEX ) {
                        bIndexRangeError = true;
                    }                           
                    else if ( iStatusNumber == RecordObject.I_EMPTY_RECORD ) {
                        stMessage = "There are currently no objects to select "
                            + "using an index.";
                        objThisObject = new Status( I_NO_DECLARATIONS, stMessage );
                    }
                    else {
                        stMessage = "getFieldAtIndex: Expecting either "
                            + "I_INVALID_RECORD_INDEX or I_EMPTY_RECORD or I_OK.";
                        objThisObject = new Status( I_UNKNOWN, stMessage );
                    }
                }
                else {
                    bMajorError = true;
                }
                break;

            case I_RECORD_OBJECT_FIELDS:
                objThisObject = roGlobalObjects.getField( stRecordName );
                
                if ( objThisObject instanceof DataObject ) {
                    doThisDataObject = (DataObject)objThisObject;
                    objThisObject = doThisDataObject.getRecordObject();
                    
                    if ( objThisObject instanceof RecordObject ) {                        
                        roThisRecord = (RecordObject)objThisObject;
                        objThisObject = roThisRecord.getFieldAtIndex( iIndex );
                        
                        if ( objThisObject instanceof NameAndDataObjectWrapper ) {
                            bIndexRangeError = false;
                            // will be stored in objThisObject                        
                        }
                        else if ( objThisObject instanceof Status ) {
                            bRecordError = true;
                        }
                        else {
                            bMajorError = true;
                        }
                    }
                    else if ( objThisObject instanceof Status ) {
                        bRecordError = true;
                    }
                    else {
                        bMajorError = true;
                    }
                }
                else if ( objThisObject instanceof Status ) {
                    bRecordError = true;
                }
                else {
                    bMajorError = true;
                }
                
                if ( bRecordError ) {
                    statInformation = (Status)objThisObject;
                    stMessage = "  The record is named " + stRecordName + ".";
                    objThisObject = new Status( statInformation.getStatusNumber(),
                        statInformation.getStatusMessage() + stMessage );
                }
                break;
                
            default:
                objThisObject = new Status( I_ILLEGAL_SCOPE,
                    "getFieldAtIndex: Incorrect scope issued." );                
                break;
        }
        
        if ( bIndexRangeError == true ) {
            stMessage = "***Declarations.getFieldAtIndex: The index value " + iIndex + " does not correspond " +
                "to any declared object.";                            
            objThisObject = new Status( I_INDEX_OUT_OF_RANGE, stMessage );
        }
        else if ( bMajorError == true ) {
             stMessage = "getFieldAtIndex: Expecting Status or "
                + "DataObject.";
             objThisObject = new Status( 
                statInformation.getStatusNumber(), stMessage );
        }

        return objThisObject;
    }

    public Object getIndexOfObject( int iScope, String stRecordObjectName, 
        String stObjectName ) {
            
        Object objThisObject;
        String stMessage;
        int iStatusNumber, iIndexNumber;
        boolean bObjectFound, bStatusObjectCreated;
        Status statInformation;
        
        bObjectFound = false;
        bStatusObjectCreated = false;
        objThisObject = new Object();
        statInformation = new Status( I_OK );
        stMessage = new String( ST_NOTHING );
        
        switch ( iScope ) {
            case I_NON_RECORD_OBJECTS:
                iIndexNumber = alNonRecordObjectNames.indexOf( stObjectName );
                if ( iIndexNumber == I_OBJECT_NOT_FOUND_IN_LIST ) {
                    stMessage = stObjectName + " has not been defined or does "
                        + " not exist as a non-record object.";
                }
                else {
                    objThisObject = new Integer( iIndexNumber );                    
                    bObjectFound = true;
                }
                break;
                
            case I_RECORD_OBJECTS:
                iIndexNumber = alRecordObjectNames.indexOf( stObjectName );
                if ( iIndexNumber == I_OBJECT_NOT_FOUND_IN_LIST ) {
                    stMessage = stObjectName + " has not been defined or does "
                        + " not exist as a record object.";
                }
                else {
                    objThisObject = new Integer( iIndexNumber );                    
                    bObjectFound = true;
                }
                break;
                
            case I_ALL_OBJECTS:
                iIndexNumber = roGlobalObjects.getIndexOfObject( stObjectName );
                if ( iIndexNumber == I_OBJECT_NOT_FOUND_IN_LIST ) {
                    stMessage = stObjectName + " has not been defined or does "
                        + " not exist as an object.";
                }
                else {
                    objThisObject = new Integer( iIndexNumber );
                    bObjectFound = true;                    
                }
                break;
                
            case I_RECORD_OBJECT_FIELDS:
                RecordObject roThisRecord;
                DataObject doThisDataObject;
                ArrayList alRecordFieldNames;
                
                objThisObject = getDeclaration( stRecordObjectName );
                
                if ( objThisObject instanceof DataObject ) {
                    doThisDataObject = (DataObject)objThisObject;
                    
                    if ( doThisDataObject.isRecord() == B_IS_RECORD ) {
                        objThisObject = doThisDataObject.getRecordObjectFieldNames();
                        
                        if ( objThisObject instanceof ArrayList ) {
                            alRecordFieldNames = (ArrayList)objThisObject;
                            iIndexNumber = alRecordObjectNames.indexOf( stObjectName );

                            if ( iIndexNumber == I_OBJECT_NOT_FOUND_IN_LIST ) {
                                stMessage = stObjectName + " has not been defined or does "
                                    + " not exist as a record object.";
                                objThisObject =
                                    new Status( I_OBJECT_NOT_IN_SCOPE_OR_UNDECLARED,
                                        stMessage );
                                bStatusObjectCreated = true;
                            }
                            else {
                                objThisObject = new Integer( iIndexNumber );                    
                                bObjectFound = true;
                            }                            
                        }
                        else if ( objThisObject instanceof Status ) {
                            statInformation = (Status)objThisObject;
                            stMessage = statInformation.getStatusMessage() + 
                                "  The record name is " + stRecordObjectName + ".";
                            objThisObject = new Status( I_EMPTY_RECORD, stMessage );
                            bStatusObjectCreated = true;
                        }
                        else {
                            stMessage = "getIndexOfObject: Expecting ArrayRecord "
                                + " or Status object.";
                            objThisObject = new Status( I_UNKNOWN, stMessage );
                            bStatusObjectCreated = true;
                        }                            
                    }
                    else {
                        stMessage = stRecordObjectName + " has been defined "
                            + ", but it is not of type record.";
                        objThisObject = new Status( I_OBJECT_NOT_A_RECORD, stMessage );
                        bStatusObjectCreated = true;
                    }
                }
                else {
                    // already have objThisObject with correct Status object information
                    bStatusObjectCreated = true;  
                }
                break;
        }
 
        if ( ( bObjectFound = false )  &&  ( bStatusObjectCreated == false ) ) {
            objThisObject =
                new Status( I_OBJECT_NOT_IN_SCOPE_OR_UNDECLARED, stMessage );
        }
        
        return objThisObject;
    }
    
    public void setDeclarationsListener( DeclarationsListener delListener ) {
        delDeclarationsListener = delListener;
    }
    
    public void setProgramDeclarationsListener( ProgramListener plistListener ) {
        plistProgramListener = plistListener;
    }
    
    public Object getGlobalObjectNames() {
        Object objThisObject;
        String stMessage;
        ArrayList alAllObjects;
        
        if ( alRecordObjectNames.size() > 0 ) {
            alAllObjects = new ArrayList( alRecordObjectNames );
            
            if  ( alNonRecordObjectNames.size() > 0 ) {
                alAllObjects.addAll( alNonRecordObjectNames );
            }                
        }
        else {
            if  ( alNonRecordObjectNames.size() > 0 ) {
                alAllObjects = new ArrayList( alNonRecordObjectNames );
            }
            else {
                alAllObjects = new ArrayList();
            }
        }
        
        if ( alAllObjects.size() > 0 ) {
            objThisObject = alAllObjects;
        }
        else {
            stMessage = "There are currently no objects to select "
                + "using an index.";
            objThisObject = new Status( I_NO_DECLARATIONS, stMessage );
        }
        
        return objThisObject;
    }
    
    public Vector exportState() {
        Vector vData;
        
        vData = new Vector();
        vData.setSize( I_TOTAL_FIELD_COUNT );
        vData.set( I_RECORD_OBJECT_DATA, roGlobalObjects.exportState() );
        //System.out.println( "DECLARATIONS:  roGlobalObjects.exportState()=" +  roGlobalObjects.exportState() );
        //vData.set( I_RECORD_OBJECT_NAMES, alRecordObjectNames );
        //System.out.println( "DECLARATIONS:  alRecordObjectNames=" + alRecordObjectNames );
        //vData.set( I_NON_RECORD_OBJECT_NAMES, alNonRecordObjectNames );
        //System.out.println( "DECLARATIONS: alNonRecordObjectNames=" + alNonRecordObjectNames );
        vData.set( I_PROGRAM_NAME, stProgramName );
        //System.out.println( "DECLARATIONS:  stProgramName=" + stProgramName );
        //System.out.println( "DECLARATIONS:  vData=" + vData );
        return vData;
    }
    
    public void importState( Vector vData ) {
        Vector vDeclarations, vDataObject;
        String stName;
        DataObject doThisObject;

        // revision allows proper updating
        
        vDeclarations = (Vector)vData.get( I_RECORD_OBJECT_DATA );
        
        for ( int iIndex = 0; iIndex < vDeclarations.size(); iIndex++ ) {
            vDataObject = (Vector)vDeclarations.get( iIndex );
            stName = (String)vDataObject.get( RecordObject.I_DATA_OBJECT_NAME );
            doThisObject = new DataObject();
            System.out.println( "NAME = " +stName );
            doThisObject.importState( (Vector)
                vDataObject.get( RecordObject.I_DATA_OBJECT ) );
            System.out.println( "Before adding declaration..." );
            doThisObject.validationTest();
            addDeclaration( stName, doThisObject );
            System.out.println( "After adding declaration..." );
            doThisObject.validationTest();
        }
        stProgramName = (String)vData.get( I_PROGRAM_NAME );
        //roGlobalObjects.importState( (Vector)vData.get( I_RECORD_OBJECT_DATA ) );
        //alRecordObjectNames = (ArrayList)vData.get( I_RECORD_OBJECT_NAMES );
        //alNonRecordObjectNames = (ArrayList)vData.get( I_NON_RECORD_OBJECT_NAMES );        
    }

    /////////////////////// PRIVATE FUNCTIONS \\\\\\\\\\\\\\\\\\\\\\\\\\
    private Status checkNamesInDataObjects( ArrayList alNamesForChecking,
                                            boolean bListHasDataObjectName ) {
        Status statInformation;
        String stMessage, stRecordName, stNameToCheck, stDuplicatedField, stRecordWithDuplicatedField;
        int iNamesIndex, iRecordIndex, iArraySize, iNumberOfRecords;
        boolean bNameAlreadyDeclared, bMatchesSystemDeclarationsName, bMatchesProgramName;
        DataObject doRecordObject;
                
        iArraySize = alNamesForChecking.size();
        iNumberOfRecords = alRecordObjectNames.size();        
        bNameAlreadyDeclared = false;
        bMatchesSystemDeclarationsName = false;
        bMatchesProgramName = false;
        stDuplicatedField = new String();
        stRecordWithDuplicatedField = new String();
        
        // checking names with names in fields of records
        for ( iNamesIndex = 0; ( ( iNamesIndex < iArraySize )  &&  
                            ( bNameAlreadyDeclared == false ) ); iNamesIndex++ ) {

            stNameToCheck = (String)alNamesForChecking.get( iNamesIndex );                                
                                
            for ( iRecordIndex = 0; ( ( iRecordIndex < iNumberOfRecords )  &&  
                                      ( bNameAlreadyDeclared == false ) );
                                    iRecordIndex++ ) {
                                        
                stRecordName = (String)alRecordObjectNames.get( iRecordIndex );

                doRecordObject = 
                    (DataObject)roGlobalObjects.getField( stRecordName );
                
                if ( doRecordObject.isFieldInRecordObject( stNameToCheck )
                    == true ) {
                        stDuplicatedField = stNameToCheck;
                        stRecordWithDuplicatedField = stRecordName;
                        bNameAlreadyDeclared = true;
                }
            }
             
            ///// ADDED AFTER ORIGINAL TEST \\\\
            if ( SystemDeclarations.isObjectDeclaredAsASystemObject(
                stNameToCheck ) == true ) {

                stDuplicatedField = stNameToCheck;
//                stRecordWithDuplicatedField = stRecordName;
                bNameAlreadyDeclared = true;
                bMatchesSystemDeclarationsName = true;
            }
            //// ANOTHER ADDITION \\\\
            else if ( stProgramName.equals( stNameToCheck ) == true ) {
                stDuplicatedField = stNameToCheck;
//                stRecordWithDuplicatedField = stRecordName;
                bNameAlreadyDeclared = true;
                bMatchesProgramName = true;
            }                    
        }
        
        if ( bNameAlreadyDeclared == false ) {
            statInformation = new Status( I_OK );
        }
        else {
            if ( ( iNamesIndex == I_DATA_OBJECT_NAME )  &&
                 ( bListHasDataObjectName == true ) ) {                    
                stMessage = "The object named ";
            }
            else { 
                stMessage = "The field named ";
            }
            
            if ( bMatchesProgramName == true ) {
                stMessage = "The object named " + stDuplicatedField + " already exists "
                    + "as the name of the program.";                
            }
            else {
                if ( bMatchesSystemDeclarationsName == false ) {
                    stMessage = "The object named " + stDuplicatedField + " already exists "
                        + "in an object with a record field named " + stRecordWithDuplicatedField
                        + ".";
                }
                else {
                    stMessage = stMessage + " is used as an internal system object.";
                }
            }       
            statInformation = new Status( I_OBJECT_DECLARED, stMessage );
        }
        
        return statInformation;
    }
    
    private void updateNameLists( String stObjectName, String stNewObjectName,
        int iActionCode, boolean bIsRecord ) {

        int iObjectIndex;

        // unless internal object data is corrupted, no error checking
        // should be necessary
        
        switch ( iActionCode ) {
            case I_REMOVE_NAME_FROM_LIST:
                if ( bIsRecord )
                    alRecordObjectNames.remove( stObjectName );
                else
                    alNonRecordObjectNames.remove( stObjectName );
             break;
        
            case I_ADD_NAME_TO_LIST:
                if ( bIsRecord )
                    alRecordObjectNames.add( stObjectName );
                else
                    alNonRecordObjectNames.add( stObjectName );
                break;
                    
            case I_UPDATE_NAME_IN_LIST:
                if ( bIsRecord ) {
                    iObjectIndex = alRecordObjectNames.indexOf( stObjectName );
                    alRecordObjectNames.set( iObjectIndex, stNewObjectName );
                }
                else {
                    iObjectIndex = alNonRecordObjectNames.indexOf( stObjectName );
                    alNonRecordObjectNames.set( iObjectIndex, stNewObjectName );
                }
                break;
                
            default:
                System.out.println("Code not available for updating name lists." );
                break;
        }        
    }        
    
    public void addDefaultDeclarations() {
/*        DataObject doTemp;
        // GET RID OF DOs FOR REAL PROGRAM...DONE TO ELIMINATE REPEATED DEFINITIONS
        doTemp = new DataObject();
        doTemp.setRestriction( DataObject.I_RESTRICTION_VARIABLE );
        doTemp.setType( DataObject.I_TYPE_INTEGER );
        doTemp.setStructure( DataObject.I_STRUCTURE_SINGLE_VALUE );
        this.addDeclaration( "KATINT", doTemp );

        doTemp = new DataObject();
        doTemp.setRestriction( DataObject.I_RESTRICTION_CONSTANT );
        doTemp.setType( DataObject.I_TYPE_INTEGER );
        doTemp.setStructure( DataObject.I_STRUCTURE_SINGLE_VALUE );
        doTemp.setValue( "25" );
        this.addDeclaration( "KATINT2", doTemp );
        
        doTemp = new DataObject();
        doTemp.setRestriction( DataObject.I_RESTRICTION_CONSTANT );
        doTemp.setType( DataObject.I_TYPE_INTEGER );
        doTemp.setStructure( DataObject.I_STRUCTURE_SINGLE_VALUE );
        doTemp.setValue( "50" );
        this.addDeclaration( "KATINT3", doTemp );        
        
        doTemp = new DataObject();
        doTemp.setRestriction( DataObject.I_RESTRICTION_CONSTANT );
        doTemp.setType( DataObject.I_TYPE_CHARACTER );
        doTemp.setStructure( DataObject.I_STRUCTURE_SINGLE_VALUE );
        doTemp.setValue( "'k'" );
        this.addDeclaration( "KATCHAR", doTemp );
        
        doTemp = new DataObject();
        doTemp.setRestriction( DataObject.I_RESTRICTION_VARIABLE );
        doTemp.setType( DataObject.I_TYPE_CHARACTER );
        doTemp.setStructure( DataObject.I_STRUCTURE_SINGLE_VALUE );
        this.addDeclaration( "KATCHAR2", doTemp );
        
        doTemp = new DataObject();
        doTemp.setRestriction( DataObject.I_RESTRICTION_CONSTANT );
        doTemp.setType( DataObject.I_TYPE_CHARACTER );
        doTemp.setStructure( DataObject.I_STRUCTURE_1D_ARRAY );
        doTemp.setValue( "Kathleen Marie Spale" );
        doTemp.setArrayDimensions( 20, 1 );
        this.addDeclaration( "KATARAY", doTemp );
        
        doTemp = new DataObject();
        doTemp.setRestriction( DataObject.I_RESTRICTION_VARIABLE );
        doTemp.setType( DataObject.I_TYPE_CHARACTER );
        doTemp.setStructure( DataObject.I_STRUCTURE_1D_ARRAY );
        doTemp.setArrayDimensions( 35, 1 );
        this.addDeclaration( "KATARAY2", doTemp );
        
        doTemp = new DataObject();
        doTemp.setType( DataObject.I_TYPE_LABEL );
        this.addDeclaration( "KATLAB", doTemp );
        
        doTemp = new DataObject();
        doTemp.setType( DataObject.I_TYPE_LABEL );
        this.addDeclaration( "KATLAB2", doTemp );
        
        doTemp = new DataObject();
        doTemp.setType( DataObject.I_TYPE_LABEL );
        this.addDeclaration( "KATLAB3", doTemp );
        
        doTemp = new DataObject();
        doTemp.setType( DataObject.I_TYPE_LABEL );
        this.addDeclaration( "KATLAB4", doTemp );

        doTemp = new DataObject();
        doTemp.setType( DataObject.I_TYPE_LABEL );
        this.addDeclaration( "KATLAB5", doTemp );
        
        doTemp = new DataObject();
        doTemp.setType( DataObject.I_TYPE_LABEL );
        this.addDeclaration( "KATLAB6", doTemp );
        
        // HHOOOORAY*/
    }        
    
    
    
    // instance variables
    private RecordObject roGlobalObjects;
    private ArrayList alRecordObjectNames, alNonRecordObjectNames;
    private DeclarationsListener delDeclarationsListener;
    private ProgramListener plistProgramListener;
    private String stProgramName;
    
    // class constants
    private static final int I_RECORD_OBJECT_DATA = 0;
    //private static final int I_RECORD_OBJECT_NAMES = 1;
    //private static final int I_NON_RECORD_OBJECT_NAMES = 2;
    private static final int I_PROGRAM_NAME = 1;
    private static final int I_TOTAL_FIELD_COUNT = 2;
    
    private static final int I_DATA_OBJECT_NAME = 0;
    private static final int I_ADD_NAME_TO_LIST = 1;
    private static final int I_REMOVE_NAME_FROM_LIST = 2;
    private static final int I_UPDATE_NAME_IN_LIST = 3;
    
    private static final boolean B_LIST_CONTAINS_DATA_OBJECT_NAME = true;
    private static final boolean B_LIST_HAS_NO_DATA_OBJECT_NAME = false;
    private static final boolean B_IS_RECORD = true;
    private static final boolean B_IS_NOT_RECORD = false;    

    private static final String ST_NOTHING = "";
    
        // scope codes
        public static final int I_NON_RECORD_OBJECTS = 1;
        public static final int I_RECORD_OBJECTS = 2;
        public static final int I_ALL_OBJECTS = 3;
        public static final int I_RECORD_OBJECT_FIELDS = 4;
        
        
        // error codes
        public static final int I_OK = 0;
        public static final int I_OBJECT_DECLARED = -1;
        public static final int I_OBJECT_NOT_FOUND_IN_LIST = -1;
        public static final int I_OBJECT_UNDECLARED = -2;
        public static final int I_ILLEGAL_SCOPE = -3;
        public static final int I_INDEX_OUT_OF_RANGE = -4;
        public static final int I_NO_DECLARATIONS = -5;
        public static final int I_OBJECT_NOT_IN_SCOPE_OR_UNDECLARED = -6;       
        public static final int I_EMPTY_RECORD = -7;
        public static final int I_OBJECT_NOT_A_RECORD = -8;
        
        public static final int I_UNKNOWN = -8888;
}