/*
 * DeclarationsTableModel.java
 *
 * Created on February 2, 2002, 8:39 PM
 */

//package panelsAndTools;

import javax.swing.event.*;
import javax.swing.table.*;
import javax.swing.*;
import java.util.*;
import java.lang.*;
import java.io.*;

/**
 *
 * @author  _
 * @version 
 */


public class DeclarationsTableModel extends AbstractTableModel
    implements DeclarationsListener, Cloneable, Serializable {

    public DeclarationsTableModel() {
        // passes reference to itself that will be notified
        // when Declarations is updated
        deDeclarationModel = new Declarations();
        iModelNumber = I_UNDEFINED;
        stRecordName = ST_NO_RECORD;
        iColumnToTrack = I_COLUMN_NAME;
        alPackedInformationList = new ArrayList();
        alObjectNames = new ArrayList();
        bSuppressTableUpdates = false;
    }
    
    protected Object clone() {
        // strucutre based on Sun's online Java tutorial for
        // the Object.clone() method and pp.235-6 C of Core Java 2: ISBN 0-13-081933-6        
        try {
            DeclarationsTableModel dtmObject = (DeclarationsTableModel)super.clone();
            dtmObject.deDeclarationModel = (Declarations)this.deDeclarationModel.clone();
            dtmObject.alPackedInformationList = 
                (ArrayList)this.alPackedInformationList.clone();
            dtmObject.alObjectNames = (ArrayList)this.alObjectNames.clone();

            return dtmObject;
        }
        catch ( CloneNotSupportedException cnseException ) {
            System.out.println( "Clone of DeclarationsTableModel is not supported." );
            return null;
        }
    }
    
    public Declarations getDeclarationsModel() {
        return (Declarations)deDeclarationModel.clone();
    }
    
    public Status setDeclarationsModel( Declarations deDeclarationsObject ) {
        deDeclarationModel = deDeclarationsObject;
        deDeclarationModel.setDeclarationsListener( this );
        
        return new Status( I_OK );
    }
    
    public int getModelNumber() {
        return iModelNumber;
    }
    
    public Status setModelNumber( int iThisModelNumber ) {
        String stMessage;
        Status statInformation;
        
        if  ( ( iThisModelNumber >= I_MODEL_MINIMUM_NUMBER )  &&
              ( iThisModelNumber <= I_MODEL_MAXIMUM_NUMBER ) ) {
            iModelNumber = iThisModelNumber;
            
            if ( alPackedInformationList.size() > 0 )
                alPackedInformationList.clear();
            
            statInformation = updatePackedInformationList();            
        }
        else {
            stMessage = "The number " + iThisModelNumber + " does not correspond"
                + " to a valid model number.";
            statInformation = new Status( I_ILLEGAL_MODEL, stMessage );
        }

        return statInformation;
    }
    
    public String getRecordName() {
        return new String( stRecordName );
    }
    
    public Status setRecordName( String stName ) {
        Status statInformation;
        
        stRecordName = stName;
        statInformation = updatePackedInformationList();
        
        return statInformation;
    }
        
    public Status setColumnToTrack( int iColumn ) {
        Status statInformation;
        String stMessage, stNameIndexBuffer;
        
        if ( ( iColumn >= I_COLUMN_ICON )  &&  ( iColumn <= I_COLUMN_SIZE ) ) {
            iColumnToTrack = iColumn;
            stNameIndexBuffer = ST_COL_NAME_BUF;
            
            //stNameIndexBuffer = ST_COL_NAME_BUF + ST_INDEX_NUMBER_BUF;
            
            switch ( iColumn ) {
                case I_COLUMN_ICON:
                    stBufferInitializer = ST_COL_ICON_BUF + stNameIndexBuffer;
                    break;
                
                case I_COLUMN_NAME:
                    stBufferInitializer = stNameIndexBuffer;
                    break;

                case I_COLUMN_TYPE:
                    stBufferInitializer = ST_COL_TYPE_BUF + stNameIndexBuffer;
                    //System.out.println( "Set column to track as TYPE." );
                    break;

                case I_COLUMN_RESTRICTION:
                    stBufferInitializer = ST_COL_RESTRICTION_BUF + stNameIndexBuffer;
                    break;

                case I_COLUMN_STRUCTURE:
                    stBufferInitializer = ST_COL_STRUCTURE_BUF + stNameIndexBuffer;
                    break;

                case I_COLUMN_SIZE:
                    stBufferInitializer = ST_COL_SIZE_BUF + stNameIndexBuffer;
                    break;

                default:
                    stBufferInitializer = ST_NOTHING;
                    break;
            }
            
            iPackedInformationLength = stBufferInitializer.length();
            statInformation = updatePackedInformationList();
        }
        else {
            stMessage = "Column number " + iColumn + " cannot be tracked.";
            statInformation = new Status( I_CANNOT_TRACK_COLUMN, stMessage );
        }
        
        return statInformation;
    }
    
    public int getColumnToTrack() {
        return iColumnToTrack;
    }
    
    public ArrayList getPackedInformationList() {
        return (ArrayList)alPackedInformationList.clone();
    }
    
    public void setTableModelListener( TableModelListener tmlListener ) {
        addTableModelListener( tmlListener );
    }
    
    public Object getDataObjectFromRow( int iRow ) {        
        // receives cloned/new object
        return SerializedClone.cloneObject( 
            deDeclarationModel.getFieldAtIndex( iModelNumber, iRow, stRecordName ) );
    }
    
    
    /* Because this part of the class works in conjunction with Declarations,
       it is assumed that all parameters validly exist in Declarations.
       Also, since model validity is checked, it is assumed all models are 
       valid. */
    
    ////////////// DECLARATIONS LISTENER IMPLEMENTATION \\\\\\\\\\\
    public void addedOrUpdatedDeclaration( String stThisObjectName, 
        DataObject doADataObject, boolean bDeclarationAction ) {        
        
        String stObjectName, stTrackedColumnValue;
        Object objThisObject;
        int iIndex, iNumberOfFields, iPropertyValue;
        DataObject doThisDataObject;
        BitSet bsModelEvaluation;
        Integer intObjectIndex;
        ArrayList alRecordFieldNames, alRecordFieldObjects;
        
        //System.out.println( "DeclarationsTableModel.addedOrUpdatedDeclaration: " );
        //System.out.println( "Calling addedOrUpdatedDeclaration in DeclarationsTableModel." );        
        doThisDataObject = new DataObject();
        iNumberOfFields = 0;
        
        bsModelEvaluation = 
            evaluateModelNumberAndObject( stThisObjectName, doADataObject );
        
        alRecordFieldObjects = new ArrayList();
        alRecordFieldNames = new ArrayList();
        
        if ( ( bsModelEvaluation.get( I_IS_RECORD_OBJECT ) == true )  && 
             ( bsModelEvaluation.get( I_OBJECT_INCLUDED_IN_MODEL ) == true ) ) {

            objThisObject = doADataObject.getRecordObjectFieldNames();
   
            if ( objThisObject instanceof ArrayList ) {
                // fields in record objects are changed as a complete unit
                if ( alPackedInformationList.size() > 0 )
                    alPackedInformationList.clear();
                
                alRecordFieldNames = (ArrayList)objThisObject;
                iNumberOfFields = alRecordFieldNames.size();
                        
                if ( iNumberOfFields > 0 ) {
                    alRecordFieldObjects = new ArrayList( 
                        (Collection)doADataObject.getRecordObjectFieldObjects() );
                }
                // else empty list...do nothing                        
            }
            // else if Status...do nothing...no field info
        }
        
        if ( bsModelEvaluation.get( I_OBJECT_INCLUDED_IN_MODEL ) == true ) {
            //System.out.println( "Entering switch..." );
           
            switch ( iColumnToTrack ) {
                case I_COLUMN_ICON:
                    ImageIcon iiIcon;
                    
                    if ( bsModelEvaluation.get( I_IS_RECORD_OBJECT ) == true ) {
			bSuppressTableUpdates = true;
                        for ( iIndex = 0; iIndex < iNumberOfFields; iIndex++ ) {
                            doThisDataObject = (DataObject)alRecordFieldObjects.get( iIndex );
                            stObjectName = (String)alRecordFieldNames.get( iIndex );
                            iiIcon = doThisDataObject.getIcon();
                            stTrackedColumnValue = iiIcon.getDescription();
                            addOrUpdateSingleObject( stTrackedColumnValue, stObjectName,
                                I_AFTER_ICON_INDEX, bDeclarationAction,
                                DeclarationsListener.B_RECORD_FIELD_DATA_OBJECT );
                        }
			bSuppressTableUpdates = false;
                        updateTableView();
                    }
                    else {
                        iiIcon = doADataObject.getIcon();
                        stTrackedColumnValue = iiIcon.getDescription();
                        addOrUpdateSingleObject( stTrackedColumnValue, stThisObjectName,
                            I_AFTER_ICON_INDEX, bDeclarationAction,
                            DeclarationsListener.B_GLOBAL_DATA_OBJECT );
                    }
                    break;
                    
                case I_COLUMN_NAME:
                    if ( bsModelEvaluation.get( I_IS_RECORD_OBJECT ) == true ) {
			bSuppressTableUpdates = true;
                        for ( iIndex = 0; iIndex < iNumberOfFields; iIndex++ ) {                      
                            stObjectName = (String)alRecordFieldNames.get( iIndex );
                            addOrUpdateSingleObject( ST_NOTHING, stObjectName,
                                I_AFTER_NAME_INDEX, bDeclarationAction,
                                DeclarationsListener.B_RECORD_FIELD_DATA_OBJECT );
                        }
			bSuppressTableUpdates = false;
                        //fireTableDataChanged();
                        updateTableView();
                    }
                    else {
                        // I_AFTER_NAME_INDEX = 0                     
                        addOrUpdateSingleObject( ST_NOTHING, stThisObjectName,
                            I_AFTER_NAME_INDEX, bDeclarationAction,
                            DeclarationsListener.B_GLOBAL_DATA_OBJECT );
                    }
                    break;

                case I_COLUMN_TYPE:
                    //System.out.println( "About to add object..." );
                    if ( bsModelEvaluation.get( I_IS_RECORD_OBJECT ) == true ) {
			bSuppressTableUpdates = true;
                        for ( iIndex = 0; iIndex < iNumberOfFields; iIndex++ ) {
                            doThisDataObject = (DataObject)alRecordFieldObjects.get( iIndex );
                            stObjectName = (String)alRecordFieldNames.get( iIndex );
                            iPropertyValue = doThisDataObject.getType();
                            stTrackedColumnValue = 
                                DataObjectPropertyTextView.getTypeName( iPropertyValue );
                            addOrUpdateSingleObject( stTrackedColumnValue, stObjectName,
                                I_AFTER_TYPE_INDEX, bDeclarationAction,
                                DeclarationsListener.B_RECORD_FIELD_DATA_OBJECT );
                        }
			bSuppressTableUpdates = false;
                        //fireTableDataChanged();
                        updateTableView();
                    }
                    else {
                        iPropertyValue = doADataObject.getType();
                        stTrackedColumnValue = 
                            DataObjectPropertyTextView.getTypeName( iPropertyValue );
                        addOrUpdateSingleObject( stTrackedColumnValue, stThisObjectName
                            ,I_AFTER_TYPE_INDEX, bDeclarationAction,
                            DeclarationsListener.B_GLOBAL_DATA_OBJECT );
                    }
                    break;

                case I_COLUMN_RESTRICTION:
                    if ( bsModelEvaluation.get( I_IS_RECORD_OBJECT ) == true ) {
			bSuppressTableUpdates = true;
                        for ( iIndex = 0; iIndex < iNumberOfFields; iIndex++ ) {
                            doThisDataObject = (DataObject)alRecordFieldObjects.get( iIndex );
                            stObjectName = (String)alRecordFieldNames.get( iIndex );
                            iPropertyValue = doThisDataObject.getRestriction();
                            stTrackedColumnValue = 
                                DataObjectPropertyTextView.getRestrictionName( iPropertyValue );
                            addOrUpdateSingleObject( stTrackedColumnValue, stObjectName,
                                I_AFTER_RESTRICTION_INDEX, bDeclarationAction,
                                DeclarationsListener.B_RECORD_FIELD_DATA_OBJECT );
                        }
			bSuppressTableUpdates = false;
                        updateTableView();
                    }
                    else {
                        iPropertyValue = doADataObject.getRestriction();
                        stTrackedColumnValue = 
                            DataObjectPropertyTextView.getRestrictionName( iPropertyValue );
                        addOrUpdateSingleObject( stTrackedColumnValue, stThisObjectName, 
                            I_AFTER_RESTRICTION_INDEX, bDeclarationAction,
                            DeclarationsListener.B_GLOBAL_DATA_OBJECT );
                    }
                    break;

                case I_COLUMN_STRUCTURE:                                     
                    if ( bsModelEvaluation.get( I_IS_RECORD_OBJECT ) == true ) {
			bSuppressTableUpdates = true;
                        for ( iIndex = 0; iIndex < iNumberOfFields; iIndex++ ) {
                            doThisDataObject = (DataObject)alRecordFieldObjects.get( iIndex );
                            stObjectName = (String)alRecordFieldNames.get( iIndex );
                            stTrackedColumnValue = 
                                prepareStructureTrackedValue( doThisDataObject );
                            addOrUpdateSingleObject( stTrackedColumnValue, stObjectName,
                                I_AFTER_STRUCTURE_INDEX, bDeclarationAction,
                                DeclarationsListener.B_RECORD_FIELD_DATA_OBJECT );
                        }
			bSuppressTableUpdates = false;
                        updateTableView();
                    }
                    else {
                        stTrackedColumnValue = 
                            prepareStructureTrackedValue( doADataObject );
                        addOrUpdateSingleObject( stTrackedColumnValue, stThisObjectName,
                            I_AFTER_STRUCTURE_INDEX, bDeclarationAction,
                            DeclarationsListener.B_GLOBAL_DATA_OBJECT );
                    }
                    break;

                case I_COLUMN_SIZE:
                    int iThisSize;
                    
                    if ( bsModelEvaluation.get( I_IS_RECORD_OBJECT ) == true ) {
			bSuppressTableUpdates = true;
                        for ( iIndex = 0; iIndex < iNumberOfFields; iIndex++ ) {
                            doThisDataObject = (DataObject)alRecordFieldObjects.get( iIndex );
                            stObjectName = (String)alRecordFieldNames.get( iIndex );
                            iThisSize = doADataObject.getSize();
                            stTrackedColumnValue = Integer.toString( iThisSize );
                            addOrUpdateSingleObject( stTrackedColumnValue, stObjectName,
                                I_AFTER_SIZE_INDEX, bDeclarationAction,
                                DeclarationsListener.B_RECORD_FIELD_DATA_OBJECT );
                        }
			bSuppressTableUpdates = false;
                        updateTableView();
                    }
                    else {
                        iThisSize = doADataObject.getSize();
                        stTrackedColumnValue = Integer.toString( iThisSize );
                        addOrUpdateSingleObject( stTrackedColumnValue, stThisObjectName,
                            I_AFTER_SIZE_INDEX, bDeclarationAction,
                            DeclarationsListener.B_GLOBAL_DATA_OBJECT );
                    }
                    break;                    
            } // end switch
        }
        // ignore this listener event
        // else do not get extra information since it is not needed
    }
    
    public void removedDeclaration( String stDeletedObjectName, 
        DataObject doADataObject ) {
        
        int iIndexToDelete;
        BitSet bsModelEvaluation;        

        // does not matter if this is a record...record fields in
        // a DataObject would be deleted in addedOrUpdatedDeclaration
        
        bsModelEvaluation = 
            evaluateModelNumberAndObject( stDeletedObjectName, doADataObject );

        // non-record object
        if ( ( bsModelEvaluation.get( I_IS_RECORD_OBJECT ) == false )  &&
             ( bsModelEvaluation.get( I_OBJECT_INCLUDED_IN_MODEL ) == true ) ) {

            //iIndexToDelete = alObjectNames.indexOf( stDeletedObjectName );
            //alObjectNames.remove( iIndexToDelete );
            //alPackedInformationList.remove( iIndexToDelete );
            //fireTableRowsDeleted( iIndexToDelete, iIndexToDelete );            
            updatePackedInformationList();
            updateTableView();
        
        }
        // record object...delete all record fields
        else if ( ( bsModelEvaluation.get( I_IS_RECORD_OBJECT ) == true )  &&
                  ( bsModelEvaluation.get( I_OBJECT_INCLUDED_IN_MODEL ) == true ) ) { 
            
            //if ( alObjectNames.size() > 0 )
            //    alObjectNames.clear();
            
            //fireTableDataChanged();
            updatePackedInformationList();
	    updateTableView();
            
        }
            // TableModelEvent issued to TableModelListener
            // ^-from DeclarationsTableModel    ^-to SortableTableModel          
    }
    
    public void updatedDeclarationName( String stOldObjectName, 
        String stNewObjectName, DataObject doADataObject ) {

        int iIndexToUpdate, iBufferPosition;
        BitSet bsModelEvaluation;
        String stPackedInformationString;
        StringBuffer sbPackedInformation;

        bsModelEvaluation = 
            evaluateModelNumberAndObject( stOldObjectName, doADataObject );        
        
        // only relevant for non-record objects
        if ( ( bsModelEvaluation.get( I_IS_RECORD_OBJECT ) == false )  &&
             ( bsModelEvaluation.get( I_OBJECT_INCLUDED_IN_MODEL ) == true ) ) {
                
            iIndexToUpdate = alObjectNames.indexOf( stOldObjectName );           
            
            // update name in sbPackedInformation
            stPackedInformationString = (String)
                alPackedInformationList.get( iIndexToUpdate );
            
            sbPackedInformation = new StringBuffer( stPackedInformationString );
                           
            switch ( iColumnToTrack ) {
                case I_COLUMN_ICON:
                    iBufferPosition = I_AFTER_ICON_INDEX;
                    break;
                
                case I_COLUMN_NAME:
                    iBufferPosition = 0;
                    break;

                case I_COLUMN_TYPE:
                    iBufferPosition = I_AFTER_TYPE_INDEX;
                    break;

                case I_COLUMN_RESTRICTION:
                    iBufferPosition = I_AFTER_RESTRICTION_INDEX;
                    break;

                case I_COLUMN_STRUCTURE:
                    iBufferPosition = I_AFTER_STRUCTURE_INDEX;
                    break;

                case I_COLUMN_SIZE:
                    iBufferPosition = I_AFTER_SIZE_INDEX;
                    break;

                default:
                    iBufferPosition = 0;
                    break;
            }
                        
            // clear the old name and overwrite the new
            sbPackedInformation.replace( iBufferPosition, 
                iBufferPosition + ST_COL_NAME_BUF.length(), ST_COL_NAME_BUF );    

            sbPackedInformation.replace( iBufferPosition, 
                iBufferPosition + stNewObjectName.length(), stNewObjectName );    

            alPackedInformationList.set( iIndexToUpdate, 
                sbPackedInformation.toString() );

            alObjectNames.set( iIndexToUpdate, stNewObjectName );            

            
            // TableModelEvent issued to TableModelListener
            // ^-from DeclarationsTableModel    ^-to SortableTableModel
            
            // non-record object...only update one cell of a row
            // no action for name change in record object
            // taken care of in global record object
            
            //fireTableCellUpdated( iIndexToUpdate, I_COLUMN_NAME );
            updateTableView();
        }
        // else ignore message from updating name model
    }
    
    //////////////////// CLASS PRIVATE HELPER FUNCTIONS \\\\\\\\\\\\\\\
    private void addOrUpdateSingleObject( String stTrackedColumnValue, 
        String stNameValue, int iBufferPosition, boolean bObjectAction,
        boolean bObjectIsFieldInRecord ) {
        
        String stPosition, stPackedInformationString;
        int iPositionNumberLength, iNumberIndexShift, iTotalIndexShift, 
            iObjectIndex;
        boolean bValidObjectIndex;
        Object objThisObject;
        Integer intObjectIndex;
        Status statInformation;
        StringBuffer sbPackedInformation;
        
        sbPackedInformation = new StringBuffer( stBufferInitializer );        
        bValidObjectIndex = false;
        
        //System.out.print("Before: " + sbPackedInformation.toString() );
        // ** "tracked value" **  appears first in buffer
        if ( iColumnToTrack == I_COLUMN_SIZE ) {
            iPositionNumberLength = stTrackedColumnValue.length();
            iNumberIndexShift = 
                I_MAXIMUM_NUMBER_SIZE - iPositionNumberLength;                        
            //iTotalIndexShift = iBufferPosition
            //    + I_NAME_BUFFER_LENGTH + iNumberIndexShift;                            

            sbPackedInformation.replace( iNumberIndexShift, 
                I_MAXIMUM_NUMBER_SIZE, stTrackedColumnValue );
        }
        else {
            sbPackedInformation.replace( 0, stTrackedColumnValue.length(), 
                stTrackedColumnValue );
        }
            
        // ** object name **  appears second in buffer
        sbPackedInformation.replace( iBufferPosition, 
            iBufferPosition + stNameValue.length(), stNameValue );    

        // ** array index **  appears third (last) in buffer
        objThisObject = deDeclarationModel.getIndexOfObject( iModelNumber, 
            stRecordName, stNameValue );
            
        if ( objThisObject instanceof Integer ) {
            intObjectIndex = (Integer)objThisObject;
            iObjectIndex = intObjectIndex.intValue();
            stPosition = intObjectIndex.toString();
            bValidObjectIndex = true;
        }
        else if ( objThisObject instanceof Status ) {
            statInformation = (Status)objThisObject;                
            System.out.println( "addOrUpdateSingleObject: "
                + statInformation.getStatusMessage() );
            iObjectIndex = I_UNDEFINED;
            stPosition = new String( "0" );
        }
        else {
            System.out.println( "addOrUpdateSingleObject: "
                + "Expected Integer or Status object." );                
            iObjectIndex = I_UNDEFINED;
            stPosition = new String( "0" );
        }
            
        if ( bValidObjectIndex == true )
        {
            stPackedInformationString = 
                sbPackedInformation.toString() + stPosition;
        }
        else {
            stPackedInformationString = ST_NOTHING;
            
            System.out.println( "addOrUpdateSingleObject: "
                + "Invalid index number of object." );                
        }
        
        // TableModelEvent issued to TableModelListener
        // ^-from DeclarationsTableModel    ^-to SortableTableModel          
            
       
        // put buffer and object name into respective ArrayLists and clear buffer
        // index irrelevant in ArrayList...index must be in packedInformation
                
        if ( bObjectAction == DeclarationsListener.B_ADD_OBJECT ) {
            alPackedInformationList.add( stPackedInformationString ); 
            alObjectNames.add( stNameValue );
            
            if ( bObjectIsFieldInRecord == false ) 
                //fireTableRowsInserted( iObjectIndex, iObjectIndex ); 
                updateTableView();
            
            //System.out.println( "Added object..." );
        }
        // else update object
        else {
            // since ObjectName and PackedInformation objects are inserted in
            // operated on together, getting the ObjectName index should 
            // be sufficient; ObjectName stays the same
            
            if ( iObjectIndex == I_UNDEFINED )
                iObjectIndex = alObjectNames.indexOf( stNameValue );

            alPackedInformationList.set( iObjectIndex, 
                stPackedInformationString );
            
            if ( bObjectIsFieldInRecord == false ) 
                //fireTableRowsUpdated( iObjectIndex, iObjectIndex );             
                updateTableView();
        }        
        //System.out.println("   Position: " + stPosition + "  After: " + stPackedInformationString );
    }
    
    private String prepareStructureTrackedValue( DataObject doCurrentDataObject ) {
        StringBuffer sbTrackedValueBuffer;
        String stStructureName, stDimensionValue;
        ArrayDimensions adArrayDimensions;
        int iDimension, iDimensionLength, iNumberIndexShift, iTotalIndexShift;
        StructureAndDimensionsWrapper sadwStructureAndDimensions;
        
        sbTrackedValueBuffer = new StringBuffer( ST_COL_STRUCTURE_BUF );

        sadwStructureAndDimensions = doCurrentDataObject.getStructureAndDimensions();
        stStructureName = sadwStructureAndDimensions.getStructureName();
        adArrayDimensions = sadwStructureAndDimensions.getArraySize();

        // STRUCTURE NAME
        sbTrackedValueBuffer.replace( 0, stStructureName.length(), stStructureName );
                        
        // X-DIMENSION VALUE
        iDimension = adArrayDimensions.getXDimension();
        stDimensionValue = Integer.toString( iDimension );
        iDimensionLength = stDimensionValue.length();
        iNumberIndexShift = ( I_MAXIMUM_NUMBER_SIZE - 1 ) - iDimensionLength;
        iTotalIndexShift = I_X_DIMENSION_START_INDEX + iNumberIndexShift;
                        
        sbTrackedValueBuffer.replace( iTotalIndexShift,
            iTotalIndexShift + iDimensionLength, stDimensionValue );

        // Y-DIMENSION VALUE
        iDimension = adArrayDimensions.getYDimension();
        stDimensionValue = Integer.toString( iDimension );                                                
        iDimensionLength = stDimensionValue.length();
        iNumberIndexShift = ( I_MAXIMUM_NUMBER_SIZE - 1 ) - iDimensionLength;
        iTotalIndexShift = I_Y_DIMENSION_START_INDEX + iNumberIndexShift;
                        
        sbTrackedValueBuffer.replace( iTotalIndexShift,
            iTotalIndexShift + iDimensionLength, stDimensionValue );
         
        // Convert buffer to string
        return sbTrackedValueBuffer.toString();
    }
    
    private BitSet evaluateModelNumberAndObject( String stNameOfObject, 
        DataObject doDataObject ) {
            
        BitSet bsModelInformation;

        // all bits initially set to false
        bsModelInformation = new BitSet( 2 );
        
        switch ( iModelNumber ) {
            case Declarations.I_ALL_OBJECTS:
                bsModelInformation.set( I_OBJECT_INCLUDED_IN_MODEL );
                break;
                
            case Declarations.I_RECORD_OBJECTS:
                if ( doDataObject.isRecord() == true )
                    bsModelInformation.set( I_OBJECT_INCLUDED_IN_MODEL );
                break;
               
            case Declarations.I_NON_RECORD_OBJECTS:
                if ( doDataObject.isRecord() == false )
                    bsModelInformation.set( I_OBJECT_INCLUDED_IN_MODEL );                    
                break;

            case Declarations.I_RECORD_OBJECT_FIELDS:
                if ( stRecordName.compareTo( stNameOfObject ) == I_SAME_STRING ) {
                    bsModelInformation.set( I_IS_RECORD_OBJECT );
                    bsModelInformation.set( I_OBJECT_INCLUDED_IN_MODEL );
                }
                break;
        }
        
        return bsModelInformation;
    }
            
    /////////////////  HELPER FUNCTIONS \\\\\\\\\\\\\\\\\
    private Status updatePackedInformationList() {               
        DataObject doThisDataObject;
        Object objThisObject;
        int iNumberOfObjects, iIndex;
        Status statInformation;
        String stMessage, stObjectName;
        ArrayList alGlobalObjectNames;

        if ( alPackedInformationList.size() > 0 ) {
           alPackedInformationList.clear(); 
        }
        
        if ( alObjectNames.size() > 0 ) {
           alObjectNames.clear(); 
        }        
 
        if ( iModelNumber == Declarations.I_RECORD_OBJECT_FIELDS ) {
            objThisObject = deDeclarationModel.getDeclaration( stRecordName );            
            
            if ( objThisObject instanceof DataObject ) {
                bSuppressTableUpdates = true;
                
                doThisDataObject = (DataObject)objThisObject;

                // adds all fields from the record to alPacked... and alObjectNames
                addedOrUpdatedDeclaration( stRecordName, 
                    doThisDataObject, DeclarationsListener.B_ADD_OBJECT );                
                
                bSuppressTableUpdates = false;
                
                //fireTableDataChanged();
                updateTableView();
                statInformation = new Status( I_OK );
            }
            else if ( objThisObject instanceof Status ) {
                statInformation = (Status)objThisObject;
            }
            else {
                stMessage = "DeclarationsTableModel.updatePackedInformationList: "
                    + "Did not receive object of DataObject or Status.";
                statInformation = new Status( I_UNKNOWN, stMessage );
            }
        }
        else { // any other objects that are not fields in a record
            objThisObject = deDeclarationModel.getGlobalObjectNames();
            
            if ( objThisObject instanceof ArrayList ) {
                alGlobalObjectNames = (ArrayList)objThisObject;
                iNumberOfObjects = alGlobalObjectNames.size();

                // if Status was not returned from getFieldNames,
                // then DataObjects should exist in Declarations
                // let added... filter what is important
                
                bSuppressTableUpdates = true;
                
                for ( iIndex = 0; iIndex < iNumberOfObjects; iIndex++ ) {
                    stObjectName = (String)alGlobalObjectNames.get( iIndex );

                    // assuming DataObjects properly exist, can cast immediately
                    // without a test
                    doThisDataObject = (DataObject)
                        deDeclarationModel.getDeclaration( stObjectName );
                    addedOrUpdatedDeclaration( stObjectName, 
                        doThisDataObject, DeclarationsListener.B_ADD_OBJECT );                
                }
                
                bSuppressTableUpdates = false;
                
                //fireTableDataChanged();
                updateTableView();
                statInformation = new Status( I_OK );
            }
            else if ( objThisObject instanceof Status ) {
                statInformation = (Status)objThisObject;
            }
            else {
                stMessage = "DeclarationsTableModel.updatePackedInformationList: "
                    + "Did not receive object of ArrayList or Status.";
                statInformation = new Status( I_UNKNOWN, stMessage );
            }
        }

        //get list of all data objects relevant to model/scope
        //record object fields only needs the record name & its data object
        //  containing its record fields
        
        return (Status)statInformation.clone();
    }
    
    private void updateTableView() {
        if ( bSuppressTableUpdates == false )
            fireTableDataChanged();
    }
    
    ///////////// ABSTRACT MODEL INTERFACE FUNCTIONS \\\\\\\\\\\\\\
    public int getRowCount() {
        int iRows;
        Integer intRows;
        Object objThisObject;
        String stMessage;
        boolean bValidModel;
        Status statInformation;


        bValidModel = true;
        
        
        switch ( iModelNumber ) {
            case Declarations.I_ALL_OBJECTS:
                objThisObject = 
                    deDeclarationModel.getNumberOfObjects( Declarations.I_ALL_OBJECTS,
                        ST_NO_RECORD );                               
                break;
                
            case Declarations.I_RECORD_OBJECTS:
                objThisObject = 
                    deDeclarationModel.getNumberOfObjects( Declarations.I_RECORD_OBJECTS, 
                        ST_NO_RECORD );
                break;
               
            case Declarations.I_NON_RECORD_OBJECTS:
                objThisObject = 
                    deDeclarationModel.getNumberOfObjects( Declarations.I_NON_RECORD_OBJECTS, 
                        ST_NO_RECORD );
                break;
                
            case Declarations.I_RECORD_OBJECT_FIELDS:
                objThisObject = 
                    deDeclarationModel.getNumberOfObjects( Declarations.I_RECORD_OBJECT_FIELDS,
                        stRecordName );
                break;
                
            default:
                bValidModel = false;
                objThisObject = new Object();
                break;
        }
        
        if ( objThisObject instanceof Integer ) {
            intRows = (Integer)objThisObject;
            iRows = intRows.intValue();
        }
        else if ( objThisObject instanceof Status ) {
            statInformation = (Status)objThisObject;
            iRows = I_UNDEFINED;
        }
        else if ( bValidModel == false ) {
            //stMessage = "getRowCount: Invalid model.";
            //statInformation = new Status( I_ILLEGAL_MODEL, stMessage );
            iRows = I_UNDEFINED;
        }
        else {
            //stMessage = "getRowCount: Expected Status or Integer.";
            //statInformation = new Status( I_UNDEFINED, stMessage );
            iRows = I_UNDEFINED;
        }
        
        return iRows; 
    }
       
    public int getColumnCount() {
        return I_COLUMN_COUNT;
    }
    
    public Object getValueAt( int iRow, int iColumn ) {
        Object objThisObject;
        NameAndDataObjectWrapper nadowWrappedObject;
        DataObject doThisObject;
        Status statInformation;
        String stMessage;
        boolean bValidModel;
        int iPropertyValue;

        bValidModel = true;


        switch ( iModelNumber ) {
            case Declarations.I_ALL_OBJECTS:
                objThisObject = 
                    deDeclarationModel.getFieldAtIndex( Declarations.I_ALL_OBJECTS,
                        iRow, ST_NO_RECORD );                               
                    /**DEBUGGING**/
//                    NameAndDataObjectWrapper nadow = (NameAndDataObjectWrapper)
///                        objThisObject;
//                    DataObject doObj = nadow.getObject();
//                    System.out.println("IN DECLARATIONS_TABLE_MODEL...row = " +iRow+", column = " + iColumn + ": " );
//                    doObj.printDataObject();                    
                    /**END EBUG**/
                break;
                
            case Declarations.I_RECORD_OBJECTS:
                objThisObject = 
                    deDeclarationModel.getFieldAtIndex( Declarations.I_RECORD_OBJECTS, 
                        iRow, ST_NO_RECORD );
                break;
               
            case Declarations.I_NON_RECORD_OBJECTS:
                objThisObject = 
                    deDeclarationModel.getFieldAtIndex( Declarations.I_NON_RECORD_OBJECTS, 
                        iRow, ST_NO_RECORD );
                break;
                
            case Declarations.I_RECORD_OBJECT_FIELDS:
                objThisObject = 
                    deDeclarationModel.getFieldAtIndex( Declarations.I_RECORD_OBJECT_FIELDS,
                        iRow, stRecordName );
                break;

            default:
                bValidModel = false;
                objThisObject = new Object();
        }
       
        if ( objThisObject instanceof NameAndDataObjectWrapper ) {
            nadowWrappedObject = (NameAndDataObjectWrapper)objThisObject;
            doThisObject = nadowWrappedObject.getObject();
                        
            switch ( iColumn ) {
                case I_COLUMN_ICON:
                    objThisObject = doThisObject.getMenuIcon();
                    //objThisObject = doThisObject.getIcon();
                    //objThisObject = (doThisObject.getIcon()).getDescription();
                    break;

                case I_COLUMN_NAME:
                    objThisObject = nadowWrappedObject.getName();
                    break;
                        
                case I_COLUMN_TYPE:
                    iPropertyValue = doThisObject.getType();                    
                    objThisObject = DataObjectPropertyTextView.getTypeName( iPropertyValue );
                    break;
                        
                case I_COLUMN_RESTRICTION:
                    iPropertyValue = doThisObject.getRestriction();                    
                    objThisObject = DataObjectPropertyTextView.getRestrictionName( iPropertyValue );
                    break;
                        
                case I_COLUMN_STRUCTURE:
                    // convert wrapped data object to String type
                    
                    StructureAndDimensionsWrapper sadwColumnData;
                    String stCombinedData;
                    ArrayDimensions adColumnDataDimensions;
                    int iDimension;
                                        
                    sadwColumnData = doThisObject.getStructureAndDimensions();
                    stCombinedData = sadwColumnData.getStructureName();
                    /*adColumnDataDimensions = sadwColumnData.getArraySize();
                    
                    //System.out.println( "Name = " + nadowWrappedObject.getName() + "   Strucutre = " + stCombinedData );
                    if ( ( doThisObject.getStructure() != 
                        DataObject.I_STRUCTURE_SINGLE_VALUE )  && 
                         ( doThisObject.getType() != DataObject.I_TYPE_LABEL ) ) {
                        iDimension = adColumnDataDimensions.getXDimension();
                        //stCombinedData = stCombinedData + ": " +
                        //    Integer.toString( iDimension ) + "x";

                        iDimension = adColumnDataDimensions.getYDimension();
                        stCombinedData = stCombinedData + 
                            Integer.toString( iDimension );
                    }*/
                    
                    objThisObject = stCombinedData;
                    break;
                        
                case I_COLUMN_SIZE:
                    //objThisObject = new Integer( doThisObject.getSize() );
                    objThisObject = Integer.toString( doThisObject.getSize() );
                    break;
                        
                default:
                    stMessage = "The column number " + iColumn + " does "
                        + "not correspond to an object property.";
                    objThisObject = new Status( I_ILLEGAL_COLUMN, stMessage );
                    break;                        
            }                
        }
        // else if...Status will already be in objThisObject

        else if ( bValidModel == false ) {
            stMessage = "getValueAt: Invalid model.";
            objThisObject = new Status( I_ILLEGAL_MODEL, stMessage );
        }
        else {
            stMessage = "DeclarationsTableModel.getValueAt: Expected Status or NameAndDataObjectWrapper.";
            objThisObject = new Status( I_UNDEFINED, stMessage );
        }

        return objThisObject;
    }
       
    public String getColumnName( int iColumn ) {
        String stColumnName;
        
        switch ( iColumn ) {
            case I_COLUMN_ICON:
                stColumnName = "Icon";
                break;
                
            case I_COLUMN_NAME:
                stColumnName = "Name";
                break;

            case I_COLUMN_TYPE:
                stColumnName = "Type";
                break;

            case I_COLUMN_RESTRICTION:
                stColumnName = "Restriction";
                break;

            case I_COLUMN_STRUCTURE:
                stColumnName = "Structure";
                break;

            case I_COLUMN_SIZE:
                stColumnName = "Size";
                break;

            default:
                stColumnName = "?";
                break;
        }
    
        return stColumnName;
    }
    
    // not required but recommended in the Java Tutorial from Sun in
    // order to allow the correct cell renderer to be called for 
    // the JTable
    public Class getColumnClass( int iColumn ) {
        return getValueAt( 0, iColumn ).getClass();
    }
    
    
    // instance variables
    private Declarations deDeclarationModel; 
    private int iModelNumber, iColumnToTrack, iPackedInformationLength;
    private String stRecordName, stBufferInitializer;
    private ArrayList alPackedInformationList, alObjectNames;
    private boolean bSuppressTableUpdates;
        
    // class variables
        public static final int I_INDEX_STRING_LENGTH = 4;
    
        private static final int I_COLUMN_COUNT = 6;    
        private static final String ST_NO_RECORD = "";
        private static final int I_SAME_STRING = 0;
            
        // model numbers
        public static int I_MODEL_MINIMUM_NUMBER = Declarations.I_NON_RECORD_OBJECTS;
        public static int I_MODEL_MAXIMUM_NUMBER = Declarations.I_RECORD_OBJECT_FIELDS;        
    
        // column codes
        public static final int I_COLUMN_ICON = 0;
        public static final int I_COLUMN_NAME = 1;
        public static final int I_COLUMN_TYPE = 2;
        public static final int I_COLUMN_RESTRICTION = 3;
        public static final int I_COLUMN_STRUCTURE = 4;
        public static final int I_COLUMN_SIZE = 5;        

/*chars*/ // column type buffer initializers
/*21*/  private static final String ST_COL_ICON_BUF = ".....................|";
/* 8*/  private static final String ST_COL_NAME_BUF = "........|";
/*24*/  private static final String ST_COL_TYPE_BUF = "........................|";
/* 8*/  private static final String ST_COL_RESTRICTION_BUF = "........|";
/*21*/  private static final String ST_COL_STRUCTURE_BUF = "............:.000x000|";
/* 4*/  private static final String ST_COL_SIZE_BUF = "0000|";
        
        // ST_INDEX_NUMBER_BUF used to be here but is not needed
        // because it is used to store a number and is not for
        // sorting, so no leading zeroes are needed "0000";
        // plus Integer function does not extract the number correctly
            
/* 0*/  private static final String ST_NOTHING = "";

        // column starting points after non-NAME field
        private static final int I_X_DIMENSION_START_INDEX = 14;
        private static final int I_Y_DIMENSION_START_INDEX = 18;
        private static final int I_MAXIMUM_NUMBER_SIZE = 4;
        
        private static int I_NAME_BUFFER_LENGTH;

        private static int I_AFTER_ICON_INDEX;
        private static int I_AFTER_TYPE_INDEX;
        private static int I_AFTER_RESTRICTION_INDEX;
        private static int I_AFTER_STRUCTURE_INDEX;
        private static int I_AFTER_SIZE_INDEX;
        private static final int I_AFTER_NAME_INDEX = 0;
        
        // error codes
        public static final int I_OK = 0;
        public static final int I_ILLEGAL_COLUMN = -1;
        public static final int I_ILLEGAL_MODEL = -2;
        public static final int I_CANNOT_TRACK_COLUMN = -3;
        public static final int I_UNKNOWN = -4;
        public static final int I_UNDEFINED = -8888;
        
        // ModelInformation BitSet codes (based on integer indices of BitSet)
        public static final int I_IS_RECORD_OBJECT = 0;
        public static final int I_OBJECT_INCLUDED_IN_MODEL = 1;
        
        
        // class initializer
        static {
            I_AFTER_ICON_INDEX = ST_COL_ICON_BUF.length();
            I_AFTER_TYPE_INDEX = ST_COL_TYPE_BUF.length();
            I_AFTER_RESTRICTION_INDEX = ST_COL_RESTRICTION_BUF.length();
            I_AFTER_STRUCTURE_INDEX = ST_COL_STRUCTURE_BUF.length();
            I_AFTER_SIZE_INDEX = ST_COL_SIZE_BUF.length();

            I_NAME_BUFFER_LENGTH = ST_COL_NAME_BUF.length();
            
            I_MODEL_MINIMUM_NUMBER = Declarations.I_NON_RECORD_OBJECTS;
            I_MODEL_MAXIMUM_NUMBER = Declarations.I_RECORD_OBJECT_FIELDS;                    
        }
}