/*
 * LiteralDialog.java
 *
 * Created on June 3, 2002, 8:10 PM
 */

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;

/**
 *
 * @author  _
 * @version 
 */
public class LiteralDialog extends JDialog implements ActionListener {

    public LiteralDialog( JFrame jfParentFrame ) {
        super( jfParentFrame, "Literal Object", B_MODAL );
        
        Container cPane;
        GridBagLayout gblLayoutGrid;
        
        gblLayoutGrid = new GridBagLayout();
        cPane = this.getContentPane();
        jpDialogPanel = new JPanel();       
        cPane.add( jpDialogPanel );
        jpDialogPanel.setLayout( gblLayoutGrid );
        
        fontLabel = new Font( "SansSerif", Font.BOLD, 14 );
        fontField = new Font( "monospaced", Font.PLAIN, 14 );
        
        jcbTypeAndStructure = new JComboBox();        
        jcbTypeAndStructure.setFont( fontField );
        
        jbOK = createButton( "OK" );
        jbCancel = createButton( "Cancel" );
        jbHelp = createButton( "Help" );
        
        jtfLiteral = new JTextField();
        jtfLiteral.setFont( fontField );
        
        jlLiteral = createLabel( "Literal value" );
        jlLiteral.setForeground( Color.black );

        jlTypeAndStructure = createLabel( "Type and structure" );        
        
        ptvhdHelp = new PTVHelpDialog( jfParentFrame, "Help >> General >> Literals" );
        ptvhdHelp.setHelpFile( "help_general_literal.html" );
        
        this.setSize( 335, 150 );
        createPanel();
        
        this.addWindowListener( new WindowAdapter() {
            public void windowClosing( WindowEvent weEvent ) {
                hide();
                dispose();
            }
        } );        
    }        
        
    public void setRestrictionFlags( BitSet bsFlags ) {
        bsRestrictionFlags = bsFlags;
        jcbTypeAndStructure.removeAllItems();

        if ( ( bsRestrictionFlags.get( AllPurposeMenu.I_FLAG_CHARACTER_ARRAY ) ) ||
             ( bsRestrictionFlags.get( AllPurposeMenu.I_FLAG_DISPLAY_OUTPUT ) ) ||
             ( bsRestrictionFlags.get( AllPurposeMenu.I_FLAG_KEYBOARD_INPUT ) ) ) {
            jcbTypeAndStructure.addItem( st1dListItemNames[ I_FLAG_ARRAY_CHARACTER ] );
            System.out.println( "Adding character array." );
        }
        
        if ( bsRestrictionFlags.get( AllPurposeMenu.I_FLAG_CHARACTER_SINGLE ) ) {
            jcbTypeAndStructure.addItem( st1dListItemNames[ I_FLAG_SINGLE_CHARACTER ] );
            System.out.println( "Adding single character." );
        }
        
        if ( bsRestrictionFlags.get( AllPurposeMenu.I_FLAG_INT_SINGLE ) ) {
            jcbTypeAndStructure.addItem( st1dListItemNames[ I_FLAG_SINGLE_INTEGER ] );
            System.out.println( "Adding integer." );
        }        
    }
    
    public Object displayDialog() {
        jtfLiteral.setText( "" );
        this.show();  // does not return until OK or Cancel
        return objReturnObject;
    }
    
    //------------------ ActionListener Implementation -----------------\\
    public void actionPerformed( ActionEvent aeEvent ) {
        Object objEventSource;
        String stMenuItem, stLiteral;
        Status statInformation;
        boolean bValid;
        DataObject doLiteral;
        
        objEventSource = aeEvent.getSource();
        
        if ( objEventSource == jbOK ) {
            bValid = true;
            doLiteral = new DataObject();
            doLiteral.setRestriction( DataObject.I_RESTRICTION_LITERAL );
            stMenuItem = (String)jcbTypeAndStructure.getSelectedItem();

            if ( stMenuItem.equals( st1dListItemNames[ I_FLAG_SINGLE_INTEGER ] ) ) {
                doLiteral.setStructure( DataObject.I_STRUCTURE_SINGLE_VALUE );
                doLiteral.setType( DataObject.I_TYPE_INTEGER );
            }            
            else if ( stMenuItem.equals( st1dListItemNames[ I_FLAG_SINGLE_CHARACTER ] ) ) {
                doLiteral.setStructure( DataObject.I_STRUCTURE_SINGLE_VALUE );
                doLiteral.setType( DataObject.I_TYPE_CHARACTER );
            }            
            else if ( stMenuItem.equals( st1dListItemNames[ I_FLAG_ARRAY_CHARACTER ] ) ) {
                doLiteral.setStructure( DataObject.I_STRUCTURE_1D_ARRAY );
                doLiteral.setType( DataObject.I_TYPE_CHARACTER );
            }
            else {
                bValid = false;
            }
            
            if ( bValid == true ) {
                stLiteral = jtfLiteral.getText();
                doLiteral.setValue( stLiteral );
                
                if ( ((String)(jcbTypeAndStructure.getSelectedItem())).equals(
                    st1dListItemNames[ I_FLAG_ARRAY_CHARACTER ] ) ) {
                    doLiteral.setArrayDimensions( stLiteral.length() + 2, 1 );
                }
                    
                statInformation = doLiteral.validateValue();
            
                if ( statInformation.getStatusNumber() == I_OK ) {
                    objReturnObject = doLiteral;
                    System.out.println( "Literal = " + doLiteral.getValue() );
                    System.out.println( doLiteral.getArrayDimensions().getXDimension() );
                    
/*                    if ( ((String)(jcbTypeAndStructure.getSelectedItem())).equals(
                        st1dListItemNames[ I_FLAG_ARRAY_CHARACTER ] ) ) {
                        objReturnObject = doLiteral.getValidatedCharacterArray();
                        stLiteral = (String)objReturnObject;
                        doLiteral.setArrayDimensions( stLiteral.length(), 1 );
                    }
                    else {
                        objReturnObject = doLiteral.getValidatedValue( 0 );
                    }*/

                    this.hide();
                    this.dispose();
                }
                else {
                    objReturnObject = new Object();
                    JOptionPane.showMessageDialog( this, 
                        statInformation.getStatusMessage(),
                        ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );
                }
            } 
            else {  // invalid menu selection
                JOptionPane.showMessageDialog( this, 
                    "Did not make a valid menu selection.  Please try again.",
                    ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );
            }                
        }
        else if ( objEventSource == jbCancel ) {
            objReturnObject = new Status( I_CANCELLED_OPERATION, "Operation cancelled." );
            this.hide();
            this.dispose();
        }
        else if ( objEventSource == jbHelp ) {
            ptvhdHelp.show();
        }
    }
        
    /////////////////// PRIVATE METHODS \\\\\\\\\\\\\\\\\\\\\\\\\
    private void createPanel() {
        addComponentToPanel( 0, 0, 2, 1, jlTypeAndStructure );
        addComponentToPanel( 2, 0, 3, 1, jcbTypeAndStructure );
        addComponentToPanel( 0, 1, 2, 1, jlLiteral );
        addComponentToPanel( 2, 1, 3, 1, jtfLiteral );
        addComponentToPanel( 0, 2, 1, 1, new JLabel( "     " ) );
        addComponentToPanel( 1, 2, 1, 1, jbOK );
        addComponentToPanel( 2, 2, 1, 1, jbCancel );
        addComponentToPanel( 3, 2, 1, 1, jbHelp );
        addComponentToPanel( 4, 2, 1, 1, new JLabel( "     " ) );
    }
    
    private void addComponentToPanel( int iGridXPos, int iGridYPos, int iGridWidth,
                                      int iGridHeight, Component coComponent ) {
        
        GridBagConstraints gbcLayoutConstraints;

        gbcLayoutConstraints = new GridBagConstraints();
        gbcLayoutConstraints.weightx = 100;
        gbcLayoutConstraints.weighty = 100;
        gbcLayoutConstraints.gridx = iGridXPos;
        gbcLayoutConstraints.gridy = iGridYPos;
        gbcLayoutConstraints.gridwidth = iGridWidth;
        gbcLayoutConstraints.gridheight = iGridHeight;
        gbcLayoutConstraints.anchor = GridBagConstraints.SOUTHWEST;
        gbcLayoutConstraints.insets = new Insets( 2, 5, 2, 3 );
        gbcLayoutConstraints.fill =  GridBagConstraints.HORIZONTAL;
        jpDialogPanel.add( coComponent, gbcLayoutConstraints );
    }
    
    private JButton createButton( String stText ) {
        JButton jbThisButton;
        
        jbThisButton = new JButton( stText );
        jbThisButton.addActionListener( this );
        return jbThisButton;
    }
    
    private JLabel createLabel( String stText ) {
        JLabel jlThisLabel;
        
        jlThisLabel = new JLabel( stText );
        jlThisLabel.setFont( fontLabel );
        jlThisLabel.setForeground( Color.black );
        return jlThisLabel;
    }
    
    // instance variables
    private String stProgramName;
    private JButton jbOK, jbCancel, jbHelp;
    private JLabel jlTypeAndStructure, jlLiteral;
    private JTextField jtfLiteral;
    private JComboBox jcbTypeAndStructure;
    private JPanel jpDialogPanel;
    private BitSet bsRestrictionFlags;
    private Font fontLabel, fontField;
    private Object objReturnObject;
    private PTVHelpDialog ptvhdHelp;
    
    // class constants
    private static final boolean B_MODAL = true;
    private static final String ST_APPLICATION_NAME = "Panels :: Tools :: Views";
    
    public static String[] st1dListItemNames;
    public static final int I_FLAG_SINGLE_INTEGER = 0;
    public static final int I_FLAG_SINGLE_CHARACTER = 1;
    public static final int I_FLAG_ARRAY_CHARACTER = 2;
    
    public static final int I_OK = 0;
    public static final int I_CANCELLED_OPERATION = -1;
    
    static { 
        st1dListItemNames = new String[ 3 ];
        st1dListItemNames[ I_FLAG_SINGLE_INTEGER ] = "Single Integer";
        st1dListItemNames[ I_FLAG_SINGLE_CHARACTER ] = "Single Character";
        st1dListItemNames[ I_FLAG_ARRAY_CHARACTER ] = "Array of Characters";
    }
}