/*
 * MainPanel.java
 *
 * Created on September 14, 2001, 2:14 PM
 */

//package panelsAndTools;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;

/**
 *
 * @author  family
 * @version 
 */
public class MainPanel extends JPanel 
    implements ActionListener, CodeTableListener {

    /** Creates new form MainPanel */
    public MainPanel( JFrame jfThisFrame, Program progProgram ) {
        bDoneWithInitialization = false;    
        vctCodeTable = progProgram.getVisualCodeTable();
        progThisProgram = progProgram;
        deDeclarations = progThisProgram.getDeclarations();
        
        codeThisCode = progThisProgram.getCode();
        jfParentFrame = jfThisFrame;
        
        vctCodeTable.setCodeTableListener( this );
        
        sdtmObjects = new SortableDeclarationsTableModel();        
        sdtmObjects.setDeclarationsModel( deDeclarations );
        sdtmObjects.setColumnToTrack( DeclarationsTableModel.I_COLUMN_TYPE );
        sdtmObjects.setModelNumber( Declarations.I_ALL_OBJECTS );

        apmMenu = new AllPurposeMenu( this );
        apmMenu.setSortableDeclarationsModel( sdtmObjects );
        apmMenu.setCodeObject( progThisProgram.getCode() );
        
        if ( bDoneWithInitialization == false ) {
            apmMenu.setMenuForInitalStartup();
        }
        
        setLayout( new BorderLayout() );
        
        addMouseListener( new MouseAdapter() {
            public void mouseReleased( MouseEvent meEvent ) {                
                apmMenu.updateCodeEditingMode();
                System.out.println( "Main Panel: Mouse Pressed" );                

                if ( meEvent.isPopupTrigger() ) {
                    jfParentFrame.toFront();
                    iCurrentXLocation = meEvent.getX();
                    iCurrentYLocation = meEvent.getY();                
                    apmMenu.show( meEvent.getComponent(), iCurrentXLocation, 
                        iCurrentYLocation );
                }
                else {
                    showAllActivePanels();
                }
            }
            
            /*public void mousePressed( MouseEvent meEvent ) {
                //showAllActivePanels();
                if ( meEvent.isPopupTrigger() == false ) {
                    showAllActivePanels();
                }
            }*/
        });
        
        initializePanels();        
    }
    
    public void showAllActivePanels() {
        if ( jdMath.isVisible() == true ) {
           jdMath.show();
        }
        
        if ( jdBranchAndCompare.isVisible() == true ) {
           jdBranchAndCompare.show();
        }

        if ( jdMoveParts.isVisible() == true ) {
           jdMoveParts.show();
        }

        if ( jdMoveAll.isVisible() == true ) {
           jdMoveAll.show();
        }

        if ( jdSectionComments.isVisible() == true ) {
           jdSectionComments.show();
        }

        if ( jdObjectEditor.isVisible() == true ) {
           jdObjectEditor.show();
        }        
    }    
    
    //------------------- ActionListener Implementation -------------------\\   
    public void actionPerformed( ActionEvent event ) {
        String stCommandString, stProgramName;

        System.out.println( "MAIN PANEL: actionPerformed" );
        stCommandString = event.getActionCommand();
        
        if ( stCommandString.equals( AllPurposeMenu.ST_MENUITEM__PANEL_BRANCH_COMPARE ) ) {
            showPanel( I_PANEL__BRANCH_AND_COMPARE, null, false );
        }
        else if ( stCommandString.equals( AllPurposeMenu.ST_MENUITEM__PANEL_MATH ) ) {                
            showPanel( I_PANEL__MATH, null, false );
        }
        else if ( stCommandString.equals( AllPurposeMenu.ST_MENUITEM__PANEL_MOVE ) ) {
            showPanel( I_PANEL__MOVE_ALL, null, false );
        }
        else if ( stCommandString.equals( 
            AllPurposeMenu.ST_MENUITEM__PANEL_MOVE_PARTS ) ) {
            showPanel( I_PANEL__MOVE_PARTS, null, false );
        }
        else if ( stCommandString.equals( 
            AllPurposeMenu.ST_MENUITEM__PANEL_OBJECT_EDITOR ) ) {
            showPanel( I_PANEL__OBJECT_EDITOR, null, false );
        }
        else if ( stCommandString.equals( AllPurposeMenu.ST_MENUITEM__PANEL_COMMENTS ) ) {
            showPanel( I_PANEL__SECTION_COMMENTS, null, false );
        }
        else {
            // ignore return value...this is the last option
            progThisProgram.actionPerformedInProgram( stCommandString );
            System.out.println( "Program name = " + progThisProgram.getProgramName() );
            if ( bDoneWithInitialization == false ) {
                if ( progThisProgram.getProgramName().equals( "" ) == false ) {
                    bDoneWithInitialization = true;
                    apmMenu.setMenuForEmptyArea();
                    System.out.println( "Done with initializataion." );
                }
            }
            else {
                if ( progThisProgram.getProgramName().equals( "" ) == true ) {
                    apmMenu.setMenuForInitalStartup();                    
                    System.out.println( "No program name." );
                }
                else {
                    apmMenu.setMenuForEmptyArea();
                }
            }
        }
    }
        
    public void paintComponent( Graphics currentGraphics ) {
        super.paintComponent( currentGraphics );
    }

    public void mouseMoved( MouseEvent mouseEvent ) {}
    public void mouseDragged( MouseEvent mouseEvent ) {}

    //----------------- CodeTableListener Implementation -----------\\
    public void showPanelForLineOfCode( Instruction instThisInstruction ) {
        Opcode opThisOpcode;
        int iOpcodeID;
        
        opThisOpcode = instThisInstruction.getOpcode();
        iOpcodeID = opThisOpcode.getOpcode();
        System.out.println( "Trying to move the cursor...with Instruction..." );
        
        System.out.println( "OPCODE = " + iOpcodeID );
        System.out.println( "LEFT = " + instThisInstruction.getLeftOperand().getOperandName() );
        System.out.println( "RIGHT = " + instThisInstruction.getRightOperand().getOperandName() );
        System.out.println( "COMMENTS = " + instThisInstruction.getComments() );
        
        if ( ( iOpcodeID >= Opcode.I_OPCODE__MATH_FIRST )  &&  
             ( iOpcodeID <= Opcode.I_OPCODE__MATH_LAST ) ) {
            showPanel( I_PANEL__MATH, instThisInstruction, true );
        }
        else if ( ( iOpcodeID >= Opcode.I_OPCODE__BRANCH_FIRST )  &&  
                  ( iOpcodeID <= Opcode.I_OPCODE__BRANCH_LAST ) ) {
            showPanel( I_PANEL__BRANCH_AND_COMPARE, instThisInstruction, true );
        }
        else if ( iOpcodeID == Opcode.I_OPCODE__MOVE_ALL ) {
            showPanel( I_PANEL__MOVE_ALL, instThisInstruction, true );
        }
        else if ( iOpcodeID == Opcode.I_OPCODE__MOVE_PARTS ) {
            showPanel( I_PANEL__MOVE_PARTS, instThisInstruction, true );
        }
        else if ( iOpcodeID == Opcode.I_OPCODE__COMMENTS ) {
            showPanel( I_PANEL__SECTION_COMMENTS, instThisInstruction, true );
        }
    }
    
    /////////////////////// PRIVATE METHODS \\\\\\\\\\\\\\\\\\\\\\
    private void showPanel( int iPanelID, Instruction instInstruction,
        boolean bShowCurrentInstruction ) {
                
        showAllActivePanels();
        
        switch ( iPanelID ) {            
            case I_PANEL__MATH:
                System.out.println( "SHOW PTV_PANEL" );
                jdMath.showDialog( iCurrentXLocation, iCurrentYLocation );
                if ( bShowCurrentInstruction == true ) {
                    jdMath.setInstruction( instInstruction );
                }
                break;
                
            case I_PANEL__BRANCH_AND_COMPARE:
                System.out.println( "SHOW PTV_PANEL" );
                jdBranchAndCompare.showDialog( iCurrentXLocation, iCurrentYLocation );
                if ( bShowCurrentInstruction == true ) {
                    jdBranchAndCompare.setInstruction( instInstruction );
                }
                break;
                
            case I_PANEL__MOVE_PARTS:
                System.out.println( "SHOW move parts" );
                jdMoveParts.showDialog( iCurrentXLocation, iCurrentYLocation );
                if ( bShowCurrentInstruction == true ) {
                    jdMoveParts.setInstruction( instInstruction );
                }
                break;
                
            case I_PANEL__MOVE_ALL:
                System.out.println( "SHOW move all" );
                jdMoveAll.showDialog( iCurrentXLocation, iCurrentYLocation );
                if ( bShowCurrentInstruction == true ) {
                    jdMoveAll.setInstruction( instInstruction );
                }

                break;
                
            case I_PANEL__SECTION_COMMENTS:
                System.out.println( "section comments,,,SHOW PTV_PANEL" );
                jdSectionComments.showDialog( iCurrentXLocation, iCurrentYLocation );
                if ( bShowCurrentInstruction == true ) {
                    jdSectionComments.setInstruction( instInstruction );
                }
                break;
                
            case I_PANEL__OBJECT_EDITOR:
                System.out.println( "SHOW PTV_PANEL" );
                jdObjectEditor.showDialog( iCurrentXLocation, iCurrentYLocation );
                break;
        }
    }
    
    private void initializePanels() {
        jdMath = new MathPanel( JFrameFactory.createJFrame(  
            ST_HOME_DIRECTORY + "math-icon.jpg" ), "Math", 
            new Dimension( 649, 190 ), progThisProgram, sdtmObjects );
        
        jdBranchAndCompare = new BranchAndComparePanel( JFrameFactory.createJFrame( 
            ST_HOME_DIRECTORY + "branch-and-compare-icon.jpg" ), "Compare & Branch",
            new Dimension( 654, 228 ), progThisProgram, sdtmObjects );
 
        jdMoveParts = new MovePartsPanel( JFrameFactory.createJFrame( 
            ST_HOME_DIRECTORY + "move-icon.jpg" ), "Move Characters", 
            new Dimension( 581, 213 ), progThisProgram, sdtmObjects );

        jdMoveAll = new MoveAllPanel( JFrameFactory.createJFrame(  
            ST_HOME_DIRECTORY + "move-icon.jpg" ), "Move", 
            new Dimension( 552, 190 ), progThisProgram, sdtmObjects );

        jdSectionComments = new SectionCommentsPanel( JFrameFactory.createJFrame(  
            ST_HOME_DIRECTORY + "comments.jpg" ), "Section Comments", 
            new Dimension( 299, 183 ), progThisProgram, sdtmObjects ); /*height = 217*/
        
        jdObjectEditor = 
            new ObjectEditorPanel( JFrameFactory.createJFrame(
                ST_HOME_DIRECTORY + "objects.jpg" ), "Object Editor",
                new Dimension( 600, 371 ), progThisProgram, sdtmObjects );
    }        


    
    // instance variables
    private int iCurrentXLocation, iCurrentYLocation, iStartingXLocation,
        iStartingYLocation, iEndingXLocation, iEndingYLocation;
 
    private VisualCodeTable vctCodeTable;
    private BranchAndComparePanel jdBranchAndCompare;
    private MathPanel jdMath;
    private MoveAllPanel jdMoveAll;
    private MovePartsPanel jdMoveParts;
    private ObjectEditorPanel jdObjectEditor;
    private SectionCommentsPanel jdSectionComments;
    private Declarations deDeclarations;
    private SortableDeclarationsTableModel sdtmObjects;   
    private JFrame jfParentFrame;  
    private AllPurposeMenu apmMenu;
    private Program progThisProgram;
    private boolean bDoneWithInitialization;
    private Code codeThisCode;
    
    // class constants
    private final String ST_HOME_DIRECTORY = "D:\\forte4j\\Development\\panelsAndTools\\";
    
    private static final int I_PANEL__MATH = 0;
    private static final int I_PANEL__BRANCH_AND_COMPARE = 1;
    private static final int I_PANEL__MOVE_PARTS = 2;
    private static final int I_PANEL__MOVE_ALL = 3;
    private static final int I_PANEL__SECTION_COMMENTS = 4;
    private static final int I_PANEL__OBJECT_EDITOR = 5;    
}