/*
 * MovePartsPanel.java
 *
 * Created on June 14, 2002, 9:25 PM
 */

import java.awt.event.*;
import javax.swing.*;
import java.util.*;
import java.awt.*;

/**
 *
 * @author  _
 * @version 
 */
public class MovePartsPanel extends PTVPanel
    implements ActionListener, DataObjectAreaListener, MouseListener {
        
    public MovePartsPanel( JFrame jfParentFrame, String stDialogTitle, 
        Dimension diPanelSize, Program progThisProgram, 
        SortableDeclarationsTableModel sdtmModel ) {
        
        super( jfParentFrame, stDialogTitle, diPanelSize, progThisProgram, sdtmModel );

        iFirstClickedOperandID = I_UNDEFINED;
        super.setActionListener( this );
        super.setDataObjectAreaListener( this );
        
        createObjects();
        drawPanel();
        finishPanelConfiguration();
    }
    
    public void setInstruction( Instruction instInstruction ) {
        instThisInstruction = (Instruction)instInstruction.clone();
        operDestination = instThisInstruction.getLeftOperand();
        operSource = instThisInstruction.getRightOperand();
        operDestination.clearFailedOperandInformation();
        operSource.clearFailedOperandInformation();
        
        // set up the panel appropriately
        dataObjectAreaChanged( iDestinationOperandID, operDestination );
        super.updateDataObjectArea( iDestinationOperandID, operDestination );
        dataObjectAreaChanged( iSourceOperandID, operSource );
        super.updateDataObjectArea( iSourceOperandID, operSource );        
        jlLabelText.setText( instThisInstruction.getLabel().getName() );
        jtaComments.setText( instThisInstruction.getComments() );

        plDestinationArrayRange.setMaximumArrayRange( 
            operDestination.getOperandArrayLength() );
        plDestinationArrayRange.setArrayRange( operDestination.getStartingArrayIndex(),
            operDestination.getEndingArrayIndex() );
        
        plSourceArrayRange.setMaximumArrayRange( operSource.getOperandArrayLength() );
        plSourceArrayRange.setArrayRange( operSource.getStartingArrayIndex(),
            operSource.getEndingArrayIndex() );
    }
    
    //------------------ ActionListener Implementation -----------------\\
    public void actionPerformed( ActionEvent aeEvent ) {
        Object objActionSource, objTemp;
        JButton jbThisButton;
        boolean bFound;
        int iIndex, iCurrentLine, iLabelUsedStatus;
        String stTemp;
        DataObject doThisDataObject;
        Status statInformation;
        
        iLabelUsedStatus = super.I_OK;
        objActionSource = aeEvent.getSource();
        statInformation = new Status( iLabelUsedStatus );
        
        if ( objActionSource == jbApply ) {
            // since the Instruction parts were likely defined legally, it is
            // not necessary to do error checking...unless in "debugging mode"            
            stTemp = this.getLabel().getText();

            if ( stTemp.equals( "" ) == false ) {
                statInformation = super.isLabelUsed( stTemp );
                iLabelUsedStatus = statInformation.getStatusNumber();

                if ( iLabelUsedStatus == super.I_OK ) {                
                    doThisDataObject = (DataObject)deDeclarations.getDeclaration( stTemp );
                    instThisInstruction.setLabel( 
                        new NameAndDataObjectWrapper( stTemp, doThisDataObject ) );
                }
                else {
                    JOptionPane.showMessageDialog( this, 
                        statInformation.getStatusMessage(), 
                        ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );                    
                }                                                                
            }
            else {
                doThisDataObject = new DataObject();
                doThisDataObject.setType( DataObject.I_TYPE_LABEL );
                instThisInstruction.setLabel( 
                    new NameAndDataObjectWrapper( "", doThisDataObject ) );
            }


            stTemp = this.getComments().getText();

            if ( stTemp.equals( "" ) == false ) {
                instThisInstruction.setComments( stTemp );
            }
            //  else, ignore

            if ( iLabelUsedStatus == super.I_OK ) {
                // operands are updated as the panel is clicked
                if ( operDestination.getOperandType() != DataObject.I_UNDEFINED ) {
                    instThisInstruction.setLeftOperand( operDestination );

                    if ( operSource.getOperandType() != DataObject.I_UNDEFINED ) {
                        instThisInstruction.setRightOperand( operSource );
                        statInformation = instThisInstruction.validateInstruction();

                        if ( statInformation.getStatusNumber() == Instruction.I_OK ) {
                            iCurrentLine = vctCodeTable.getCurrentRow();

                            if ( codeThisCode.getEditingMode() == Code.B_MODE_INSERT ) {
                                statInformation = codeThisCode.insertSingleLine( 
                                    iCurrentLine, instThisInstruction );
                            }
                            else {  // overwrite
                                statInformation = codeThisCode.changeSingleLine( 
                                    iCurrentLine, instThisInstruction );
                            }

                            if ( statInformation.getStatusNumber() != Instruction.I_OK ) {
                                JOptionPane.showMessageDialog( this, 
                                    statInformation.getStatusMessage(), 
                                    ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );                    
                            }  
                            else {
    //                                System.out.println( "25" );
                                instThisInstruction = new Instruction();
                                instThisInstruction.setOpcode( opMovePartsOpcode );
    //                                System.out.println( "26" );
                                clearAll();
    //                                System.out.println( "27" );
                            }                            

                        }
                        else {
                            JOptionPane.showMessageDialog( this, 
                                statInformation.getStatusMessage(),
                                ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );                    
                        }                           
                    }
                    else {
                        JOptionPane.showMessageDialog( this, 
                            "A source operand is required for this instruction.", 
                            ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );                    
                    }
                }
                else {
                    JOptionPane.showMessageDialog( this, 
                        "A destination operand is required for this instruction.", 
                        ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );                    
                }
            }
        }
        else {
            super.actionPerformed( aeEvent );
        
            if ( objActionSource == super.jbClear ) {
                clearAll();
            }        
        }            
    }
    
    //------------- DataObjectAreaListener Implementation ------------\\
    public void dataObjectAreaChanged( int iID, Operand operData ) {
        int iOperandType, iOperandStructure;
        
        iOperandType = operData.getOperandType();
        iOperandStructure = operData.getOperandStructure();
        
        if ( iID == iDestinationOperandID ) {
            operDestination = operData;
            updateRestrictionFlagsAndLengthLimit( iID );
            instThisInstruction.setLeftOperand( operDestination );
            
            if ( ( iOperandType == DataObject.I_SYSTEM_TYPE_KEYBOARD )  ||
                 ( ( iOperandStructure == DataObject.I_STRUCTURE_1D_ARRAY )  &&
                   ( iOperandType == DataObject.I_TYPE_CHARACTER ) ) ) {            
                plDestinationArrayRange.setMaximumArrayRange( 
                    operDestination.getOperandArrayLength() );
            }
            else if ( ( iOperandStructure == DataObject.I_STRUCTURE_SINGLE_VALUE )  &&
                      ( iOperandType == DataObject.I_TYPE_CHARACTER ) ) {                        
                plDestinationArrayRange.setArrayRangeForSingleCharacter();
            }
        }
        else if ( iID == iSourceOperandID ) {
            operSource = operData;
            updateRestrictionFlagsAndLengthLimit( iID );
            instThisInstruction.setRightOperand( operSource );

            if ( ( iOperandType == DataObject.I_SYSTEM_TYPE_KEYBOARD )  ||
                 ( ( iOperandStructure == DataObject.I_STRUCTURE_1D_ARRAY )  &&
                   ( iOperandType == DataObject.I_TYPE_CHARACTER ) ) ) {            
                plSourceArrayRange.setMaximumArrayRange( 
                    operSource.getOperandArrayLength() );
            }
            else if ( ( iOperandStructure == DataObject.I_STRUCTURE_SINGLE_VALUE )  &&
                      ( iOperandType == DataObject.I_TYPE_CHARACTER ) ) {                        
                plSourceArrayRange.setArrayRangeForSingleCharacter();
            }
        }
        // else ignore
    }
    
    //-----------------MouseListener Implementation ------------------\\
    public void mouseReleased( MouseEvent meEvent ) {
        Object objThisClickedObject;
        int iOperandType, iOperandStructure;
        
        if ( meEvent.isPopupTrigger() == true ) {
            System.out.println( "MOVE PARTS: Right mouse button pressed." );
            super.mouseReleased( meEvent );
        }
        else {
            System.out.println( "MOVE PARTS: LEFT mouse button pressed." );
            objThisClickedObject = meEvent.getSource();
            System.out.println( "clicked: " + objThisClickedObject );

            if ( objThisClickedObject == plSourceArrayRange ) {
                System.out.println( "SOURCE" );
                iOperandType = operSource.getOperandType();
                iOperandStructure = operSource.getOperandStructure();
                
                if ( ( iOperandType == DataObject.I_SYSTEM_TYPE_KEYBOARD )  ||
                     ( ( iOperandStructure == DataObject.I_STRUCTURE_1D_ARRAY )  &&
                       ( iOperandType == DataObject.I_TYPE_CHARACTER ) ) ) {
                   processArrayRangeDialog( operSource, operDestination,
                        plSourceArrayRange );
                }        
            }            
            else if ( objThisClickedObject == plDestinationArrayRange ) {
                System.out.println( "DESTINATION" );
                iOperandType = operDestination.getOperandType();
                iOperandStructure = operDestination.getOperandStructure();

                if ( ( iOperandType == DataObject.I_SYSTEM_TYPE_KEYBOARD )  ||
                     ( ( iOperandStructure == DataObject.I_STRUCTURE_1D_ARRAY )  &&
                       ( iOperandType == DataObject.I_TYPE_CHARACTER ) ) ) {
                   processArrayRangeDialog( operDestination, operSource,
                        plDestinationArrayRange );
                }
            }
        }
    }

    // unused methods
    public void mouseExited( MouseEvent meEvent ) {}
    public void mousePressed( MouseEvent meEvent ) {}
    public void mouseClicked( MouseEvent meEvent ) {}
    public void mouseEntered( MouseEvent meEvent ) {}    
    
    //////////////////////// PRIVATE METHODS \\\\\\\\\\\\\\\\\\\\\\\\\\
    private void createObjects() {
        ptvhdHelp = new PTVHelpDialog( jfParentFrame, "Help >> Panel >> Move Parts" );
        ptvhdHelp.setHelpFile( "help_panel_move_characters.html" );
        super.setHelpDialog( ptvhdHelp );
        
        jbApply = super.getApplyInstructionButton();
        jbApply.addActionListener( this );
        
        // just use display and keyboard and constant for one operand
        // ...no point in comparing two constants, two keyboards, two displays
        bsDestinationOpRestrictionFlags = super.getNewDataObjectAreaRestrictionList();
        bsDestinationOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_VARIABLE );
        bsDestinationOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_CHARACTER_ARRAY );
        bsDestinationOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_CHARACTER_SINGLE );
        
        bsSourceOpRestrictionFlags = super.getNewDataObjectAreaRestrictionList();
        bsSourceOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_CONSTANT );        
        bsSourceOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_VARIABLE );
        bsSourceOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_CHARACTER_ARRAY );
        bsSourceOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_CHARACTER_SINGLE );        

        bsDestinationOpBackupRestrictionFlags = (BitSet)bsDestinationOpRestrictionFlags.clone();
        bsSourceOpBackupRestrictionFlags = (BitSet)bsSourceOpRestrictionFlags.clone();
                                        
        operDestination = new Operand();
        operSource = new Operand();
        opMovePartsOpcode = new Opcode();
        opMovePartsOpcode.setOpcode( Opcode.I_OPCODE__MOVE_PARTS );
        instThisInstruction = new Instruction();
        instThisInstruction.setOpcode( opMovePartsOpcode );
    }
    
    private void finishPanelConfiguration() {
        // assume that everything worked ok...since IDs should be valid
        super.setDataObjectAreaRestrictionList( iDestinationOperandID, bsDestinationOpRestrictionFlags );
        super.setDataObjectAreaRestrictionList( iSourceOperandID, bsSourceOpRestrictionFlags );
    }
    
    private void drawPanel() {
        iSourceOperandID = super.drawDataObjectArea( new Point( 10, 7 ), operDestination );
        iDestinationOperandID = super.drawDataObjectArea( new Point( 216, 7 ), operSource );

        plSourceArrayRange = new PaintableLabel( 9, 114, 145, 56 );
        plSourceArrayRange.setGraphicsType( PaintableLabel.I_ARRAY_RANGE );
        plSourceArrayRange.setMaximumArrayRange( 0 );
        plSourceArrayRange.setArrayRange( 0, 0 );
        plSourceArrayRange.addMouseListenerToPaintableLabel( this );
        
        plDestinationArrayRange = new PaintableLabel( 214, 114, 145, 56 );
        plDestinationArrayRange.setGraphicsType( PaintableLabel.I_ARRAY_RANGE );
        plDestinationArrayRange.setMaximumArrayRange( 0 );
        plDestinationArrayRange.setArrayRange( 0, 0 );
        plDestinationArrayRange.addMouseListenerToPaintableLabel( this );
                
        super.drawLabelAndCommentsAndActionButtons();
        super.drawLine( new Rectangle( 154, 39, 44, 3 ) );  // line from Destination operand
        super.drawArrowhead( new Rectangle( 198, 34, 15, 14 ), PTVPanel.I_RIGHT_TRIANGLE );
        super.addPaintableLabel( plSourceArrayRange );
        super.addPaintableLabel( plDestinationArrayRange );
    }
    
    private void processArrayRangeDialog( Operand operThisOperand,
        Operand operOtherOperand, PaintableLabel plThisLabel ) {
            
        ArrayRangeDialog ardArrayRange;
        Object objReturnedObject;
        Vector vArrayRange;
        Status statInformation;
        int iStartingIndex, iEndingIndex, iCharacterLength;
        boolean bError;
        
        if ( ( operOtherOperand.getOperandType() == DataObject.I_TYPE_CHARACTER )  &&
             ( operOtherOperand.getOperandStructure() == 
                DataObject.I_STRUCTURE_SINGLE_VALUE ) ) {
            iCharacterLength = 1;
        }
        else {
            iCharacterLength = ArrayRangeDialog.I_NO_LIMIT;
        }
        
        ardArrayRange = new ArrayRangeDialog( jfParentFrame, operThisOperand,
            iCharacterLength );
        objReturnedObject = ardArrayRange.displayDialog();
        
        if ( objReturnedObject instanceof Vector ) {
            vArrayRange = (Vector)objReturnedObject;
            iStartingIndex = ((Integer)
                vArrayRange.get( ArrayRangeDialog.I_INDEX_STARTING )).intValue();
            iEndingIndex = ((Integer)
                vArrayRange.get( ArrayRangeDialog.I_INDEX_ENDING )).intValue();
            System.out.println( "RECEIVED...start=" + iStartingIndex
                + "   end=" + iEndingIndex );
            operThisOperand.setArrayIndices( iStartingIndex, iEndingIndex );
            plThisLabel.setArrayRange( iStartingIndex, iEndingIndex );
            bError = false;
        }
        else if ( objReturnedObject instanceof Status ) {
            statInformation = (Status)objReturnedObject;
            
            if ( statInformation.getStatusNumber() == ArrayRangeDialog.I_ILLEGAL_VALUE ) {
                bError = true;
            }
            else {
                bError = false;
            }
        }
        else {
            bError = true;
        }
        
        if ( bError == true ) {
            JOptionPane.showMessageDialog( this, "Internal error creating "
                + "array range.", ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );        
        }
    }

    private void updateRestrictionFlagsAndLengthLimit( int iClickedID ) {
        Operand operClickedOperand;
        int iOperandType, iOperandStructure, iOperandRestriction, iOperandArrayElement;
        
        // BEFORE CLICK: one operand blank; AFTER CLICK: both operands blank
        if ( ( operDestination.getOperandType() == DataObject.I_UNDEFINED )  && 
             ( operSource.getOperandType() == DataObject.I_UNDEFINED ) ) {
            iFirstClickedOperandID = I_UNDEFINED;
            bsDestinationOpRestrictionFlags = (BitSet)
                bsDestinationOpBackupRestrictionFlags.clone();
            bsSourceOpRestrictionFlags = (BitSet)bsSourceOpBackupRestrictionFlags.clone();
            super.setDataObjectAreaRestrictionList( iDestinationOperandID, 
                bsDestinationOpRestrictionFlags );
            super.setDataObjectAreaRestrictionList( iSourceOperandID, 
                bsSourceOpRestrictionFlags );            
        }
        else {
            if ( iFirstClickedOperandID == I_UNDEFINED ) {
                iFirstClickedOperandID = iClickedID;
            }

            // clicked operand affects the unclicked operand
            if ( iFirstClickedOperandID == iClickedID ) {
                if ( iClickedID == iDestinationOperandID ) {
                    iOperandType = operDestination.getOperandType();
                    iOperandStructure = operDestination.getOperandStructure();
                    iOperandRestriction = operDestination.getOperandRestriction();                    
                    iOperandArrayElement = operDestination.getArrayElementIndexSetting();
                    
                    bsSourceOpRestrictionFlags = 
                        super.getNewDataObjectAreaRestrictionList();
                    bsSourceOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_VARIABLE );
                    bsSourceOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_CONSTANT );
                    bsSourceOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_LITERAL );
                    
                    switch ( iOperandType ) {                        
                        case DataObject.I_TYPE_CHARACTER:                        
                            if ( iOperandStructure == DataObject.I_STRUCTURE_1D_ARRAY ) {
                                if ( iOperandArrayElement == Operand.I_UNDEFINED ) {
                                    bsSourceOpRestrictionFlags.set( 
                                        AllPurposeMenu.I_FLAG_CHARACTER_ARRAY );
                                    bsSourceOpRestrictionFlags.set( 
                                        AllPurposeMenu.I_FLAG_CHARACTER_SINGLE );                            
                                }
                                else {  // treat as a single character
                                    bsSourceOpRestrictionFlags.set( 
                                        AllPurposeMenu.I_FLAG_CHARACTER_ARRAY );
                                }
                            }
                            else if ( iOperandStructure == 
                                DataObject.I_STRUCTURE_SINGLE_VALUE ) {
                                bsSourceOpRestrictionFlags.set( 
                                    AllPurposeMenu.I_FLAG_CHARACTER_ARRAY );
                            }
                            break;
                    }
                    super.setDataObjectAreaRestrictionList( iSourceOperandID, 
                        bsSourceOpRestrictionFlags );                    
                }
                else if ( iClickedID == iSourceOperandID ) {
                    iOperandType = operSource.getOperandType();
                    iOperandStructure = operSource.getOperandStructure();
                    iOperandRestriction = operSource.getOperandRestriction();
                    iOperandArrayElement = operSource.getArrayElementIndexSetting();

                    bsDestinationOpRestrictionFlags = 
                        super.getNewDataObjectAreaRestrictionList();
                    bsDestinationOpRestrictionFlags.set( AllPurposeMenu.I_FLAG_VARIABLE );
                    
                    switch ( iOperandType ) {
                        case DataObject.I_TYPE_CHARACTER:
                            if ( iOperandStructure == DataObject.I_STRUCTURE_1D_ARRAY ) {
                                if ( iOperandArrayElement == Operand.I_UNDEFINED ) {
                                    bsDestinationOpRestrictionFlags.set( 
                                        AllPurposeMenu.I_FLAG_CHARACTER_ARRAY );
                                    bsDestinationOpRestrictionFlags.set( 
                                        AllPurposeMenu.I_FLAG_CHARACTER_SINGLE );
                                }
                                else {  // treat as a single character
                                    bsDestinationOpRestrictionFlags.set( 
                                        AllPurposeMenu.I_FLAG_CHARACTER_ARRAY );
                                }
                            }                                                                
                            else if ( iOperandStructure == 
                                DataObject.I_STRUCTURE_SINGLE_VALUE ) {
                                bsDestinationOpRestrictionFlags.set( 
                                    AllPurposeMenu.I_FLAG_CHARACTER_ARRAY );
                            }
                            break;
                    }                    
                    super.setDataObjectAreaRestrictionList( iDestinationOperandID, 
                        bsDestinationOpRestrictionFlags );                                        
                }  // end of source operand
            } // else, no restriction change...ignore
        }  // end of else        
    }
        
    private void clearAll() {
        instThisInstruction = new Instruction();
        instThisInstruction.setOpcode( opMovePartsOpcode );
        plDestinationArrayRange.setArrayRange( 0, 0 );
        plSourceArrayRange.setArrayRange( 0, 0 );
        super.clearForm();
    }
    
            
    // instance variables
    private Instruction instThisInstruction;
    private Opcode opMovePartsOpcode;
    private Operand operDestination, operSource;
    private JButton jbApply;
    private int iDestinationOperandID, iSourceOperandID, iFirstClickedOperandID;
    private BitSet bsDestinationOpRestrictionFlags, bsDestinationOpBackupRestrictionFlags,
        bsSourceOpRestrictionFlags, bsSourceOpBackupRestrictionFlags;
    private PTVHelpDialog ptvhdHelp;
    private ArrayRangeDialog ardDialog;
    private PaintableLabel plSourceArrayRange, plDestinationArrayRange;
}