/*
 * ObjectFormEditorPanel.java
 *
 * Created on March 14, 2002, 6:50 PM
 */

//package panelsAndTools;

import javax.swing.*;
import javax.swing.event.*;
import java.lang.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

/**
 *
 * @author  _
 * @version 
 */

public class ObjectFormEditorPanel extends JPanel
    implements RecordObjectTableListener, DocumentListener, ActionListener, ItemListener {
        
    public ObjectFormEditorPanel( Program progThisProgram,
        RecordObjectTable rotThisDeclarationsTable, JFrame jfThisParent ) {

        GridBagLayout gblLayoutGrid;
        
        deAllDeclarations = progThisProgram.getDeclarations();
        codeThisCode = progThisProgram.getCode();
        rotDeclarationsTable = rotThisDeclarationsTable;
        gblLayoutGrid = new GridBagLayout();
        setLayout( gblLayoutGrid );
        
        alStructures = DataObjectPropertyTextView.getStructureNamesList();
        alTypes = DataObjectPropertyTextView.getTypeNamesList();
        alRestrictions = DataObjectPropertyTextView.getRestrictionNamesList();
        
        jfParent = jfThisParent;
                
        createPanel();
        addListeners();
        clearForm();
    }
    
    /////////////////// LISTENER FUNCTIONS \\\\\\\\\\\\\\\\\\\\
    
    //--------------- ActionListener ------------------\\
    public void actionPerformed( ActionEvent aeEvent ) {
        Object objEventSource, objThisObject;
        boolean bRequestCompleted, bContinue, bModifiedName;
        DataObject doDataObject;
        Status statInformation;
        int iDialogChoice, iErrorCount;
        String stFormObjectName;
        
        objEventSource = aeEvent.getSource(); 
        stFormObjectName = jtfName.getText();
        
        if ( objEventSource == jbAdd ) {
            // add object to declarations
            System.out.println( "Add: Preparing data from form." );
            objThisObject = prepareDataObjectFromForm();
            
            if ( objThisObject instanceof DataObject ) {
                doDataObject = (DataObject)objThisObject;
                statInformation = deAllDeclarations.addDeclaration( 
                    stFormObjectName, doDataObject );
                
                if ( statInformation.getStatusNumber() == I_OK ) {
                    JOptionPane.showMessageDialog( jfParent, 
                        stFormObjectName + " was added to the list of declarations.",
                        ST_APPLICATION_NAME, JOptionPane.INFORMATION_MESSAGE );                    
                    clearForm();
                }
                else {
                    JOptionPane.showMessageDialog( jfParent, 
                        statInformation.getStatusMessage(),
                        ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );
                }                
            }
            else if ( objThisObject instanceof Status ) {
                statInformation = (Status)objThisObject;
                JOptionPane.showMessageDialog( jfParent, 
                    statInformation.getStatusMessage(),
                    ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );
            }
            else {
                JOptionPane.showMessageDialog( jfParent, 
                    "Expected an object of type Status or DataObject.  Blame "
                    + "the programmer for this.", ST_APPLICATION_NAME,
                     JOptionPane.ERROR_MESSAGE );                
            }
        }
        else if ( objEventSource == jbModify ) {
            // modify current data object
            DataObject doBackupDataObject;
            String stBackupObjectName, stNewObjectName;
            boolean bError;
                        
            stNewObjectName = stFormObjectName;
            System.out.println( "Name from text field = " + stNewObjectName );
            iDialogChoice = JOptionPane.showConfirmDialog( jfParent, 
                "Are you sure that you want to modify the object named " 
                + stCurrentObjectName + "?", ST_APPLICATION_NAME, 
                JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE );
            
            if ( iDialogChoice == I_OPTION_YES ) {     
                // perform an atomic transaction...either everything is committed
                // or it is rolled back to its previous state
                //System.out.println( "Checking to modify..." );
                stBackupObjectName = stCurrentObjectName;
                //System.out.println( "Backup object name = " + stBackupObjectName );
                doBackupDataObject = doCurrentDataObject;
                //System.out.println( "Backup object = " +   doBackupDataObject );
                doBackupDataObject.printDataObject();
                if ( stNewObjectName.equals( stCurrentObjectName ) == false ) {
                    // names do not match
                    System.out.println( "Names do not match." );
                    
                    statInformation = deAllDeclarations.setDeclarationName( 
                        stCurrentObjectName, stNewObjectName );
                    System.out.println( "Old name = " + stCurrentObjectName + "    New name = " + stNewObjectName );
                    if ( statInformation.getStatusNumber() == I_OK ) {
                        bError = false;
                        bModifiedName = true;
                        System.out.println( "NO ERROR...Name modified." );
                    }
                    else {
                        JOptionPane.showMessageDialog( jfParent, 
                            statInformation.getStatusMessage(),
                            ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );
                        bError = true;
                        bModifiedName = false;
                    }
                }
                else {
                    // names match
                    System.out.println( "Names are the same." );
                    bModifiedName = false;
                    bError = false;                    
                }
                
                if ( bError == false ) {
                    System.out.println( "Modify: Getting data from form..." );
                    objThisObject = prepareDataObjectFromForm();

                    if ( objThisObject instanceof DataObject ) {
                        doDataObject = new DataObject();
                        doDataObject = (DataObject)objThisObject;
                        //doDataObject.printDataObject();
                        
                        if ( bModifiedName == true ) {
                            System.out.println( "Name modified." );  

                            /* INSERT CODE VERIFICATION HERE */
                            bContinue = allowDeclarationChanges( Code.I_ACTION_CHANGE, 
                                stCurrentObjectName, new NameAndDataObjectWrapper( 
                                stNewObjectName, doDataObject ) );
                        
                            if ( bContinue == true ) {
                                statInformation = 
                                    deAllDeclarations.setDeclarationProperties( 
                                        stNewObjectName, doDataObject );                        

                                if ( statInformation.getStatusNumber() == I_OK ) {
                                    JOptionPane.showMessageDialog( jfParent, 
                                        stNewObjectName + " is the new name of the " 
                                        + "object previously named " + stBackupObjectName,
                                        ST_APPLICATION_NAME, 
                                        JOptionPane.INFORMATION_MESSAGE );
                                    codeThisCode.commitCodeChangesFromDeclarationChange();
                                    clearForm();
                                }
                                else {
                                    JOptionPane.showMessageDialog( jfParent, 
                                        statInformation.getStatusMessage(),
                                        ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );
                                    bError = true;
                                }
                            }  // else...do not apply the change
                        }  // end of name not modified
                        else {  // name modified
                            /* INSERT CODE VERIFICATION HERE */
                            bContinue = allowDeclarationChanges( Code.I_ACTION_CHANGE, 
                                stCurrentObjectName, new NameAndDataObjectWrapper( 
                                stCurrentObjectName, doDataObject ) );

                            if ( bContinue == true ) {
                                statInformation = 
                                    deAllDeclarations.setDeclarationProperties( 
                                        stBackupObjectName, doDataObject );

                                if ( statInformation.getStatusNumber() == I_OK ) {
                                    JOptionPane.showMessageDialog( jfParent, 
                                        "The properties of the object named "
                                        + jtfName.getText() + " were modified.",
                                        ST_APPLICATION_NAME, 
                                        JOptionPane.INFORMATION_MESSAGE );
                                    codeThisCode.commitCodeChangesFromDeclarationChange();
                                    clearForm();
                                }
                                else {
                                    JOptionPane.showMessageDialog( jfParent, 
                                        statInformation.getStatusMessage(),
                                        ST_APPLICATION_NAME, 
                                        JOptionPane.ERROR_MESSAGE );
                                    bError = true;
                                }                                            
                            } // end of continue
                        }                                                
                    }
                    else if ( objThisObject instanceof Status ) {
                        statInformation = (Status)objThisObject;
                        JOptionPane.showMessageDialog( jfParent, 
                            statInformation.getStatusMessage(),
                            ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );
                        bError = true;
                    }
                    else {
                        JOptionPane.showMessageDialog( jfParent, 
                            "Expected an object of type Status or DataObject.  Blame "
                            + "the programmer for this.", ST_APPLICATION_NAME,
                            JOptionPane.ERROR_MESSAGE );
                        bError = true;
                    }
                    
                    if ( bError == true ) {
                        System.out.println( "Error occurred...returning to old data." );
                        if ( bModifiedName == true ) {
                            System.out.println( "Old name = " + stCurrentObjectName + "    New name = " + stNewObjectName );
                            deAllDeclarations.setDeclarationName( 
                                stNewObjectName, stCurrentObjectName );
                        }

                        deAllDeclarations.setDeclarationProperties(
                            stCurrentObjectName, doBackupDataObject );
                        System.out.println( "Backup object named " + stCurrentObjectName + "..." );
                        //doBackupDataObject.printDataObject();
                        //System.out.println( "Checking..." );
                        //doBackupDataObject = (DataObject)deAllDeclarations.getDeclaration( stCurrentObjectName );
                        //doBackupDataObject.printDataObject();
                    }
                }
                // else: do nothing...error occurred with name change...do nothing
            }// end if selection of OK
        }
        else if ( objEventSource == jbDelete ) {
            // delete current data object
            iDialogChoice = JOptionPane.showConfirmDialog( jfParent, 
                "Are you sure that you want to delete the object named "
                + stFormObjectName + "?", ST_APPLICATION_NAME, 
                JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE );
            System.out.println( "Dialog choice: " + iDialogChoice );
            
            if ( iDialogChoice == I_OPTION_YES ) {
                System.out.println( "Name text: " + stFormObjectName );
                
                /* INSERT CODE VERIFICATION HERE */
                bContinue = allowDeclarationChanges( Code.I_ACTION_DELETE, 
                    stFormObjectName, null );
                
                if ( bContinue == true ) {
                    statInformation = deAllDeclarations.removeDeclaration( stFormObjectName );

                    if ( statInformation.getStatusNumber() == I_OK ) {
                        JOptionPane.showMessageDialog( jfParent, 
                            "The object named " + stFormObjectName 
                            + " has been deleted.", ST_APPLICATION_NAME,
                            JOptionPane.INFORMATION_MESSAGE );                    
                        codeThisCode.commitCodeChangesFromDeclarationChange();
                        clearForm();                    
                    }
                    else {
                        JOptionPane.showMessageDialog( jfParent, 
                            statInformation.getStatusMessage(), ST_APPLICATION_NAME, 
                            JOptionPane.ERROR_MESSAGE );                  
                    }
                }  // end of continue
            }
            // else: do nothing
        }
        else if ( objEventSource == jbClear ) {
            // clear form
            clearForm();
        }
        else if ( objEventSource == jbHelp ) {
            ptvhdHelp.show();
        }
    }
        
    //------------------ RecordObjectTableListener -------------------\\
    public void updatedDataObjectAtCurrentRow( Object objSelectedDataObject ) {
        Object objThisObject;
        
        if ( objSelectedDataObject instanceof NameAndDataObjectWrapper ) {
            String stObjectName, stTemp;
            ArrayDimensions adArraySize;
            NameAndDataObjectWrapper nadowSelectedRowObject;
            StructureAndDimensionsWrapper sadwStructureAndDimensions;

            nadowSelectedRowObject = (NameAndDataObjectWrapper)objSelectedDataObject;                
            stCurrentObjectName = nadowSelectedRowObject.getName();
            doCurrentDataObject = nadowSelectedRowObject.getObject();
            sadwStructureAndDimensions = doCurrentDataObject.getStructureAndDimensions();
            adArraySize = sadwStructureAndDimensions.getArraySize();

            // fires DocumentListener.<something>updated()
            jtfName.setText( stCurrentObjectName );
            jtfArrayLength.setText( Integer.toString( adArraySize.getXDimension() ) );
            //jtfArrayWidth.setText( Integer.toString( adArraySize.getYDimension() ) );
            jtfDescription.setText( doCurrentDataObject.getDescription() );

            objThisObject = doCurrentDataObject.getValue();               
            if ( objThisObject instanceof String ) {
                jtfLiteral.setText( String.valueOf( (String)objThisObject ) );
            }
            else if ( objThisObject instanceof Integer ) {
                jtfLiteral.setText( String.valueOf( (Integer)objThisObject ) );
            }
            else if ( objThisObject instanceof Float ) {
                jtfLiteral.setText( String.valueOf( (Float)objThisObject ) );
            }
            else if ( objThisObject instanceof Character ) {
                jtfLiteral.setText( String.valueOf( (Character)objThisObject ) );
            }                
            else {
                jtfLiteral.setText( "" );
            }

            // fires ItemListener.itemStateChanged()
            stTemp = 
                DataObjectPropertyTextView.getTypeName( doCurrentDataObject.getType() );
            jcbTypeList.setSelectedIndex( alTypes.indexOf( stTemp ) );
            stTemp = 
                DataObjectPropertyTextView.getRestrictionName( doCurrentDataObject.getRestriction() );
            jcbRestrictionList.setSelectedIndex( alRestrictions.indexOf( stTemp ) );
            stTemp = 
                DataObjectPropertyTextView.getStructureName( doCurrentDataObject.getStructure() );
            jcbStructureList.setSelectedIndex( alStructures.indexOf( stTemp ) );
            settingsForSelectedTableRow();
        }
        else {   
            JOptionPane.showMessageDialog( this, "This object does not exist.",
                ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );
        }
        rotDeclarationsTable.requestFocus();
    }
    
    //------------- DocumentListener -------------------\\
    public void insertUpdate( DocumentEvent doceEvent ) {
        settingsForPropertyChange();
    }
    
    public void removeUpdate( DocumentEvent doceEvent ) {
        settingsForPropertyChange();
    }
    
    public void changedUpdate( DocumentEvent doceEvent ) {
        // not needed for this implementation
    }
    
    //-------------------- ItemListener ----------------\\
    public void itemStateChanged( ItemEvent ieEvent ) {
        Object objEventSource;
        int iSelectedIndex;
        ItemSelectable isItemSelected;
        
        //System.out.println( "ITEM STATE CHANGED" );
        objEventSource = ieEvent.getItemSelectable();  
        isItemSelected = (ItemSelectable)objEventSource;
        //System.out.println("Event Object = " +ieEvent.paramString()) ;
        if ( isItemSelected == jcbTypeList ) {
            //System.out.println("Type list changed");
            iSelectedIndex = jcbTypeList.getSelectedIndex();
            iSelectedIndex = DataObject.I_FIRST_NON_SYSTEM_TYPE_INDEX + iSelectedIndex;
            
            switch ( iSelectedIndex ) {
                case DataObject.I_TYPE_RECORD:
                    settingsForTypeRecord();
                    break;
                    
                case DataObject.I_TYPE_LABEL:
                    settingsForTypeLabel();
                    break;
                    
                case DataObject.I_TYPE_INTEGER:
                    settingsForTypeInteger();
                    
                default:
                    settingsForTypeCharacter();
                    break;
            }
        }
        else if ( isItemSelected == jcbRestrictionList ) {
            iSelectedIndex = jcbRestrictionList.getSelectedIndex();
            //System.out.println("Restriction list changed");
            switch ( iSelectedIndex ) {
                case I_RESTRICTION_CONSTANT:
                    settingsForRestrictionConstant();
                    break;
                    
                case I_RESTRICTION_VARIABLE:
                    settingsForRestrictionVariable();
                    break;
                    
                default:
                    break;
            }            
        }
        else if ( isItemSelected == jcbStructureList ) {
            //System.out.println("Structure list changed");
            settingsForArrayDimensions( true, true );
        }
        
        settingsForPropertyChange();
    }
    
    //////////////////////// PRIVATE METHODS \\\\\\\\\\\\\\\\\\\    
    private boolean allowDeclarationChanges( int iAction, 
        String stOldObjectName, NameAndDataObjectWrapper nadowNewData ) {
    
        boolean bContinue;
        int iDialogChoice, iErrorCount;
        String stAction, stPlural;
        
        iErrorCount = codeThisCode.getErrorCountWithDeclarationChange( 
            iAction, stOldObjectName, nadowNewData, deAllDeclarations );

        if ( iErrorCount > 0 ) {
            if ( iAction == Code.I_ACTION_CHANGE ) {
                stAction = "Modifying ";
            }
            else if ( iAction == Code.I_ACTION_DELETE ) {
                stAction = "Deleting ";
            }
            else {
                stAction = "";
            }
            
            if ( iErrorCount > 1 ) {
                stPlural = "s";
            }
            else {
                stPlural = "";
            }
            
            if ( iAction == Code.I_ACTION_CHANGE ) {
                iDialogChoice = JOptionPane.showConfirmDialog( new JFrame(), 
                    "Modifying " + stOldObjectName + " will cause " + iErrorCount 
                    + " line" + stPlural + " of code to have errors.  " 
                    + "Continue?", ST_APPLICATION_NAME, 
                    JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE );
            }
            else {
                iDialogChoice = JOptionPane.showConfirmDialog( new JFrame(), 
                    stAction + stOldObjectName + " will cause " + iErrorCount
                    + " line" + stPlural + " of code to have errors.  Continue?", 
                    ST_APPLICATION_NAME, JOptionPane.YES_NO_OPTION, 
                    JOptionPane.QUESTION_MESSAGE );
            }
            
            if ( iDialogChoice == I_OPTION_YES ) {
                bContinue = true;
            }
            else {
                bContinue = false;
            }
        }  // else, no errors...allow the changes
        else {
            bContinue = true;
        }
        return bContinue;
    }
    /////////////////////// PANEL CREATION \\\\\\\\\\\\\\\\\\\
    private void addListeners() {
        jtfName.getDocument().addDocumentListener( this );
        jtfArrayLength.getDocument().addDocumentListener( this );
        //jtfArrayWidth.getDocument().addDocumentListener( this );
        jtfLiteral.getDocument().addDocumentListener( this );
        jtfDescription.getDocument().addDocumentListener( this );
                
        jcbTypeList.addItemListener( this );
        jcbRestrictionList.addItemListener( this );
        jcbStructureList.addItemListener( this );
        
        rotDeclarationsTable.setRecordTableListener( this );
    }
    
    private void createPanel() {
        Font fontLabel, fontField;        
        
        fontLabel = new Font( "SansSerif", Font.BOLD, 14 );
        fontField = new Font( "monospaced", Font.PLAIN, 14 );

        JLabel jlFiller1 = createLabel( "      ", fontLabel );
        addComponentToPanel( 0, 0, 5, 2, jlFiller1 );
        
        JLabel jlName = createLabel( "Name", fontLabel );
        addComponentToPanel( 0, 2, 1, 1, jlName );
        
        jtfName = createField( 9, fontField );
        addComponentToPanel( 1, 2, 1, 1, jtfName );

        JLabel jlStructure = createLabel( "Structure", fontLabel );
        addComponentToPanel( 3, 3, 1, 1, jlStructure );

        jcbStructureList = 
            createComboBox( alStructures.toArray(), fontField );
        addComponentToPanel( 4, 3, 1, 1, jcbStructureList );
        
        JLabel jlType = createLabel( "Type", fontLabel );
        addComponentToPanel( 0, 3, 1, 1, jlType );
        
        jcbTypeList = 
            createComboBox( alTypes.toArray(), fontField );
        addComponentToPanel( 1, 3, 1, 1, jcbTypeList );
        
        JLabel jlXSize = createLabel( "Array Length", fontLabel );
        addComponentToPanel( 3, 4, 1, 1, jlXSize );
        
        jtfArrayLength = createField( 5, fontField );
        addComponentToPanel( 4, 4, 1, 1, jtfArrayLength );
        
        JLabel jlRestriction = createLabel( "Restriction", fontLabel );
        addComponentToPanel( 0, 4, 1, 1, jlRestriction );
        
        jcbRestrictionList = 
            createComboBox( alRestrictions.toArray(), fontField );
        addComponentToPanel( 1, 4, 1, 1, jcbRestrictionList );
        
        //JLabel jlYSize = createLabel( "Array Width", fontLabel );
        //addComponentToPanel( 3, 4, 1, 1, jlYSize );
        
        //jtfArrayWidth = createField( 5, fontField );
        //addComponentToPanel( 4, 4, 1, 1, jtfArrayWidth );
                
        JLabel jlValue = createLabel( "Literal Value", fontLabel );
        addComponentToPanel( 0, 5, 1, 1, jlValue );
        
        jtfLiteral = createField( 50, fontField );
        addComponentToPanel( 1, 5, 4, 1, jtfLiteral );

        JLabel jlDescription = createLabel( "Description", fontLabel );
        addComponentToPanel( 0, 6, 1, 1, jlDescription );
        
        jtfDescription = createField( 50, fontField );
        addComponentToPanel( 1, 6, 4, 1, jtfDescription );
        
        JLabel jlFiller2 = createLabel( "      ", fontLabel );
        addComponentToPanel( 0, 7, 5, 1, jlFiller2 );
        
        jbAdd = createButton( "Add" );
        addComponentToPanel( 0, 8, 1, 1, jbAdd );
        
        jbModify = createButton( "Modify" );
        addComponentToPanel( 1, 8, 1, 1, jbModify );

        jbDelete = createButton( "Delete" );
        addComponentToPanel( 2, 8, 1, 1, jbDelete );

        jbClear = createButton( "Clear" );
        addComponentToPanel( 3, 8, 1, 1, jbClear );

        jbHelp = createButton( "Help" );
        addComponentToPanel( 4, 8, 1, 1, jbHelp ); 
        
        ptvhdHelp = new PTVHelpDialog( jfParent, "Help >> Panel >> Object Editor" );
        ptvhdHelp.setHelpFile( ST_HOME_DIRECTORY + "help_panel_object_editor.html" );               
    }        
    
    private void addComponentToPanel( int iGridXPos, int iGridYPos, int iGridWidth,
                                      int iGridHeight, Component coComponent ) {
        
        GridBagConstraints gbcLayoutConstraints;

        gbcLayoutConstraints = new GridBagConstraints();
        gbcLayoutConstraints.weightx = 100;
        gbcLayoutConstraints.weighty = 100;
        gbcLayoutConstraints.gridx = iGridXPos;
        gbcLayoutConstraints.gridy = iGridYPos;
        gbcLayoutConstraints.gridwidth = iGridWidth;
        gbcLayoutConstraints.gridheight = iGridHeight;
        gbcLayoutConstraints.anchor = GridBagConstraints.SOUTHWEST;
        gbcLayoutConstraints.insets = new Insets( 2, 5, 2, 3 );
        gbcLayoutConstraints.fill =  GridBagConstraints.HORIZONTAL;
        add( coComponent, gbcLayoutConstraints );
    }
    
    private JLabel createLabel( String stLabelText, Font fontLabel ) {
        JLabel jlLabel;
        
        jlLabel = new JLabel( stLabelText, SwingConstants.LEFT );
        jlLabel.setFont( fontLabel );
        jlLabel.setForeground( Color.black );
        
        return jlLabel;
    }
    
    private JTextField createField( int iColumns, Font fontField ) {
        JTextField jtfTextField;
        
        jtfTextField = new JTextField( iColumns );
        jtfTextField.setFont( fontField );
        
        return jtfTextField;
    }
    
    private JComboBox createComboBox( Object[] obj1dData, Font fontList ) {
        JComboBox jcbList;
        
        jcbList = new JComboBox( obj1dData );
        jcbList.setFont( fontList );
        jcbList.setEditable( false );
        
        return jcbList;
    }
    
    private JButton createButton( String stButtonLabel ) {
        JButton jbButton;
        
        jbButton = new JButton( stButtonLabel );
        jbButton.addActionListener( this );
        
        return jbButton;
    }

    ////////////////////////// FORM SETTINGS \\\\\\\\\\\\\\\\\\\\
    private void settingsForLiteralValue( boolean bFieldSetting ) {
        int iCurrentRestrictionIndex;
        
        iCurrentRestrictionIndex = jcbRestrictionList.getSelectedIndex();
        
        switch ( iCurrentRestrictionIndex ) {
            case I_RESTRICTION_CONSTANT:
                jtfLiteral.setEnabled( bFieldSetting );
                break;
            
            case I_RESTRICTION_VARIABLE:
            default:
                jtfLiteral.setEnabled( false );
                jtfLiteral.setText( "" );
                break;
        }
    }
    
    private void settingsForArrayDimensions( boolean bArrayLengthSetting,
                                             boolean bArrayWidthSetting )  {
        int iCurrentStructureIndex;
        
        iCurrentStructureIndex = jcbStructureList.getSelectedIndex();
        iCurrentStructureIndex = DataObject.I_FIRST_STRUCTURE_INDEX
            + jcbStructureList.getSelectedIndex();
        
        switch ( iCurrentStructureIndex ) {
            case DataObject.I_STRUCTURE_1D_ARRAY:
                jtfArrayLength.setEnabled( bArrayLengthSetting );
                //jtfArrayWidth.setText( "1" );
                //jtfArrayWidth.setEnabled( false );
                break;
            
            case DataObject.I_STRUCTURE_2D_ARRAY:
               jtfArrayLength.setEnabled( bArrayLengthSetting );
               //jtfArrayWidth.setEnabled( bArrayWidthSetting );
               break;

            case DataObject.I_STRUCTURE_SINGLE_VALUE:
            default:
               jtfArrayLength.setText( "1" );               
               jtfArrayLength.setEnabled( false );
               //jtfArrayWidth.setText( "1" );
               //jtfArrayWidth.setEnabled( false );
               break;
        }
    }
    
    private void clearForm() {
        jtfName.setText( "" );
        jtfArrayLength.setText( "1" );
        //jtfArrayWidth.setText( "1" );
        jtfDescription.setText( "" );
        jtfLiteral.setText( ""  );
        jcbTypeList.setSelectedIndex( DataObject.I_TYPE_CHARACTER 
            - DataObject.I_FIRST_NON_SYSTEM_TYPE_INDEX );
        jcbRestrictionList.setSelectedIndex( I_RESTRICTION_CONSTANT );
        jcbStructureList.setSelectedIndex( DataObject.I_STRUCTURE_SINGLE_VALUE
            - DataObject.I_FIRST_STRUCTURE_INDEX );        

        jbAdd.setEnabled( false );
        jbModify.setEnabled( false );
        jbDelete.setEnabled( false );
    }
           
    private void settingsForSelectedTableRow() {
        if ( jbAdd.isEnabled() == true )
            jbAdd.setEnabled( false );
        
        if ( jbModify.isEnabled() == true )
            jbModify.setEnabled( false );
        
        if ( jbDelete.isEnabled() == false )
            jbDelete.setEnabled( true );
    }
    
    private void settingsForPropertyChange() {
        if ( jbAdd.isEnabled() == false )
            jbAdd.setEnabled( true );
        
        if ( jbModify.isEnabled() == false )
            jbModify.setEnabled( true );
        
        if ( jbDelete.isEnabled() == true )
            jbDelete.setEnabled( false );
    }
    
    private void settingsForTypeLabel() {        
        jcbRestrictionList.setEnabled( false );
        jcbStructureList.setEnabled( false );
        settingsForArrayDimensions( false, false );
        settingsForLiteralValue( false );
        jtfLiteral.setText( "" );
        jtfDescription.setEnabled( false );        
        jtfDescription.setText( "" );
    }    
    
    private void settingsForTypeRecord() {
        jcbRestrictionList.setEnabled( false );
        jcbRestrictionList.setSelectedIndex( I_RESTRICTION_VARIABLE );
        jcbStructureList.setEnabled( true );
        settingsForArrayDimensions( true, true );
        settingsForLiteralValue( false );
        jtfDescription.setEnabled( true );
    }
    
    private void settingsForTypeCharacter() {
        jcbRestrictionList.setEnabled( true );
        jcbStructureList.setEnabled( true );
        settingsForArrayDimensions( true, true );
        settingsForLiteralValue( true );
        jtfDescription.setEnabled( true );    
    }
    
    private void settingsForTypeInteger() {
        jcbRestrictionList.setEnabled( false );
        jcbStructureList.setEnabled( false );
        settingsForArrayDimensions( false, false );
        settingsForLiteralValue( true );
        jcbStructureList.setSelectedIndex( DataObject.I_FIRST_STRUCTURE_INDEX -
            DataObject.I_STRUCTURE_SINGLE_VALUE );
        jtfDescription.setEnabled( true );        
    }
            
    private void settingsForRestrictionVariable() {
        jcbTypeList.setEnabled( true );
        jcbStructureList.setEnabled( true );
        settingsForArrayDimensions( true, true );
        settingsForLiteralValue( false );
        jtfDescription.setEnabled( true );
    }
    
    private void settingsForRestrictionConstant() {       
        jcbTypeList.setEnabled( true );
        jcbStructureList.setEnabled( true );
        settingsForArrayDimensions( true, true );
        settingsForLiteralValue( true );
        jtfDescription.setEnabled( true );
    }
    
    /////////////////////// DECLARATIONS HELPERS \\\\\\\\\\\\\\\\\\\\\\
    private Object prepareDataObjectFromForm() {
        DataObject doThisDataObject;
        Status statInformation;
        String stMessage;
        int iArrayLength, iArrayWidth, iStructureListSelection, 
            iTypeListSelection, iRestrictionListSelection;
        boolean bError;
        Object objThisObject;

        doThisDataObject = new DataObject();
            
        iTypeListSelection = jcbTypeList.getSelectedIndex() +
            DataObject.I_FIRST_NON_SYSTEM_TYPE_INDEX;
        
        if ( jcbStructureList.isEnabled() == true ) {
            iStructureListSelection = jcbStructureList.getSelectedIndex() +
                DataObject.I_FIRST_STRUCTURE_INDEX;
        }
        else {
            iStructureListSelection = DataObject.I_UNDEFINED;
        }
        
        if ( jcbRestrictionList.isEnabled() == true ) {            
            iRestrictionListSelection = jcbRestrictionList.getSelectedIndex()
                + DataObject.I_FIRST_RESTRICTION_INDEX;
        }
        else {            
            iRestrictionListSelection = DataObject.I_UNDEFINED;
        }

        doThisDataObject.setType( iTypeListSelection );
        doThisDataObject.setRestriction( iRestrictionListSelection );
        doThisDataObject.setStructure( iStructureListSelection );
        
        if ( jtfLiteral.isEnabled() == true ) {
            doThisDataObject.setValue( jtfLiteral.getText() );
        }
        else {
            doThisDataObject.setValue( "" );
        }
        
        if ( jtfDescription.isEnabled() == true ) {            
            doThisDataObject.setDescription( jtfDescription.getText() );
        }
        else {
            doThisDataObject.setDescription( "" );
        }

        try {
            if ( jtfArrayLength.isEnabled() == true ) {
                iArrayLength = Integer.parseInt( jtfArrayLength.getText() );
            }
            else {
                iArrayLength = 1;
            }
            
            iArrayWidth = 1;
            /*if ( jtfArrayLength.isEnabled() == true ) {
                iArrayWidth = Integer.parseInt( jtfArrayWidth.getText() );
            }
            else {
                iArrayWidth = 1;
            }*/

            // if made it this far, numbers are validly typed
            if ( iArrayLength <= 35 ) {
                statInformation = 
                    doThisDataObject.setArrayDimensions( iArrayLength, iArrayWidth );
            }
            else {
                stMessage = "No array can have more than 35 elements.";
                statInformation = new Status( I_ILLEGAL_ARRAY_SIZE, stMessage );
            }
            
            if ( statInformation.getStatusNumber() == I_OK ) {
                bError = false;
            }
            else {
                bError = true;
            }            
        }
        catch ( NumberFormatException nfeException ) {
            stMessage = "Array dimensions length and width must be numbers.";
            statInformation = new Status( I_ILLEGAL_ARRAY_SIZE, stMessage );
            bError = true;
        }
        
        System.out.println( "Structure = " + iStructureListSelection + "    Restriction = " + iRestrictionListSelection );
        
        if ( ( iRestrictionListSelection == DataObject.I_RESTRICTION_CONSTANT )  && 
             ( jtfLiteral.getText().equals( "" ) ) ) {
            stMessage = "Constant objects require a literal value.";
            statInformation = new Status( I_MISSING_LITERAL_VALUE ,stMessage );
            bError = true;
        }
        else if ( ( iTypeListSelection != DataObject.I_TYPE_LABEL )  &&
                  ( iStructureListSelection == I_NO_STRUCTURE_SELECTED ) ) {
            stMessage = "Objects that are not a label require a structure to be selected.";
            statInformation = new Status( I_NO_STRUCTURE, stMessage );
            bError = true;
        }
        else if ( ( iTypeListSelection != DataObject.I_TYPE_LABEL )  &&
                  ( iRestrictionListSelection == I_NO_RESTRICTION_SELECTED ) ) {
            stMessage = "Objects that are not a label require a restriction to be selected.";
            statInformation = new Status( I_NO_STRUCTURE, stMessage );
            bError = true;
        }
                        
        if ( bError == false ) {
            objThisObject = doThisDataObject;
        }
        else {
            objThisObject = statInformation;
        }
        
        return objThisObject;
    }
        
    
    // instance variables
    private JTextField jtfName, jtfArrayLength, jtfArrayWidth, jtfLiteral, jtfDescription;
    private JComboBox jcbTypeList, jcbRestrictionList, jcbStructureList;
    private JButton jbAdd, jbModify, jbDelete, jbClear, jbHelp;
    private int iCurrentRow;
    private Declarations deAllDeclarations;
    private Code codeThisCode;
    private DataObject doCurrentDataObject;
    private NameAndDataObjectWrapper nadowSelectedRowObject;
    private ListSelectionModel lsmTableSelectionModel;
    private ArrayList alTypes, alRestrictions, alStructures;
    private RecordObjectTable rotDeclarationsTable;
    private String stCurrentObjectName;
    private JFrame jfParent;
    private PTVHelpDialog ptvhdHelp;
    
        // class constants
        public final static int I_RESTRICTION_CONSTANT = 0;
        public final static int I_RESTRICTION_VARIABLE = 1;    
        public final static int I_TEMP_RESTRICTION_LITERAL = 1;

        public final static int I_OPTION_YES = 0;
        public final static int I_OPTION_NO = 1;
                
        // error code
        public final static int I_OK = 0;
        public final static int I_ILLEGAL_ARRAY_SIZE = -1;
        public final static int I_MISSING_LITERAL_VALUE = -2;        
        public final static int I_NO_STRUCTURE = -3;
        
        private final static int I_NO_STRUCTURE_SELECTED;
        private final static int I_NO_RESTRICTION_SELECTED;
        private final static String ST_APPLICATION_NAME = "Panels - Tools - Views";
         protected static final String ST_HOME_DIRECTORY = "";
            /*"D:\\forte4j\\Development\\panelsAndTools\\";*/
       
    // initializer
    static {
        I_NO_STRUCTURE_SELECTED = DataObject.I_FIRST_STRUCTURE_INDEX - 1;
        I_NO_RESTRICTION_SELECTED = DataObject.I_FIRST_RESTRICTION_INDEX - 1;
    }
}