/*
 * Operand.java
 *
 * Created on April 16, 2002, 11:19 AM
 */

//package panelsAndTools;

import java.util.*;
import java.io.*;

/**
 *
 * @author  _
 * @version 
 */

public class Operand implements Cloneable, Serializable {
    public Operand() {
        nadowData = new NameAndDataObjectWrapper();
        stDataObjectArrayIndex = "";
        iStartingArrayIndex = 0;
        iEndingArrayIndex = 0;
        iArrayLength = 1;
        iArrayIndexState = I_UNDEFINED;
        bIllegalArrayIndex = false;
        stIllegalArrayIndexMessage = "";        
    }
    
    protected Object clone() {
        // strucutre based on Sun's online Java tutorial for
        // the Object.clone() method and pp.235-6 C of Core Java 2: ISBN 0-13-081933-6        
        try {
            Operand operObject = (Operand)super.clone();
            operObject.nadowData = (NameAndDataObjectWrapper)this.nadowData.clone();
            operObject.stDataObjectArrayIndex = new String( this.stDataObjectArrayIndex );
            operObject.stIllegalArrayIndexMessage = new String( 
                this.stIllegalArrayIndexMessage );
            //operObject.stDataObjectArrayIndex = new String( this.stDataObjectArrayIndex );
            
            return operObject;
        }
        catch ( CloneNotSupportedException cnseException ) {
            System.out.println( "Clone of NameAndDataObjectWrapper is not supported." );
            return null;
        }
    }
    
    public void setOperandData( NameAndDataObjectWrapper nadow ) {
        nadowData = nadow;
    }
    
    public NameAndDataObjectWrapper getOperandData() {
        return (NameAndDataObjectWrapper)nadowData.clone();
    }

    // for DataObject representing a numerical array index
    public Status setArrayElementDataObject( String stName, int iType,
        int iStructure ) {
        
        Status statInformation;
        String stMessage;                    
       
        System.out.println( "Setting Array Element: " + stName + " of type "
           + DataObjectPropertyTextView.getTypeName( iType ) + " and structure "
           + DataObjectPropertyTextView.getStructureName( iStructure ) );
        
        if ( SystemDeclarations.isObjectDeclaredAsARegister( stName ) == true ) {
            iArrayIndexState = I_ARRAY_INDEX_REGISTER;
            statInformation = new Status( I_OK );
            stDataObjectArrayIndex = stName;
        }
        else {
            if ( iType == DataObject.I_TYPE_INTEGER )  {
                if ( iStructure == DataObject.I_STRUCTURE_SINGLE_VALUE ) {
                    iArrayIndexState = I_ARRAY_INDEX_DATA_OBJECT;
                    statInformation = new Status( I_OK );
                    stDataObjectArrayIndex = stName;
                }
                else {
                    stMessage = "The array index must be a number or single, "
                        + "integer data object.";
                    statInformation = new Status( I_ILLEGAL_ARRAY_INDEX, stMessage );
                }
            }
            else {
                stMessage = "The array index must be a number or single, "
                    + "integer data object.";
                statInformation = new Status( I_ILLEGAL_ARRAY_INDEX, stMessage );
            }                
        }
        return statInformation;
    }
    
    // for DataObject representing a numerical array index
    public String getArrayElementDataObject() {
        return new String( stDataObjectArrayIndex );
    }
    
    public Status setArrayElementIndex( int iIndex ) {
        String stMessage;
        
        if ( ( iIndex >= 0 )  &&  
             ( iIndex < nadowData.getObject().getArrayDimensions().getXDimension() ) ) {
            iArrayElementIndex = iIndex;
            iArrayIndexState = I_ARRAY_INDEX_NUMBER;
            return new Status( I_OK );
        }
        else {
            stMessage = "The array index is outside of the legal range of values.";
            return new Status( I_ILLEGAL_ARRAY_INDEX, stMessage );
        }
    }
    
    public int getArrayElementIndex() {
        return iArrayElementIndex;
    }
    
    public void clearArrayElementIndexSetting() {
        stDataObjectArrayIndex = "";
        iStartingArrayIndex = 0;
        iEndingArrayIndex = 0;
        iArrayLength = 1;
        iArrayIndexState = I_UNDEFINED;
        bIllegalArrayIndex = false;
        stIllegalArrayIndexMessage = "";                
    }
    
    public int getArrayElementIndexSetting() {
        return iArrayIndexState;
    }
               
    public Status setArrayIndex( int iIndex ) {
        Status statInformation;
        String stMessage;
        int iMaximumLength;
        
        if ( nadowData != null ) {
            iMaximumLength = nadowData.getObject().getArrayDimensions().getXDimension();
            if ( ( iIndex >= 0 )  &&  ( iIndex < iMaximumLength ) ) {
                iStartingArrayIndex = iIndex;
                iEndingArrayIndex = I_UNDEFINED;
                iArrayLength = I_UNDEFINED;
                statInformation = new Status( I_OK );
            }
            else {
                stMessage = "The array index cannot be greater than one less "
                    + "than the total array size.";
                statInformation = new Status( I_ILLEGAL_ARRAY_INDEX, stMessage );
            }
        }
        else {
            stMessage = "No operand data has been defined for this operand.";
            statInformation = new Status( I_NO_DATA, stMessage );
        }
        return statInformation;
    }
    
    public Status setArrayIndices( int iStart, int iEnd ) {
        Status statInformation;
        String stMessage;
        int iMaximumSize;
        
        if ( iStart >= 0 ) {
            if ( nadowData != null ) {
                iMaximumSize = nadowData.getObject().getArrayDimensions().getXDimension();

                // array indices are 0 to ( iMaximumSize - 1 )
                if ( iEnd < iMaximumSize ) {                
                    if ( ( iEnd == I_UNDEFINED )  ||  ( iEnd >= iStart ) ) {
                        iStartingArrayIndex = iStart;
                        iEndingArrayIndex = iEnd;
                        iArrayLength = iEndingArrayIndex - iStartingArrayIndex + 1;
                        statInformation = new Status( I_OK );
                        iArrayIndexState = I_ARRAY_INDEX_RANGE;
                    }
                    else {
                        stMessage = "The ending array index must be greater than the " 
                            + "starting array index and at least 0, or be left undefined.";
                        statInformation = new Status( I_ILLEGAL_ARRAY_INDEX, stMessage );
                    }
                }
                else {
                    stMessage = "The ending array index cannot be greater than one less "
                        + "than the total array size.";
                    statInformation = new Status( I_ILLEGAL_ARRAY_INDEX, stMessage );
                }
            }
            else {
                stMessage = "No operand data has been defined for this operand.";
                statInformation = new Status( I_NO_DATA, stMessage );
            }
        }
        else {
            stMessage = "The starting array index must be at least 0.";
            statInformation = new Status( I_ILLEGAL_ARRAY_INDEX, stMessage );
        }
        return statInformation;
    }
    
    public int getStartingArrayIndex() {
        return iStartingArrayIndex;
    }
    
    public int getEndingArrayIndex() {
        return iEndingArrayIndex;
    }

    public int getArrayLength() {
        return iArrayLength;
    }
    
    public int getOperandType() {
        if ( nadowData != null )
            return nadowData.getObject().getType();
        else
            return DataObject.I_UNDEFINED;
    }
    
    public String getOperandName() {
        if ( nadowData != null )
            return nadowData.getName();
        else
            return new String( "" );
    }
    
    public int getOperandStructure() {
        if ( nadowData != null )
            return nadowData.getObject().getStructure();
        else
            return DataObject.I_UNDEFINED;
    }

    public int getOperandRestriction() {
        if ( nadowData != null )
            return nadowData.getObject().getRestriction();
        else
            return DataObject.I_UNDEFINED;
    }
    
    public int getOperandSize() {
        if ( nadowData != null )
            return nadowData.getObject().getSize();
        else
            return DataObject.I_UNDEFINED;
    }

    public int getOperandArrayLength() {
        if ( nadowData != null )
            return nadowData.getObject().getArrayDimensions().getXDimension();
        else
            return I_UNDEFINED;
    }
    
    public int getOperandTypeSize() {
        if ( nadowData != null )
            return nadowData.getObject().getTypeSize();
        else
            return I_UNDEFINED;
    }
    
    // depends on iStartingArrayIndex
    public Object getOperandValue() {
        Object objThisObject;
        int iType, iStructure;
        DataObject doThisObject;
        String stMessage;

        if ( nadowData != null ) {
            doThisObject = nadowData.getObject();
            iType = doThisObject.getType();
            iStructure = doThisObject.getStructure();

            if ( doThisObject.getRestriction() != DataObject.I_RESTRICTION_VARIABLE ) {
                if ( iStructure == DataObject.I_STRUCTURE_SINGLE_VALUE ) {
                    if ( ( iType == DataObject.I_TYPE_CHARACTER )  ||
                         ( iType == DataObject.I_TYPE_INTEGER ) ) {
                        objThisObject = doThisObject.getValidatedValue( 0 );
                    }
                    else {
                        stMessage = "Only non-variable integer and character " 
                            + "non-array objects can have a predefined value.";
                        objThisObject = new Status( I_NO_DATA, stMessage );
                    }                        
                }
                else if ( iStructure == DataObject.I_STRUCTURE_1D_ARRAY ) {
                    if ( iType == DataObject.I_TYPE_CHARACTER ) {
                        objThisObject = doThisObject.getValidatedCharacterArray();
                    }
                    else if ( iType == DataObject.I_TYPE_INTEGER ) {
                        objThisObject = 
                            doThisObject.getValidatedValue( iStartingArrayIndex );
                    }
                    else {
                        stMessage = "Internal error.";
                        objThisObject = new Status( I_INTERNAL_ERROR, stMessage );
                    }
                }
                else {
                    stMessage = "Only single variable or 1-D array structured "
                        + "objects may have a non-variable value.";
                    objThisObject = new Status( I_NO_DATA, stMessage );                    
                }
            }
            else {
                stMessage = "Variable objects do not have any predefined value.";
                objThisObject = new Status( I_NO_DATA, stMessage );
            }
        }
        else {
            stMessage = "No operand data has been defined for this operand.";
            objThisObject = new Status( I_NO_DATA, stMessage );
        }
        return objThisObject;
    }
    
    public boolean isSameOperandFromLastComparison( Operand operComparedOperand ) {
        boolean bSame, bContinue;
        String stThisString, stComparedString;
        Object objComparedValueObject, objThisValueObject;
        
        bSame = false;
        stThisString = this.getOperandName();
        stComparedString = operComparedOperand.getOperandName();
        
        // compare names (since names are unique, they represent type & structure)
        // operand position is important
        if ( stThisString.equals( stComparedString ) == true ) {
            // compare literals
            objThisValueObject = this.getOperandValue();
            objComparedValueObject = operComparedOperand.getOperandValue();
            
            if ( ( objThisValueObject instanceof String )  &&
                 ( objComparedValueObject instanceof String ) ) {
                stThisString = (String)objThisValueObject;
                stComparedString = (String)objComparedValueObject;

                if ( stThisString.equals( stComparedString ) == true ) {
                    bContinue = true;
                }
                else {
                    bContinue = false;
                }
            }
            else if ( ( objThisValueObject instanceof Status ) &&
                      ( objComparedValueObject instanceof Status ) ) {
                bContinue = true;
            }
            else {
                bContinue = false;
            }
                
            if ( bContinue == true ) {
                // compare array settings
                if ( iArrayIndexState == 
                     operComparedOperand.getArrayElementIndexSetting() ) {
                    // since both are equal, just check one state
                    if ( iArrayIndexState == I_UNDEFINED ) {
                        bSame = true;
                    }
                    else if ( iArrayIndexState == I_ARRAY_INDEX_NUMBER ) {
                        if ( iArrayElementIndex ==
                             operComparedOperand.getArrayElementIndex() ) {
                            bSame = true;
                        }
                    }
                    else {
                        stComparedString = 
                            operComparedOperand.getArrayElementDataObject();

                        if ( stDataObjectArrayIndex.equals( stComparedString ) == true ) {
                            bSame = true;
                        }
                    }
                }
            }
        }
        return bSame;
    }
    
    public void setFailedOperandInformation( String stMessage ) {
        bIllegalArrayIndex = true;
        stIllegalArrayIndexMessage = stMessage;
    }
    
    public void clearFailedOperandInformation() {
        bIllegalArrayIndex = false;
        stIllegalArrayIndexMessage = "";
    }
    
    public boolean getIllegalArrayIndexFlag() {
        return bIllegalArrayIndex;
    }
    
    public String getFailedOperandInformation() {
        return new String( stIllegalArrayIndexMessage );
    }
    
    ////////////////////////////// FILE I/O METHODS \\\\\\\\\\\\\\\\\\
    public Vector exportState() {
        Vector vData;
        
        vData = new Vector();
        vData.setSize( I_TOTAL_FIELD_COUNT );
        
        vData.set( I_NADOW_DATA, nadowData.exportState() );
        vData.set( I_ARRAY_ELEMENT_INDEX, new Integer( iArrayElementIndex ) );
        vData.set( I_ARRAY_ELEMENT_DATA_OBJECT, new String( stDataObjectArrayIndex ) );
        vData.set( I_STARTING_INDEX, new Integer( iStartingArrayIndex ) );
        vData.set( I_ENDING_INDEX, new Integer( iEndingArrayIndex ) );
        vData.set( I_ARRAY_LENGTH, new Integer( iArrayLength ) );
        vData.set( I_ARRAY_INDEX_STATE, new Integer( iArrayIndexState ) );
        vData.set( I_ILLEGAL_ARRAY_INDEX_FLAG, new Boolean( bIllegalArrayIndex ) );
        vData.set( I_ILLEGAL_ARRAY_INDEX_MESSAGE, new String( stIllegalArrayIndexMessage ) );
        return vData;
    }
    
    public void importState( Vector vData ) {
        nadowData.importState((Vector)vData.get( I_NADOW_DATA ) );
        iArrayElementIndex = ((Integer)vData.get( I_ARRAY_ELEMENT_INDEX )).intValue();
        stDataObjectArrayIndex = new String( 
            (String)vData.get( I_ARRAY_ELEMENT_DATA_OBJECT ) );
        iStartingArrayIndex = ((Integer)vData.get( I_STARTING_INDEX )).intValue();
        iEndingArrayIndex = ((Integer)vData.get( I_ENDING_INDEX )).intValue();
        iArrayLength = ((Integer)vData.get( I_ARRAY_LENGTH )).intValue();
        iArrayIndexState = ((Integer)vData.get( I_ARRAY_INDEX_STATE )).intValue();
        bIllegalArrayIndex = ((Boolean)vData.get( I_ILLEGAL_ARRAY_INDEX_FLAG )).booleanValue();
        stIllegalArrayIndexMessage = new String(
            (String)vData.get( I_ILLEGAL_ARRAY_INDEX_MESSAGE ) );
    }
        
// Because 2-D arrays are not implemented, iStartingArrayIndex will
// refer to any object in an array.  When both iStartingArrayIndex and 
// iEndingArrayIndex will only have valid values when a range of array values are needed
    
    // instance variables
    private NameAndDataObjectWrapper nadowData;
    private String stDataObjectArrayIndex, stIllegalArrayIndexMessage;
    private int iStartingArrayIndex, iEndingArrayIndex, iArrayLength,
        iArrayElementIndex, iArrayIndexState;
    private boolean bIllegalArrayIndex;

    // class constants
    private static final int I_NADOW_DATA = 0;
    private static final int I_ARRAY_ELEMENT_INDEX = 1;
    private static final int I_ARRAY_ELEMENT_DATA_OBJECT = 2;
    private static final int I_STARTING_INDEX = 3;
    private static final int I_ENDING_INDEX = 4;
    private static final int I_ARRAY_LENGTH = 5;
    private static final int I_ARRAY_INDEX_STATE = 6;
    private static final int I_ILLEGAL_ARRAY_INDEX_FLAG = 7;
    private static final int I_ILLEGAL_ARRAY_INDEX_MESSAGE = 8;
    private static final int I_TOTAL_FIELD_COUNT = 9;
    
    public static final int I_UNDEFINED = -99999;
    public static final int I_ARRAY_INDEX_NUMBER = 0;
    public static final int I_ARRAY_INDEX_DATA_OBJECT = 1;
    public static final int I_ARRAY_INDEX_REGISTER = 2;
    public static final int I_ARRAY_INDEX_RANGE = 3;
    
    public static final int I_STARTING_ARRAY_INDEX = 0;
    public static final int I_ENDING_ARRAY_INDEX = 1;
    
        // error codes
        public static final int I_OK = 0;
        public static final int I_ILLEGAL_ARRAY_INDEX = -1;
        public static final int I_NO_DATA = -2;
        public static final int I_INTERNAL_ERROR = -3;
}