/*
 * PaintableLabel.java  
 *
 * Created on April 23, 2002, 6:34 PM
 * Major Revision on June 30, 2002, 4:25 PM
 *
 */

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

/**
 *
 * @author  _
 * @version 
 */

public class PaintableLabel extends JLabel {
    public PaintableLabel( int iXLocation, int iYLocation, int iWidth, 
        int iHeight ) {

        //this.setOpaque( true );
        setDimensions( iXLocation, iYLocation, iWidth, iHeight );
        
        iDrawingType = I_SOLID_LINE;        
        iStartingArrayIndex = I_UNDEFINED;
        iEndingArrayIndex = I_UNDEFINED;
        iStartingRangeLocation = I_UNDEFINED;
        iRangeWidth = I_UNDEFINED;       
        
        jlInfoLabel = new JLabel();
        jlInfoLabel.setFont( foArrayRangeFont );
        jlInfoLabel.setForeground( Color.white );
        jlInfoLabel.setBounds( 4, 2, I_METER_LENGTH, 14 );

        jlArrayRangeBackground = new JLabel();        
        jlArrayRangeBackground.setOpaque( false );
        //jlArrayRangeBackground.setBackground( Color.white );
        jlArrayRangeBackground.setBorder( BorderFactory.createLineBorder( Color.white ) );        
        jlArrayRangeBackground.setBounds( 4, 20, I_METER_LENGTH, 14 );

        jlArrayRangeForeground = new JLabel();        
        jlArrayRangeForeground.setOpaque( true );
        jlArrayRangeForeground.setBackground( Color.white );
        
        jlArrayRangeForeground.setBounds( 4 + iStartingRangeLocation,
            20, iRangeWidth, 14 );     
        
        jlInfoLabel2 = new JLabel();
        jlInfoLabel2.setFont( foArrayRangeFont );
        jlInfoLabel2.setForeground( Color.white );
        jlInfoLabel2.setBounds( 4, 38, I_METER_LENGTH, 14 );        
   }
        
    public void setGraphicsType( int iGraphicsType ) {
        iDrawingType = iGraphicsType;
        this.setOpaque( false );

        switch( iDrawingType ) {
            case I_DOTTED_LINE:
                this.setBorder( BorderFactory.createMatteBorder( 0, 0, 5, 0, iiDot ) );
                //System.out.println( "Dashed line." );
                break;
                    
            case I_SOLID_LINE:
                this.setBorder( BorderFactory.createMatteBorder( 0, 0, 3, 0, Color.black ) );
                //System.out.println( "Solid line." );
                break;
                  
            case I_DOTTED_BORDER:
                this.setBorder( BorderFactory.createMatteBorder( 5, 5, 5, 5, iiDot ) );
                //System.out.println( "Dashed border." );
                break;

            case I_SOLID_BORDER:
                this.setBorder( BorderFactory.createLineBorder( Color.black, 3 ) );
                //System.out.println( "Solid border." );
                break;
                                
            case I_ARRAY_RANGE:
                this.setOpaque( true );
                this.setBackground( PTVPanel.colorDarkBlue );
                this.add( jlInfoLabel );
                this.add( jlInfoLabel2 );
                this.add( jlArrayRangeBackground );
                this.add( jlArrayRangeForeground );                
                break;
                
            case I_NONE:
                this.setBorder( null );
                this.setOpaque( false );
             
            default:
                break;
        }                
    }
    
    public void setDimensions( int iXLocation, int iYLocation, int iWidth, 
        int iHeight ) {

        iXPosition = iXLocation;
        iYPosition = iYLocation;
        iThisWidth = iWidth;
        iThisHeight = iHeight;            
        this.setBounds( iXPosition, iYPosition, iThisWidth, iThisHeight );
    }
    
    public void setArrayRangeForSingleCharacter() {
        setArrayRange( I_SINGLE_CHARACTER, I_SINGLE_CHARACTER );
    }
    
    // used to draw rectangle area to show which part of array is used
    public void setMaximumArrayRange( int iMax ) {
        iMaximumIndex = iMax;
        setArrayRange( iStartingArrayIndex, iEndingArrayIndex );
    }

    // used to draw rectangle area to show which part of array is used    
    public void setArrayRange( int iStartingIndex, int iEndingIndex ) {
        double dRatio;

        // because GUI controls the error checking...no need to duplicate efforts
        if ( ( iStartingIndex == I_SINGLE_CHARACTER )  &&
             ( iEndingIndex == I_SINGLE_CHARACTER ) ) {
            iStartingRangeLocation = 4;
            iRangeWidth = I_METER_LENGTH;
        }
        else {
            iStartingArrayIndex = iStartingIndex;
            iEndingArrayIndex = iEndingIndex;

            jlInfoLabel.setText( "Start = " + iStartingArrayIndex 
                + ", End = " + iEndingArrayIndex );
            
            jlInfoLabel2.setText( "Selection Length = " 
                + ( iEndingArrayIndex - iStartingIndex + 1 ) );

            dRatio = (double)( (double)I_METER_LENGTH / ( iMaximumIndex - 1 ) );        
            iStartingRangeLocation = 4 + (int)( dRatio * iStartingArrayIndex );
            iRangeWidth = (int)( dRatio * iEndingArrayIndex ) - iStartingRangeLocation
                + 2;  // WAS +1
        }

        System.out.println( "Start: " + iStartingRangeLocation
                + "   Width: " + iRangeWidth + "   Max: " + iMaximumIndex );
        jlArrayRangeForeground.setBounds( iStartingRangeLocation,
            20, iRangeWidth, 14 );
        //jlArrayRangeForeground.setText( "###" );
        
    }
    
    public void addMouseListenerToPaintableLabel( MouseListener mlListener ) {
        this.addMouseListener( mlListener );
    }

        
    // instance variables
    private int iXPosition, iYPosition, iThisWidth, iThisHeight, iDrawingType,
        iStartingArrayIndex, iEndingArrayIndex, iStartingRangeLocation,
        iRangeWidth, iMaximumIndex;
    private JLabel jlInfoLabel, jlInfoLabel2, jlArrayRangeBackground, 
        jlArrayRangeForeground;
   
    // class constants
    private static final int I_METER_LENGTH = 135;
    
    public static final int I_DOTTED_LINE = 0;
    public static final int I_SOLID_LINE = 1;
    public static final int I_DOTTED_BORDER = 2;
    public static final int I_SOLID_BORDER = 3;
    public static final int I_ARRAY_RANGE = 4;
    public static final int I_NONE = 5;
    
    private static final int I_SINGLE_CHARACTER = -10;
    
    public static final int I_UNDEFINED = -9999;
    
        // error codes
        public static final int I_OK = 0;
        public static final int I_ILLEGAL_ROTATION = -1;
            
    private static Font foArrayRangeFont;
    private static final ImageIcon iiDot;
    
    static {
        iiDot = new ImageIcon( "dot-part.jpg" );
        foArrayRangeFont = new Font( "SansSerif", Font.BOLD, 12 );
    }        
}