/*
 * Program.java
 *
 * Created on May 19, 2002, 4:36 PM
 */

import java.io.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;

/**
 *
 * @author  _
 * @version 
 */

// Do not return copies of objects...always use original references
public class Program implements ProgramListener {
    public Program() {
        bProgramChangedFlag = false;
        stFileName = "";
        stProgramName = "";
        codeProgramCode = new Code();
        deProgramDeclarations = new Declarations();
        codeProgramCode.setProgramCodeListener( this );
        deProgramDeclarations.setProgramDeclarationsListener( this );
        
        jfcFileChooser = new JFileChooser( "." );  // set to current directory
        jfcFileChooser.setMultiSelectionEnabled( false );        
        jfcFileChooser.setFileFilter( new PTVFileFilter() );
        
        jfcAssemblyFileChooser = new JFileChooser( "." );  // set to current directory
        jfcAssemblyFileChooser.setMultiSelectionEnabled( false );        
    }
    
    public void clearData() {
        codeProgramCode.clearData();
        deProgramDeclarations.clearData();
        AllPurposeMenu.clearData();
    }
    
    public String getFileName() {
        return stFileName;
    }
    
    public String getProgramName() {
        return stProgramName;
    }
    
    public void setDeclarations( Declarations deDeclarations ) {
        deProgramDeclarations = deDeclarations;
        deProgramDeclarations.setProgramDeclarationsListener( this );
    }
    
    // return original object
    public Declarations getDeclarations() {
        return deProgramDeclarations;        
    }
    
    public void setVisualCodeTable( VisualCodeTable vctThisVisualCodeTable ) {
        vctCodeTable = vctThisVisualCodeTable;
        codeProgramCode = vctCodeTable.getCode();
        codeProgramCode.setVisualCodeTable( vctCodeTable );
    }
    
    public VisualCodeTable getVisualCodeTable() {
        return vctCodeTable;
    }
    
    // return original object
    public Code getCode() {
        return codeProgramCode;
    }
    
    public void setParentFrame( JFrame jfFrame ) {
        jfParentFrame = jfFrame;

        pndProgramNameDialog = new ProgramNameDialog( jfParentFrame, 
            ST_APPLICATION_NAME, this );  
        System.out.println( "Frame in set parent frame = " + jfParentFrame );
        jfParentFrame.setTitle( ST_APPLICATION_NAME + "  >>  " + stProgramName );
    }
        
    //----------------- Menu-type ActionListener calls ----------------\\
    public boolean actionPerformedInProgram( String stMenuCommand ) {
        boolean bFound;
        
        bFound = true;
        
        if ( stMenuCommand.equals( AllPurposeMenu.ST_MENUITEM__MAIN_FILE_NEW ) ) {
            performMenuItemNew();
        }
        else if ( stMenuCommand.equals( AllPurposeMenu.ST_MENUITEM__MAIN_FILE_RENAME ) ) {
            performMenuItemRenameProgram();
        }
        else if ( stMenuCommand.equals( AllPurposeMenu.ST_MENUITEM__MAIN_FILE_OPEN ) ) {
            performMenuItemOpen();
        }
        else if ( stMenuCommand.equals( AllPurposeMenu.ST_MENUITEM__MAIN_FILE_SAVE ) ) {
            performMenuItemSave();
        }
        else if ( stMenuCommand.equals( AllPurposeMenu.ST_MENUITEM__MAIN_FILE_SAVE_AS ) ) {
            performMenuItemSaveAs();
        }
        else if ( stMenuCommand.equals( AllPurposeMenu.ST_MENUITEM__MAIN_FILE_GENERATE_CODE ) ) {
            performMenuGenerateCode();
        }
        else if ( stMenuCommand.equals( AllPurposeMenu.ST_MENUITEM__MAIN_FILE_EXIT ) ) {
            performMenuItemExit();
        }
        else {
            bFound = false;
        }        
        return bFound;
    }
    
    public void performMenuItemNew() {
        Status statInformation;
        int iStatusNumber;
        
        statInformation = checkIfFileShouldBeSaved();
        iStatusNumber = statInformation.getStatusNumber();
        
        if ( ( iStatusNumber == I_PROGRAM_WAS_NOT_CHANGED )  ||
             ( iStatusNumber == I_OK )  || 
             ( iStatusNumber == I_DO_NOT_SAVE_PROGRAM ) ) {
            // if name changed, ProgramListener will change the name of the program
            clearData();
            stFileName = "";
            System.out.println( "Clear data" );
            jfParentFrame.setTitle( ST_APPLICATION_NAME + "  >>  " );
            bProgramChangedFlag = false;
            statInformation = pndProgramNameDialog.showDialog();
            
            if ( statInformation.getStatusNumber() == 
                   ProgramNameDialog.I_CANCELLED_OPERATION ) {
                stProgramName = "";
            }
        }
    }
    
    public void performMenuItemRenameProgram() {
        // if name changed, ProgramListener will change the name of the program
        pndProgramNameDialog.showDialog();
    }
    
    public void performMenuItemOpen() {
        Status statInformation;
        int iStatusNumber;
        
        statInformation = checkIfFileShouldBeSaved();
        iStatusNumber = statInformation.getStatusNumber();
        
        if ( ( iStatusNumber == I_PROGRAM_WAS_NOT_CHANGED )  ||
             ( iStatusNumber == I_OK ) || 
             ( iStatusNumber == I_DO_NOT_SAVE_PROGRAM ) ) {
            openProgram();
        }
    }
    
    public void performMenuItemSaveAs() {
        saveProgram();
    }

    public void performMenuItemSave() {    
        Status statInformation;
        
        if ( stFileName.equals( "" ) == true ) {
            saveProgram();
        }
        else {
            statInformation = writeToFile();

            if ( statInformation.getStatusNumber() == I_OK ) {
                bProgramChangedFlag = false;
                jfParentFrame.setTitle( ST_APPLICATION_NAME + "  >>  " + stProgramName );
            }
            else {
                JOptionPane.showMessageDialog( jfParentFrame, 
                    statInformation.getStatusMessage(), ST_APPLICATION_NAME, 
                    JOptionPane.ERROR_MESSAGE );                
            }
        }
    }   
    
    public void performMenuItemExit() {
        int iDialogChoice, iStatusNumber;
        Status statInformation;
        
        iDialogChoice = JOptionPane.showConfirmDialog( jfParentFrame, 
            "Exit  Panels :: Tools :: Views?", ST_APPLICATION_NAME, 
            JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE );
        System.out.println( "Dialog Choice = " + iDialogChoice + "   OK_OPTION = "
            + JOptionPane.OK_OPTION );
        if ( iDialogChoice == JOptionPane.OK_OPTION ) {
            statInformation = checkIfFileShouldBeSaved();
            iStatusNumber = statInformation.getStatusNumber();
            
            if ( ( iStatusNumber == I_PROGRAM_WAS_NOT_CHANGED )  ||
                 ( iStatusNumber == I_OK ) || 
                 ( iStatusNumber == I_DO_NOT_SAVE_PROGRAM ) ) {
                System.exit( 0 );
            }
        }
    }
    
    //----------------- ProgramListener Implementation -----------\\
    public void programChanged() {
        bProgramChangedFlag = true;
        jfParentFrame.setTitle( ST_APPLICATION_NAME + "  >>  " + stProgramName
            + " [unsaved]" );
    }

    public void programNameChanged( String stName ) {
        stProgramName = stName;
        deProgramDeclarations.setProgramName( stProgramName );
        jfParentFrame.setTitle( ST_APPLICATION_NAME + "  >>  " + stProgramName
            + " [unsaved]" );
        System.out.println( "JFRame = " + jfParentFrame );
        System.out.println( "Program name changed to..." + stProgramName );
        bProgramChangedFlag = true;
    }

    //////////////////////// PRIVATE METHODS \\\\\\\\\\\\\\\\\\\\\\\\
    private Status writeToFile() {
        ObjectOutputStream oosSerializedObjectOutput;
        FileOutputStream fosOutputFile;
        String stMessage;

        if ( stFileName.equals( "" ) == false ) {            
            if ( ( jfcFileChooser.getFileFilter() instanceof PTVFileFilter )  &&
                 ( stFileName.endsWith( ".ptv" ) == false ) ) {
                stFileName = stFileName + ".ptv";
            }
            
            try {
                fosOutputFile = new FileOutputStream( stFileName );
                oosSerializedObjectOutput = new ObjectOutputStream( fosOutputFile );
                oosSerializedObjectOutput.writeObject( stProgramName );
                System.out.println( "write: stProgramName = " + stProgramName );
                
                oosSerializedObjectOutput.writeObject( codeProgramCode.exportState() );
                System.out.println( "write: codeProgramCode = " + codeProgramCode );                
    
                oosSerializedObjectOutput.writeObject( deProgramDeclarations.exportState() );                
                System.out.println( "write: deProgramDeclarations = " + deProgramDeclarations );
                
                oosSerializedObjectOutput.close();
                return new Status( I_OK );
            }
            catch ( FileNotFoundException fnfeException ) {
                stMessage = "The file named \"" + stFileName + "\" does not exist.";
                return new Status( I_FILE_ERROR, stMessage );
            }
            catch ( IOException ioeException ) {
                System.out.println( "write: ERROR = I/O" );
                return new Status( I_SERIALIZATION_ERROR, ioeException.getMessage() );
            }
            catch ( SecurityException seException ) {
                System.out.println( "write: ERROR = Security" );
                return new Status( I_FILE_ERROR, seException.getMessage() );
            }
/*            catch ( InvalidClassException iceException ) {
                return new Status( I_SERIALIZATION_ERROR, iceException.getMessage() );
            }
            catch ( NotSerializableException nseException ) {
                return new Status( I_SERIALIZATION_ERROR, nseException.getMessage() );
            }*/
        }
        else {
            stMessage = "The program does not have a name.";
            return new Status( I_NO_PROGRAM_NAME, stMessage );
        }
    }

    private Status saveProgram() {       
        Status statInformation;
        File fileOutput;
        boolean bContinue;
        int iResult, iDialogChoice;

        bContinue = false;

        try {
            do {
                iResult = jfcFileChooser.showSaveDialog( jfParentFrame );
                fileOutput = jfcFileChooser.getSelectedFile().getCanonicalFile();
                stFileName = fileOutput.getCanonicalPath();

                if ( ( iResult == JFileChooser.CANCEL_OPTION )  ||
                     ( iResult == JFileChooser.ERROR_OPTION ) ) {
                    bContinue = true;
                }            
                else if ( fileOutput.exists() == true ) {
                    iDialogChoice = JOptionPane.showConfirmDialog( jfParentFrame, 
                        stFileName + " already exists.  Should it be replaced?", 
                        ST_APPLICATION_NAME, JOptionPane.YES_NO_OPTION, 
                        JOptionPane.QUESTION_MESSAGE );

                    if ( iDialogChoice == I_OPTION_YES ) {
                        bContinue = true;
                    }
                }
                else {
                    bContinue = true;
                }
            }
            while ( bContinue == false );
        }
        catch ( IOException ioeException ) {
            return new Status( I_FILE_ERROR, "File path could not be created." );
        }
        
        if ( iResult == JFileChooser.APPROVE_OPTION ) {
            statInformation = writeToFile();

            if ( statInformation.getStatusNumber() == I_OK ) {
                bProgramChangedFlag = false;
                jfParentFrame.setTitle( ST_APPLICATION_NAME + "  >>  " + stProgramName );
                System.out.println( "=== Successful save ===" );
            }
            else {
                JOptionPane.showMessageDialog( jfParentFrame, 
                    statInformation.getStatusMessage(),
                    ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );
                System.out.println( "@@@ SAVE FAILED @@@" );
            }
        }
        else {
            System.out.println( "# # #   Save cancelled." );
            statInformation = new Status( I_CANCELLED_OPERATION );
        }
        return statInformation;
    }
        
    private Status readFromFile() {
        String stMessage;
        ObjectInputStream oisSerializedObjectInput;
        FileInputStream fisInputFile;
        Vector vCode, vDeclarations;

        try {            
            if ( ( jfcFileChooser.getFileFilter() instanceof PTVFileFilter )  &&
                 ( stFileName.endsWith( ".ptv" ) == false ) ) {
                stFileName = stFileName + ".ptv";
            }

            fisInputFile = new FileInputStream( stFileName );
            oisSerializedObjectInput = new ObjectInputStream( fisInputFile );
            stProgramName = (String)oisSerializedObjectInput.readObject();
            System.out.println( "read: stProgramName = " + stProgramName );
            
            vCode = (Vector)oisSerializedObjectInput.readObject();
            codeProgramCode.importState( vCode );
            System.out.println( "read: codeProgramCode = " + codeProgramCode );
            
            vDeclarations = (Vector)oisSerializedObjectInput.readObject();
            deProgramDeclarations.importState( vDeclarations );
            System.out.println( "read: deProgramDeclarations = " + deProgramDeclarations );
            
            oisSerializedObjectInput.close();
            return new Status( I_OK );
        }
        catch ( StreamCorruptedException sceException ) {
            System.out.println( "read: ERROR = Stream Corrupted" );
            return new Status( I_SERIALIZATION_ERROR, sceException.getMessage() );
        }
        catch ( IOException ioeException ) {
            System.out.println( "read: ERROR = I/O" );
            return new Status( I_SERIALIZATION_ERROR, ioeException.getMessage() );
        }
        catch ( SecurityException seException ) {
            System.out.println( "read: ERROR = Security" );
            return new Status( I_FILE_ERROR, seException.getMessage() );
        }
        catch ( ClassNotFoundException cnfeException ) {
            System.out.println( "read: ERROR = Class Not Found" );
            return new Status( I_SERIALIZATION_ERROR, cnfeException.getMessage() );
        }
/*        catch ( InvalidClassException iceException ) {
            return new Status( I_SERIALIZATION_ERROR, iceException.getMessage() );
        }
        catch ( OptionalDataException odeException ) {
            return new Status( I_SERIALIZATION_ERROR, odeException.getMessage() );
        }
        catch ( FileNotFoundException fnfeExcpetion ) {
            stMessage = "The file named \"" + stFileName + "\" does not exist.";
            return new Status( I_FILE_ERROR, stMessage );
        }*/
    }

    private Status openProgram() {
        Status statInformation;
        File fileInput;

        if ( jfcFileChooser.showOpenDialog( jfParentFrame ) == 
            JFileChooser.APPROVE_OPTION ) {

            try {
                fileInput = jfcFileChooser.getSelectedFile().getCanonicalFile();                
                stFileName = fileInput.getCanonicalPath();
                clearData();
                statInformation = readFromFile();

                if ( statInformation.getStatusNumber() == I_OK ) {
                    bProgramChangedFlag = false;
                    jfParentFrame.setTitle( ST_APPLICATION_NAME + "  >>  " + stProgramName );
                    System.out.println( "=== File opened ===" );
                }
                else {
                    JOptionPane.showMessageDialog( jfParentFrame, 
                        statInformation.getStatusMessage(),
                        ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );
                    System.out.println( "@@@ OPEN FAILED @@@" );
                    stProgramName = "";
                }

                bProgramChangedFlag = false;
                jfParentFrame.setTitle( ST_APPLICATION_NAME + "  >>  " + stProgramName );            
            }
            catch ( IOException ioeException ) {
                return new Status( I_FILE_ERROR, "File path could not be created." );
            }
        }
        else {
           statInformation = new Status( I_CANCELLED_OPERATION );
        }
        return statInformation;
    }
            
    private Status checkIfFileShouldBeSaved() {
        int iDialogChoice;
        Status statInformation;
        
        if ( bProgramChangedFlag == true ) {
            iDialogChoice = JOptionPane.showConfirmDialog( jfParentFrame, 
                "Save changes made to " + stProgramName + "?",
                ST_APPLICATION_NAME, JOptionPane.YES_NO_CANCEL_OPTION, 
                JOptionPane.QUESTION_MESSAGE );

            if ( iDialogChoice == JOptionPane.OK_OPTION ) {
                statInformation = saveProgram();
            }
            else if ( iDialogChoice == JOptionPane.NO_OPTION ) {
                statInformation = new Status( I_DO_NOT_SAVE_PROGRAM );
            }
            else {
                statInformation = new Status( I_CANCELLED_OPERATION );
            }
        }
        else {
           statInformation = new Status( I_PROGRAM_WAS_NOT_CHANGED );
        }
        return statInformation;
    }
    
    private void performMenuGenerateCode() {
        Status statInformation;
        String stMessage, stFilePath, stActualName;
        int iStatusNumber, iResult, iDialogChoice;
        File fileAssembly;
        boolean bContinue;
        
        bContinue = false;
        
        do {
            try {
                iResult = jfcAssemblyFileChooser.showSaveDialog( jfParentFrame );
                fileAssembly = jfcAssemblyFileChooser.getSelectedFile().getCanonicalFile();
                stFilePath = fileAssembly.getCanonicalPath();
                stActualName = fileAssembly.getName();
                
                if ( stActualName.length() <= 8 ) {
                    if ( ( iResult == JFileChooser.CANCEL_OPTION )  ||
                         ( iResult == JFileChooser.ERROR_OPTION ) ) {
                        bContinue = true;
                    }
                    else if ( fileAssembly.exists() == true ) {
                        iDialogChoice = JOptionPane.showConfirmDialog( jfParentFrame, 
                            stFilePath + " already exists.  Should it be replaced?", 
                            ST_APPLICATION_NAME, JOptionPane.YES_NO_OPTION, 
                            JOptionPane.QUESTION_MESSAGE );

                        if ( iDialogChoice == I_OPTION_YES ) {
                            bContinue = true;
                        }
                    }
                    else {
                        bContinue = true;
                    }
                }
                else {
                    JOptionPane.showMessageDialog( jfParentFrame, 
                        "File name must have 8 characters or less.  No file extension is needed.",
                        ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );                    
                }
            }
            catch ( IOException ioeException ) {
                System.out.println( "File path could not be created." );
                iResult = JFileChooser.ERROR_OPTION;
                stFilePath = "";
            }                
        }
        while ( bContinue == false );
        
        if ( iResult == JFileChooser.APPROVE_OPTION ) {                
            statInformation = AssemblyProgramView.translateCodeToAssemblyView( 
                stFilePath, this );
            iStatusNumber = statInformation.getStatusNumber();
            
            if ( iStatusNumber != AssemblyProgramView.I_OK ) {
                JOptionPane.showMessageDialog( jfParentFrame, 
                    statInformation.getStatusMessage(),
                    ST_APPLICATION_NAME, JOptionPane.ERROR_MESSAGE );                
            }
            else {
                //  file created
                JOptionPane.showMessageDialog( jfParentFrame, 
                    "Assembly language view of the program was written to file.",
                    ST_APPLICATION_NAME, JOptionPane.INFORMATION_MESSAGE );                
            }
        }
        //  else, ignore
    }
    
    
    // instance variables
    private String stFileName, stProgramName;
    private VisualCodeTable vctCodeTable;
    private Code codeProgramCode;
    private Declarations deProgramDeclarations;
    private JFrame jfParentFrame;
    private JFileChooser jfcFileChooser, jfcAssemblyFileChooser;
    private ProgramNameDialog pndProgramNameDialog;
    private boolean bProgramChangedFlag;
    
    // class constants
    private final static String ST_APPLICATION_NAME = "Panels :: Tools :: Views";    
    private static final int I_OPTION_YES = 0;

    public static final int I_OK = 0;
    public static final int I_FILE_ERROR = -1;
    public static final int I_SERIALIZATION_ERROR = -2;
    public static final int I_NO_PROGRAM_NAME = -3;        
    public static final int I_CANCELLED_OPERATION = -4;
    public static final int I_PROGRAM_WAS_NOT_CHANGED = -5;
    public static final int I_DO_NOT_SAVE_PROGRAM = -6;
}