/*
 * RecordObject.java
 *
 * Created on January 31, 2002, 10:52 AM
 */

//package panelsAndTools;

import java.util.*;
import java.lang.*;
import java.io.*;

/**
 *
 * @author  _
 * @version 
 */

public class RecordObject implements Cloneable, Serializable {
    public RecordObject() {
        hmFieldMap = new HashMap();
        alFieldNamesList = new ArrayList();
        iRecordSize = 0;        
    }    
    
    public void clearData() {
        hmFieldMap.clear();
        alFieldNamesList.clear();
        iRecordSize = 0;        
    }        

    protected Object clone() {
        // strucutre based on Sun's online Java tutorial for
        // the Object.clone() method and pp.235-6 C of Core Java 2: ISBN 0-13-081933-6        
        try {
            RecordObject roObject = (RecordObject)super.clone();
            roObject.hmFieldMap = (HashMap)this.hmFieldMap.clone();
            roObject.alFieldNamesList = (ArrayList)this.alFieldNamesList.clone();
            
            return roObject;
        }
        catch ( CloneNotSupportedException cnseException ) {
            System.out.println( "Clone of RecordObject is not supported." );
            return null;
        }
    }
        
    public RecordObject getRecord() {
        return (RecordObject)this.clone();
    }
    
    public Object getFieldAtIndex( int iIndex ) {
        Object objThisObject;
        DataObject doThisDataObject;
        String stMessage, stFieldName;
        int iNumberOfFields;
        
        iNumberOfFields = alFieldNamesList.size();
        //System.out.println( "RecordObject.getFieldAtIndex: iNumberOfFields = " + iNumberOfFields );
        
        if ( iNumberOfFields > 0 ) {
            if ( ( iIndex >= 0 )  &&  ( iIndex < iNumberOfFields ) ) {
                //System.out.println( "RecordObject.getFieldAtIndex:  VALID INDEX!!!" );
                stFieldName = (String)alFieldNamesList.get( iIndex );
                //System.out.println( "RecordObject.getFieldAtIndex: stFieldName = " +stFieldName );
                doThisDataObject = (DataObject)hmFieldMap.get( stFieldName );
                //System.out.println( "---PRINTING OBJECT---" );
                //doThisDataObject.printDataObject();

                objThisObject = new NameAndDataObjectWrapper( new String( stFieldName ),
                        (DataObject)doThisDataObject.clone() );
            }
            else {
                stMessage = "RecordObject.getFieldAtIndex: The index value " + iIndex + " does not correspond " +
                    "to a field in this record.";                
                objThisObject = new Status( I_INVALID_RECORD_INDEX, stMessage );        
            }
        }
        else {
            stMessage = "This record does not have any fields defined.";
            objThisObject = new Status( I_EMPTY_RECORD, stMessage );
        }

        //if ( objThisObject instanceof NameAndDataObjectWrapper )
            //System.out.println( "Leaving RecordObject with objThisObject as a NameAndDataObjectWrapper." );

        return objThisObject;
    }
    
    public Status addField( String stFieldName, DataObject doThisDataObject ) {
        Status statInformation;
        String stMessage;
        
        statInformation = validateObjectName( stFieldName );
        
        if ( statInformation.getStatusNumber() == I_OK ) {        
            if ( hmFieldMap.containsKey( stFieldName ) == false ) {
                hmFieldMap.put( stFieldName, doThisDataObject );
                alFieldNamesList.add( stFieldName );
                iRecordSize += doThisDataObject.getSize();
                statInformation = new Status( I_OK );
            }
            else {            
                stMessage = "The field named " + stFieldName + " already exists in "
                    + "this record.";
                statInformation = new Status( I_FIELD_DECLARED, stMessage );
            }
        }
        // else error already present in statInformation [invalid name]
            
        return statInformation;
    }
    
    public Status removeField( String stFieldName ) {
        Status statInformation;
        String stMessage;
        DataObject doThisDataObject;
        int iIndexOfName;
        
        if ( hmFieldMap.containsKey( stFieldName ) ) {
            doThisDataObject = (DataObject)hmFieldMap.get( stFieldName );
            hmFieldMap.remove( stFieldName );
            iIndexOfName = alFieldNamesList.indexOf( stFieldName );
            alFieldNamesList.remove( iIndexOfName );
            iRecordSize -= doThisDataObject.getSize();            
            statInformation = new Status( I_OK );
        }
        else {            
            stMessage = "The field named " + stFieldName + " does not exist in "
                + "this record.";
            statInformation = new Status( I_FIELD_UNDECLARED, stMessage );
        }
        
        return statInformation;                            
    }
    
    public Object getField( String stFieldName ) {
        Object objThisObject;
        Status statInformation;
        String stMessage;
        DataObject doTemp;
        
        doTemp = new DataObject();
        
        if ( hmFieldMap.containsKey( stFieldName ) ) {
            doTemp = (DataObject)hmFieldMap.get( stFieldName );
            objThisObject = doTemp.clone();
        }
        else {            
            stMessage = "The field named " + stFieldName + "is not declared "
                + "in this record.";
            objThisObject = new Status( I_FIELD_UNDECLARED, stMessage );
        }

        return objThisObject;
    }
    
    public Status setFieldName( String stOldFieldName, String stNewFieldName ) {
        Status statInformation;
        String stMessage;
        int iIndexOfName;
        DataObject doThisDataObject;

        statInformation = validateObjectName( stNewFieldName );
        
        if ( statInformation.getStatusNumber() == I_OK ) {        
            if ( hmFieldMap.containsKey( stOldFieldName ) == true ) {
                if ( hmFieldMap.containsKey( stNewFieldName ) == false ) {
                    
                    // replace the old name with the new name
                    iIndexOfName = alFieldNamesList.indexOf( stOldFieldName );
                    alFieldNamesList.set( iIndexOfName, stNewFieldName );

                    // remove the old key and insert the new key (same object)
                    doThisDataObject = (DataObject)hmFieldMap.get( stOldFieldName );
                    hmFieldMap.remove( stOldFieldName );
                    hmFieldMap.put( stNewFieldName, doThisDataObject );
                    statInformation = new Status( I_OK );                    
                }
                else {
                    stMessage = "The name " + stNewFieldName + " already is a field "
                        + "in this record.";
                    statInformation = new Status( I_FIELD_DECLARED, stMessage );
                }                
            }
            else {
                stMessage = "The field named " + stOldFieldName + " is not declared "
                    + "in this record.";
                statInformation = new Status( I_FIELD_UNDECLARED, stMessage );
            }
        }
        // else...error exists in statInformation [invalid name]

        return statInformation;
    }
    
    public Status setFieldProperties( String stFieldName, 
                                      DataObject doNewDataObject ) {
        Status statInformation;
        String stMessage;
        DataObject doOldDataObject;
        
        if ( hmFieldMap.containsKey( stFieldName ) ) {
            doOldDataObject = (DataObject)hmFieldMap.get( stFieldName );
            iRecordSize -= doOldDataObject.getSize();
            hmFieldMap.put( stFieldName, doNewDataObject );
            iRecordSize += doNewDataObject.getSize();
            statInformation = new Status( I_OK );
        }
        else {
            stMessage = "The field named " + stFieldName + " is not declared "
                + "in this record.";
            statInformation = new Status( I_FIELD_UNDECLARED, stMessage );
        }

        return statInformation;
    }
    
    public boolean isFieldInRecord( String stFieldName ) {        
        return hmFieldMap.containsKey( stFieldName );
    }
    
    public int getRecordSize() {
        return iRecordSize;
    }   
    
    public Object getFieldNames() {
        Object objThisObject;
        String stMessage;
        
        if ( alFieldNamesList.size() > 0 ) {
            objThisObject = ((ArrayList)(alFieldNamesList)).clone();
        }
        else {
            stMessage = "This record does not have any fields defined.";
            objThisObject = new Status( I_EMPTY_RECORD, stMessage );
        }

        return objThisObject;
    }
    
    public Object getFieldObjects() {
        Object objThisObject;
        String stMessage;
        
        if ( hmFieldMap.size() > 0 ) {
            objThisObject = hmFieldMap.values();
        }
        else {
            stMessage = "This record does not have any fields defined.";
            objThisObject = new Status( I_EMPTY_RECORD, stMessage );
        }

        return SerializedClone.cloneObject( objThisObject );
    }
        
    
    public Object writeRecordObject() {
        return new Object();
    } 
    
    public Status readIntoRecordObject( StringBuffer sbData ) {
        return new Status( I_OK );
    }
    
    public int getNumberOfFields() {
        return hmFieldMap.size();
    }
    
    public int getIndexOfObject( String stObjectName ) {
        return alFieldNamesList.indexOf( stObjectName );
    }        
    
    /////////////////// FILE I/O METHODS \\\\\\\\\\\\\\\\\\\\\\\\
    public Vector exportState() {
        Vector vRecord, vData, vDataObject;
        DataObject doThisDataObject;
        String stDataObjectName;
        
        vData = new Vector();
        vData.setSize( hmFieldMap.size() );        
        
        for ( int iIndex = 0; iIndex <  hmFieldMap.size(); iIndex++ ) {
            vRecord = new Vector(); 
            vRecord.setSize( I_TOTAL_FIELD_COUNT );
            
            stDataObjectName = (String)alFieldNamesList.get( iIndex );
            vRecord.set( I_DATA_OBJECT_NAME, stDataObjectName );
            System.out.println( "RECORD OBJECT: stDataObjectName=" + stDataObjectName );
            
            doThisDataObject = (DataObject)getField( stDataObjectName ); 
            vDataObject = doThisDataObject.exportState();
            vRecord.set( I_DATA_OBJECT,  vDataObject );
            System.out.println( "RECORD OBJECT: vDataObject=" + vDataObject );
            
            vData.set( iIndex, vRecord );
            System.out.println( "RECORD OBJECT: vRecord=" + vRecord );            
        }
        System.out.println( "RECORD OBJECT: vData=" + vData );
        return vData;
    }
    
    public void importState( Vector vData ) {
        Vector vRecord, vDataObject;
        String stDataObjectName;
        DataObject doThisDataObject;
  
        for ( int iIndex = 0; iIndex < vData.size(); iIndex++ ) {
            vRecord = (Vector)vData.get( iIndex );
            stDataObjectName = (String)vRecord.get( I_DATA_OBJECT_NAME );
       
            vDataObject = (Vector)vRecord.get( I_DATA_OBJECT );
            doThisDataObject = new DataObject();
            doThisDataObject.importState( vDataObject );
            
            addField( stDataObjectName, doThisDataObject );
        }
    }
    
    ///////////////// CLASS METHOD \\\\\\\\\\\\\\\\\\\\\\\    
    public static Status validateObjectName( String stName ) {
        char chFirstCharacter, chOtherCharacter;
        int iNameLength, iNameCharacterIndex;
        boolean bValidCharacter;
        Status statInformation;
        String stMessage;
        
        chFirstCharacter = stName.charAt( 0 );      
        iNameLength = stName.length();
        
        if ( ( iNameLength > 0 )  &&  ( iNameLength <= 8 ) ) {
        
            if ( ( ( chFirstCharacter >= 'A' )  && ( chFirstCharacter <= 'Z' ) )  ||
                   ( chFirstCharacter == '$' ) ||
                   ( chFirstCharacter == '#' ) ||
                   ( chFirstCharacter == '@' ) ) {
            
            
                for ( iNameCharacterIndex = 1, bValidCharacter = true;
                      ( ( iNameCharacterIndex < iNameLength )  && 
                        ( bValidCharacter == true ) );
                      iNameCharacterIndex++ ) {

                    chOtherCharacter = stName.charAt( iNameCharacterIndex );

                    if ( ( ( chOtherCharacter >= 'A' )  && 
                           ( chOtherCharacter <= 'Z' ) )  ||
                         ( ( chOtherCharacter >= '0' )  && 
                           ( chOtherCharacter <= '9' ) )  ||
                         ( chOtherCharacter == '$' ) ||
                         ( chOtherCharacter == '#' ) ||
                         ( chOtherCharacter == '@' ) ) {
                        bValidCharacter = true;
                    }
                    else {
                        bValidCharacter = false;
                    }
                }

                if ( bValidCharacter == true ) {
                    statInformation = new Status( I_OK );                
                }            
                else {            
                    stMessage = "A character other than the first character of "                
                        + "the name " + stName + " is not a valid character.";               
                    statInformation = new Status( I_INVALID_NAME, stMessage );                
                }            
            }
            else {
                stMessage = "The first character of the name "
                    + stName + " is not a valid character.";               
                statInformation = new Status( I_INVALID_NAME, stMessage );                            
            }
        }
        else {
            stMessage = "An object name must have between 1 and 8 characters.";
            statInformation = new Status( I_INVALID_NAME, stMessage );                            
        }
        
        return statInformation;
    }

    
    // instance variables
    private HashMap hmFieldMap; 
    private ArrayList alFieldNamesList;
    private int iRecordSize;


    // class variables
    private static final int I_COLUMN_COUNT = 5; 
    
    public static final int I_TOTAL_FIELD_COUNT = 2;
    public static final int I_DATA_OBJECT_NAME = 0;
    public static final int I_DATA_OBJECT = 1;
    
    
        // column codes
        public static final int I_COLUMN_NAME = 0;
        public static final int I_COLUMN_TYPE = 1;
        public static final int I_COLUMN_RESTRICTION = 2;
        public static final int I_COLUMN_STRUCTURE = 3;
        public static final int I_COLUMN_SIZE = 4;
    
        // error codes
        public static final int I_OK = 0;
        public static final int I_INVALID_TABLE_CELL = -1;
        public static final int I_EMPTY_RECORD = -2;
        public static final int I_INVALID_RECORD_INDEX = -3;        
        public static final int I_FIELD_DECLARED = -4;        
        public static final int I_FIELD_UNDECLARED = -5;
        public static final int I_INVALID_NAME = -6;
}