/*
 * SerializedClone.java
 *
 * Created on March 21, 2002, 6:19 PM
 */

// package panelsAndTools;

import java.io.*;

/**
 *
 * @author  _
 * @version 
 */

public class SerializedClone {
    public static Object cloneObject( Object objOriginal ) {
        // based on the structure given on p.677 of Core Java 2: ISBN 0-13-081933-6        
        Object objCloned;
        Class clsObjectClass;
        ObjectOutputStream oosSerializationOut;
        ObjectInputStream oisSerializationIn;
        ByteArrayInputStream baisSerializationIn;
        ByteArrayOutputStream baisSerializationOut;

        clsObjectClass = objOriginal.getClass();
        //System.out.println( "SerializedClone: Received " + clsObjectClass + " object" );
        
        try {
            if ( objOriginal instanceof Integer ) {
                objCloned = new Integer( ((Integer)objOriginal).toString() );
            }
            else if ( objOriginal instanceof Float ) {
                objCloned = new Float( ((Float)objOriginal).toString() );                
            }
            else if ( objOriginal instanceof Character ) {
                objCloned = new Character( ((Character)objOriginal).charValue() );
            }
            else if ( objOriginal instanceof NameAndDataObjectWrapper ) {
                NameAndDataObjectWrapper nadowTemp;
                
                nadowTemp = (NameAndDataObjectWrapper)objOriginal;
                objCloned = new NameAndDataObjectWrapper( 
                    nadowTemp.getName(), nadowTemp.getObject() );
            }
            else if ( objOriginal instanceof Status ) {
                Status statInformation;
                
                statInformation = (Status)objOriginal;
                objCloned = new Status( statInformation.getStatusNumber(),
                    statInformation.getStatusMessage() );
            }
            else {
                baisSerializationOut = new ByteArrayOutputStream();
                oosSerializationOut = new ObjectOutputStream( baisSerializationOut );

                objCloned = clsObjectClass.newInstance();
                oosSerializationOut.writeObject( objOriginal );
                oosSerializationOut.close();

                baisSerializationIn = 
                    new ByteArrayInputStream( baisSerializationOut.toByteArray() );            
                oisSerializationIn = new ObjectInputStream( baisSerializationIn );
                objCloned = oisSerializationIn.readObject();
                oisSerializationIn.close();
            }
            return objCloned;
        }
        catch ( IllegalAccessException iaeException ) {
            System.out.println( "Calling SerializedClone with class " + clsObjectClass + ": Illegal Access Exception" );
            return null;
        }
        catch ( InstantiationException ineException ) {
            System.out.println( "Calling SerializedClone with class " + clsObjectClass + ": Instantiation Exception" );
            return null;
        }            
        catch ( IOException ioeException ) {
            System.out.println( "Calling SerializedClone with class " + clsObjectClass + ": I/O Exception" );
            return null;
        }
        catch ( ClassNotFoundException cnfeException ) {
            System.out.println( "Calling SerializedClone with class " + clsObjectClass + ": Class Not Found Exception" );
            return null;
        }
    }
}