/*
 * SortableDeclarationsTableModel.java
 *
 * Created on February 22, 2002, 8:30 PM
 */

//package panelsAndTools;

import javax.swing.event.*;
import javax.swing.table.*;
import java.util.*;
import java.lang.*;
import java.io.*;


/**
 *
 * @author  _
 * @version 
 */


public class SortableDeclarationsTableModel extends AbstractTableModel
    implements TableModelListener, Cloneable, Serializable {

    public SortableDeclarationsTableModel() {
        dtmDataObjects = new DeclarationsTableModel();
        iNumberOfObjects = 0;
        iIndexArraySize = 5;
        i1dSortedIndices = new int[ iIndexArraySize ];
        alPackedDataObjectStrings = new ArrayList();
    }
    
    protected Object clone() {
        // strucutre based on Sun's online Java tutorial for
        // the Object.clone() method and pp.235-6 C of Core Java 2: ISBN 0-13-081933-6
        try {
            SortableDeclarationsTableModel sdtmObject = 
                (SortableDeclarationsTableModel)super.clone();
            sdtmObject.dtmDataObjects = (DeclarationsTableModel)this.dtmDataObjects.clone();
            sdtmObject.alPackedDataObjectStrings = 
                (ArrayList)this.alPackedDataObjectStrings.clone();
            
            return sdtmObject;
        }
        catch ( CloneNotSupportedException cnseException ) {
            System.out.println( "Clone of SortableDeclarationsTableModel is not supported." );
            return null;
        }
    }
        
    public Status setColumnToTrack( int iColumnNumber ) {
        System.out.println( "SORTABLE_TABLE_MODEL...Column to track =  " + iColumnNumber );
        //call DecTabModel getRelevantDecs
        return dtmDataObjects.setColumnToTrack( iColumnNumber );
    }
    
    public Status setDeclarationsModel( Declarations deDeclarationsObject ) {        
        dtmDataObjects.setTableModelListener( this );
        //call update data in table

        return dtmDataObjects.setDeclarationsModel( deDeclarationsObject );
    }
    
    public Status setRecordName( String stRecordName ) {
        //call update data in table
        return dtmDataObjects.setRecordName( stRecordName );
    }
    
    public Status setModelNumber( int iModelNumber ) {
        Status statInformation;
        
        statInformation = dtmDataObjects.setModelNumber( iModelNumber );
        
        if ( statInformation.getStatusNumber() == DeclarationsTableModel.I_OK ) {
            //call update data in table
            alPackedDataObjectStrings = dtmDataObjects.getPackedInformationList();
            clearIndexArray();
            setNumberOfObjects();
            sortData();        
        }
        
        return statInformation;
    }

    public Object getDataFromRow( int iRow ) {
        return dtmDataObjects.getDataObjectFromRow( i1dSortedIndices[ iRow ] );        
    }
    
    public void setListener( SortableDeclarationsTableModelListener sdtml ) {
        System.out.println(" Set listener to... " + sdtml );
        sdtmlListener = sdtml;
        
        /*if ( sdtmlListener == null ) {
            System.out.println( "SO WHAT...IT'S NULL!!!" );
        }
        else {
            System.out.println( ">>> HEY <<<  IT'S NOT NULL " );
        }*/
    }
    
    public int getNumberOfObjects() {
        return iNumberOfObjects;
    }
    
    //////////////////////// PRIVATE FUNCTIONS \\\\\\\\\\\\\\\\\\\\\\\\
    private void sortData() {
        int iNumber, iStartIndexNumberPosition;
        String stTemp, stIndex;
        
        System.out.println("Sorting...");
        Collections.sort( alPackedDataObjectStrings );
        
        // extract the original indices to give the mapping from unsorted to sorted
        for ( int iIndex = 0; iIndex < iNumberOfObjects; iIndex++ ) {
            stTemp = (String)alPackedDataObjectStrings.get( iIndex );            
            //System.out.println( "index " + iIndex + " = " + stTemp );
            iStartIndexNumberPosition = stTemp.lastIndexOf( '|' ) + 1;
            
            stIndex = stTemp.substring( iStartIndexNumberPosition, stTemp.length() );
            
            /*stIndex = stTemp.substring( iStartIndexNumberPosition, 
                iStartIndexNumberPosition + DeclarationsTableModel.I_INDEX_STRING_LENGTH );*/
            
            iNumber = Integer.parseInt( stIndex );            
            i1dSortedIndices[ iIndex ] = iNumber;
        }
        fireTableDataChanged();

        //System.out.println( "SDTML Listener = " + sdtmlListener );

        if ( sdtmlListener != null ) {
            sdtmlListener.updateDeclarationsMenus( AllPurposeMenu.B_USE_PARTIAL_LABEL_LIST );
        }
    }
    
    private void updateDataInTable() {
        alPackedDataObjectStrings = dtmDataObjects.getPackedInformationList();
        clearIndexArray();
        setNumberOfObjects();
        sortData();  // fires an update to JTable
    }

        
    ///////////////////// ARRAY FUNCTIONS \\\\\\\\\\\\\\\\\\\\\

    private void clearIndexArray() {
        int iIndex;
        
        for ( iIndex = 0; iIndex < iNumberOfObjects; iIndex++ )
            i1dSortedIndices[ iIndex ] = I_UNKNOWN;
    }
        
    private void expandIndexArray() {
        int iIndex, iNewIndexArraySize;
        int[] i1dNewSortedIndices;
        
        iNewIndexArraySize = iIndexArraySize * 2;
        i1dNewSortedIndices = new int[ iNewIndexArraySize ];
        
        for ( iIndex = 0; iIndex < iIndexArraySize; iIndex++ )
            i1dNewSortedIndices[ iIndex ] = i1dSortedIndices[ iIndex ];
        
        i1dSortedIndices = i1dNewSortedIndices;
        iIndexArraySize = iNewIndexArraySize;
    }
    
    private void setNumberOfObjects() {
        int iCurrentObjectCount;
                
        iCurrentObjectCount = alPackedDataObjectStrings.size();
        if ( iCurrentObjectCount > iIndexArraySize ) {
            expandIndexArray();
        }
        
        iNumberOfObjects = iCurrentObjectCount;        
    }
            
    
    ///////////// ABSTRACT TABLE MODEL METHODS \\\\\\\\\\\\\\\\\
    public int getColumnCount() { 
        return dtmDataObjects.getColumnCount();
    }
    
    public String getColumnName( int iColumn ) {        
        return dtmDataObjects.getColumnName( iColumn );
    }
    
    public Object getValueAt( int iRow, int iColumn ) {
        Object objThisObject;
        /*        System.out.print( "SortedDeclarationModel.getValueAt:: " );
        System.out.println( "iRow = " + iRow + "  iColumn = " + iColumn +
            "  i1dSortedIndices[ iRow ] = " + i1dSortedIndices[ iRow ] );*/
        
        objThisObject = dtmDataObjects.getValueAt( i1dSortedIndices[ iRow ], iColumn );        
        /**DEBUGGING**/
//        System.out.println("IN SORTABLE_DECLARATIONS_TABLE_MODEL...row = " +iRow+", column = " + iColumn + ": " );
//        System.out.println( objThisObject );
        /**END EBUG**/

        return objThisObject;
    }
    
    public int getRowCount() { 
        return dtmDataObjects.getRowCount();
    }
    
    public void tableChanged( TableModelEvent tmeEvent ) {
        alPackedDataObjectStrings = dtmDataObjects.getPackedInformationList();
        clearIndexArray();
        setNumberOfObjects();
        sortData();
        System.out.println( "Table changed." );
    }       

    // not required but recommended in the Java Tutorial from Sun in
    // order to allow the correct cell renderer to be called for 
    // the JTable
    public Class getColumnClass( int iColumn ) {
        return dtmDataObjects.getColumnClass( iColumn );        
    }
               
    
    // instance variables
    private SortableDeclarationsTableModelListener sdtmlListener;
    private DeclarationsTableModel dtmDataObjects;
    private ArrayList alPackedDataObjectStrings;
    private int[] i1dSortedIndices;
    private int iNumberOfObjects, iIndexArraySize;
    
        // class variables
        private static int I_UNKNOWN = -8888;      
}