/*
 * VisualCodeTableModel.java
 *
 * Created on April 19, 2002, 12:31 PM
 */

//package panelsAndTools;

import javax.swing.event.*;
import javax.swing.table.*;
import javax.swing.*;
import java.util.*;
import java.lang.*;
import java.io.*;

/**
 *
 * @author  _
 * @version 
 */


public class VisualCodeTableModel extends AbstractTableModel
    implements CodeListener, Cloneable, Serializable {

    public VisualCodeTableModel() {
        codeModel = new Code();
    }
    
    protected Object clone() {
        // strucutre based on Sun's online Java tutorial for
        // the Object.clone() method and pp.235-6 C of Core Java 2: ISBN 0-13-081933-6        
        try {
            VisualCodeTableModel vctmObject = (VisualCodeTableModel)super.clone();
            vctmObject.codeModel = (Code)this.codeModel.clone();            
            return vctmObject;
        }
        catch ( CloneNotSupportedException cnseException ) {
            System.out.println( "Clone of VisualCodeTableModel is not supported." );
            return null;
        }        
    }
    
    public void setCodeModel( Code codeThisModel ) {
        codeModel = codeThisModel;
        codeModel.setCodeListener( this );
    }
    
    ////////////////// CODE LISTENER IMPLEMENTATION \\\\\\\\\\\\\\\\
    public void insertLinesOfCode( int iFirstLine, int iLastLine ) {
        fireTableRowsInserted( iFirstLine, iLastLine );
    }

    public void insertSingleLineOfCode( int iLine ) {
        fireTableRowsInserted( iLine, iLine );
    }
    
    public void changeSingleLineOfCode( int iLine ) {
        fireTableRowsUpdated( iLine, iLine ); 
    }
    
    public void deleteLinesOfCode( int iFirstLine, int iLastLine ) {
        fireTableRowsDeleted( iFirstLine, iLastLine );
    }
    
    public void deleteSingleLineOfCode( int iLine ) {
        fireTableRowsDeleted( iLine, iLine );
    }

    //////////// ABSTRACT TABLE MODEL IMPLEMENTATION \\\\\\\\\\\\\\\
    public int getRowCount() {
        return codeModel.getNumberOfLines();
    }
      
    public int getColumnCount() {
        return I_NUMBER_OF_COLUMNS;
    }
    
    public Object getValueAt( int iRow, int iColumn ) {
        Instruction inst;
        Object objThisObject;

        objThisObject = codeModel.getSingleLine( iRow );
        
        if ( objThisObject instanceof Instruction ) {
            inst = (Instruction)objThisObject;
            //inst = (Instruction)codeModel.getSingleLine( iRow );
            /*System.out.println( "Get value at row " + iRow + "..." );
            System.out.println( "LABEL = " + inst.getLabel().getName() );
            System.out.println( "OPCODE = " + inst.getOpcode().getOpcode() );
            System.out.println( "LEFT = " + inst.getLeftOperand().getOperandName() );
            System.out.println( "RIGHT = " + inst.getRightOperand().getOperandName() );
            System.out.println( "COMMENTS = " + inst.getComments() );*/
        }
        return codeModel.getSingleLine( iRow );
    }

        /*
         * JTable uses this method to determine the default renderer/
         * editor for each cell.  If we didn't implement this method,
         * then the last column would contain text ("true"/"false"),
         * rather than a check box.
         * TAKEN FROM JAVA TUTORIAL: How To Use Tables (12-2001)
         */
    
    public Class getColumnClass( int iColumn ) {
        return getValueAt( 0, iColumn ).getClass();
    }
    
    
    // instance variables
    private Code codeModel;
    
    // class constants
    public static final int I_NUMBER_OF_COLUMNS = 1;
}